//
//  AI.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "AI.h"
#import "entities.h"
#import "ResourceManager.h"


@implementation AI

- (id) init
{    
    self = [super init];
    //
	ai_stack = [[NSMutableArray alloc] initWithCapacity:8];
	//
	pendingMessages = [[NSMutableArray alloc] initWithCapacity:16]; // alloc retains
	//
	nextThinkTime = 0.0;
	//
	aiLock = [[NSLock alloc] init];
	//
	thinkTimeInterval = AI_THINK_INTERVAL;
	//
    return self;
}

- (void) dealloc
{
	if (owner_desc)			[owner_desc release];
	if (ai_stack)			[ai_stack release];
    if (stateMachine)		[stateMachine release];
    if (currentState)		[currentState release];
	if (pendingMessages)	[pendingMessages release];
	if (aiLock)				[aiLock release];
	[super dealloc];
}

- (id) initWithStateMachine:(NSString *) smName andState:(NSString *) stateName
{    
    self = [super init];
    //
	pendingMessages = [[NSMutableArray alloc] initWithCapacity:16]; // alloc retains
	//
	aiLock = [[NSLock alloc] init];
	//
	[self setStateMachine:smName];
	//
	currentState = [stateName retain];
	//
    return self;
}

- (void) setOwner:(ShipEntity *)ship
{
	owner = ship;   // now we assume this is retained elsewhere!
	if (owner_desc)			[owner_desc release];
	owner_desc = [[NSString stringWithFormat:@"%@ %d", [owner name], [owner universal_id]] retain];
}

- (void) preserveCurrentStateMachine
{
	NSMutableDictionary *pickledMachine = [NSMutableDictionary dictionaryWithCapacity:3];
	[pickledMachine setObject:stateMachine forKey:@"stateMachine"];
	[pickledMachine setObject:currentState forKey:@"currentState"];
	[pickledMachine setObject:pendingMessages forKey:@"pendingMessages"];
	
	if (!ai_stack)
		ai_stack = [[NSMutableArray alloc] initWithCapacity:8];
	
	[ai_stack insertObject:pickledMachine atIndex:0];	//  PUSH
}

- (void) restorePreviousStateMachine
{
	if (!ai_stack)
		return;
	if ([ai_stack count] < 1)
		return;
	NSMutableDictionary *pickledMachine = [ai_stack objectAtIndex:0];
	
	//debug
	//NSLog(@"restoring pickled ai :\n%@",[pickledMachine description]);
	
	[aiLock lock];
	if (stateMachine)   [stateMachine release];
	stateMachine = [[NSDictionary dictionaryWithDictionary:(NSDictionary *)[pickledMachine objectForKey:@"stateMachine"]] retain];
	if (currentState)   [currentState release];
	currentState = [[NSString stringWithString:(NSString *)[pickledMachine objectForKey:@"currentState"]] retain];
	if (pendingMessages)   [pendingMessages release];
	pendingMessages = [[NSMutableArray arrayWithArray:(NSArray *)[pickledMachine objectForKey:@"pendingMessages"]] retain];  // restore a MUTABLE array
	//NSLog(@"debug restorePreviousStateMachine");
	[aiLock unlock];
	
	[ai_stack removeObjectAtIndex:0];   //  POP
}

- (void) exitStateMachine
{
	if ([ai_stack count] > 0)
	{
		if ((owner)&&([owner reportAImessages]))   NSLog(@"Popping previous state machine for %@",self);
		[self restorePreviousStateMachine];
	}
}

- (void) setStateMachine:(NSString *) smName
{
    //NSString *filepath;
	
	//
	//filepath = [[[[[NSBundle mainBundle] bundlePath] stringByAppendingPathComponent:@"Contents"] stringByAppendingPathComponent:@"Resources"] stringByAppendingPathComponent:smName];
    //
    //NSLog(@"Loading AI data from %@",filepath);
    //
	[aiLock lock];
	//
    if (stateMachine)
	{
		[self preserveCurrentStateMachine];
		[stateMachine release];
    }
	stateMachine = [[ResourceManager dictionaryFromFilesNamed:smName inFolder:@"AIs" andMerge:NO] retain];
	//stateMachine = [[NSDictionary alloc] initWithContentsOfFile:filepath]; // alloc retains
	//
	[aiLock unlock];
	//
	[self setState:@"GLOBAL"];
    //
    //NSLog(@"AI Loaded:\n%@",[stateMachine description]);
    //
	
	//  refresh name
	//
	if (owner_desc)			[owner_desc release];
	owner_desc = [[NSString stringWithFormat:@"%@ %d", [owner name], [owner universal_id]] retain];
}

- (void) setState:(NSString *) stateName
{
	if ([stateMachine objectForKey:stateName])
	{
		//if ((owner)&&([owner universal_id])&&([owner reportAImessages])) NSLog(@"AI for %@ enters state %@", owner_desc, stateName);
		//
		[self reactToMessage:@"EXIT"];
		if (currentState)		[currentState release];
		currentState = [stateName retain];
		[self reactToMessage:@"ENTER"];
	}
}

- (void) reactToMessage:(NSString *) message
{
	int i;
	NSArray* actions;
	NSDictionary* messagesForState;
	
	if ([owner universal_id] == NO_TARGET)  // don't think until launched
		return;
	//

	[aiLock lock];
	//
	messagesForState = [NSDictionary dictionaryWithDictionary:[stateMachine objectForKey:currentState]];
	//
	if ((currentState)&&(![message isEqual:@"UPDATE"])&&((owner)&&([owner reportAImessages])))
		NSLog(@"AI for %@ in state '%@' receives message '%@'", owner_desc, currentState, message);
	//
	actions = [NSArray arrayWithArray:[messagesForState objectForKey:message]];
	//
	[aiLock unlock];

	if ((actions)&&([actions count] > 0))
	{
		//
		for (i = 0; i < [actions count]; i++)
			[self takeAction:[actions objectAtIndex:i]];
		//
	}
	else
	{
		if (currentState)
		{
			//if ((owner)&&([owner reportAImessages]))   NSLog(@"AI for %@ has no response to '%@' in state '%@'", owner_desc, message, currentState);
			if ([owner respondsToSelector:NSSelectorFromString(@"interpretAIMessage:")])
				[owner performSelector:NSSelectorFromString(@"interpretAIMessage:") withObject:message];
		}
		return;
	}
}

- (void) takeAction:(NSString *) action
{
	NSArray*	tokens = [action componentsSeparatedByString:@" "];
	NSString*	dataString = nil;
	NSString*   my_selector;
	SEL			_selector;
	
	if ((owner)&&([owner reportAImessages]))   NSLog(@"%@ to take action %@", owner_desc, action);
	
	if ([tokens count] < 1)
	{
		if ([owner reportAImessages])   NSLog(@"No action '%@'",action);
		return;
	}
	
	my_selector = (NSString *)[tokens objectAtIndex:0];
	
	if ([tokens count] > 1)
	{
		dataString = (NSString *)[tokens objectAtIndex:1];
	}
	
	_selector = NSSelectorFromString(my_selector);
	
	if (![owner respondsToSelector:_selector])
	{
		if ([my_selector isEqual:@"setStateTo:"])
			[self setState:dataString];
		else
			NSLog(@"***** %@ does not respond to %@",owner_desc, my_selector);
	}
	else
	{
		if (dataString)
			[owner performSelector:_selector withObject:dataString];
		else
			[owner performSelector:_selector];
	}
}

- (void) think
{
	
	NSArray *ms_list = nil;
	
	if ([owner universal_id] == NO_TARGET)  // don't think until launched
		return;
	//
	
	[self reactToMessage:@"UPDATE"];

	[aiLock lock];
	if (pendingMessages)
	{
		//NSLog(@"debug1");
		ms_list = [NSArray arrayWithArray:pendingMessages];
		//NSLog(@"debug2");
		[pendingMessages removeAllObjects];
	}
	[aiLock unlock];
	
	if (ms_list)
	{
		int i;
		for (i = 0; i < [ms_list count]; i++)
			[self reactToMessage:(NSString *)[ms_list objectAtIndex:i]];
	}
}

- (void) message:(NSString *) ms
{
	if ([owner universal_id] == NO_TARGET)  // don't think until launched
		return;
	//

	[pendingMessages addObject:ms];
	//[self think];
}

- (void) setNextThinkTime:(double) ntt
{
	nextThinkTime = ntt;
}

- (double) nextThinkTime
{
	return nextThinkTime;
}

- (void) setThinkTimeInterval:(double) tti
{
	thinkTimeInterval = tti;
}

- (double) thinkTimeInterval
{
	return thinkTimeInterval;
}

@end
