//
//  Entity.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#define MAX_VERTICES_PER_ENTITY	256
#define MAX_FACES_PER_ENTITY	256
#define MAX_VERTICES_PER_FACE	16

#define PI	3.141592653589793

#define NO_DRAW_DISTANCE_FACTOR 640.0

#define STATUS_EXPERIMENTAL			99
#define STATUS_EFFECT				10
#define STATUS_DEMO					2
#define STATUS_TEST					1
#define STATUS_INACTIVE				0
#define STATUS_DEAD					-1
#define STATUS_IN_FLIGHT			100
#define STATUS_DOCKED				200
#define STATUS_AUTOPILOT_ENGAGED	300
#define STATUS_DOCKING				401
#define STATUS_LAUNCHING			402
#define STATUS_WITCHSPACE_COUNTDOWN 410
#define STATUS_ENTERING_WITCHSPACE  411
#define STATUS_EXITING_WITCHSPACE   412
#define STATUS_ESCAPE_SEQUENCE		500
#define STATUS_IN_HOLD				600

#define CLASS_NO_DRAW	0
#define CLASS_NEUTRAL	1
#define CLASS_STATION	3
#define CLASS_TARGET	4
#define CLASS_CARGO		5
#define CLASS_MISSILE	6
#define CLASS_THARGOID	9
#define CLASS_POLICE	999
#define CLASS_BUOY		666

#define NO_TARGET		0

#define SCANNER_MAX_RANGE   25600.0
#define SCANNER_MAX_RANGE2  655360000.0


#define MODEL_FILE @"CORIOLIS.DAT"

#import <Cocoa/Cocoa.h>
#include "vector.h"
#include "legacy_random.h"

@class Universe;

struct face
{
	GLfloat red;
	GLfloat green;
	GLfloat blue;
        
        Vector	normal;
        
        int	n_verts;
        
        GLint	vertex[MAX_VERTICES_PER_FACE];
        
        NSString	*textureFile;
        GLuint	texName;
        GLfloat	s[MAX_VERTICES_PER_FACE];
        GLfloat	t[MAX_VERTICES_PER_FACE];
		
};

typedef struct face Face;

@interface Entity : NSObject
{
    // the base object for ships/stations/anything actually
    //
    Universe	*universe;
	int			universal_id;				// used to reference the entity
    //
    Vector	position;
	double  zero_distance;
	double  no_draw_distance;  //  10 km
	//
    Matrix	rotation;
    Quaternion	q_rotation;
    gl_matrix	rotMatrix;
    //
	Vector		velocity;
	//
	int		scan_class;
    int		status;
	//
	double  energy;
	double  max_energy;
	//
	double  collision_radius;
	//
	NSMutableArray  *collidingEntities;
    //
    int			n_vertices, n_faces;
    Vector		vertices[MAX_VERTICES_PER_ENTITY];
	//
	//new...
    Vector		vertex_normal[MAX_VERTICES_PER_ENTITY];
	BOOL		is_smooth_shaded;
	//
    Face		faces[MAX_FACES_PER_ENTITY];
	BoundingBox boundingBox;
    GLuint		displayListName;
    //
    NSString	*basefile;
	//
	int			owner;
	//
}

+ (void) setDataStore:(Universe *)univ; // class methods, they set the underlying data_storage universe
+ (Universe *) dataStore;

- (id) init;
- (void) dealloc;

- (void) warnAboutHostiles;

- (void) setUniverse:(Universe *)univ;

- (void) setUniversal_id:(int)uid;
- (int) universal_id;

- (BOOL) isSmoothShaded;
- (void) setSmoothShaded:(BOOL) value;

- (void) setOwner:(Entity *) ent;
- (Entity *) owner;

- (void) setModel:(NSString *) modelName;
- (NSString *) getModel;
- (void) setPosition:(Vector) posn;
- (void) setPosition:(GLfloat) x:(GLfloat) y:(GLfloat) z;
- (Vector) getPosition;
- (double) getZeroDistance;
- (NSComparisonResult) compareZeroDistance:(Entity *)otherEntity;

- (BoundingBox) getBoundingBox;

- (void) setQRotation:(Quaternion) quat;
- (Quaternion) QRotation;

- (void) setVelocity:(Vector) vel;
- (Vector) getVelocity;
- (double) getVelocityAsSpeed;

- (void) setStatus:(int) stat;
- (int) getStatus;

- (void) setScanClass:(int) s_class;
- (int) scanClass;

- (void) setEnergy:(double) amount;
- (double) getEnergy;

- (void) applyRoll:(GLfloat) roll andClimb:(GLfloat) climb;
- (void) applyRoll:(GLfloat) roll climb:(GLfloat) climb andYaw:(GLfloat) yaw;
- (void) moveForward:(double) amount;

- (GLfloat *) rotationMatrix;

- (BOOL) canCollide;
- (double) collisionRadius;
- (void) setCollisionRadius:(double) amount;
- (NSMutableArray *) collisionArray;

- (void) drawEntity:(BOOL) immediate :(BOOL) translucent;
- (void) initialiseTextures;
- (void) regenerateDisplayList;
- (void) generateDisplayList;

- (void) update:(double) delta_t;

- (void) loadData:(NSString *) filename;
- (void) checkNormalsAndAdjustWinding;
- (void) calculateVertexNormals;

- (double) findCollisionRadius;

- (BoundingBox) findBoundingBoxRelativeTo:(Entity *)other InVectors:(Vector) _i :(Vector) _j :(Vector) _k;

- (BOOL) checkCloseCollisionWith:(Entity *)other;

- (void) takeEnergyDamage:(double) amount from:(Entity *) ent becauseOf:(Entity *) other;

- (void) fakeTexturesWithImageFile: (NSString *) textureFile andMaxSize:(NSSize) maxSize;

- (NSString *) toString;
/*--
- (NSDictionary *) toDictionary;
--*/
@end
