//
//  GuiDisplayGen.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "GuiDisplayGen.h"
#import "OpenGLSprite.h"
#import "ResourceManager.h"

@implementation GuiDisplayGen

- (void) click
{
	if ([guiclick isPlaying])
		[guiclick stop];
	[guiclick play];
}

- (void) setColor:(NSColor *) color forRow:(int) row
{
	if ((row >= rowRange.location)&&(row < rowRange.location + rowRange.length))
		[rowColor replaceObjectAtIndex:row withObject:color];
}

- (id) objectForRow:(int) row
{
	if ((row >= rowRange.location)&&(row < rowRange.location + rowRange.length))
		return [rowText objectAtIndex:row];
	else
		return NULL;
}

- (int) selectedRow
{
	return selectedRow;
}

- (BOOL) setSelectedRow:(int) row
{
	if ((row >= selectableRange.location) && (row < selectableRange.location+selectableRange.length))
	{
		selectedRow = row;
		//NSLog(@"New selectedRow = %d",selectedRow);
		[self generateDisplay];
		return YES;
	}
	else
		return NO;
}

- (void) setNoSelectedRow
{
	selectedRow = -1;
	[self generateDisplay];
}

- (NSString *) selectedRowText
{
	if ([[rowText objectAtIndex:selectedRow] isKindOfClass:[NSString class]])
		return (NSString *)[rowText objectAtIndex:selectedRow];
	if ([[rowText objectAtIndex:selectedRow] isKindOfClass:[NSArray class]])
		return (NSString *)[[rowText objectAtIndex:selectedRow] objectAtIndex:0];
	return NULL;
}

- (NSString *) selectedRowKey
{
	return (NSString *)[rowKey objectAtIndex:selectedRow];
}


- (NSRange) selectableRange
{
	return selectableRange;
}

- (void) setSelectableRange:(NSRange) range
{
	selectableRange = range;
}

- (void) setTabStops:(int *)stops
{
	int i = 0;
	for (i = 0; i < GUI_MAX_COLUMNS; i++)
		tabStops[i] = stops[i];
}


- (void) clear
{
	int i;
	[self setTitle:@""];
	for (i = 0; i < GUI_MAX_ROWS; i++)
	{
		[self setText:@"" forRow:i align:GUI_ALIGN_LEFT];
		[self setColor:[NSColor yellowColor] forRow:i];
	}
	[self setSelectableRange:NSMakeRange(0,0)];
}

- (void) setTitle: (NSString *) str
{
	if (title)  [title release];
	title = [str retain];
}

- (void) setKey: (NSString *) str forRow:(int) row
{
	if ((row >= rowRange.location)&&(row < rowRange.location + rowRange.length))
		[rowKey replaceObjectAtIndex:row withObject:str];
}

- (void) setText: (NSString *) str forRow:(int) row
{
	if ((row >= rowRange.location)&&(row < rowRange.location+rowRange.length))
	{
		[rowText replaceObjectAtIndex:row withObject:str];
	}
}

- (void) setText: (NSString *) str forRow:(int) row align:(int) alignment
{
	if ((row >= rowRange.location)&&(row < rowRange.location+rowRange.length))
	{
		[rowText replaceObjectAtIndex:row withObject:str];
		rowAlignment[row] = alignment;
	}
}

- (int) addLongText: (NSString *) str startingAtRow:(int) row align:(int) alignment
{
    NSMutableDictionary *textAttributes = [NSMutableDictionary dictionaryWithObjectsAndKeys:
        [NSFont boldSystemFontOfSize:0.75*GUI_PIXEL_ROW_HEIGHT], NSFontAttributeName,NULL];
		
	NSSize strsize = [str sizeWithAttributes:textAttributes];
	if (strsize.width < GUI_PIXEL_WIDTH)
	{
		[self setText:str forRow:row align:alignment];
		return row + 1;
	}
	else
	{
		NSMutableArray*	words = [NSMutableArray arrayWithArray:[str componentsSeparatedByString:@" "]];
		NSMutableString* string1 = [NSMutableString stringWithString:@""];
		NSMutableString* string2 = [NSMutableString stringWithString:@""];
		strsize.width = 0.0;
		while ((strsize.width < GUI_PIXEL_WIDTH)&&([words count] > 0))
		{
			[string1 appendString:(NSString *)[words objectAtIndex:0]];
			[string1 appendString:@" "];
			[words removeObjectAtIndex:0];
			strsize = [string1 sizeWithAttributes:textAttributes];
			if ([words count] > 0)
				strsize.width += [(NSString *)[words objectAtIndex:0] sizeWithAttributes:textAttributes].width;
		}
		[string2 appendString:[words componentsJoinedByString:@" "]];
		[self setText:string1		forRow:row			align:alignment];
		return  [self addLongText:string2   startingAtRow:row+1	align:alignment];
	}
}


- (void) setArray: (NSArray *) arr forRow:(int) row
{
	if ((row >= rowRange.location)&&(row < rowRange.length))
		[rowText replaceObjectAtIndex:row withObject:arr];
}

/////////////////////////////////////////////////////

- (id) init
{
	int stops[6] = {0, 192, 256, 320, 384, 448};
	int i;
	
	self = [super init];
	
	image = [[NSImage alloc] initWithSize:NSMakeSize(GUI_PIXEL_WIDTH, GUI_PIXEL_HEIGHT)];   // retained

	rowRange = NSMakeRange(0,GUI_MAX_ROWS);

	rowText =   [[NSMutableArray alloc] initWithCapacity:GUI_MAX_ROWS];   // alloc retains
	rowKey =	[[NSMutableArray alloc] initWithCapacity:GUI_MAX_ROWS];   // alloc retains
	rowColor =	[[NSMutableArray alloc] initWithCapacity:GUI_MAX_ROWS];   // alloc retains
	
	for (i = 0; i < GUI_MAX_ROWS; i++)
	{
		[rowText addObject:@"."];
		[rowKey addObject:[NSString stringWithFormat:@"%d",i]];
		[rowColor addObject:[NSColor yellowColor]];
		rowPosition[i].x = 0.0;
		rowPosition[i].y = GUI_PIXEL_HEIGHT - (GUI_PIXEL_ROW_START + i * GUI_PIXEL_ROW_HEIGHT);
		rowAlignment[i] = GUI_ALIGN_LEFT;
	}
	
	for (i = 0; i < GUI_MAX_COLUMNS; i++)
	{
		tabStops[i] = stops[i];
	}
	
	selectedRow = 0;
	selectableRange = NSMakeRange(0,0);
	
	title = @"Test Page";
	
	openGLSprite =  [[OpenGLSprite alloc]   initWithImage:[self generateDisplay]
							cropRectangle:NSMakeRect(0.0,0.0,GUI_PIXEL_WIDTH,GUI_PIXEL_HEIGHT)
							size:NSMakeSize(GUI_PIXEL_WIDTH,GUI_PIXEL_HEIGHT)];		// retained
	
	guiclick =  [[ResourceManager soundNamed:@"guiclick.aiff" inFolder:@"Sounds"] retain];
	return self;
}

- (void) dealloc
{
	if (image)			[image release];
	if (openGLSprite)   [openGLSprite release];
	if (title)			[title release];
	if (rowText)		[rowText release];
	if (rowKey)			[rowKey release];
	if (rowColor)		[rowColor release];
	[super dealloc];
}

- (NSImage *) generateDisplay
{
	NSSize  strsize;
	NSPoint drawpos;
	int i;
	
    NSMutableDictionary *textAttributes = [NSMutableDictionary dictionaryWithObjectsAndKeys:
        [NSFont boldSystemFontOfSize:0.75*GUI_PIXEL_ROW_HEIGHT], NSFontAttributeName,
        [NSColor yellowColor], NSForegroundColorAttributeName, NULL];

    NSMutableDictionary *selectedTextAttributes = [NSMutableDictionary dictionaryWithObjectsAndKeys:
        [NSFont boldSystemFontOfSize:0.75*GUI_PIXEL_ROW_HEIGHT], NSFontAttributeName,
        [NSColor blackColor], NSForegroundColorAttributeName,
        [NSColor redColor], NSBackgroundColorAttributeName, NULL];

    NSMutableDictionary *titleAttributes = [NSMutableDictionary dictionaryWithObjectsAndKeys:
        [NSFont systemFontOfSize:0.75*GUI_PIXEL_TITLE_HEIGHT], NSFontAttributeName,
        [NSColor redColor], NSForegroundColorAttributeName, NULL];

	
    [image lockFocus];
    [[NSColor clearColor] set];
    NSRectFill(NSMakeRect(0,0,GUI_PIXEL_WIDTH,GUI_PIXEL_HEIGHT));
	
	if ((selectedRow < selectableRange.location)||(selectedRow >= selectableRange.location + selectableRange.length))
		selectedRow = -1;   // out of Range;
	
    ////
	// drawing operations here
	
	//NSLog(@"rowText :\n%@selectedRow : %d",[rowText description],selectedRow);
	
	// draw the title
	//
	//NSLog(@"title :%@",title);
	
    strsize = [title sizeWithAttributes:titleAttributes];
	drawpos = NSMakePoint(GUI_PIXEL_ROW_CENTER - strsize.width/2.0,GUI_PIXEL_HEIGHT - GUI_PIXEL_TITLE_HEIGHT);
	[title drawAtPoint:drawpos withAttributes:titleAttributes];
		
	// draw a horizontal divider
	//
    [[NSColor lightGrayColor] set];
    NSRectFill(NSMakeRect(0,GUI_PIXEL_HEIGHT - GUI_PIXEL_TITLE_HEIGHT - 2,GUI_PIXEL_WIDTH,2));
	
	// draw each row of text
	//
	for (i = 0; i < GUI_MAX_ROWS; i++)
	{
		[textAttributes setObject:(NSColor *)[rowColor objectAtIndex:i] forKey:NSForegroundColorAttributeName];
		NSDictionary* attributes = (NSDictionary *)textAttributes;
		if (i == selectedRow)
			attributes = (NSDictionary *)selectedTextAttributes;
		if ([[rowText objectAtIndex:i] isKindOfClass:[NSString class]])
		{
			NSString*   text = (NSString *)[rowText objectAtIndex:i];
			strsize = [text sizeWithAttributes:attributes];
			switch (rowAlignment[i])
			{
				case GUI_ALIGN_LEFT :
					rowPosition[i].x = 0.0;
					break;
				case GUI_ALIGN_RIGHT :
					rowPosition[i].x = GUI_PIXEL_WIDTH - strsize.width;
					break;
				case GUI_ALIGN_CENTER :
					rowPosition[i].x = (GUI_PIXEL_WIDTH - strsize.width)/2.0;
					break;
			}
			[text drawAtPoint:rowPosition[i] withAttributes:attributes];
		}
		if ([[rowText objectAtIndex:i] isKindOfClass:[NSArray class]])
		{
			int j;
			NSArray*	array = (NSArray *)[rowText objectAtIndex:i];
			if (i == selectedRow)
				attributes = (NSDictionary *)selectedTextAttributes;
			for (j = 0; ((j < [array count])&&(j < GUI_MAX_COLUMNS)) ; j++)
			{
				if ([array objectAtIndex:j])
				{
					NSString*   text = (NSString *)[array objectAtIndex:j];
					rowPosition[i].x = tabStops[j];
					[text drawAtPoint:rowPosition[i] withAttributes:attributes];
				}
			}
		}
	}
		
	//
	////
    [image unlockFocus];

    return image;
}

- (void) updateGui
{
	[openGLSprite makeTextureFromImage:image
									cropRectangle:NSMakeRect(0.0,0.0,GUI_PIXEL_WIDTH,GUI_PIXEL_HEIGHT)
									size:NSMakeSize(GUI_PIXEL_WIDTH,GUI_PIXEL_HEIGHT)];
}

- (OpenGLSprite *) openGLSprite
{
	return openGLSprite;
}

@end
