/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/
#import "MyOpenGLView.h"
#import <OpenGL/gl.h>
#import <OpenGL/glu.h>
//#import <GLUT/glut.h>

#import "GameController.h"
#import "AppDelegate.h"
#import "Universe.h"
#import "TextureStore.h"
#import "Entity.h"
#import "PlanetEntity.h"

@interface MyOpenGLView(Internal)

- (int) translateKeyCode: (int) input;

@end


@implementation MyOpenGLView

- (id) initWithFrame:(NSRect)frameRect
{
    
	//NSLog(@"-- initWithFrame MyOpenGLView");
	
	// Pixel Format Attributes for the View-based (non-FullScreen) NSOpenGLContext
    NSOpenGLPixelFormatAttribute attrs[] = {

        // Specifying "NoRecovery" gives us a context that cannot fall back to the software renderer.
		//This makes the View-based context a compatible with the fullscreen context, enabling us to use the "shareContext"
		// feature to share textures, display lists, and other OpenGL objects between the two.
        NSOpenGLPFANoRecovery,

        // Attributes Common to FullScreen and non-FullScreen
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        NSOpenGLPFAAccelerated,
        0
    };
    long rendererID;

    // Create our non-FullScreen pixel format.
    NSOpenGLPixelFormat* pixelFormat = [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];

    // Just as a diagnostic, report the renderer ID that this pixel format binds to.  CGLRenderers.h contains a list of known renderers
	// and their corresponding RendererID codes.
    [pixelFormat getValues:&rendererID forAttribute:NSOpenGLPFARendererID forVirtualScreen:0];
    //NSLog(@" init! NSOpenGLView pixelFormat RendererID = %08x", (unsigned)rendererID);

    self = [super initWithFrame:frameRect pixelFormat:pixelFormat];

    glClearColor(0.0, 0.0, 0.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT);
    [[self openGLContext] flushBuffer];
    glClear(GL_COLOR_BUFFER_BIT);
    [[self openGLContext] flushBuffer];
	
	virtualJoystickPosition = NSMakePoint(0.0,0.0);
	
	typedString = [[NSMutableString alloc] initWithString:@""];
	allowingStringInput = NO;
	isAlphabetKeyDown = NO;
	
    return self;
}

- (void) dealloc
{
	if (typedString)
		[typedString release];
	[super dealloc];
}

- (void) allowStringInput: (BOOL) value
{
	allowingStringInput = value;
}

-(BOOL) allowingStringInput
{
	return allowingStringInput;
}

- (NSString *) typedString
{
	return typedString;
}

- (void) resetTypedString
{
	[typedString setString:@""];
}

- (NSSize) viewSize
{
	return viewSize;
}

- (GameController *) gameController
{
	return gameController;
}

- (void) setGameController:(GameController *) controller
{
	gameController = controller;
}

- (void)drawRect:(NSRect)rect
{    
	AppDelegate	*deleg = (AppDelegate *)[NSApp delegate];
	
	if ((viewSize.width != [self frame].size.width)||(viewSize.height != [self frame].size.height)) // resized
	{
		m_glContextInitialized = NO;
		viewSize = [self frame].size;
	}
	
    if (!m_glContextInitialized)
		[self initialiseGLWithSize:viewSize];
    
	// do all the drawing!
	//
	[[deleg universe] drawFromEntity:0];
    //
	////
	
	[[self openGLContext] flushBuffer];
}

- (void) initialiseGLWithSize:(NSSize) v_size
{
	AppDelegate	*deleg = (AppDelegate *)[NSApp delegate];
	GLfloat	sun_ambient[] =	{0.1, 0.1, 0.1, 1.0};
	GLfloat	sun_diffuse[] =	{1.0, 1.0, 1.0, 1.0};
	GLfloat	sun_specular[] = 	{1.0, 1.0, 1.0, 1.0};
	GLfloat	sun_center_position[] = {4000000.0, 0.0, 0.0, 1.0};
	
	viewSize = v_size;
	
	float	ratio = 0.5;
	float   aspect = viewSize.height/viewSize.width;
	
	glShadeModel(GL_FLAT);
	glClearColor(0.0, 0.0, 0.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT);
    [[self openGLContext] flushBuffer];
    glClear(GL_COLOR_BUFFER_BIT);
    [[self openGLContext] flushBuffer];
	
	glClearDepth(MAX_CLEAR_DEPTH);
	glViewport( 0, 0, viewSize.width, viewSize.height);
	
	squareX = 0.0;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();	// reset matrix
	glFrustum(-ratio, ratio, -aspect*ratio, aspect*ratio, 1.0, MAX_CLEAR_DEPTH);	// set projection matrix

	glMatrixMode(GL_MODELVIEW);
	
	glEnable(GL_DEPTH_TEST);		// depth buffer
	glDepthFunc(GL_LESS);			// depth buffer
	
	glFrontFace(GL_CCW);			// face culling - front faces are AntiClockwise!
	glCullFace(GL_BACK);			// face culling
	glEnable(GL_CULL_FACE);			// face culling
	
	glEnable(GL_BLEND);									// alpha blending
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	// alpha blending
	
	if ([deleg universe])
	{
		Vector sun_pos = [[[deleg universe] sun] getPosition];
		sun_center_position[0] = sun_pos.x;
		sun_center_position[1] = sun_pos.y;
		sun_center_position[2] = sun_pos.z;
	}
	
	glLightfv(GL_LIGHT1, GL_AMBIENT, sun_ambient);
	glLightfv(GL_LIGHT1, GL_SPECULAR, sun_specular);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, sun_diffuse);
	glLightfv(GL_LIGHT1, GL_POSITION, sun_center_position);

	glEnable(GL_LIGHTING);		// lighting
	glEnable(GL_LIGHT1);		// lighting
	
	m_glContextInitialized = YES;
}

- (void) snapShot
{
    NSRect boundsRect = [self bounds];
    int w = viewSize.width;
    int h = viewSize.height;
    long nPixels = w*h + 1;	

	unsigned char   *red = (unsigned char *) malloc( nPixels);
	unsigned char   *green = (unsigned char *) malloc( nPixels);
	unsigned char   *blue = (unsigned char *) malloc( nPixels);
	
	NSString	*filepath = [[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent];
	int imageNo = 1;
	NSString	*pathToPic = [filepath stringByAppendingPathComponent:[NSString stringWithFormat:@"oolite-%03d.tiff",imageNo]];
	while ([[NSFileManager defaultManager] fileExistsAtPath:pathToPic])
	{
		imageNo++;
		pathToPic = [filepath stringByAppendingPathComponent:[NSString stringWithFormat:@"oolite-%03d.tiff",imageNo]];
	}
	NSLog(@">>>>> Snapshot file path chosen = %@",pathToPic);
	
    NSBitmapImageRep* bitmapRep = 
        [[NSBitmapImageRep alloc]
            initWithBitmapDataPlanes:	nil		// --> let the class allocate it
            pixelsWide:			w
            pixelsHigh:			h
            bitsPerSample:		8		// each component is 8 bits (1 byte)
            samplesPerPixel:		3		// number of components (R, G, B)
            hasAlpha:			NO		// no transparency
            isPlanar:			NO		// data integrated into single plane
            colorSpaceName:		NSDeviceRGBColorSpace
            bytesPerRow:		0		// --> let the class figure it out
            bitsPerPixel:		0		// --> let the class figure it out
        ];

    unsigned char *pixels = [bitmapRep bitmapData];
		
	glReadPixels(0,0, w,h, GL_RED,   GL_UNSIGNED_BYTE, red);
	glReadPixels(0,0, w,h, GL_GREEN, GL_UNSIGNED_BYTE, green);
	glReadPixels(0,0, w,h, GL_BLUE,  GL_UNSIGNED_BYTE, blue);

	int x,y;
	for (y = 0; y < h; y++)
	{
		for (x = 0; x < w; x++)		// set bitmap pixels
		{
			long index = (h - y - 1)*w + x;
			*pixels++ = red[index];
			*pixels++ = green[index];
			*pixels++ = blue[index];
		}
	}
	NSImage *image = [[NSImage alloc] initWithSize:boundsRect.size];
	[image addRepresentation:bitmapRep];
														// save TIFF representation of image
	[[image TIFFRepresentation] writeToFile:pathToPic atomically:YES]; 
		
	[image release];         
	[bitmapRep release];         
	
	// free allocated memory
	free(red);
	free(green);
	free(blue);
	
}

- (BOOL) acceptsFirstResponder
{
	return YES;
} 

- (void) keyUp:(NSEvent *)theEvent
{
	int key = [[theEvent charactersIgnoringModifiers] characterAtIndex:0];
	key = [self translateKeyCode: key];
	isAlphabetKeyDown = NO;
	if ( key >= 0 && key < [self numKeys] )
	{
		keys[key] = NO;
	}
	else
	{
		NSLog(@"***** translated key: %d out of range\n", key);
	}
}

- (void) keyDown:(NSEvent *)theEvent
{
	int key = [[theEvent charactersIgnoringModifiers] characterAtIndex:0];
	key = [self translateKeyCode: key];
	if ((key >= 0)&&(key < [self numKeys]))
	{
		//NSLog( @"key : %d", key );
		keys[key] = YES;
		
		if (allowingStringInput)
		{
			if (((key > 64)&&(key < 91))||((key > 96)&&(key < 123)))
			{
				// alphanumeric
				isAlphabetKeyDown = YES;
				// convert to lowercase
				[typedString appendString:[[[theEvent charactersIgnoringModifiers] lowercaseString] substringToIndex:1]];
				//NSLog(@"accumulated string '%@'",typedString);
			}
			else
				isAlphabetKeyDown = NO;
			if (key == NSDeleteCharacter)
			{
				//delete
				[typedString setString:@""];
			}
		}
	}
	else
	{
		NSLog(@"***** translated key: %d out of range\n", key);
	}
} 

/*     Capture shift, ctrl, opt and command press & release */
- (void)flagsChanged:(NSEvent *)theEvent
{
	int flags = [theEvent modifierFlags];
	opt = (flags & NSAlternateKeyMask) ? YES : NO;
	ctrl = (flags & NSControlKeyMask) ? YES : NO;
	command = (flags & NSCommandKeyMask) ? YES : NO;
	shift = ( flags & NSShiftKeyMask ) ? YES : NO;

}

- (void)mouseDown:(NSEvent *)theEvent
{
    keys[gvMouseLeftButton] = YES; // 'a' down
}

- (void)mouseUp:(NSEvent *)theEvent
{
	keys[gvMouseLeftButton] = NO;  // 'a' up
}

- (void)mouseDragged:(NSEvent *)theEvent
{
    squareX = [theEvent locationInWindow].x - mouseDragStartPoint.x;
    squareY = [theEvent locationInWindow].y - mouseDragStartPoint.y;
    [self setNeedsDisplay:YES];
}

/////////////////////////////////////////////////////////////
/*     Turn the Cocoa ArrowKeys into our arrow key constants. */
- (int) translateKeyCode: (int) input
{
	int key = input;
	switch ( input )
	{
		case NSUpArrowFunctionKey:
			key = gvArrowKeyUp;
			break;
		
		case NSDownArrowFunctionKey:
			key = gvArrowKeyDown;
			break;
			
		case NSLeftArrowFunctionKey:
			key = gvArrowKeyLeft;
			break;
			
		case NSRightArrowFunctionKey:
			key = gvArrowKeyRight;
			break;
		
		case NSF1FunctionKey:
			key = gvFunctionKey1;
			break;
			
		case NSF2FunctionKey:
			key = gvFunctionKey2;
			break;
			
		case NSF3FunctionKey:
			key = gvFunctionKey3;
			break;
			
		case NSF4FunctionKey:
			key = gvFunctionKey4;
			break;
			
		case NSF5FunctionKey:
			key = gvFunctionKey5;
			break;
			
		case NSF6FunctionKey:
			key = gvFunctionKey6;
			break;
			
		case NSF7FunctionKey:
			key = gvFunctionKey7;
			break;
			
		case NSF8FunctionKey:
			key = gvFunctionKey8;
			break;
			
		case NSF9FunctionKey:
			key = gvFunctionKey9;
			break;
			
		case NSF10FunctionKey:
			key = gvFunctionKey10;
			break;
			
		case NSF11FunctionKey:
			key = gvFunctionKey11;
			break;
			
		default:
			break;
	}
	return key;
}

- (void) setVirtualJoystick:(double) vmx :(double) vmy
{
	virtualJoystickPosition.x = vmx;
	virtualJoystickPosition.y = vmy;
}

- (NSPoint) virtualJoystickPosition
{
	return virtualJoystickPosition;
}

/////////////////////////////////////////////////////////////

- (void) clearKeys
{
	int i;
	for (i = 0; i < [self numKeys]; i++)
		keys[i] = NO;
}

- (BOOL) isAlphabetKeyDown
{
	return isAlphabetKeyDown = NO;;
}

- (BOOL) isDown: (int) key
{
	if ( key < 0 )
		return NO;
	if ( key >= [self numKeys] )
		return NO;
	return keys[key];
}

- (BOOL) isOptDown
{
	return opt;
}

- (BOOL) isCtrlDown
{
	return ctrl;
}

- (BOOL) isCommandDown
{
	return command;
}

- (BOOL) isShiftDown
{
	return shift;
}

- (int) numKeys
{
	return NUM_KEYS;
} 

@end
