//
//  ParticleEntity.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <Foundation/Foundation.h>
#import "Entity.h"

#import "Universe.h"
#import "vector.h"

#define PARTICLE_TEST				1
#define PARTICLE_SHOT_EXPIRED		200
#define PARTICLE_EXPLOSION			201
#define PARTICLE_SHOT_GREEN_PLASMA	100
#define PARTICLE_SHOT_YELLOW_PLASMA	101
#define PARTICLE_SHOT_PLASMA		110
#define PARTICLE_LASER_BEAM_RED		150
#define PARTICLE_LASER_BEAM			160
#define PARTICLE_EXHAUST			300
#define PARTICLE_ECM_MINE			400
#define PARTICLE_HYPERRING			800

#define PARTICLE_LASER_DURATION		0.20
#define PARTICLE_LASER_LENGTH		10000.0
#define PARTICLE_LASER_RANGE_LIMIT	1000000000.0

@class Entity;

@interface ParticleEntity : Entity {

    NSString*	textureNameString;
    GLuint		texName;
	
	NSColor*	color;
	GLfloat		color_fv[4];
	
	double		alpha;
	double		time_counter;
	
	double		duration;
	double		activation_time;
		
	double		ring_inner_radius, ring_outer_radius;
		
	int			particle_type;
	
	double		alpha_for_vertex[MAX_VERTICES_PER_ENTITY];
	
	Vector		exhaust_offset;
	
    NSSize	size;
}

- (id) initLaserFromShip:(ShipEntity *) ship view:(int) view;
- (id) initExhaustFromShip:(ShipEntity *) ship offsetVector:(Vector) offset scaleVector:(Vector) scale;
- (id) initExhaustFromShip:(ShipEntity *) ship details:(NSString *) details;
- (id) initECMMineFromShip:(ShipEntity *) ship;
- (id) initHyperringFromShip:(ShipEntity *) ship;

- (void) updateExplosion:(double) delta_t;
- (void) updateECMMine:(double) delta_t;
- (void) updateShot:(double) delta_t;
- (void) updateLaser:(double) delta_t;
- (void) updateHyperring:(double) delta_t;

- (void) setTexture:(NSString *) filename;
- (void) setColor:(NSColor *) a_color;

- (void) setParticleType:(int) p_type;
- (int) particleType;

- (void) setDuration:(double) dur;
- (void) setSize:(NSSize) siz;
- (NSSize) size;

- (void) initialiseTexture: (NSString *) name;

- (void) drawParticle;
- (void) drawLaser;
- (void) drawExhaust:(BOOL) immediate;
- (void) drawHyperring;

@end
