//
//  ParticleEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "SkyEntity.h"
#import "entities.h"

#import "Universe.h"
#import "AI.h"
#import "TextureStore.h"

@implementation ParticleEntity

static Vector   circleVertex[65];		// holds vector coordinates for a unit circle

- (id) init
{    
    self = [super init];
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
	//
	status = STATUS_EFFECT;
	time_counter = 0.0;
	//
	particle_type = PARTICLE_TEST;
	//
	basefile = @"Particle";
	textureNameString   = @"blur256.png";
	[self setColor:[NSColor greenColor]];
	//
	texName = 0;
	[self initialiseTexture: textureNameString];
	size = NSMakeSize(32.0,32.0);
	//
	owner = NO_TARGET;
	//
	collision_radius = 32.0;
	//
	int i;
	for (i = 0; i < 65; i++)
	{
		circleVertex[i].x = sin(i * PI / 32.0);
		circleVertex[i].y = cos(i * PI / 32.0);
		circleVertex[i].z = 0.0;
	}
	//
    return self;
}

- (id) initLaserFromShip:(ShipEntity *) ship view:(int) view
{
    self = [super init];
    //
	status = STATUS_EFFECT;
    position = [ship getPosition];
	q_rotation = [ship QRotation];
	if ([ship isKindOfClass:[PlayerEntity class]])
		q_rotation.w = -q_rotation.w;   //reverse view direction for the player
	Vector v_up = vector_up_from_quaternion(q_rotation);
	Vector v_forward = vector_forward_from_quaternion(q_rotation);
	Vector v_right = vector_right_from_quaternion(q_rotation);
	double distance;
	switch (view)
	{
		case VIEW_FORWARD :
			distance = [ship getBoundingBox].max_z;
			position.x += distance * v_forward.x;	position.y += distance * v_forward.y;	position.z += distance * v_forward.z;
			break;
		case VIEW_AFT :
			quaternion_rotate_about_axis(&q_rotation, v_up, PI);
			distance = [ship getBoundingBox].min_z;
			position.x += distance * v_forward.x;	position.y += distance * v_forward.y;	position.z += distance * v_forward.z;
			break;
		case VIEW_PORT :
			quaternion_rotate_about_axis(&q_rotation, v_up, PI/2.0);
			distance = [ship getBoundingBox].min_x;
			position.x += distance * v_right.x;	position.y += distance * v_right.y;	position.z += distance * v_right.z;
			break;
		case VIEW_STARBOARD :
			quaternion_rotate_about_axis(&q_rotation, v_up, -PI/2.0);
			distance = [ship getBoundingBox].max_x;
			position.x += distance * v_right.x;	position.y += distance * v_right.y;	position.z += distance * v_right.z;
			break;
	}
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
    //
	if ([ship isKindOfClass:[PlayerEntity class]])
	{
		position.x += -20.0 * v_up.x;	position.y += -20.0 * v_up.y;	position.z += -20.0 * v_up.z;	// offset below the view line
	}
	//
	time_counter = 0.0;
	//
	particle_type = PARTICLE_LASER_BEAM_RED;
	//
	[self setColor:[NSColor redColor]];
	//
	duration = PARTICLE_LASER_DURATION;
	//
	[self setOwner:ship];
	//
	collision_radius = PARTICLE_LASER_LENGTH;
	//
    return self;
}

- (id) initExhaustFromShip:(ShipEntity *) ship offsetVector:(Vector) offset scaleVector:(Vector) scale
{
    int i;
	
	self = [super init];
    //
	status = STATUS_EFFECT;

	// set a smooth shaded model :
	is_smooth_shaded = YES;
	[self setModel:@"exhaust.dat"];
	
	// adjust vertices according to offset and scale
	//
	for (i = 0; i < n_vertices; i++)
	{
		alpha_for_vertex[i] = 1.0 + (vertices[i].z / 10.0);	// gives the alpha value at that point...
		vertices[i].x *= scale.x; //vertices[i].x += offset.x;
		vertices[i].y *= scale.y; //vertices[i].y += offset.y;
		vertices[i].z *= scale.z; //vertices[i].z += offset.z;
	}
	exhaust_offset.x = offset.x;
	exhaust_offset.y = offset.y;
	exhaust_offset.z = offset.z;
	particle_type = PARTICLE_EXHAUST;
	//
	[self setOwner:ship];
	//
    return self;
}

- (id) initExhaustFromShip:(ShipEntity *) ship details:(NSString *) details
{
    int i;
	NSArray *values = [details componentsSeparatedByString:@" "];
	if ([values count] != 6)
		return nil;
	Vector offset, scale;
	offset.x = [(NSString *)[values objectAtIndex:0] doubleValue];
	offset.y = [(NSString *)[values objectAtIndex:1] doubleValue];
	offset.z = [(NSString *)[values objectAtIndex:2] doubleValue];
	scale.x = [(NSString *)[values objectAtIndex:3] doubleValue];
	scale.y = [(NSString *)[values objectAtIndex:4] doubleValue];
	scale.z = [(NSString *)[values objectAtIndex:5] doubleValue];
	
	self = [super init];
    //
	status = STATUS_EFFECT;
	
	//NSLog(@"Adding an exhaust to a %@ at ( %3.1f, %3.1f, %3.1f)", [ship name], offset.x, offset.y, offset.z);
	
	// set a smooth shaded model :
	is_smooth_shaded = YES;
	[self setModel:@"exhaust.dat"];
	
	// adjust vertices according to offset and scale
	//
	for (i = 0; i < n_vertices; i++)
	{
		alpha_for_vertex[i] = 1.0 + (vertices[i].z / 10.0);	// gives the alpha value at that point...
		vertices[i].x *= scale.x;// vertices[i].x += offset.x;
		vertices[i].y *= scale.y;// vertices[i].y += offset.y;
		vertices[i].z *= scale.z;// vertices[i].z += offset.z;
	}
	exhaust_offset.x = offset.x;
	exhaust_offset.y = offset.y;
	exhaust_offset.z = offset.z;
	particle_type = PARTICLE_EXHAUST;
	//
	[self setOwner:ship];
	//	
    return self;
}

- (id) initECMMineFromShip:(ShipEntity *) ship
{
	self = [super init];
    //
    time_counter = 0.0;
	activation_time = 0.5;
	duration = 2.0;
	[self setPosition:[ship getPosition]];
	//
	status = STATUS_EFFECT;
	scan_class = CLASS_NO_DRAW;
	//
	//NSLog(@"```firing ECM at ( %3.1f, %3.1f, %3.1f)", position.x, position.y, position.z);
	//
	particle_type = PARTICLE_ECM_MINE;
	//
	[self setOwner:ship];
	//	
    return self;
}

- (id) initHyperringFromShip:(ShipEntity *) ship
{
	self = [super init];
    //
    time_counter = 0.0;
	duration = 2.0;
	size.width = [ship collisionRadius] * 0.5;
	size.height = size.width * 1.25;
	ring_inner_radius = size.width;
	ring_outer_radius = size.height;
	[self setPosition:[ship getPosition]];
	[self setQRotation:[ship QRotation]];
	[self setVelocity:[ship getVelocity]];
	//
	status = STATUS_EFFECT;
	scan_class = CLASS_NO_DRAW;
	//
	particle_type = PARTICLE_HYPERRING;
	int i;
	for (i = 0; i < 65; i++)
	{
		circleVertex[i].x = sin(i * PI / 32.0);
		circleVertex[i].y = cos(i * PI / 32.0);
		circleVertex[i].z = 0.0;
	}
	//
	[self setOwner:ship];
	//	
    return self;
}

- (void) dealloc
{
    if (textureNameString)	[textureNameString release];
    if (color)				[color release];
    [super dealloc];
}

- (BOOL) canCollide
{
	switch (particle_type)
	{
		case PARTICLE_TEST :
		case PARTICLE_LASER_BEAM_RED :
		case PARTICLE_LASER_BEAM :
		case PARTICLE_EXPLOSION :
		case PARTICLE_SHOT_EXPIRED :
		case PARTICLE_EXHAUST :
		case PARTICLE_ECM_MINE :
			return NO;
			break;
		default :
			return YES;
			break;
	}
}

- (void) setTexture:(NSString *) filename
{
    if (filename)
	{
		if (textureNameString)	[textureNameString release];
		textureNameString = filename;
		[textureNameString retain];
		[self initialiseTexture: textureNameString];
	}
}

- (void) setColor:(NSColor *) a_color
{
	if (!a_color)
		return;
	NSColor *rgbColor = [a_color colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
	
	if (color)				[color release];
	color = [rgbColor retain];

	color_fv[0] = [color redComponent];
	color_fv[1] = [color greenComponent];
	color_fv[2] = [color blueComponent];
}



- (void) setParticleType:(int) p_type
{
	particle_type = p_type;
}


- (int) particleType
{
	return particle_type;
}


- (void) setDuration:(double) dur
{
	duration = dur;
	time_counter = 0.0;
}

- (void) setSize:(NSSize) siz
{
	size = siz;
	collision_radius = sqrt (size.width * size.width + size.height * size.height);
}

- (NSSize) size
{
	return size;
}

- (void) initialiseTexture: (NSString *) name
{
    if (universe)
	{
         texName = [[universe textureStore] getTextureNameFor:name];
	}
}

- (void) update:(double) delta_t
{
	[super update:delta_t];
	
	time_counter += delta_t;
	
	if (universe)
	{
		switch (particle_type)
		{
			case PARTICLE_TEST :
			case PARTICLE_SHOT_EXPIRED :
			case PARTICLE_SHOT_YELLOW_PLASMA :
			case PARTICLE_SHOT_GREEN_PLASMA :
			case PARTICLE_SHOT_PLASMA :
			case PARTICLE_EXPLOSION :
				if (!texName)
					[self initialiseTexture: textureNameString];
				q_rotation = [[universe entityZero] QRotation];
				q_rotation.w = -q_rotation.w;
				quaternion_into_gl_matrix(q_rotation, rotMatrix);
				break;
		}
		switch (particle_type)
		{
			case PARTICLE_TEST :
				alpha = (sin(time_counter) + 2.0) / 3.0;
				break;
			
			case PARTICLE_EXPLOSION :
				[self updateExplosion:delta_t];
				break;
			
			case PARTICLE_HYPERRING :
				[self updateHyperring:delta_t];
				break;
			
			case PARTICLE_LASER_BEAM_RED :
			case PARTICLE_LASER_BEAM :
				[self updateLaser:delta_t];
				break;
			
			case PARTICLE_EXHAUST :
				break;
			
			case PARTICLE_ECM_MINE :
				[self updateECMMine:delta_t];
				break;
			
			case PARTICLE_SHOT_EXPIRED :
			case PARTICLE_SHOT_YELLOW_PLASMA :
			case PARTICLE_SHOT_GREEN_PLASMA :
			case PARTICLE_SHOT_PLASMA :
				[self updateShot:delta_t];
				break;
		}
	}

}

- (void) updateExplosion:(double) delta_t
{
	[self setSize:NSMakeSize((1.0 + time_counter)*64.0,(1.0 + time_counter)*64.0)];
	alpha = (duration - time_counter);
	if (time_counter > duration)
		[universe removeEntity:self];
}

- (void) updateECMMine:(double) delta_t
{
	if (time_counter > activation_time)
	{
		// do ecm stuff
		double radius = 0.5 * activation_time * SCANNER_MAX_RANGE;
		if (radius > SCANNER_MAX_RANGE)
			radius = SCANNER_MAX_RANGE;
		NSArray* targets = [universe getEntitiesWithinRange:radius ofEntity:self];
		if ([targets count] > 0)
		{
			int i;
			for (i = 0; i < [targets count]; i++)
			{
				Entity *e2 = [targets objectAtIndex:i];
				if ([e2 isKindOfClass:[ShipEntity class]])
					[[(ShipEntity *)e2 getAI] reactToMessage:@"ECM"];
			}
		}
		activation_time += 0.5; // go off every half second
	}
	if (time_counter > duration)	// until the timer runs out!
		[universe removeEntity:self];
}

- (void) updateShot:(double) delta_t
{
	if ([collidingEntities count] > 0)
	{
		int i;
		for (i = 0; i < [collidingEntities count]; i++)
		{
			Entity *	e = (Entity *)[collidingEntities objectAtIndex:i];
			[e takeEnergyDamage:energy from:self becauseOf:[universe entityForUniversalID:owner]];
		}
		velocity.x = 0.0;
		velocity.y = 0.0;
		velocity.z = 0.0;
		[self setColor:[NSColor redColor]];
		[self setSize:NSMakeSize(64.0,64.0)];
		duration = 2.0;
		time_counter = 0.0;
		particle_type = PARTICLE_EXPLOSION;
	}
	position.x += velocity.x * delta_t;
	position.y += velocity.y * delta_t;
	position.z += velocity.z * delta_t;
	alpha = (duration - time_counter);
	if (time_counter > duration)
		[universe removeEntity:self];
}

- (void) updateLaser:(double) delta_t
{
	position.x += velocity.x * delta_t;
	position.y += velocity.y * delta_t;
	position.z += velocity.z * delta_t;
	alpha = (duration - time_counter) / PARTICLE_LASER_DURATION;
	if (time_counter > duration)
		[universe removeEntity:self];
}

- (void) updateHyperring:(double) delta_t
{
	position.x += velocity.x * delta_t;
	position.y += velocity.y * delta_t;
	position.z += velocity.z * delta_t;
	alpha = (duration - time_counter) / PARTICLE_LASER_DURATION;
	ring_inner_radius += delta_t * size.width * 1.1;
	ring_outer_radius += delta_t * size.height;
	if (time_counter > duration)
		[universe removeEntity:self];
}


- (void) drawEntity:(BOOL) immediate :(BOOL) translucent;
{
	if ([universe breakPatternHide])   return; // DON'T DRAW
	
	if (translucent)
	{
		switch (particle_type)
		{
			case PARTICLE_LASER_BEAM_RED :
				[self drawLaser];
				break;
			
			case PARTICLE_EXHAUST :
				[self drawExhaust: immediate];
				break;
				
			case PARTICLE_HYPERRING :
				[self drawHyperring];
				break;
				
			case PARTICLE_ECM_MINE :
				// not a visible entity
				break;
				
			default :
				[self drawParticle];
				break;
		}
	}
}

- (void) drawParticle
{
    int viewdir;
	double  xx = size.width / 2.0;
	double  yy = size.height / 2.0;
	
	if (alpha < 0.0)
        alpha = 0.0;	// clamp the alpha value
    if (alpha > 1.0)
        alpha = 1.0;	// clamp the alpha value

    color_fv[3] = alpha;
    
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glPushMatrix();
	
	glColor4f(1.0, 1.0, 1.0, alpha);
	
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, color_fv);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);

	glBindTexture(GL_TEXTURE_2D, texName);
	glBegin(GL_QUADS);
	
	viewdir = [universe viewDir];
	
	switch (viewdir)
	{
		case	VIEW_AFT :
			glTexCoord2f(0.0, 1.0);
			glVertex3f(xx, -yy, 0);

			glTexCoord2f(1.0, 1.0);
			glVertex3f(-xx, -yy, 0);

			glTexCoord2f(1.0, 0.0);
			glVertex3f(-xx, yy, 0);
			
			glTexCoord2f(0.0, 0.0);
			glVertex3f(xx, yy, 0);
			break;
			
		case	VIEW_STARBOARD :
			glTexCoord2f(0.0, 1.0);
			glVertex3f(0, -yy, xx);

			glTexCoord2f(1.0, 1.0);
			glVertex3f(0, -yy, -xx);

			glTexCoord2f(1.0, 0.0);
			glVertex3f(0, yy, -xx);
			
			glTexCoord2f(0.0, 0.0);
			glVertex3f(0, yy, xx);
			break;
			
		case	VIEW_PORT :
			glTexCoord2f(0.0, 1.0);
			glVertex3f(0, -yy, -xx);

			glTexCoord2f(1.0, 1.0);
			glVertex3f(0, -yy, xx);

			glTexCoord2f(1.0, 0.0);
			glVertex3f(0, yy, xx);
			
			glTexCoord2f(0.0, 0.0);
			glVertex3f(0, yy, -xx);
			break;
			
		default :
			glTexCoord2f(0.0, 1.0);
			glVertex3f(-xx, -yy, 0);

			glTexCoord2f(1.0, 1.0);
			glVertex3f(xx, -yy, 0);

			glTexCoord2f(1.0, 0.0);
			glVertex3f(xx, yy, 0);
			
			glTexCoord2f(0.0, 0.0);
			glVertex3f(-xx, yy, 0);
			break;
	}

	glEnd();
	
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
}

- (void) drawLaser
{
	GLfloat flat_ambdiff[4]		= {1.0, 1.0, 1.0, 1.0};   // for alpha
	GLfloat lase_ambdiff[4]		= {0.0, 0.0, 0.0, 1.0};   // for alpha
    GLfloat lase_zero[4]		= {0.0, 0.0, 0.0, 1.0};   // nothing
	double  lase_alpha = 0.75;
	
	color_fv[3]		= lase_alpha;  // set alpha
	lase_ambdiff[3] = lase_alpha;  // set alpha
	
	glDisable(GL_CULL_FACE);			// face culling
	//glShadeModel(GL_SMOOTH);
		
	//NSLog(@"drawing laser from %3.1f distant to %3.1f distant with alphas (%3.1f, %3.1f)", sqrt(d0), sqrt(d1), alpha0, alpha1);
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, lase_ambdiff);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color_fv);
	
	glBegin(GL_QUADS);
	
	glVertex3f(0.25, 0.0, 0.0);
	glVertex3f(0.25, 0.0, collision_radius);
	glVertex3f(-0.25, 0.0, collision_radius);
	glVertex3f(-0.25, 0.0, 0.0);
	
	glVertex3f(0.0, 0.25, 0.0);
	glVertex3f(0.0, 0.25, collision_radius);
	glVertex3f(0.0, -0.25, collision_radius);
	glVertex3f(0.0, -0.25, 0.0);
	
	glEnd();
		
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, lase_zero);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, flat_ambdiff);
	
	//glShadeModel(GL_FLAT);
	glEnable(GL_CULL_FACE);			// face culling
}

- (void) drawExhaust:(BOOL) immediate
{
    int fi, vi;
	
	GLfloat flat_ambdiff[4] = {1.0, 1.0, 1.0, 1.0};   // flat
	GLfloat ex_ambdiff[4]	= {0.0, 0.0, 0.0, 1.0};   // pale blue
	GLfloat ex_emissive[4]	= {0.6, 0.8, 1.0, 0.9};   // pale blue
    GLfloat ex_zero[4]		= {0.0, 0.0, 0.0, 1.0};   // nothing
	
	ShipEntity  *ship =(ShipEntity *)[universe entityForUniversalID:owner];
	int dam = [ship damage];
	double flare_length = [ship speed_factor];
	
	if (flare_length == 0.0)	// don't draw if there's no fire!
		return;
	
	double flare_factor = flare_length * ex_emissive[3];
	double red_factor = flare_length * ex_emissive[0] * (rand() % 11) * 0.1;	// random fluctuations
	double green_factor = flare_length * ex_emissive[1];
	//double blue_factor = flare_length * ex_emissive[2];

	if ((rand() % 50) < dam - 50)   // flicker the damaged engines
		red_factor = 0.0;
	if ((rand() % 40) < dam - 60)
		green_factor = 0.0;
	if ((rand() % 25) < dam - 75)
		flare_factor = 0.0;
	
	
	if (flare_length < 0.1)   flare_length = 0.1;	

	glPushMatrix();
	glDisable(GL_CULL_FACE);			// face culling
	glShadeModel(GL_SMOOTH);

	//NSLog(@"... drawing exhaust for ship %@ %d with flare_length %.2f", [ship name], owner, flare_length);
	glTranslated( exhaust_offset.x, exhaust_offset.y, exhaust_offset.z);
	
	if (basefile)
	{
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ex_zero);
		glBegin(GL_TRIANGLES);
		for (fi = 0; fi < n_faces; fi++)
		{
			for (vi = 0; vi < 3; vi++)
			{
				int v = faces[fi].vertex[vi];
				ex_ambdiff[3] = flare_factor * alpha_for_vertex[v];		// fade alpha towards rear of exhaust
				ex_emissive[3] = flare_factor * alpha_for_vertex[v];	// fade alpha towards rear of exhaust
				ex_emissive[1] = green_factor * alpha_for_vertex[v];	// diminish green part towards rear of exhaust
				ex_emissive[0] = red_factor * alpha_for_vertex[v];		// diminish red part towards rear of exhaust
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ex_ambdiff);
				glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, ex_emissive);
				glNormal3f(vertex_normal[v].x,vertex_normal[v].y,vertex_normal[v].z);
				glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z * flare_length);
			}
			
		}
		glEnd();
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, ex_zero);
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, flat_ambdiff);
	}

	glEnable(GL_CULL_FACE);			// face culling
	glPopMatrix();
}

- (void) drawHyperring
{
    int i;
	GLfloat aleph = (alpha*0.5 < 1.0) ? alpha*0.5 : 1.0;
	
    GLfloat ex_zero[4]		= {0.0, 0.0, 0.0, 1.0};		// black
	GLfloat flat_ambdiff[4] = {1.0, 1.0, 1.0, 1.0};		// flat
	GLfloat ex_ambdiff[4]	= {0.0, 0.0, 0.0, aleph};		// black
	GLfloat ex_em_hi[4]		= {0.6, 0.8, 1.0, aleph};   // pale blue
	GLfloat ex_em_lo[4]		= {0.2, 0.0, 1.0, 0.0};		// purplish-blue-black
	
	glPushMatrix();
	glDisable(GL_CULL_FACE);			// face culling
	glShadeModel(GL_SMOOTH);

	//NSLog(@"... drawing hyppering inner_radius:%.1f  alpha:%.2f", ring_inner_radius, aleph);
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ex_zero);
	glBegin(GL_TRIANGLE_STRIP);
	for (i = 0; i < 65; i++)
	{
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ex_em_lo);
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, ex_em_lo);
		glVertex3f( ring_inner_radius*circleVertex[i].x, ring_inner_radius*circleVertex[i].y, ring_inner_radius*circleVertex[i].z );
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ex_ambdiff);
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, ex_em_hi);
		glVertex3f( ring_outer_radius*circleVertex[i].x, ring_outer_radius*circleVertex[i].y, ring_outer_radius*circleVertex[i].z );
	}
	glEnd();
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, ex_zero);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, flat_ambdiff);

	glEnable(GL_CULL_FACE);			// face culling
	glPopMatrix();
}

@end
