//
//  PlanetEntity.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <Foundation/Foundation.h>

#import "entities.h"

#define PLANET_TYPE_GREEN		100
#define PLANET_TYPE_SUN			200
#define PLANET_TYPE_ATMOSPHERE  300

@interface PlanetEntity : Entity {
	
	int planet_type;
    int r_seed[MAX_VERTICES_PER_ENTITY];
	GLuint		displayListNames[6];	// 0 -> 20 verts
										// 1 -> 80 verts
										// 2 -> 320 verts
										// 3 -> 1280 verts
										// 4 -> 5120 verts
										// 5 -> 20480 verts !!
	
	int planet_seed;
	double		polar_color_factor;
	double		sin_value[360];
	
	GLfloat		amb_land[4];
	GLfloat		amb_polar_land[4];
	GLfloat		amb_sea[4];
	GLfloat		amb_polar_sea[4];
	
	PlanetEntity*   atmosphere;		// secondary sphere used to show atmospheric details
	
	int			shuttles_on_ground;			// starting number of shuttles
	double		last_launch_time;			// space launches out by about 15 minutes
	double		shuttle_launch_interval;	// space launches out by about 15 minutes

}

- (id) initAsAtmosphereForPlanet:(PlanetEntity *) planet;
- (id) initWithSeed:(Random_Seed) p_seed fromUniverse:(Universe *) uni;

- (void) drawCircle:(double) radius :(int) step;

- (double) polar_color_factor;
- (GLfloat *) amb_land;
- (GLfloat *) amb_polar_land;
- (GLfloat *) amb_sea;
- (GLfloat *) amb_polar_sea;

- (int) getPlanetType;
- (void) setPlanetType:(int) pt;

- (double) getRadius;

- (void) setRadius:(double) rad;

- (void) rescaleTo:(double) rad;

- (void) drawModelSubdividingBy:(int) subdivide;

/*--
- (void) drawTriangle: (Vector) v0: (Vector) v1: (Vector) v2
			withIndices: (int) r0: (int) r1: (int) r2
			andSeed:(int) seed1
			subdividingBy:(int) subdivide;
		
- (void) doGLColorVertex: (int) r : (Vector) v: (int) seed;
--*/

void triangle ( PlanetEntity * planet, Vector v0, Vector v1, Vector v2, int r0, int r1, int r2, int seed1, int subdivide);
void colorVertex( PlanetEntity * planet,  int r, Vector v, int seed);

- (void) launchShuttle;

- (void) welcomeShuttle:(ShipEntity *) shuttle;

@end
