//
//  PlanetEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <OpenGL/gl.h>

#import "PlanetEntity.h"
#import "entities.h"

#import "AI.h"
#import "Universe.h"
#import "MyOpenGLView.h"

#define LIM500  500.0*500.0 * NO_DRAW_DISTANCE_FACTOR*NO_DRAW_DISTANCE_FACTOR
#define LIM4K   4000.0*4000.0 * NO_DRAW_DISTANCE_FACTOR*NO_DRAW_DISTANCE_FACTOR
#define LIM8K   8000.0*8000.0 * NO_DRAW_DISTANCE_FACTOR*NO_DRAW_DISTANCE_FACTOR
#define LIM16K  16000.0*16000.0 * NO_DRAW_DISTANCE_FACTOR*NO_DRAW_DISTANCE_FACTOR

@implementation PlanetEntity


- (id) init
{
	int		i;
	int		percent_land;
	double  aleph =  1.0 / sqrt(2.0);
	//
	self = [super init];
	
    //
    collision_radius = 25000.0; //  25km across
	//
	for (i = 0; i < 360; i++)
		sin_value[i] = sin(i * PI / 180);
	//
	scan_class = CLASS_NO_DRAW;
	//
	q_rotation.w =  aleph;		// represents a 90 degree rotation around x axis
	q_rotation.x =  aleph;		// (I hope!)
	q_rotation.y =  0.0;
	q_rotation.z =  0.0;
	//
	planet_type =   PLANET_TYPE_GREEN;
	//
	shuttles_on_ground = 0;
	last_launch_time = 0.0;
	shuttle_launch_interval = 60 * 60;
	//
	for (i = 0; i < 5; i++)
		displayListNames[i] = 0;	// empty for now!
	//
	[self setModel:@"icosahedron.dat"];
	//
	[self rescaleTo:1.0];
	//
	planet_seed = 54321;
	//
	srand(planet_seed);
	percent_land = (rand() % 50);
	//NSLog(@"Planet surface is %d percent land.",percent_land);
	for (i = 0; i < n_vertices; i++)
	{
		if (rand() % 100 < percent_land)
			r_seed[i] = 0;  // land
		else
			r_seed[i] = 1;  // sea
	}
	//
	polar_color_factor = 1.0;
	//
	amb_land[0] = 0.0;
	amb_land[1] = 1.0;
	amb_land[2] = 0.0;
	amb_land[3] = 1.0;
	amb_sea[0] = 0.0;
	amb_sea[1] = 0.0;
	amb_sea[2] = 1.0;
	amb_sea[3] = 1.0;
	amb_polar_land[0] = 0.9;
	amb_polar_land[1] = 0.9;
	amb_polar_land[2] = 0.9;
	amb_polar_land[3] = 1.0;
	amb_polar_sea[0] = 1.0;
	amb_polar_sea[1] = 1.0;
	amb_polar_sea[2] = 1.0;
	amb_polar_sea[3] = 1.0;
	
    return self;
}

- (id) initAsAtmosphereForPlanet:(PlanetEntity *) planet
{    
    int		i;
	int		percent_land;
	double  aleph =  1.0 / sqrt(2.0);
	//
	self = [super init];
    //
	position = [planet getPosition];
	q_rotation = [planet QRotation];
    collision_radius = [planet collisionRadius] + 500.0; //  atmosphere is 500m deep only
	//
	shuttles_on_ground = 0;
	last_launch_time = 0.0;
	shuttle_launch_interval = 60 * 60;
	
	scan_class = CLASS_NO_DRAW;
	//
	q_rotation.w =  aleph;		// represents a 90 degree rotation around x axis
	q_rotation.x =  aleph;		// (I hope!)
	q_rotation.y =  0.0;
	q_rotation.z =  0.0;
	//
	planet_type =   PLANET_TYPE_ATMOSPHERE;
	//
	for (i = 0; i < 5; i++)
		displayListNames[i] = 0;	// empty for now!
	//
	[self setModel:@"icosahedron.dat"];
	//
	[self rescaleTo:1.0];
	//
	//srand(12345);
	percent_land = (gen_rnd_number() % 33)+(gen_rnd_number() % 34);
	//NSLog(@"Atmosphere is %d percent clear.",percent_land);
	percent_land = percent_land * 2.56;
	for (i = 0; i < n_vertices; i++)
	{
		if (gen_rnd_number() < percent_land)
			r_seed[i] = 0;  // land
		else
			r_seed[i] = 100;  // sea
	}
	//
	polar_color_factor = 1.0;
	//
	amb_land[0] = gen_rnd_number() / 256.0;
	amb_land[1] = gen_rnd_number() / 256.0;
	amb_land[2] = gen_rnd_number() / 256.0;
	amb_land[3] = 0.05;  // bluesky .. zero clouds
	amb_sea[0] = 0.9 + gen_rnd_number() / 2560.0;
	amb_sea[1] = 0.9 + gen_rnd_number() / 2560.0;
	amb_sea[2] = 0.9 + gen_rnd_number() / 2560.0;
	amb_sea[3] = 0.50;  // 50% opaque clouds
	amb_polar_land[0] = gen_rnd_number() / 256.0;
	amb_polar_land[1] = gen_rnd_number() / 256.0;
	amb_polar_land[2] = gen_rnd_number() / 256.0;
	amb_polar_land[3] = 0.25;	// 25% gray clouds
	amb_polar_sea[0] = 0.9 + gen_rnd_number() / 2560.0;
	amb_polar_sea[1] = 0.9 + gen_rnd_number() / 2560.0;
	amb_polar_sea[2] = 0.9 + gen_rnd_number() / 2560.0;
	amb_polar_sea[3] = 0.75;	// 75% clouds
	
	atmosphere = nil;
	
    return self;
}

- (id) initWithSeed:(Random_Seed) p_seed fromUniverse:(Universe *) uni
{    
    int		i;
	int		percent_land;
	double  aleph =  1.0 / sqrt(2.0);
	//
	self = [super init];
    //
	seed_for_planet_description(p_seed);
	//
	NSDictionary*   systemInfo = [uni generateSystemData:p_seed];
	int radius_km =		[(NSNumber *)[systemInfo objectForKey:@"radius"] intValue];
	int techlevel =		[(NSNumber *)[systemInfo objectForKey:@"techlevel"] intValue];
	//NSLog(@"Generating planet %@ with radius %dkm",[systemInfo objectForKey:@"name"],radius_km);
	
	shuttles_on_ground = 1 + floor(techlevel * 0.5);
	last_launch_time = 0.0;
	shuttle_launch_interval = 3600.0 / shuttles_on_ground; // all are launched in an hour
	last_launch_time = -(rand() % 60) * shuttle_launch_interval/60.0;
	last_launch_time = 30.0 - shuttle_launch_interval;   // debug - launch 30s after player enters universe
	
	//NSLog(@"shuttles on ground:%d launch_interval:%.1f minutes", shuttles_on_ground, shuttle_launch_interval/60);
	
    //collision_radius = 25000.0; //  25km across
	collision_radius = radius_km * 10.0; // scale down by a factor of 100 !
	//
	scan_class = CLASS_NO_DRAW;
	//
	q_rotation.w =  aleph;		// represents a 90 degree rotation around x axis
	q_rotation.x =  aleph;		// (I hope!)
	q_rotation.y =  0.0;
	q_rotation.z =  0.0;
	//
	planet_type =   PLANET_TYPE_GREEN;  // generic planet type
	//
	for (i = 0; i < 5; i++)
		displayListNames[i] = 0;	// empty for now!
	//
	[self setModel:@"icosahedron.dat"];
	//
	[self rescaleTo:1.0];
	//
	percent_land = (gen_rnd_number() % 48);
	//NSLog(@"Planet surface is %d percent land.",percent_land);
	for (i = 0; i < n_vertices; i++)
	{
		if (gen_rnd_number() < 256 * percent_land / 100)
			r_seed[i] = 0;  // land
		else
			r_seed[i] = 100;  // sea
	}
	//
	polar_color_factor = 1.0;
	//
	Vector land_hsb, sea_hsb, land_polar_hsb, sea_polar_hsb;
	
	land_hsb.x = gen_rnd_number() / 256.0;  land_hsb.y = gen_rnd_number() / 256.0;  land_hsb.z = 0.5 + gen_rnd_number() / 512.0;  
	sea_hsb.x = gen_rnd_number() / 256.0;  sea_hsb.y = gen_rnd_number() / 256.0;  sea_hsb.z = 0.5 + gen_rnd_number() / 512.0;  
	while (dot_product(land_hsb,sea_hsb) > .80) // make sure land and sea colors differ significantly
	{
	sea_hsb.x = gen_rnd_number() / 256.0;  sea_hsb.y = gen_rnd_number() / 256.0;  sea_hsb.z = 0.5 + gen_rnd_number() / 512.0;  
	}
	
	// polar areas are brighter but have less color (closer to white)
	//
	land_polar_hsb.x = land_hsb.x;  land_polar_hsb.y = (land_hsb.y / 5.0);  land_polar_hsb.z = 1.0 - (land_hsb.z / 10.0);  
	sea_polar_hsb.x = sea_hsb.x;  sea_polar_hsb.y = (sea_hsb.y / 5.0);  sea_polar_hsb.z = 1.0 - (sea_hsb.z / 10.0);  

	amb_land[0] = [[NSColor colorWithCalibratedHue:land_hsb.x saturation:land_hsb.y brightness:land_hsb.z alpha:1.0] redComponent];
	amb_land[1] = [[NSColor colorWithCalibratedHue:land_hsb.x saturation:land_hsb.y brightness:land_hsb.z alpha:1.0] blueComponent];
	amb_land[2] = [[NSColor colorWithCalibratedHue:land_hsb.x saturation:land_hsb.y brightness:land_hsb.z alpha:1.0] greenComponent];
	amb_land[3] = 1.0;
	amb_sea[0] = [[NSColor colorWithCalibratedHue:sea_hsb.x saturation:sea_hsb.y brightness:sea_hsb.z alpha:1.0] redComponent];
	amb_sea[1] = [[NSColor colorWithCalibratedHue:sea_hsb.x saturation:sea_hsb.y brightness:sea_hsb.z alpha:1.0] blueComponent];
	amb_sea[2] = [[NSColor colorWithCalibratedHue:sea_hsb.x saturation:sea_hsb.y brightness:sea_hsb.z alpha:1.0] greenComponent];
	amb_sea[3] = 1.0;
	amb_polar_land[0] = [[NSColor colorWithCalibratedHue:land_polar_hsb.x saturation:land_polar_hsb.y brightness:land_polar_hsb.z alpha:1.0] redComponent];
	amb_polar_land[1] = [[NSColor colorWithCalibratedHue:land_polar_hsb.x saturation:land_polar_hsb.y brightness:land_polar_hsb.z alpha:1.0] blueComponent];
	amb_polar_land[2] = [[NSColor colorWithCalibratedHue:land_polar_hsb.x saturation:land_polar_hsb.y brightness:land_polar_hsb.z alpha:1.0] greenComponent];
	amb_polar_land[3] = 1.0;
	amb_polar_sea[0] = [[NSColor colorWithCalibratedHue:sea_polar_hsb.x saturation:sea_polar_hsb.y brightness:sea_polar_hsb.z alpha:1.0] redComponent];
	amb_polar_sea[1] = [[NSColor colorWithCalibratedHue:sea_polar_hsb.x saturation:sea_polar_hsb.y brightness:sea_polar_hsb.z alpha:1.0] blueComponent];
	amb_polar_sea[2] = [[NSColor colorWithCalibratedHue:sea_polar_hsb.x saturation:sea_polar_hsb.y brightness:sea_polar_hsb.z alpha:1.0] greenComponent];
	amb_polar_sea[3] = 1.0;
	
	// do atmosphere
	//
	atmosphere = [[PlanetEntity alloc] initAsAtmosphereForPlanet:self];
	[atmosphere setUniverse:universe];
    return self;
}

- (void) setUniverse:(Universe *)univ
{
    if (univ)
    {
        if (universe)	[universe release];
        universe = [univ retain];
		if (atmosphere)
			[atmosphere setUniverse:univ];
    }
}

- (void) dealloc
{
    if (atmosphere) [atmosphere dealloc];
	[super dealloc];
}

- (BOOL) canCollide
{
	return YES;
}

- (BOOL) checkCloseCollisionWith:(Entity *)other
{
	//NSLog(@"PLANET Collision!");
	if ([other isKindOfClass:[ShipEntity class]])
	{
		if ([(ShipEntity *)other reportAImessages])
		{
			ShipEntity *ship = (ShipEntity *)other;
			Vector p1 = [ship getPosition];
			NSLog(@"%@ %d collided with planet at (%.1f,%.1f,%.1f)",[ship name], [ship universal_id], p1.x,p1.y,p1.z);
		}
	}
	
	return YES;
}


- (void) update:(double) delta_t
{	
	[super update:delta_t];
	
	if (planet_type == PLANET_TYPE_GREEN)
	{
		double ugt = [universe getTime];

		if ((shuttles_on_ground > 0)&&(ugt > last_launch_time + shuttle_launch_interval))
		{
			[self launchShuttle];
			shuttles_on_ground--;
			last_launch_time = ugt;
		}
		
		// normal planetary rotation
		quaternion_rotate_about_y( &q_rotation, 0.02*delta_t);
		quaternion_normalise(&q_rotation);
		quaternion_into_gl_matrix(q_rotation, rotMatrix);
		
		if (atmosphere)
			[atmosphere update:delta_t]; // spin clouds 3x faster
	}

	if (planet_type == PLANET_TYPE_ATMOSPHERE)
	{
		// atmospheric rotation
		quaternion_rotate_about_y( &q_rotation, 0.05*delta_t);
		quaternion_normalise(&q_rotation);
		quaternion_into_gl_matrix(q_rotation, rotMatrix);
	}
	
	if (planet_type == PLANET_TYPE_SUN)
	{
		// billboard routine
		q_rotation = [[universe entityZero] QRotation];
		q_rotation.w = -q_rotation.w;
		quaternion_into_gl_matrix(q_rotation, rotMatrix);
	}
}


- (void) setPosition:(Vector) posn
{
	position = posn;
	if (atmosphere)
		[atmosphere setPosition:posn];
}

- (void) setPosition:(GLfloat) x:(GLfloat) y:(GLfloat) z
{
	position.x = x;
	position.y = y;
	position.z = z;
	if (atmosphere)
		[atmosphere setPosition:x:y:z];
}

- (void) setModel:(NSString *) modelName
{
    double  old_collision_radius = collision_radius;
    [super setModel:modelName];
	collision_radius = old_collision_radius;	// preserve the radius
	//NSLog(@"Planet collision radius preserved!");
}


- (void) drawEntity:(BOOL) immediate :(BOOL) translucent;
{
	int		subdivideLevel =	2;		// 4 is probably the maximum!
	//double  drawRatio =			no_draw_distance / zero_distance;
	double  drawFactor = [(MyOpenGLView *)[universe gameView] viewSize].width / 100.0;
	double  drawRatio2 = drawFactor * collision_radius / sqrt(zero_distance); // equivalent to size on screen in pixels

	if ([universe breakPatternHide])   return; // DON'T DRAW

	if (zero_distance > 0.0)
	{
		subdivideLevel = 2 + floor(drawRatio2);
		if (subdivideLevel > 4)
			subdivideLevel = 4;
	}
	
	
	glFrontFace(GL_CW);			// face culling - front faces are AntiClockwise!
    	
	/*
	
	The depth test gets disabled in parts of this and instead
	we rely on the painters algorithm instead.
	
	The depth buffer isn't granular enough to cope with huge objects at vast
	distances.
	
	*/
	
	switch (planet_type)
	{
		case PLANET_TYPE_ATMOSPHERE :
		case PLANET_TYPE_GREEN :
			if (!translucent)
			{
				GLfloat amb_1[]		= {1.0, 1.0,	1.0,	1.0 };
				
				glColor4f(1.0, 1.0, 1.0, 1.0);
				glShadeModel(GL_SMOOTH);
				
				glFrontFace(GL_CCW);
				if (displayListNames[subdivideLevel] != 0)
				{
					glCallList(displayListNames[subdivideLevel]);
				}
				else
				{
					displayListNames[subdivideLevel] = glGenLists(1);
					if (displayListNames[subdivideLevel] != 0)
					{
						//NSLog(@"Generating planet data for subdivide %d",subdivideLevel);
						glNewList(displayListNames[subdivideLevel], GL_COMPILE);
						[self drawModelSubdividingBy:subdivideLevel];
						glEndList();
					}
				}
				glFrontFace(GL_CW);
				
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb_1);

				if (atmosphere)
				{
					glDisable(GL_DEPTH_TEST);
					[atmosphere drawEntity:immediate :translucent];
					glEnable(GL_DEPTH_TEST);
				}

			}
			break;
		
		case PLANET_TYPE_SUN :
			if (!translucent)
			{
				GLfloat amb_green[] = {0.05, 0.8,	0.1,	1.0 };
				GLfloat amb_1[]		= {1.0, 1.0,	1.0,	1.0 };
				double  alpha;
				
				glDisable(GL_LIGHTING);

				glColor4f(1.0, 1.0, 0.75, 1.0);
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb_green);
				
				[self drawCircle:collision_radius :6 - subdivideLevel];
												
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb_1);
				

				glDisable(GL_DEPTH_TEST);
				
				if ((zero_distance < LIM4K)&&(zero_distance > collision_radius + 4000.0))
				{
					alpha = (zero_distance - collision_radius - 4000.0)/4000.0;
					if (alpha > 1.0)	alpha = 1.0;
					glColor4f(1.00, 1.00, 0.0, 0.25 * alpha);	//4km coronasphere
					[self drawCircle:collision_radius + 4000.0 :6 - subdivideLevel];
				}
				if ((zero_distance < LIM8K)&&(zero_distance > collision_radius + 8000.0))
				{
					alpha = (zero_distance - collision_radius - 8000.0)/8000.0;
					if (alpha > 1.0)	alpha = 1.0;
					glColor4f(1.00, 0.80, 0.2, 0.25 * alpha);	//8km coronasphere
					[self drawCircle:collision_radius + 8000.0 :6 - subdivideLevel];
				}
				if ((zero_distance < LIM16K)&&(zero_distance > collision_radius + 16000.0))
				{
					alpha = (zero_distance - collision_radius - 16000.0)/16000.0;
					if (alpha > 1.0)	alpha = 1.0;
					glColor4f(1.00, 0.40, 0.40, 0.25 * alpha);	//16km coronasphere
					[self drawCircle:collision_radius + 16000.0 :6 - subdivideLevel];
				}
						
				glEnable(GL_DEPTH_TEST);
				glEnable(GL_LIGHTING);
			}
			break;
	}
	
	glFrontFace(GL_CCW);			// face culling - front faces are AntiClockwise!

}

- (void) drawCircle:(double) radius :(int) step
{
	int i;
	double s, c;
	int viewdir = [universe viewDir];
	glBegin(GL_TRIANGLE_FAN);
	glVertex3i( 0, 0, 0);
	switch (viewdir)
	{
		case VIEW_FORWARD :
			for ( i = 0; i < 360; i += step )
			{
				s = radius * sin_value[i];
				c = radius * sin_value[(i + 90) % 360];
				
				glVertex3f(s,c,0.0);
			}
			glVertex3f( 0.0, radius, 0.0);	//repeat the zero value to close
			break;
		case VIEW_PORT :
			for ( i = 0; i < 360; i += step )
			{
				s = radius * sin_value[i];
				c = radius * sin_value[(i + 90) % 360];
				
				glVertex3f(0.0,c,s);
			}
			glVertex3f( 0.0, radius, 0.0);	//repeat the zero value to close
			break;
		case VIEW_STARBOARD :
			for ( i = 0; i < 360; i += step )
			{
				s = radius * sin_value[i];
				c = radius * sin_value[(i + 90) % 360];
				
				glVertex3f(0.0,c,-s);
			}
			glVertex3f( 0.0, radius, 0.0);	//repeat the zero value to close
			break;
		case VIEW_AFT :
			for ( i = 0; i < 360; i += step )
			{
				s = radius * sin_value[i];
				c = radius * sin_value[(i + 90) % 360];
				
				glVertex3f(s,-c,0.0);
			}
			glVertex3f( 0.0, -radius, 0.0);	//repeat the zero value to close
			break;
	}
	glEnd();
}

- (double) polar_color_factor
{
	return polar_color_factor;
}
- (GLfloat *) amb_land
{
	return amb_land;
}
- (GLfloat *) amb_polar_land
{
	return amb_polar_land;
}
- (GLfloat *) amb_sea
{
	return amb_sea;
}
- (GLfloat *) amb_polar_sea
{
	return amb_polar_sea;
}


- (int) getPlanetType
{
	return planet_type;
}

- (void) setPlanetType:(int) pt
{
	planet_type = pt;
}

- (double) getRadius
{
	return collision_radius;
}
- (void) setRadius:(double) rad
{
	collision_radius = rad;
}

- (void) rescaleTo:(double) rad
{
    int i;
	Vector  vert;
		
    for (i = 0; i < n_vertices; i++)
    {
        vert = unit_vector(&vertices[i]);
		vert.x *= rad;
		vert.y *= rad;
		vert.z *= rad;
		vertices[i] = vert;
	}
}

static int level;

- (void) drawModelSubdividingBy:(int) subdivide
{
    // roll out each face in turn
    //
    int fi;
    //
	level = 0;
	//
	glBegin(GL_TRIANGLES);
	for (fi = 0; fi < n_faces; fi++)
		triangle (self, vertices[faces[fi].vertex[0]],vertices[faces[fi].vertex[1]],vertices[faces[fi].vertex[2]],
						r_seed[faces[fi].vertex[0]],r_seed[faces[fi].vertex[1]],r_seed[faces[fi].vertex[2]],
						planet_seed,	subdivide);
/*--
	{
		[self   drawTriangle: vertices[faces[fi].vertex[0]]   : vertices[faces[fi].vertex[1]]   : vertices[faces[fi].vertex[2]]
				withIndices: r_seed[faces[fi].vertex[0]]	: r_seed[faces[fi].vertex[1]]   : r_seed[faces[fi].vertex[2]]
				andSeed: planet_seed
				subdividingBy: subdivide];
	}	
--*/
	glEnd();
}

/*--
// TODO
// turn the next two methods into more streamlined ordinary c function calls
// =========================================================================
//
- (void) drawTriangle:(Vector) v0:(Vector) v1:(Vector) v2
			withIndices:(int) r0:(int) r1:(int) r2
			andSeed:(int) seed1
			subdividingBy:(int) subdivide
{
	if (subdivide > 0)
	{
		Vector  v01, v12, v20;
		int		r01, r12, r20;
		
		//
		level++;
		//
								
		// create the new vertices
		
		v01.x = (v0.x + v1.x);			v01.y = (v0.y + v1.y);			v01.z = (v0.z + v1.z);
		v12.x = (v1.x + v2.x);			v12.y = (v1.y + v2.y);			v12.z = (v1.z + v2.z);
		v20.x = (v2.x + v0.x);			v20.y = (v2.y + v0.y);			v20.z = (v2.z + v0.z);
		
		// normalise
		v01 = unit_vector(&v01);		v12 = unit_vector(&v12);		v20 = unit_vector(&v20); 
		
		// create the new color values
		if (r0 == r1)
		{
			r01 = r0;
		}
		else
		{
			srand(v01.x*1000+v01.y*100+v01.z*10);
			r01 = (rand() % 2) *100; //  was  % 2
		}
		
		if (r1 == r2)
		{
			r12 = r1;
		}
		else
		{
			srand(v12.x*1000+v12.y*100+v12.z*10);
			r12 = (rand() % 2) *100; //  was  % 2
		}
		
		if (r2 == r0)
		{
			r20 = r2;
		}
		else
		{
			srand(v20.x*1000+v20.y*100+v20.z*10);
			r20 = (rand() % 2) *100; //  was  % 2
		}
		
		//draw the sub-triangles
		[self   drawTriangle:v0:v01:v20		withIndices:r0:r01:r20  andSeed:seed1   subdividingBy:subdivide-1];
		[self   drawTriangle:v01:v1:v12		withIndices:r01:r1:r12  andSeed:seed1   subdividingBy:subdivide-1];
		[self   drawTriangle:v12:v2:v20		withIndices:r12:r2:r20  andSeed:seed1	subdividingBy:subdivide-1];
		[self   drawTriangle:v01:v12:v20	withIndices:r01:r12:r20 andSeed:seed1	subdividingBy:subdivide-1];
		
		//
		level--;
		//
	}
	else
	{
		[self doGLColorVertex:r0:v0:seed1];
		[self doGLColorVertex:r1:v1:seed1];
		[self doGLColorVertex:r2:v2:seed1];
	}
}

- (void) doGLColorVertex: (int) r : (Vector) v : (int) seed
{
	int i;
	GLfloat my_land[] = { 0.2, 0.9, 0.0, 1.0};
	GLfloat my_sea[] = { 0.0, 0.2, 0.9, 1.0};
	double pole_blend = v.z * v.z * polar_color_factor;
	my_land[0] = (1.0 - pole_blend)*amb_land[0] + pole_blend*amb_polar_land[0];
	my_land[1] = (1.0 - pole_blend)*amb_land[1] + pole_blend*amb_polar_land[1];
	my_land[2] = (1.0 - pole_blend)*amb_land[2] + pole_blend*amb_polar_land[2];
	my_sea[0] = (1.0 - pole_blend)*amb_sea[0] + pole_blend*amb_polar_sea[0];
	my_sea[1] = (1.0 - pole_blend)*amb_sea[1] + pole_blend*amb_polar_sea[1];
	my_sea[2] = (1.0 - pole_blend)*amb_sea[2] + pole_blend*amb_polar_sea[2];
	if (planet_type == PLANET_TYPE_ATMOSPHERE)
	{
		// do alphas
		my_land[3] = (1.0 - pole_blend)*amb_land[3] + pole_blend*amb_polar_land[3];
		my_sea[3] = (1.0 - pole_blend)*amb_sea[3] + pole_blend*amb_polar_sea[3];
	}
	
	srand(seed+v.x*1000+v.y*100+v.z*10);
	// add a random component to the color
	for (i = 0; i < 3; i++)
	{
		double cv = (rand() % 100)*0.01;	//  0..1
		my_land[i] += (cv - 0.5)*0.1;
		my_sea[i] += (cv - 0.5)*0.1;
	}

	{
		GLfloat my_color[4];
		for (i = 0; i < 4; i++)
			my_color[i] = (r * my_sea[i])*0.01 + ((100 - r) * my_land[i])*0.01;
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, my_color);
	}
	
	//NSLog(@"	Doing point (%.2f,%.2f,%.2f) color %d radius %f",v.x,v.y,v.z, r, collision_radius);
	glNormal3f(v.x, v.y, v.z);
	glVertex3f(v.x * collision_radius, v.y * collision_radius, v.z * collision_radius);
}
--*/
void triangle ( PlanetEntity * planet, Vector v0, Vector v1, Vector v2, int r0, int r1, int r2, int seed1, int subdivide)
{
	if (subdivide > 0)
	{
		Vector  v01, v12, v20;
		int		r01, r12, r20;
		//
		level++;
		//
		// create the new vertices
		v01.x = (v0.x + v1.x);			v01.y = (v0.y + v1.y);			v01.z = (v0.z + v1.z);
		v12.x = (v1.x + v2.x);			v12.y = (v1.y + v2.y);			v12.z = (v1.z + v2.z);
		v20.x = (v2.x + v0.x);			v20.y = (v2.y + v0.y);			v20.z = (v2.z + v0.z);
		// normalise
		v01 = unit_vector(&v01);		v12 = unit_vector(&v12);		v20 = unit_vector(&v20); 
		
		// create the new color values
		if (r0 == r1)
			r01 = r0;
		else
		{
			srand(v01.x*1000+v01.y*100+v01.z*10);
			r01 = (rand() % 2) *100; //  was  % 2
		}
		
		if (r1 == r2)
			r12 = r1;
		else
		{
			srand(v12.x*1000+v12.y*100+v12.z*10);
			r12 = (rand() % 2) *100; //  was  % 2
		}
		
		if (r2 == r0)
			r20 = r2;
		else
		{
			srand(v20.x*1000+v20.y*100+v20.z*10);
			r20 = (rand() % 2) *100; //  was  % 2
		}
		
		//draw the sub-triangles
		triangle(   planet, v0,v01,v20,		r0,r01,r20,		seed1,  subdivide - 1);
		triangle(   planet, v01,v1,v12,		r01,r1,r12,		seed1,  subdivide - 1);
		triangle(   planet, v12,v2,v20,		r12,r2,r20,		seed1,  subdivide - 1);
		triangle(   planet, v01,v12,v20,	r01,r12,r20,	seed1,  subdivide - 1);
		//
		level--;
		//
	}
	else
	{
		colorVertex( planet, r0,v0,seed1);
		colorVertex( planet, r1,v1,seed1);
		colorVertex( planet, r2,v2,seed1);
	}
}

void colorVertex( PlanetEntity * planet, int r, Vector v, int seed)
{
	int i;
	GLfloat my_land[] = { 0.2, 0.9, 0.0, 1.0};
	GLfloat my_sea[] = { 0.0, 0.2, 0.9, 1.0};
	GLfloat *al = [planet amb_land];
	GLfloat *apl = [planet amb_polar_land];
	GLfloat *as = [planet amb_sea];
	GLfloat *aps = [planet amb_polar_sea];
	double  cr = [planet collisionRadius];
	double pole_blend = v.z * v.z * [planet polar_color_factor];
		GLfloat my_color[4];
	my_land[0] = (1.0 - pole_blend)*al[0] + pole_blend*apl[0];
	my_land[1] = (1.0 - pole_blend)*al[1] + pole_blend*apl[1];
	my_land[2] = (1.0 - pole_blend)*al[2] + pole_blend*apl[2];
	my_sea[0] = (1.0 - pole_blend)*as[0] + pole_blend*aps[0];
	my_sea[1] = (1.0 - pole_blend)*as[1] + pole_blend*aps[1];
	my_sea[2] = (1.0 - pole_blend)*as[2] + pole_blend*aps[2];
	if ([planet getPlanetType] == PLANET_TYPE_ATMOSPHERE)
	{
		// do alphas
		my_land[3] = (1.0 - pole_blend)*al[3] + pole_blend*apl[3];
		my_sea[3] = (1.0 - pole_blend)*as[3] + pole_blend*aps[3];
		
		// experiment with 'more fluffy' clouds
		/*--
		srand(seed+v.x*1000+v.y*100+v.z*10);
		my_land[3] *= (50 + rand() % 50)*0.01;  // 0.5 .. 1.0
		my_sea[3] *= (50 + rand() % 50)*0.01;   // 0.5 .. 1.0
		--*/
	}
	
	srand(seed+v.x*1000+v.y*100+v.z*10);
	// add a random component to the color
	for (i = 0; i < 3; i++)
	{
		double cv = (rand() % 100)*0.01;	//  0..1
		my_land[i] += (cv - 0.5)*0.1;
		my_sea[i] += (cv - 0.5)*0.1;
	}

	for (i = 0; i < 4; i++)
		my_color[i] = (r * my_sea[i])*0.01 + ((100 - r) * my_land[i])*0.01;
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, my_color);
	
	glNormal3f(v.x, v.y, v.z);
	glVertex3f(v.x * cr, v.y * cr, v.z * cr);
}


- (void) launchShuttle
{
	ShipEntity  *shuttle_ship;
	Quaternion  q1;
	Vector		launch_pos = position;
	double		start_distance = collision_radius + 125.0;
	
	quaternion_set_random(&q1);
		
	Vector vf = vector_forward_from_quaternion(q1);
	
	launch_pos.x += start_distance * vf.x;
	launch_pos.y += start_distance * vf.y;
	launch_pos.z += start_distance * vf.z;
	
	shuttle_ship = [universe getShipWithRole:@"shuttle"];   // retain count = 1
	[shuttle_ship setPosition:launch_pos];
	[shuttle_ship setQRotation:q1];
	
	[shuttle_ship setScanClass: CLASS_NEUTRAL];
	
	[shuttle_ship setCargoFlag:CARGO_FLAG_FULL_PLENTIFUL];
	
	[shuttle_ship setStatus:STATUS_IN_FLIGHT];
	
	//[shuttle_ship setReportAImessages:YES]; // debug

	[universe addEntity:shuttle_ship];
	[[shuttle_ship getAI] setStateMachine:@"risingShuttleAI.plist"];	// must happen after adding to the universe!

	//NSLog(@"Planet %@ in universe %@ Launching shuttle: %@ %d", self, universe, [shuttle_ship name], [shuttle_ship universal_id]);
		
	[shuttle_ship release];
}

- (void) welcomeShuttle:(ShipEntity *) shuttle
{
	shuttles_on_ground++;
}


@end
