//
//  PlayerEntity Additions.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "PlayerEntity.h"
#import "PlayerEntity Additions.h"
#import "GuiDisplayGen.h"
#import "Universe.h"
#import "AI.h"

@implementation PlayerEntity (Scripting)

static NSString * mission_string_value;

- (void) checkScript
{
	int i;
	
	//NSLog(@"----- checkScript");

	for (i = 0; i < [[script allKeys] count]; i++)
	{
		NSString *missionTitle = (NSString *)[[script allKeys] objectAtIndex:i];
		NSArray *mission = (NSArray *)[script objectForKey:missionTitle];
		int j;
		for (j = 0; j < [mission count]; j++)
			[self checkCouplet:(NSDictionary *)[mission objectAtIndex:j]];
	}
}

- (void) checkCouplet:(NSDictionary *) couplet
{
	NSArray *conditions = (NSArray *)[couplet objectForKey:@"conditions"];
	NSArray *actions = (NSArray *)[couplet objectForKey:@"do"];
	BOOL success = YES;
	int i;
	for (i = 0; (i < [conditions count])&&(success); i++)
		success &= [self scriptTestCondition:(NSString *)[conditions objectAtIndex:i]];
	if (success)
	for (i = 0; i < [actions count]; i++)
	{
		if ([[actions objectAtIndex:i] isKindOfClass:[NSDictionary class]])
			[self checkCouplet:(NSDictionary *)[actions objectAtIndex:i]];
		if ([[actions objectAtIndex:i] isKindOfClass:[NSString class]])
			[self scriptAction:(NSString *)[actions objectAtIndex:i]];
	}
}

- (void) scriptAction:(NSString *) scriptAction
{
	/*
	a script action takes the form of an expression:
	
	action[: string_expression]
	
	where 'action' is a  selector for PlayerEntity optionally taking a NSString object ('string_expression')
	as a variable
	
	The special action 'set: mission_variable string_expression'
	
	is used to set a mission variable to the given string_expression
	
	*/
	NSMutableArray*	tokens = [NSMutableArray arrayWithArray:[scriptAction componentsSeparatedByString:@" "]];
	NSString*   selectorString = nil;
	NSString*	valueString = nil;
	SEL			_selector;

	//NSLog(@"::::: scriptAction: \"%@\"", scriptAction);

	if ([tokens count] < 1)
	{
		NSLog(@"***** No scriptAction '%@'",scriptAction);
		return;
	}

	selectorString = (NSString *)[tokens objectAtIndex:0];
	
	if ([tokens count] > 1)
	{
		[tokens removeObjectAtIndex:0];
		valueString = [tokens componentsJoinedByString:@" "];
	}
					
	_selector = NSSelectorFromString(selectorString);
	
	if (![self respondsToSelector:_selector])
	{
		NSLog(@"***** PlayerEntity DOES NOT RESPOND TO scriptAction: \"%@\"", scriptAction);
		return;
	}
	
	if ([selectorString hasSuffix:@":"])
		[self performSelector:_selector withObject:valueString];
	else
		[self performSelector:_selector];
}

- (BOOL) scriptTestCondition:(NSString *) scriptCondition
{
	/*
	a script condition takes the form of an expression:
	
	testable_variable lessthan|equals|greaterthan constant_expression
	
	where testable_variable is an accessor selector for PlayerEntity returning an object
	that can be compared with the constant expression. They are supposed to take the form:
		variablename_type where type can be 'string', 'bool', or 'number'
	
	or where testable_variable is prefixed with 'mission_' in which case it is a 'mission variable'
	which is a string used by the script as a means of setting flags or indicating state
	
	The special test:
	
	testable_variable undefined
	
	is used only with mission variables and is true when that mission variable has yet to be used
	
	*/
	NSArray *tokens = [scriptCondition componentsSeparatedByString:@" "];
	NSString*   selectorString = nil;
	NSString*	comparisonString = nil;
	NSString*	valueString = nil;
	SEL			_selector;
	int			comparator = COMPARISON_NO;
	
	//NSLog(@"::::: scriptTestCondition: \"%@\"", scriptCondition);
	
	if ([tokens count] < 1)
	{
		NSLog(@"***** No scriptCondition '%@'",scriptCondition);
		return NO;
	}
	selectorString = (NSString *)[tokens objectAtIndex:0];
	if ([selectorString hasPrefix:@"mission_"])
	{
		//NSLog(@"..... checking mission_variable '%@'",selectorString);
		mission_string_value = (NSString *)[mission_variables objectForKey:selectorString];
		selectorString = @"mission_string";
	}
	
	if ([tokens count] > 1)
	{
		comparisonString = (NSString *)[tokens objectAtIndex:1];
		if ([comparisonString isEqual:@"equal"])
			comparator = COMPARISON_EQUAL;
		if ([comparisonString isEqual:@"lessthan"])
			comparator = COMPARISON_LESSTHAN;
		if (([comparisonString isEqual:@"greaterthan"])||([comparisonString isEqual:@"morethan"]))
			comparator = COMPARISON_GREATERTHAN;
		if ([comparisonString isEqual:@"undefined"])
			comparator = COMPARISON_UNDEFINED;
	}
	
	if ([tokens count] > 2)
		valueString = (NSString *)[tokens objectAtIndex:2];
			
	_selector = NSSelectorFromString(selectorString);
	if (![self respondsToSelector:_selector])
		return NO;
		
	// test string values (method returns NSString*)
	if ([selectorString hasSuffix:@"_string"])
	{
		NSString *result = [self performSelector:_selector];
		//NSLog(@"..... comparing \"%@\" to \"%@\"", result, valueString);
		switch (comparator)
		{
			case COMPARISON_UNDEFINED :
				return (result == nil);
			case COMPARISON_NO :
				return NO;
			case COMPARISON_EQUAL :
				return ([result isEqual:valueString]);
			case COMPARISON_LESSTHAN :
				return ([result isLessThan:valueString]);
			case COMPARISON_GREATERTHAN :
				return ([result isGreaterThan:valueString]);
		}
	}
	// test number values (method returns NSNumber*)
	if ([selectorString hasSuffix:@"_number"])
	{
		NSNumber *result = [self performSelector:_selector];
		NSNumber *value = [NSNumber numberWithDouble:[valueString doubleValue]];
		//NSLog(@"..... comparing \"%@\" to \"%@\"", result, value);
		switch (comparator)
		{
			case COMPARISON_UNDEFINED :
			case COMPARISON_NO :
				return NO;
			case COMPARISON_EQUAL :
				return ([result isEqual:value]);
			case COMPARISON_LESSTHAN :
				return ([result isLessThan:value]);
			case COMPARISON_GREATERTHAN :
				return ([result isGreaterThan:value]);
		}
	}
	// test boolean values (method returns BOOL)
	if ([selectorString hasSuffix:@"_bool"])
	{
		BOOL result = [[self performSelector:_selector] isEqual:@"YES"];
		BOOL value = [valueString isEqual:@"YES"];
		switch (comparator)
		{
			case COMPARISON_GREATERTHAN :
			case COMPARISON_LESSTHAN :
			case COMPARISON_UNDEFINED :
			case COMPARISON_NO :
				return NO;
			case COMPARISON_EQUAL :
				return (result == value);
		}
	}
	// default!
	return NO;
}


/*-----------------------------------------------------*/



- (NSString *) mission_string
{
	return mission_string_value;
}
- (NSString *) status_string
{
	switch(status)
	{
		case STATUS_AUTOPILOT_ENGAGED :
			return @"STATUS_AUTOPILOT_ENGAGED";
		case STATUS_DEAD :
			return @"STATUS_DEAD";
		case STATUS_DEMO :
			return @"STATUS_DEMO";
		case STATUS_DOCKING :
			return @"STATUS_DOCKING";
		case STATUS_DOCKED :
			return @"STATUS_DOCKED";
		case STATUS_EFFECT :
			return @"STATUS_EFFECT";
		case STATUS_ENTERING_WITCHSPACE :
			return @"STATUS_ENTERING_WITCHSPACE";
		case STATUS_ESCAPE_SEQUENCE :
			return @"STATUS_ESCAPE_SEQUENCE";
		case STATUS_EXITING_WITCHSPACE :
			return @"STATUS_EXITING_WITCHSPACE";
		case STATUS_EXPERIMENTAL :
			return @"STATUS_EXPERIMENTAL";
		case STATUS_IN_FLIGHT :
			return @"STATUS_IN_FLIGHT";
		case STATUS_IN_HOLD :
			return @"STATUS_IN_HOLD";
		case STATUS_INACTIVE :
			return @"STATUS_INACTIVE";
		case STATUS_LAUNCHING :
			return @"STATUS_LAUNCHING";
		case STATUS_TEST :
			return @"STATUS_TEST";
		case STATUS_WITCHSPACE_COUNTDOWN :
			return @"STATUS_WITCHSPACE_COUNTDOWN";
		default :
			return @"UNDEFINED";
	}
}
- (NSString *) gui_screen_string
{
	switch(gui_screen)
	{
		case GUI_SCREEN_EQUIP_SHIP :
			return @"GUI_SCREEN_EQUIP_SHIP";
		case GUI_SCREEN_INTRO1 :
			return @"GUI_SCREEN_INTRO1";
		case GUI_SCREEN_INTRO2 :
			return @"GUI_SCREEN_INTRO2";
		case GUI_SCREEN_INVENTORY :
			return @"GUI_SCREEN_INVENTORY";
		case GUI_SCREEN_LONG_RANGE_CHART :
			return @"GUI_SCREEN_LONG_RANGE_CHART";
		case GUI_SCREEN_MAIN :
			return @"GUI_SCREEN_MAIN";
		case GUI_SCREEN_MARKET :
			return @"GUI_SCREEN_MARKET";
		case GUI_SCREEN_MISSION :
			return @"GUI_SCREEN_MISSION";
		case GUI_SCREEN_OPTIONS :
			return @"GUI_SCREEN_OPTIONS";
		case GUI_SCREEN_SHORT_RANGE_CHART :
			return @"GUI_SCREEN_SHORT_RANGE_CHART";
		case GUI_SCREEN_STATUS :
			return @"GUI_SCREEN_STATUS";
		case GUI_SCREEN_SYSTEM_DATA :
			return @"GUI_SCREEN_SYSTEM_DATA";
		default :
			return @"UNDEFINED";
	}
}
- (NSNumber *) galaxy_number
{
	return [NSNumber numberWithInt:galaxy_number];
}
- (NSNumber *) planet_number
{
	return [NSNumber numberWithInt:[universe findSystemNumberAtCoords:galaxy_coordinates withGalaxySeed:galaxy_seed]];
}
- (NSNumber *) score_number
{
	return [NSNumber numberWithInt:ship_kills];
}
- (NSNumber *) scriptTimer_number
{
	return [NSNumber numberWithDouble:script_time];
}

static int shipsFound;
- (NSNumber *) shipsFound_number
{
	return [NSNumber numberWithInt:shipsFound];
}

- (NSNumber *) legalStatus_number
{
	return [NSNumber numberWithInt:legal_status];
}


- (NSNumber *) d100_number
{
	int d100 = rand() % 100;
	return [NSNumber numberWithInt:d100];
}

- (NSNumber *) pseudoFixedD100_number
{
	// set the system seed for random number generation
	seed_for_planet_description(system_seed);
	int d100 = (gen_rnd_number()+gen_rnd_number()) % 100;
	return [NSNumber numberWithInt:d100];
}

/*-----------------------------------------------------*/

- (void) setLegalStatus:(NSString *)valueString
{
	legal_status = [valueString intValue];
}

- (void) awardCredits:(NSString *)valueString
{
	int award = 10 * [valueString intValue];
	credits += award;
}

- (void) awardShipKills:(NSString *)valueString
{
	ship_kills += [valueString intValue];
}

- (void) awardEquipment:(NSString *)equipString  //eg. EQ_NAVAL_ENERGY_UNIT
{
	NSString*   eq_type		= equipString;
	
	if ([eq_type isEqual:@"EQ_FUEL"])
		fuel = 70;
	
	if ([eq_type isEqual:@"EQ_MISSILE"])
	{
		missiles++;
		missile_status = (missiles > 0) ? MISSILE_STATUS_SAFE : MISSILE_STATUS_NONE;
	}

	if ([eq_type isEqual:@"EQ_CARGO_BAY"])
	{
		// find the standard cargo limit for your craft first
		int standard_max_cargo = 20;
		if (max_cargo <= standard_max_cargo)
			max_cargo += 15;
	}

	if (([eq_type isEqual:@"EQ_ECM"]) && (!has_ecm))
		has_ecm = YES;

	if (([eq_type isEqual:@"EQ_FUEL_SCOOPS"]) && (!has_scoop))
		has_scoop = YES;

	if (([eq_type isEqual:@"EQ_ESCAPE_POD"]) && (!has_escape_pod))
		has_escape_pod = YES;

	if (([eq_type isEqual:@"EQ_ENERGY_BOMB"]) && (!has_energy_bomb))
		has_energy_bomb = YES;

	if (([eq_type isEqual:@"EQ_ENERGY_UNIT"]) && (!has_energy_unit))
	{
		has_energy_unit = YES;
		energy_unit = ENERGY_UNIT_NORMAL;
	}

	if ([eq_type isEqual:@"EQ_NAVAL_ENERGY_UNIT"])
	{
		has_energy_unit = YES;
		energy_unit = ENERGY_UNIT_NAVAL;
	}

	if (([eq_type isEqual:@"EQ_DOCK_COMP"]) && (!has_docking_computer))
		has_docking_computer = YES;

	if (([eq_type isEqual:@"EQ_GAL_DRIVE"]) && (!has_galactic_hyperdrive))
		has_galactic_hyperdrive = YES;
}

- (void) messageShipAIs:(NSString *)roles_message
{
	NSMutableArray*	tokens = [NSMutableArray arrayWithArray:[roles_message componentsSeparatedByString:@" "]];
	NSString*   roleString = nil;
	NSString*	messageString = nil;

	if ([tokens count] < 2)
	{
		NSLog(@"***** CANNOT MESSAGESHIPSAIS: '%@'",roles_message);
		return;
	}
	
	roleString = (NSString *)[tokens objectAtIndex:0];
	[tokens removeObjectAtIndex:0];
	messageString = [tokens componentsJoinedByString:@" "];
	
	[universe sendShipsWithRole:roleString messageToAI:messageString];
}

- (void) addShips:(NSString *)roles_number
{
	NSMutableArray*	tokens = [NSMutableArray arrayWithArray:[roles_number componentsSeparatedByString:@" "]];
	NSString*   roleString = nil;
	NSString*	numberString = nil;

	if ([tokens count] != 2)
	{
		NSLog(@"***** CANNOT ADDSHIPS: '%@'",roles_number);
		return;
	}
	
	roleString = (NSString *)[tokens objectAtIndex:0];
	numberString = (NSString *)[tokens objectAtIndex:1];
	
	int number = [numberString intValue];
	
	//NSLog(@"DEBUG Going to add %d ships with role '%@'", number, roleString);

	while (number--)
		[universe witchspaceShipWithRole:roleString];
}

- (void) addSystemShips:(NSString *)roles_number_position
{
	NSMutableArray*	tokens = [NSMutableArray arrayWithArray:[roles_number_position componentsSeparatedByString:@" "]];
	NSString*   roleString = nil;
	NSString*	numberString = nil;
	NSString*	positionString = nil;

	if ([tokens count] != 3)
	{
		NSLog(@"***** CANNOT ADDSYSTEMSHIPS: '%@'",roles_number_position);
		return;
	}
	
	roleString = (NSString *)[tokens objectAtIndex:0];
	numberString = (NSString *)[tokens objectAtIndex:1];
	positionString = (NSString *)[tokens objectAtIndex:2];
	
	int number = [numberString intValue];
	double posn = [positionString doubleValue];
	
	//NSLog(@"DEBUG Going to add %d ships with role '%@' at a point %.3f along route1", number, roleString, posn);

	while (number--)
		[universe addShipWithRole:roleString nearRouteOneAt:posn];
}

- (void) set:(NSString *)value
{
	NSMutableArray*	tokens = [NSMutableArray arrayWithArray:[value componentsSeparatedByString:@" "]];
	NSString*   missionVariableString = nil;
	NSString*	valueString = nil;

	if ([tokens count] < 2)
	{
		NSLog(@"***** CANNOT SET: '%@'",value);
		return;
	}
	
	missionVariableString = (NSString *)[tokens objectAtIndex:0];
	[tokens removeObjectAtIndex:0];
	valueString = [tokens componentsJoinedByString:@" "];
	
	if (![missionVariableString hasPrefix:@"mission_"])
	{
		NSLog(@"***** MISSION VARIABLE '%@' DOES NOT BEGIN WITH 'mission_'",value);
		return;
	}
	
	[mission_variables setObject:valueString forKey:missionVariableString];
}

- (void) increment:(NSString *)missionVariableString
{
	int value = 0;
	if ([mission_variables objectForKey:missionVariableString])
		value = [(NSString *)[mission_variables objectForKey:missionVariableString] intValue];
	value++;
	[mission_variables setObject:[NSString stringWithFormat:@"%d", value] forKey:missionVariableString];
}

- (void) decrement:(NSString *)missionVariableString
{
	int value = 0;
	if ([mission_variables objectForKey:missionVariableString])
		value = [(NSString *)[mission_variables objectForKey:missionVariableString] intValue];
	value--;
	[mission_variables setObject:[NSString stringWithFormat:@"%d", value] forKey:missionVariableString];
}

- (void) checkForShips: (NSString *)roleString
{
	shipsFound = [universe countShipsWithRole:roleString];
}

- (void) resetScriptTimer
{
	script_time = 0.0;
	script_time_check = SCRIPT_TIMER_INTERVAL;
	script_time_interval = SCRIPT_TIMER_INTERVAL;
}

- (void) addMissionText: (NSString *)textKey
{
	if ([textKey isEqual:lastTextKey])
		return; // don't repeatedly add the same text
	//
	GuiDisplayGen   *gui =  [universe gui];
	NSString		*text = (NSString *)[[universe missiontext] objectForKey:textKey];
	//NSLog(@"::::: Adding text '%@':\n'%@'", textKey, text);
	NSArray			*paras = [text componentsSeparatedByString:@"\\n"];
	if (text)
	{
		int i;
		for (i = 0; i < [paras count]; i++)
			missionTextRow = [gui addLongText:(NSString *)[paras objectAtIndex:i] startingAtRow:missionTextRow align:GUI_ALIGN_LEFT];
	}
	if (lastTextKey)
		[lastTextKey release];
	lastTextKey = [[NSString stringWithString:textKey] retain];  // 
	[gui generateDisplay];
	[universe guiUpdated];
}

- (void) showShipModel: (NSString *)shipKey
{
	ShipEntity		*ship;
	
	Quaternion		q2 = q_rotation;
	Vector			pos;
	Vector			vf = v_forward;
	Vector			vu = v_up;
	
	if ([universe station])
	{
		q2 = [[universe station] QRotation];
		vu = vector_up_from_quaternion(q2);
	}
	else
	{
		q2.w = -q2.w;
	}
	quaternion_rotate_about_axis(&q2,vu,PI);
	//
	ship = [universe getShipWithRole: shipKey];   // retain count = 1
	//NSLog(@"::::: showShipModel:'%@' (%@) (%@)", shipKey, ship, [ship name]);
	[ship setQRotation: q2];
	pos.x = position.x + 3.6 * [ship collisionRadius] * vf.x;
	pos.y = position.y + 3.6 * [ship collisionRadius] * vf.y;
	pos.z = position.z + 3.6 * [ship collisionRadius] * vf.z;
		
	[ship setStatus: STATUS_DEMO];
	[ship setPosition: pos];
	[ship setScanClass: CLASS_NO_DRAW];
	[ship setRoll: PI/5.0];
	[ship setPitch: PI/10.0];
	[universe addEntity: ship];
	[[ship getAI] setStateMachine: @"nullAI.plist"];

	[ship release];
	//
}



/*-----------------------------------------------------*/



- (void) setGuiToMissionScreen
{
	GuiDisplayGen* gui = [universe gui];

	// GUI stuff
	{
		[gui clear];
		[gui setTitle:@"Mission Information"];
		//
		[gui setText:@"Press Space Commander" forRow:21 align:GUI_ALIGN_CENTER];
		[gui setColor:[NSColor yellowColor] forRow:21];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	missionTextRow = 1;
	
	if (gui)
		gui_screen = GUI_SCREEN_MISSION;

	if (lastTextKey)
	{
		[lastTextKey release];
		lastTextKey = nil;
	}
	
	if (themeMusic)
	{
		GoToBeginningOfMovie ([themeMusic QTMovie]);
		StartMovie ([themeMusic QTMovie]);
	}

	// the following are necessary...
	status = STATUS_DEMO;
	[universe setDisplayText:YES];
	[universe setViewDirection:VIEW_DOCKED];
}



@end
