//
//  PlayerEntity.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <Foundation/Foundation.h>

#import <QuickTime/Movies.h>

#import "ShipEntity.h"

#define SCRIPT_TIMER_INTERVAL			10.0

#define GUI_SCREEN_MAIN					000
#define GUI_SCREEN_INTRO1				001
#define GUI_SCREEN_INTRO2				002
#define GUI_SCREEN_STATUS				101
#define GUI_SCREEN_EQUIP_SHIP			102
#define GUI_SCREEN_LONG_RANGE_CHART		103
#define GUI_SCREEN_SHORT_RANGE_CHART	104
#define GUI_SCREEN_SYSTEM_DATA			105
#define GUI_SCREEN_MARKET				106
#define GUI_SCREEN_INVENTORY			107
#define GUI_SCREEN_OPTIONS				108
#define GUI_SCREEN_MISSION				201

#define GUI_ROW_OPTIONS_SAVE		9
#define GUI_ROW_OPTIONS_LOAD		10
#define GUI_ROW_OPTIONS_OPTIONS		11
#define GUI_ROW_OPTIONS_DISPLAY		12
#define GUI_ROW_OPTIONS_SPEECH		13
#define GUI_ROW_EQUIPMENT_START		3
#define GUI_ROW_EQUIPMENT_CASH		1
#define GUI_ROW_MARKET_KEY			1
#define GUI_ROW_MARKET_START		2
#define GUI_ROW_MARKET_CASH			20

#define WEAPON_COOLING_FACTOR   2.0
#define ENERGY_RECHARGE_FACTOR  4.0
#define SHIELD_RECHARGE_FACTOR  2.0
#define ECM_ENERGY_DRAIN_FACTOR 16.0
#define ECM_DURATION			2.5

#define PLAYER_MAX_FORWARD_SHIELD   128.0
#define PLAYER_MAX_AFT_SHIELD		128.0
#define PLAYER_MAX_ENERGY			256.0
#define PLAYER_MAX_WEAPON_TEMP		256.0
#define PLAYER_MAX_FLIGHT_SPEED		160.0
#define PLAYER_MAX_CABIN_TEMP		256.0
#define PLAYER_MAX_FUEL				70.0
#define PLAYER_DIAL_MAX_ALTITUDE	40000.0
	//  ~~~~~~~~~~~~~~~~~~~~~~~~	= 40km
	
#define ALERT_CONDITION_DOCKED		0
#define ALERT_CONDITION_GREEN		1
#define ALERT_CONDITION_YELLOW		2
#define ALERT_CONDITION_RED			3

#define SHOT_RELOAD					0.25

#define HYPERSPEED_FACTOR			20.0

#define PLAYER_SHIP_DESC			@"cobra3-player"
#define PLAYER_MODEL				@"cobramk3.dat"

#define ESCAPE_SEQUENCE_TIME		10.0

#define AI_DOCKING_COMPUTER			@"dockingAI.plist"

#define MS_WITCHSPACE_SF			@"Witchspace to %@ in %.0f s."
#define MS_GAL_WITCHSPACE_F			@"Activating intergalactic jump in %.0f s."

#define MISSILE_STATUS_SAFE				0
#define MISSILE_STATUS_ARMED			1
#define MISSILE_STATUS_TARGET_LOCKED	2
#define MISSILE_STATUS_NONE				-1

#define WEAPON_FACING_NONE				0
#define WEAPON_FACING_FORWARD			1
#define WEAPON_FACING_AFT				2
#define WEAPON_FACING_PORT				3
#define WEAPON_FACING_STARBOARD			4

#define FORWARD_FACING_STRING			@"\tForward"
#define AFT_FACING_STRING				@"\tAft"
#define PORT_FACING_STRING				@"\tPort"
#define STARBOARD_FACING_STRING			@"\tStarboard"

#define ENERGY_UNIT_NONE				0
#define ENERGY_UNIT_NORMAL				15
#define ENERGY_UNIT_NAVAL				20

//@class ShipEntity;

@interface PlayerEntity : ShipEntity {
	
	NSDictionary			*script;
	NSMutableDictionary		*mission_variables;
	int						missionTextRow;
	double					script_time;
	double					script_time_check;
	double					script_time_interval;
	NSString				*lastTextKey;

	NSString				*planetSearchString;
	
	StationEntity			*docked_station;
	
	BOOL rolling, pitching;
	BOOL using_mining_laser;
	
	BOOL mouse_control_on;
	
	BOOL speech_on;

	BOOL docking_music_on;
	
	double  speed_delta, roll_delta, pitch_delta;
	
	double  forward_shield, aft_shield;
	double  weapon_temp;
	double  weapon_energy_per_shot, weapon_heat_increment_per_shot, weapon_reload_time;
	double  cabin_temp;
	
	int		chosen_weapon_facing;   // for purchasing weapons
	
	BOOL	game_over;
	BOOL	docked;
	BOOL	finished;
	BOOL	bomb_detonated;
	BOOL	autopilot_engaged;
	BOOL	hyperspeed_engaged;
	BOOL	hyperspeed_locked;

	BOOL	ecm_in_operation;
	double	ecm_start_time;
	
	NSMovie*	themeMusic;
	NSMovie*	dockingMusic;
	
	NSSound*	beepSound;
	NSSound*	boopSound;
	NSSound*	weaponSound;
	NSSound*	missileSound;
	NSSound*	damageSound;
	NSSound*	scrapeDamageSound;
	NSSound*	destructionSound;
	NSSound*	breakPatternSound;
	NSSound*	ecmSound;
	NSSound*	buySound;
	NSSound*	sellSound;
	NSSound*	warningSound;
	
	int			gui_screen;
	int			alert_condition;
	int			missile_status;

	int			current_cargo;

	NSPoint			cursor_coordinates;
	Random_Seed		system_seed;
	Random_Seed		target_system_seed;
	
	double		witchspaceCountdown;
	double		fuel_accumulator;
	
	// player commander data
	//
	NSString*		player_name;
	NSPoint			galaxy_coordinates;

	Random_Seed		galaxy_seed;

	int				credits;
	int				fuel;
	
	int				galaxy_number;
	int				forward_weapon;
	int				aft_weapon;
	int				port_weapon;
	int				starboard_weapon;
	
	//				max_cargo			// int				- from ShipEntity
	////			cargo				// NSMutableArray   - from ShipEntity
	NSArray*		shipCommodityData;
	//				has_ecm				// BOOL				- from ShipEntity
	//				has_scoop			// BOOL				- from ShipEntity
	BOOL			has_energy_bomb;
	BOOL			has_energy_unit;
	int				energy_unit;
	BOOL			has_docking_computer;
	BOOL			has_galactic_hyperdrive;
	//				has_escape_pod;		// BOOL				- from ShipEntity
	
	//				missiles			// int				- from ShipEntity
	int				legal_status;
	//NSArray*		stationCommodityData;   // no longer required
	int				market_rnd;
	int				ship_kills;
	BOOL			saved;
	
}

- (void) warnAboutHostiles;

- (void) unloadCargoPods;
- (void) loadCargoPods;

- (int)			random_factor;
- (Random_Seed) galaxy_seed;
- (NSPoint)		galaxy_coordinates;

- (NSDictionary *) commanderDataDictionary;
- (void) setCommanderDataFromDictionary:(NSDictionary *) dic;

- (void) set_up;

- (void) doBookkeeping:(double) delta_t;

- (BOOL) massLocked;
- (BOOL) atHyperspeed;
- (Vector) velocityVector;

- (double) dial_roll;
- (double) dial_pitch;
- (double) dial_speed;

- (double) dial_forward_shield;
- (double) dial_aft_shield;

- (double) dial_energy;

- (double) dial_fuel;
- (double) dial_cabin_temp;
- (double) dial_weapon_temp;
- (double) dial_altitude;

- (int) dial_missiles;
- (int) dial_missile_status;

- (int) scanAlertCondition;

- (void) pollControls:(double) delta_t;
- (void) pollApplicationControls;
- (void) pollFlightControls:(double) delta_t;
- (void) pollFlightArrowKeyControls:(double) delta_t;
- (void) pollGuiArrowKeyControls:(double) delta_t;
- (void) pollViewControls;
- (void) pollGuiScreenControls;
- (void) pollGameOverControls:(double) delta_t;
- (void) pollAutopilotControls:(double) delta_t;
- (void) pollDockedControls:(double) delta_t;
- (void) pollDemoControls:(double) delta_t;

- (BOOL) fireEnergyBomb;

- (BOOL) fireMainWeapon;

- (void) enterGalacticWitchspace;

- (void) interpretAIMessage:(NSString *)ms;

- (void) getDestroyed;
- (void) docked;

- (void) savePlayer;
- (void) loadPlayer;
- (void) loadPlayerFromFile:(NSString *)fileToOpen;
- (void) changePlayerName;

- (void) setGuiToStatusScreen;
- (void) setGuiToSystemDataScreen;
- (void) setGuiToLongRangeChartScreen;
- (void) setGuiToShortRangeChartScreen;
- (void) setGuiToLoadSaveScreen;
- (void) setGuiToEquipShipScreen;
- (void) setGuiToMarketScreen;

- (void) setGuiToIntro1Screen;
- (void) setGuiToIntro2Screen;

- (int) gui_screen;

- (BOOL) tryBuyingItem:(int) index;
- (BOOL) tryBuyingCommodity:(int) index;
- (BOOL) trySellingCommodity:(int) index;

- (BOOL) isMining;

- (BOOL) speech_on;

@end
