//
//  PlayerEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "PlayerEntity.h"
#import "PlayerEntity Additions.h"
#import "entities.h"

#import "vector.h"
#import "GameController.h"
#import "ResourceManager.h"
#import "Universe.h"
#import "AI.h"
#import "MyOpenGLView.h"

@implementation PlayerEntity

- (void) unloadCargoPods
{
	/* loads commodities from the cargo pods onto the ship's manifest */
	int i;
	NSMutableArray* localMarket = [docked_station localMarket]; 
	NSMutableArray* manifest = [[NSMutableArray arrayWithArray:localMarket] retain];  // retain
	//
	// zero the quantities in the manifest, making a mutable array of mutable arrays
	//
	for (i = 0; i < [manifest count]; i++)
	{
		NSMutableArray* commodityInfo = [NSMutableArray arrayWithArray:(NSArray *)[manifest objectAtIndex:i]];
		[commodityInfo replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:0]];
		[manifest replaceObjectAtIndex:i withObject:commodityInfo];
	}
	//
	NSArray* cargoArray = [[NSArray arrayWithArray:cargo] retain];  // retain
	//
	// step through the cargo pods adding in the quantities
	//
	for (i = 0; i < [cargoArray count]; i++)
	{
		NSMutableArray* commodityInfo;
		int co_type, co_amount, quantity;

		co_type = [(ShipEntity *)[cargoArray objectAtIndex:i] getCommodityType];
		co_amount = [(ShipEntity *)[cargoArray objectAtIndex:i] getCommodityAmount];

		//NSLog(@"unloading a %@ with %@", [(ShipEntity *)[cargoArray objectAtIndex:i] name], [universe describeCommodity:co_type amount:co_amount]);

		commodityInfo = (NSMutableArray *)[manifest objectAtIndex:co_type];
		quantity =  [(NSNumber *)[commodityInfo objectAtIndex:MARKET_QUANTITY] intValue] + co_amount;
		
		[commodityInfo replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:quantity]]; // enter the adjusted quantity
	}
	[shipCommodityData release];
	shipCommodityData = [[NSArray arrayWithArray:manifest] retain];
	[manifest release]; // release, done
	[cargoArray release]; // release, done
	for (i = 0; i < [cargoArray count]; i++)	// recycle these entities rather than leave them to be garbage collected
	{
		[(Entity *)[cargoArray objectAtIndex:i] setStatus:STATUS_DEAD];
		[universe recycleOrDiscard:(Entity *)[cargoArray objectAtIndex:i]];
	}
	[cargo removeAllObjects];   // empty the hold
}

- (void) loadCargoPods
{
	/* loads commodities from the ships manifest into individual cargo pods */
	int i,j;
	NSMutableArray* localMarket = [docked_station localMarket]; 
	NSMutableArray* manifest = [[NSMutableArray arrayWithArray:shipCommodityData] retain];  // retain
	for (i = 0; i < [manifest count]; i++)
	{
		NSMutableArray* commodityInfo = [[NSMutableArray arrayWithArray:(NSArray *)[manifest objectAtIndex:i]] retain];  // retain
		int quantity =  [(NSNumber *)[commodityInfo objectAtIndex:MARKET_QUANTITY] intValue];
		int units =		[universe unitsForCommodity:i];
		
		if (quantity > 0)
		{
			//NSLog(@"loading containers with %@",[universe describeCommodity:i amount:quantity]);

			if (units != UNITS_TONS)
			{
				// put it all in one container
				ShipEntity* container = [universe getShip:@"barrel"];   // retain count = 1
				[container setUniverse:universe];
				[container setScanClass: CLASS_CARGO];
				[container setCommodity:i andAmount:quantity];
				[cargo addObject:container];
				[container release];
			}
			else
			{
				// put each ton in a separate container
				for (j = 0; j < quantity; j++)
				{
					ShipEntity* container = [universe getShip:@"barrel"];   // retain count = 1
					[container setUniverse:universe];
					[container setScanClass: CLASS_CARGO];
					[container setStatus:STATUS_IN_HOLD];
					[container setCommodity:i andAmount:1];
					[cargo addObject:container];
					[container release];
				}
			}
			[commodityInfo setArray:(NSArray *)[localMarket objectAtIndex:i]];
			[commodityInfo replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:0]];
			[manifest replaceObjectAtIndex:i withObject:[NSArray arrayWithArray:commodityInfo]];
		}
		[commodityInfo release]; // release, done
	}
	[shipCommodityData release];
	shipCommodityData = [[NSArray arrayWithArray:manifest] retain];
	[manifest release]; // release, done
}

- (int) random_factor
{
	return market_rnd;
}

- (Random_Seed) galaxy_seed
{
	return galaxy_seed;
}

- (NSPoint) galaxy_coordinates
{
	return galaxy_coordinates;
}

- (NSDictionary *) commanderDataDictionary
{
	NSMutableDictionary *result = [[NSMutableDictionary alloc] initWithCapacity:20]; // retained

	NSString *gal_seed = [NSString stringWithFormat:@"%d %d %d %d %d %d",galaxy_seed.a, galaxy_seed.b, galaxy_seed.c, galaxy_seed.d, galaxy_seed.e, galaxy_seed.f]; 
	NSString *gal_coords = [NSString stringWithFormat:@"%d %d",(int)galaxy_coordinates.x,(int)galaxy_coordinates.y]; 

	[result setObject:gal_seed		forKey:@"galaxy_seed"];
	[result setObject:gal_coords	forKey:@"galaxy_coordinates"];
	//
	[result setObject:player_name			forKey:@"player_name"];
	//[result setObject:[self valueForKey:@"galaxy_coordinates"]  forKey:@"galaxy_coordinates"];
	[result setObject:[self valueForKey:@"credits"]				forKey:@"credits"];
	[result setObject:[self valueForKey:@"fuel"]				forKey:@"fuel"];
	//
	[result setObject:[self valueForKey:@"galaxy_number"]		forKey:@"galaxy_number"];
	[result setObject:[self valueForKey:@"forward_weapon"]		forKey:@"forward_weapon"];
	[result setObject:[self valueForKey:@"aft_weapon"]			forKey:@"aft_weapon"];
	[result setObject:[self valueForKey:@"port_weapon"]			forKey:@"port_weapon"];
	[result setObject:[self valueForKey:@"starboard_weapon"]	forKey:@"starboard_weapon"];
	//
	[result setObject:[self valueForKey:@"max_cargo"]				forKey:@"max_cargo"];
	//[result setObject:[self valueForKey:@"cargo"]					forKey:@"cargo"];
	[result setObject:shipCommodityData		forKey:@"shipCommodityData"];
	[result setObject:[self valueForKey:@"has_ecm"]					forKey:@"has_ecm"];
	[result setObject:[self valueForKey:@"has_scoop"]				forKey:@"has_scoop"];
	[result setObject:[self valueForKey:@"has_energy_bomb"]			forKey:@"has_energy_bomb"];
	[result setObject:[self valueForKey:@"has_energy_unit"]			forKey:@"has_energy_unit"];
	[result setObject:[self valueForKey:@"energy_unit"]				forKey:@"energy_unit"];
	[result setObject:[self valueForKey:@"has_docking_computer"]	forKey:@"has_docking_computer"];
	[result setObject:[self valueForKey:@"has_galactic_hyperdrive"] forKey:@"has_galactic_hyperdrive"];
	[result setObject:[self valueForKey:@"has_escape_pod"]			forKey:@"has_escape_pod"];
	//
	[result setObject:[self valueForKey:@"missiles"]				forKey:@"missiles"];
	missile_status = (missiles > 0) ? MISSILE_STATUS_SAFE : MISSILE_STATUS_NONE;
	//
	[result setObject:[self valueForKey:@"legal_status"]			forKey:@"legal_status"];
	[result setObject:[self valueForKey:@"market_rnd"]				forKey:@"market_rnd"];
	[result setObject:[self valueForKey:@"ship_kills"]				forKey:@"ship_kills"];
	[result setObject:[self valueForKey:@"saved"]					forKey:@"saved"];
	
	// mission variables
	if (mission_variables)
		[result setObject:mission_variables forKey:@"mission_variables"];
	
	//speech
	[result setObject:[self valueForKey:@"speech_on"] forKey:@"speech_on"];
	
	//NSLog(@"Player Dictionary :\n%@",[result description]);

	return [NSDictionary dictionaryWithDictionary:[result autorelease]];
}
- (void) setCommanderDataFromDictionary:(NSDictionary *) dic
{
	NSArray *seed_vals = [(NSString *)[dic objectForKey:@"galaxy_seed"] componentsSeparatedByString:@" "];
	NSArray *coord_vals = [(NSString *)[dic objectForKey:@"galaxy_coordinates"] componentsSeparatedByString:@" "];
	
	galaxy_seed.a = (unsigned char)[(NSString *)[seed_vals objectAtIndex:0] intValue];
	galaxy_seed.b = (unsigned char)[(NSString *)[seed_vals objectAtIndex:1] intValue];
	galaxy_seed.c = (unsigned char)[(NSString *)[seed_vals objectAtIndex:2] intValue];
	galaxy_seed.d = (unsigned char)[(NSString *)[seed_vals objectAtIndex:3] intValue];
	galaxy_seed.e = (unsigned char)[(NSString *)[seed_vals objectAtIndex:4] intValue];
	galaxy_seed.f = (unsigned char)[(NSString *)[seed_vals objectAtIndex:5] intValue];

	galaxy_coordinates.x = (unsigned char)[(NSString *)[coord_vals objectAtIndex:0] intValue];
	galaxy_coordinates.y = (unsigned char)[(NSString *)[coord_vals objectAtIndex:1] intValue];

	if (player_name)	[player_name release];
	player_name = [(NSString *)[dic objectForKey:@"player_name"] retain];
	//[self setValue:[dic objectForKey:@"galaxy_coordinates"] forKey:@"galaxy_coordinates"];
	[self setValue:[dic objectForKey:@"credits"]			forKey:@"credits"];
	[self setValue:[dic objectForKey:@"fuel"]				forKey:@"fuel"];
	//
	[self setValue:[dic objectForKey:@"galaxy_number"]		forKey:@"galaxy_number"];
	[self setValue:[dic objectForKey:@"forward_weapon"]		forKey:@"forward_weapon"];
	[self setValue:[dic objectForKey:@"aft_weapon"]			forKey:@"aft_weapon"];
	[self setValue:[dic objectForKey:@"port_weapon"]		forKey:@"port_weapon"];
	[self setValue:[dic objectForKey:@"starboard_weapon"]   forKey:@"starboard_weapon"];
	//
	[self setValue:[dic objectForKey:@"max_cargo"]					forKey:@"max_cargo"];
	//[self setValue:[dic objectForKey:@"cargo"]						forKey:@"cargo"];
	if (shipCommodityData)  [shipCommodityData release];
	shipCommodityData = [(NSArray *)[dic objectForKey:@"shipCommodityData"] retain];
	[self setValue:[dic objectForKey:@"has_ecm"]					forKey:@"has_ecm"];
	[self setValue:[dic objectForKey:@"has_scoop"]					forKey:@"has_scoop"];
	[self setValue:[dic objectForKey:@"has_energy_bomb"]			forKey:@"has_energy_bomb"];
	//
	[self setValue:[dic objectForKey:@"has_energy_unit"]			forKey:@"has_energy_unit"];
	energy_unit = (has_energy_unit) ? ENERGY_UNIT_NORMAL : ENERGY_UNIT_NONE;	// set default
	if ([dic objectForKey:@"energy_unit"])
		[self setValue:[dic objectForKey:@"energy_unit"]			forKey:@"energy_unit"]; // load specials
	//
	[self setValue:[dic objectForKey:@"has_docking_computer"]		forKey:@"has_docking_computer"];
	[self setValue:[dic objectForKey:@"has_galactic_hyperdrive"]	forKey:@"has_galactic_hyperdrive"];
	[self setValue:[dic objectForKey:@"has_escape_pod"]				forKey:@"has_escape_pod"];
	//
	[self setValue:[dic objectForKey:@"missiles"]				forKey:@"missiles"];
	[self setValue:[dic objectForKey:@"legal_status"]			forKey:@"legal_status"];
	[self setValue:[dic objectForKey:@"market_rnd"]				forKey:@"market_rnd"];
	[self setValue:[dic objectForKey:@"ship_kills"]				forKey:@"ship_kills"];
	[self setValue:[dic objectForKey:@"saved"]					forKey:@"saved"];
	
	// mission_variables
	[mission_variables removeAllObjects];
	if ([dic objectForKey:@"mission_variables"])
		[mission_variables addEntriesFromDictionary:(NSDictionary *)[dic objectForKey:@"mission_variables"]];
	
	// speech
	if ([dic objectForKey:@"speech_on"])
		[self setValue:[dic objectForKey:@"speech_on"]  forKey:@"speech_on"];
	
	//
	//  things...
	//
	system_seed = [universe findSystemAtCoords:galaxy_coordinates withGalaxySeed:galaxy_seed];
	target_system_seed = system_seed;
	cursor_coordinates = galaxy_coordinates;
	[universe removeAllEntitiesExceptPlayer:NO];	// reset the universe outside the station
	[universe setSystemTo:system_seed];
}

/////////////////////////////////////////////////////////////

- (id) init
{    
    self = [super init];
	//
	[self set_up];
	//
    return self;
}

static BOOL galactic_witchjump;
- (void) set_up
{    
	Random_Seed gal_seed = {0x4a, 0x5a, 0x48, 0x02, 0x53, 0xb7};
	//
	script = [[ResourceManager dictionaryFromFilesNamed:@"script.plist" inFolder:@"Config" andMerge:YES] retain];
	mission_variables =[[NSMutableDictionary dictionaryWithCapacity:16] retain];
	script_time = 0.0;
	script_time_check = SCRIPT_TIMER_INTERVAL;
	script_time_interval = SCRIPT_TIMER_INTERVAL;
	//
	speech_on = NO;
	//
	mouse_control_on = NO;
	//
	docking_music_on = YES;
	//
	name = [[NSString stringWithString:@"Player"] retain];
	rolling = NO;
	pitching = NO;
	galactic_witchjump = NO;
	//
	flight_speed =		0.0;
	max_flight_speed =  160.0;
	max_flight_roll =   2.0;
	max_flight_pitch =  1.0;
	//
	// control factors
	//
	speed_delta =		80.0;
	thrust =			speed_delta;
	roll_delta =		2.0 * max_flight_roll;
	pitch_delta =		2.0 * max_flight_pitch;
    //
    displayListName =   0;
    //
    status =			STATUS_TEST;
	forward_shield =	PLAYER_MAX_FORWARD_SHIELD;
	aft_shield =		PLAYER_MAX_AFT_SHIELD;
	energy =			PLAYER_MAX_ENERGY;
	weapon_temp =		0.0;
	cabin_temp =		60.0;
	//
	game_over =				NO;
	docked =				NO;
	finished =				NO;
	bomb_detonated =		NO;
	has_docking_computer =  NO;
	autopilot_engaged =		NO;
	hyperspeed_engaged =	NO;
	hyperspeed_locked =		NO;
	//
	ecm_in_operation =		NO;
	ecm_start_time =		0.0;
	//
	witchspaceCountdown = 0.0;
	//
	missile_status = MISSILE_STATUS_SAFE;
	//
	// release sounds
	//
    if (beepSound)				[beepSound release];
    if (boopSound)				[boopSound release];
    if (weaponSound)			[weaponSound release];
    if (damageSound)			[damageSound release];
    if (scrapeDamageSound)		[scrapeDamageSound release];
    if (destructionSound)		[destructionSound release];
    if (breakPatternSound)		[breakPatternSound release];
	//
	if (themeMusic)				[themeMusic release];
	//
	// allocate sounds
	//
	beepSound =			[[ResourceManager soundNamed:@"beep.aiff" inFolder:@"Sounds"] retain];
	boopSound =			[[ResourceManager soundNamed:@"boop.aiff" inFolder:@"Sounds"] retain];
	weaponSound =		[[ResourceManager soundNamed:@"laser.aiff" inFolder:@"Sounds"] retain];
	missileSound =		[[ResourceManager soundNamed:@"missile.aiff" inFolder:@"Sounds"] retain];
	damageSound =		[[ResourceManager soundNamed:@"hit.aiff" inFolder:@"Sounds"] retain];
	scrapeDamageSound = [[ResourceManager soundNamed:@"hullbang.aiff" inFolder:@"Sounds"] retain];
	destructionSound =  [[ResourceManager soundNamed:@"bigbang.aiff" inFolder:@"Sounds"] retain];
	breakPatternSound = [[ResourceManager soundNamed:@"breakpattern.aiff" inFolder:@"Sounds"] retain];
	ecmSound =			[[ResourceManager soundNamed:@"ecm.aiff" inFolder:@"Sounds"] retain];
	buySound =			[[ResourceManager soundNamed:@"buy.aiff" inFolder:@"Sounds"] retain];
	sellSound =			[[ResourceManager soundNamed:@"sell.aiff" inFolder:@"Sounds"] retain];
	warningSound =		[[ResourceManager soundNamed:@"warning.aiff" inFolder:@"Sounds"] retain];
	//
	themeMusic =		[[ResourceManager movieFromFilesNamed:@"OoliteTheme.mp3" inFolder:@"Music"] retain];
	dockingMusic =		[[ResourceManager movieFromFilesNamed:@"BlueDanube.mp3" inFolder:@"Music"] retain];
	//
	collision_radius =  50.0;
	//
	[self setModel:PLAYER_MODEL];
	//
	shot_time =			0.0;
	shot_counter =		0;
    //
	max_cargo =		25;
	//
	if (shipAI)		[shipAI release];
	shipAI = [[AI alloc] initWithStateMachine:AI_DOCKING_COMPUTER andState:@"GLOBAL"]; // alloc retains dealloc'd by ShipEntity
	[shipAI setOwner:self];
	//
	
	// player commander data
	//
	player_name =			[[NSString alloc] initWithString:@"Jameson"];  // alloc retains
	galaxy_coordinates =	NSMakePoint(0x14,0xAD);
	galaxy_seed =			gal_seed;
	credits =				1000;
	fuel =					PLAYER_MAX_FUEL;
	fuel_accumulator =		0.0;
	
	galaxy_number =			0;
	forward_weapon =		WEAPON_PULSE_LASER;
	aft_weapon =			WEAPON_NONE;
	port_weapon =			WEAPON_NONE;
	starboard_weapon =		WEAPON_NONE;
	
	max_cargo =				20;

	shipCommodityData = [(NSArray *)[(NSDictionary *)[ResourceManager dictionaryFromFilesNamed:@"commodities.plist" inFolder:@"Config" andMerge:YES] objectForKey:@"default"] retain];
	
	has_ecm =					NO;
	has_scoop =					NO;
	has_energy_bomb =			NO;
	has_energy_unit =			NO;
	has_docking_computer =		NO;
	has_galactic_hyperdrive =   NO;
	has_escape_pod =			NO;
	
	missiles =				3;
	legal_status =			0;

	market_rnd =			0;
	ship_kills =			0;
	saved =					NO;
	cursor_coordinates = galaxy_coordinates;
	
}

- (void) dealloc
{
    if (script)					[script release];
    if (mission_variables)		[mission_variables release];
	if (lastTextKey)			[lastTextKey release];
	
    if (beepSound)				[beepSound release];
    if (boopSound)				[boopSound release];
    if (weaponSound)			[weaponSound release];
    if (missileSound)			[missileSound release];
    if (damageSound)			[damageSound release];
    if (scrapeDamageSound)		[scrapeDamageSound release];
    if (destructionSound)		[destructionSound release];
    if (breakPatternSound)		[breakPatternSound release];
    if (ecmSound)				[ecmSound release];
    if (buySound)				[buySound release];
    if (sellSound)				[sellSound release];
    if (warningSound)			[warningSound release];

    if (themeMusic)				[themeMusic release];
    if (dockingMusic)			[dockingMusic release];

    if (player_name)			[player_name release];
    if (shipCommodityData)		[shipCommodityData release];

    [super dealloc];
}


- (void) warnAboutHostiles
{
	// make a warningSound
	//NSLog(@"player warned about hostiles!");
	if (![warningSound isPlaying])
		[warningSound play];
}

- (BOOL) canCollide
{
	switch (status)
	{
		case STATUS_IN_FLIGHT :
		case STATUS_AUTOPILOT_ENGAGED :
			return YES;
			break;
		
		case STATUS_DEAD :
		case STATUS_ESCAPE_SEQUENCE :
		default :
			return NO;
	}
}

- (NSComparisonResult) compareZeroDistance:(Entity *)otherEntity;
{
	return NSOrderedDescending;  // always the most near
}

- (void) update:(double) delta_t
{
	// update timers
	//
	shot_time += delta_t;
	script_time += delta_t;
		
	// scripting
	if (script_time > script_time_check)
	{
		[self checkScript];
		script_time_check += script_time_interval;
	}
	
	// deal with collisions
	//
	[self manageCollisions];

	[self pollControls:delta_t];
	
	if ((status == STATUS_DEMO)&&(gui_screen != GUI_SCREEN_INTRO1)&&(gui_screen != GUI_SCREEN_INTRO2)&&(gui_screen != GUI_SCREEN_MISSION))
		[self setGuiToIntro1Screen];	//set up demo mode
	
	if ((status == STATUS_AUTOPILOT_ENGAGED)||(status == STATUS_ESCAPE_SEQUENCE))
	{
		[super update:delta_t];
		[self doBookkeeping:delta_t];
		return;
	}
	
	// do flight routines
	//
	[self applyRoll:delta_t*flight_roll andClimb:delta_t*flight_pitch];
	[self moveForward:delta_t*flight_speed];
	
	if (status == STATUS_IN_FLIGHT)
	{
		[self doBookkeeping:delta_t];
	}
	
	if (status == STATUS_LAUNCHING)
	{
		if ([universe breakPatternOver])
		{
			status = STATUS_IN_FLIGHT;
			
			// time to check the script!
			[self checkScript];
		}
	}
	
	if (status == STATUS_WITCHSPACE_COUNTDOWN)
	{
		[self doBookkeeping:delta_t];
		witchspaceCountdown -= delta_t;
		if (witchspaceCountdown < 0.0)  witchspaceCountdown = 0.0;
		if (galactic_witchjump)
			[universe displayMessage:[NSString stringWithFormat:MS_GAL_WITCHSPACE_F, witchspaceCountdown] forCount:1.0];
		else
			[universe displayMessage:[NSString stringWithFormat:MS_WITCHSPACE_SF, [universe generateSystemName:target_system_seed], witchspaceCountdown] forCount:1.0];
		if (witchspaceCountdown == 0.0)
		{
			if (galactic_witchjump)
				[self enterGalacticWitchspace];
			else
				[self enterWitchspace];
		}
	}
	
	if (status == STATUS_EXITING_WITCHSPACE)
	{
		if ([universe breakPatternOver])
		{
			status = STATUS_IN_FLIGHT;

			[self resetScriptTimer];	// reset the in-system timer
			// time to check the script!
			[self checkScript];
			// announce arrival
			[universe addMessage:[NSString stringWithFormat:@" %@. ",[universe generateSystemName:system_seed]] forCount:3.0];
		}
	}
	
	if (status == STATUS_DOCKING)
	{
		if ([universe breakPatternOver])
		{
			[self docked];
		
			// time to check the script!
			[self checkScript];
		}
	}
	
	if ((status == STATUS_DEAD)&&(shot_time > 30.0))
	{
		[universe game_over];				//  we restart the universe
	}
	
	if (missile_status == MISSILE_STATUS_ARMED)
	{
		int first_target_id = [universe getFirstEntityTargettedFromEntity:self inView:[universe viewDir]];
		if (first_target_id != NO_TARGET)
		{
			Entity *first_target = [universe entityForUniversalID:first_target_id];
			if ([first_target isKindOfClass:[ShipEntity class]])
			{
				primaryTarget = first_target_id;
				missile_status = MISSILE_STATUS_TARGET_LOCKED;
				//[universe displayMessage:[NSString stringWithFormat:@"Missile locked onto %@.", [(ShipEntity *)first_target name]] forCount:4.5];
				[universe addMessage:[NSString stringWithFormat:@"Missile locked onto %@.", [(ShipEntity *)first_target name]] forCount:4.5];
				[beepSound play];
			}
		}
	}
	
	if (missile_status == MISSILE_STATUS_TARGET_LOCKED)
	{
		if ([[universe entityForUniversalID:primaryTarget] getZeroDistance] > SCANNER_MAX_RANGE2)
		{
			//[universe displayMessage:@"Target Lost." forCount:3.0];
			[universe addMessage:@"Target Lost." forCount:3.0];
			[boopSound play];
			missile_status = MISSILE_STATUS_SAFE;
		}
	}
}

- (void) doBookkeeping:(double) delta_t
{
		// Bookeeping;
		//
		if (weapon_temp > 0.0)
		{
			weapon_temp -= WEAPON_COOLING_FACTOR * delta_t;
			if (weapon_temp < 0.0)
				weapon_temp = 0.0;
		}
		if (energy < PLAYER_MAX_ENERGY)
		{
			double energy_multiplier = 1.0 + 0.1 * energy_unit; // 1.5x recharge with normal energy unit, 2x with naval!
			energy += ENERGY_RECHARGE_FACTOR * energy_multiplier * delta_t;
			if (energy > PLAYER_MAX_ENERGY)
				energy = PLAYER_MAX_ENERGY;
		}
		if (forward_shield < PLAYER_MAX_FORWARD_SHIELD)
		{
			forward_shield += SHIELD_RECHARGE_FACTOR * delta_t;
			energy -= SHIELD_RECHARGE_FACTOR * delta_t;
			if (forward_shield > PLAYER_MAX_FORWARD_SHIELD)
			{
				energy += forward_shield - PLAYER_MAX_FORWARD_SHIELD;
				forward_shield = PLAYER_MAX_FORWARD_SHIELD;
			}
		}
		if (aft_shield < PLAYER_MAX_AFT_SHIELD)
		{
			aft_shield += SHIELD_RECHARGE_FACTOR * delta_t;
			energy -= SHIELD_RECHARGE_FACTOR * delta_t;
			if (aft_shield > PLAYER_MAX_FORWARD_SHIELD)
			{
				energy += aft_shield - PLAYER_MAX_AFT_SHIELD;
				aft_shield = PLAYER_MAX_AFT_SHIELD;
			}
		}
		if (ecm_in_operation)
		{
			if (energy > 0.0)
				energy -= ECM_ENERGY_DRAIN_FACTOR * delta_t;		// drain energy because of the ECM
			else
			{
				ecm_in_operation = NO;
				[ecmSound stop];
				//[universe displayMessage:@"ECM system deactivated (no energy left)." forCount:3.0];
				[universe addMessage:@"ECM system deactivated (no energy left)." forCount:3.0];
			}
			if ([universe getTime] > ecm_start_time + ECM_DURATION)
			{
				ecm_in_operation = NO;
			}
		}
		if ((has_scoop)&&(cabin_temp > 224.0)&&(fuel < 70))
		{
			fuel_accumulator += delta_t * flight_speed * 0.025;
			while (fuel_accumulator > 1.0)
			{
				fuel ++;
				fuel_accumulator -= 1.0;
			}
			if (fuel > 70)	fuel = 70;
			[universe displayMessage:@"Fuel Scoop Active" forCount:1.0];
		}
		if (cabin_temp > PLAYER_MAX_CABIN_TEMP)
		{
			[self getDestroyed];
		}
		if ((missile_status == MISSILE_STATUS_TARGET_LOCKED)&&(![universe entityForUniversalID:primaryTarget]))
		{
			// targetting 'off' here
			missile_status = MISSILE_STATUS_SAFE;
			if (![boopSound isPlaying])
				[boopSound play];
			//[universe displayMessage:@"Target lost." forCount:2.0];
			[universe addMessage:@"Target lost." forCount:2.0];
		}
		if ((status == STATUS_ESCAPE_SEQUENCE)&&(shot_time > ESCAPE_SEQUENCE_TIME))
		{
			[self setTargetToStation];
			if ([self getPrimaryTarget])
			{
				[universe setViewDirection:VIEW_FORWARD];
				[self enterDock:(StationEntity *)[self getPrimaryTarget]];
			}
		}
}

- (void) applyRoll:(GLfloat) roll1 andClimb:(GLfloat) climb1
{	
	quaternion_rotate_about_z( &q_rotation, -roll1);
	quaternion_rotate_about_x( &q_rotation, -climb1);
	
    quaternion_normalise(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);

	v_right.x = rotMatrix[0];
	v_right.y = rotMatrix[4];
	v_right.z = rotMatrix[8];

	v_up.x = rotMatrix[1];
	v_up.y = rotMatrix[5];
	v_up.z = rotMatrix[9];

	v_forward.x = rotMatrix[2];
	v_forward.y = rotMatrix[6];
	v_forward.z = rotMatrix[10];
}

- (void) moveForward:(double) amount
{
	position.x += amount * v_forward.x;
	position.y += amount * v_forward.y;
	position.z += amount * v_forward.z;
}


- (void) drawEntity:(BOOL) immediate :(BOOL) translucent
{
    // do nothing in threedee- the player entity isn't drawn
    //
}


- (BOOL) massLocked
{
	int			i;
	NSArray *   entities = [[universe getAllEntities] retain];
	BOOL		mass_locked = NO;
	
	PlanetEntity* planet = [self findNearestLargeBody];
	if (planet)
	{
		double dist =   [planet getZeroDistance];
		double rad =	[planet collisionRadius];
		double factor = ([planet getPlanetType] == PLANET_TYPE_SUN) ? 1.44 : 4.0;
		if (dist < rad*rad*factor)
		{
			mass_locked = YES;
			NSLog(@"DEBUG Mass-locked by planet of radius :%.0f at distance %.0f",rad,sqrt(dist));
		}
	}
	
	if (!mass_locked)
	{
		for (i = 1; i < [entities count]; i++)
		{
			Entity* thing = [(Entity *)[entities objectAtIndex:i] retain];;
			double dist =   [thing getZeroDistance];
			if ((dist < SCANNER_MAX_RANGE2)&&([thing isKindOfClass:[ShipEntity class]])) // within 25 km
			{
				if ([thing scanClass] != CLASS_CARGO)
				{
					mass_locked = YES;
					NSLog(@"DEBUG Mass-locked by %@ %d",[(ShipEntity *)thing name],[thing universal_id]);
					[thing release];
					break;
				}
			}
			[thing release];
		}		
	}
	
	[entities release];
	return mass_locked;
}

- (BOOL) atHyperspeed
{
	return hyperspeed_engaged;
}

- (Vector) velocityVector
{
	Vector result = v_forward;
	result.x *= flight_speed;
	result.y *= flight_speed;
	result.z *= flight_speed;
	return result;
}

//
//			dial routines = all return 0.0 .. 1.0 or -1.0 .. 1.0
//
- (double) dial_roll
{
	return flight_roll/max_flight_roll;
}
- (double) dial_pitch
{
	return flight_pitch/max_flight_pitch;
}
- (double) dial_speed
{
	if (flight_speed < max_flight_speed)
		return flight_speed/max_flight_speed;
	else
		return 1.0;
}

- (double) dial_forward_shield
{
	return forward_shield / PLAYER_MAX_FORWARD_SHIELD;
}
- (double) dial_aft_shield
{
	return aft_shield / PLAYER_MAX_AFT_SHIELD;
}

- (double) dial_energy
{
	return energy / PLAYER_MAX_ENERGY;
}

- (double) dial_fuel
{
	return fuel / PLAYER_MAX_FUEL;
}
- (double) dial_cabin_temp
{
	return cabin_temp / PLAYER_MAX_CABIN_TEMP;
}
- (double) dial_weapon_temp
{
	return weapon_temp / PLAYER_MAX_WEAPON_TEMP;
}
- (double) dial_altitude
{
	PlanetEntity	*planet =   [universe planet];
	PlanetEntity	*sun =		[universe sun];
	double  planet_zd = [planet getZeroDistance];
	double  sun_zd = [sun getZeroDistance];
	double alt = (planet_zd < sun_zd) ? (sqrt(planet_zd) - [planet collisionRadius]) : (sqrt(sun_zd) - [sun collisionRadius]);
	alt /= PLAYER_DIAL_MAX_ALTITUDE;
	
	if (alt > 1.0)
		alt = 1.0;
	if (alt < 0.0)
		alt = 0.0;
	
	// do sun-skimming check here...
	cabin_temp = 60.0;
	if (sun_zd < planet_zd)
		cabin_temp += PLAYER_MAX_CABIN_TEMP * (1.0 - alt);
	
	return alt;
}

- (int) dial_missiles
{
	return missiles;
}
- (int) dial_missile_status
{
	if (missiles == 0)
		missile_status = MISSILE_STATUS_NONE;
	return missile_status;
}

- (int) scanAlertCondition
{
	int i;
	alert_condition = ALERT_CONDITION_GREEN;
	
	if (status == STATUS_DOCKED)
	{
		alert_condition = ALERT_CONDITION_DOCKED;
	}
	else
	{
		NSArray* entList = [[universe getAllEntities] retain];
		for (i = 0; i < [entList count]; i++)
		{
			Entity* thing = (Entity *)[entList objectAtIndex:i];
			if (([thing isKindOfClass:[ShipEntity class]])&&([(ShipEntity *)thing getPrimaryTarget] == self))
				alert_condition = ALERT_CONDITION_YELLOW;
		}
		[entList release];
		if ((alert_condition == ALERT_CONDITION_YELLOW)&&((energy < max_energy)||(forward_shield < PLAYER_MAX_FORWARD_SHIELD)||(aft_shield < PLAYER_MAX_AFT_SHIELD)))
			alert_condition = ALERT_CONDITION_RED;
		if (energy < max_energy*0.25)
			alert_condition = ALERT_CONDITION_RED;
	}
	return alert_condition;
}

/////////////////////////////////////////////////////////////////////////


- (void) pollControls:(double) delta_t
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
		
	if (gameView)
	{
		// poll the gameView keyboard things
		[self pollApplicationControls]; // quit command-f etc.
		switch (status)
		{
			case	STATUS_WITCHSPACE_COUNTDOWN :
			case	STATUS_IN_FLIGHT :
				[self pollFlightControls:delta_t];
				break;
			
			case	STATUS_DEAD :
				[self pollGameOverControls:delta_t];
				break;
				
			case	STATUS_AUTOPILOT_ENGAGED :
				[self pollAutopilotControls:delta_t];
				break;
				
			case	STATUS_DOCKED :
				[self pollDockedControls:delta_t];
				break;
								
			case	STATUS_DEMO :
				[self pollDemoControls:delta_t];
				break;
								
			case	STATUS_ESCAPE_SEQUENCE :
			default :
				break;
		}
	}
}

- (void) pollFlightControls:(double) delta_t
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	//
	// arrow keys
	//
	if ([universe displayText])
		[self pollGuiArrowKeyControls:delta_t];
	else
		[self pollFlightArrowKeyControls:delta_t];
	//
	//  view keys
	//
	[self pollViewControls];
	
	if (![gameView allowingStringInput])
	{
		//
		// flight_speed ++ = 119 'w' -- = 115 's'
		//
		if (([gameView isDown:119])&&(!hyperspeed_engaged))
		{
			if (flight_speed < max_flight_speed)
				flight_speed += speed_delta * delta_t;
			if (flight_speed > max_flight_speed)
				flight_speed = max_flight_speed;
		}
		if (([gameView isDown:115])&&(!hyperspeed_engaged))
		{
			if (flight_speed > 0.0)
				flight_speed -= speed_delta * delta_t;
			if (flight_speed < 0.0)
				flight_speed = 0.0;
		}
		//
		if (([gameView isDown:106])&&(!hyperspeed_locked))		// 'j'
		{
			hyperspeed_locked = [self massLocked];
			if (!hyperspeed_locked)
			{
				hyperspeed_engaged = YES;
				if (flight_speed < max_flight_speed * HYPERSPEED_FACTOR)
					flight_speed += speed_delta * delta_t * HYPERSPEED_FACTOR;
				if (flight_speed > max_flight_speed * HYPERSPEED_FACTOR)
					flight_speed = max_flight_speed * HYPERSPEED_FACTOR;
			}
			else
			{
				if (![boopSound isPlaying])
					[boopSound play];
				//[universe displayMessage:@"Hyperspeed mass-locked." forCount:4.5];
				[universe addMessage:@"Hyperspeed mass-locked." forCount:4.5];
			}
		}
		else
		{
			if (flight_speed > max_flight_speed)
				flight_speed -= speed_delta * delta_t * HYPERSPEED_FACTOR;
			else
			{
				hyperspeed_engaged = NO;
				hyperspeed_locked = NO;
			}
		}
		//
		//  shoot 'a'
		//
		if ((([gameView isDown:97])||((mouse_control_on)&&([gameView isDown:gvMouseLeftButton])))&&(shot_time > weapon_reload_time))
		{
			if ([self fireMainWeapon])
			{
				if (weaponSound)
					[weaponSound play];
			}
		}
		//
		//  shoot 'm'   // launch missile
		//
		if (([gameView isDown:109])&&(missile_status == MISSILE_STATUS_TARGET_LOCKED))
		{
			// launch here
			if ([self fireMissile])
			{
				missile_status = (missiles > 0) ? MISSILE_STATUS_SAFE : MISSILE_STATUS_NONE;
				[missileSound play];
			}
		}
		//
		//  shoot 't'   // switch on missile targetting
		//
		if (([gameView isDown:116])&&(missile_status == MISSILE_STATUS_SAFE))
		{
			// targetting 'on' here
			missile_status = MISSILE_STATUS_ARMED;
			if (![beepSound isPlaying])
				[beepSound play];
			//[universe displayMessage:@"Missile armed." forCount:2.0];
			[universe addMessage:@"Missile armed." forCount:2.0];
		}
		//
		//  shoot 'u'   // disarm missile targetting
		//
		if (([gameView isDown:117])&&(missile_status != MISSILE_STATUS_SAFE)&&(missile_status != MISSILE_STATUS_NONE))
		{
			// targetting 'off' here
			missile_status = MISSILE_STATUS_SAFE;
			if (![boopSound isPlaying])
				[boopSound play];
			//[universe displayMessage:@"Missiles on safety." forCount:2.0];
			[universe addMessage:@"Missiles on safety." forCount:2.0];
		}
		//
		//  shoot 'e'   // ECM
		//
		if (([gameView isDown:101])&&(has_ecm))
		{
			if (!ecm_in_operation)
			{
				if ([self fireECM])
				{
					[ecmSound play];
					//[universe displayMessage:@"ECM system activated." forCount:3.0];
					[universe addMessage:@"ECM system activated." forCount:3.0];
				}
			}
		}
		//
		//  shoot 'tab'   // Energy bomb
		//
		if (([gameView isDown:9])&&(has_energy_bomb))
		{
			[self fireEnergyBomb];
			has_energy_bomb = NO;
		}
		//
		//  shoot 'escape'   // Escape pod launch
		//
		if (([gameView isDown:27])&&(has_escape_pod))
		{
			[self launchEscapeCapsule];
		}
		//
		//  shoot 'd'   // Dump Cargo
		//
		if (([gameView isDown:100])&&([cargo count] > 0))
		{
			if ([self dumpCargo] != CARGO_NOT_CARGO)
				[beepSound play];
		}
		//
		// autopilot 'c'
		//
		if (([gameView isDown:99])&&(has_docking_computer)&&(![beepSound isPlaying]))   // look for the 'c' key
		{
			if ([self checkForAegis] == AEGIS_IN_DOCKING_RANGE)
			{
				//if (primaryTarget)  [primaryTarget release];
				primaryTarget = NO_TARGET;
				autopilot_engaged = YES;
				status = STATUS_AUTOPILOT_ENGAGED;
				[shipAI setState:@"GLOBAL"];	// restart the AI
				[beepSound play];
				//[universe displayMessage:@"Autopilot engaged" forCount:6.0];
				[universe addMessage:@"Autopilot engaged" forCount:4.5];
				//
				if (docking_music_on)
				{
					if (IsMovieDone ([dockingMusic QTMovie]))
						GoToBeginningOfMovie ([dockingMusic QTMovie]);
					StartMovie ([dockingMusic QTMovie]);
				}
				//
			}
			else
			{
				if (![boopSound isPlaying])
					[boopSound play];
				//[universe displayMessage:@"Could not contact station for docking instructions" forCount:6.0];
				[universe addMessage:@"Could not contact station for docking instructions" forCount:4.5];
			}
		}
		//
		// autopilot 'D'
		//
		if (([gameView isDown:68])&&(has_docking_computer)&&(![beepSound isPlaying]))   // look for the 'D' key
		{
			if ([self checkForAegis] == AEGIS_IN_DOCKING_RANGE)
			{
				StationEntity *the_station = [universe station];
				if (the_station)
				{
					[universe setViewDirection:VIEW_FORWARD];
					[self enterDock:the_station];
				}
			}
			else
			{
				if (![boopSound isPlaying])
					[boopSound play];
				//[universe displayMessage:@"Could not contact station for docking instructions" forCount:6.0];
				[universe addMessage:@"Could not contact station for docking instructions" forCount:4.5];
			}
		}
		//
		// hyperspace 'h'
		//
		if ([gameView isDown:104])   // look for the 'h' key
		{
			int			dx = target_system_seed.d - galaxy_coordinates.x;
			int			dy = target_system_seed.b - galaxy_coordinates.y;
			//double		distance = 4.0*sqrt(dx*dx + dy*dy*0.25); 
			double		distance = distanceBetweenPlanetPositions(target_system_seed.d,target_system_seed.b,galaxy_coordinates.x,galaxy_coordinates.y); 
			BOOL		jumpOK = YES;
			
			if ((dx == 0)&&(dy == 0))
			{
				[boopSound play];
				//[universe displayMessage:@"No target system set." forCount:3.0];
				[universe addMessage:@"No target system set." forCount:3.0];
				jumpOK = NO;
			}
			
			if (10.0 * distance > fuel)
			{
				[boopSound play];
				//[universe displayMessage:@"Insufficent fuel for witchspace jump." forCount:3.0];
				[universe addMessage:@"Insufficent fuel for witchspace jump." forCount:3.0];
				jumpOK = NO;
			}
			
			if (status == STATUS_WITCHSPACE_COUNTDOWN)
				jumpOK = NO;
			
			if (jumpOK)
			{
				galactic_witchjump = NO;
				witchspaceCountdown = 15.0;
				status = STATUS_WITCHSPACE_COUNTDOWN;
				[beepSound play];
				[universe addMessage:[NSString stringWithFormat:MS_WITCHSPACE_SF, [universe generateSystemName:target_system_seed], witchspaceCountdown] forCount:1.0];
			}
		}
		//
		// Galactic hyperspace 'g'
		//
		if (([gameView isDown:103])&&(has_galactic_hyperdrive))// look for the 'g' key
		{
			BOOL	jumpOK = YES;
			
			if (status == STATUS_WITCHSPACE_COUNTDOWN)
				jumpOK = NO;
			
			if (jumpOK)
			{
				galactic_witchjump = YES;
				witchspaceCountdown = 15.0;
				status = STATUS_WITCHSPACE_COUNTDOWN;
				[beepSound play];
				[universe addMessage:[NSString stringWithFormat:MS_GAL_WITCHSPACE_F, witchspaceCountdown] forCount:1.0];
			}
		}
	}
	//
	//  text displays
	//
	[self pollGuiScreenControls];
	//
	//
	//
}

static  BOOL	f_key_pressed;
static  BOOL	m_key_pressed;
static  BOOL	taking_snapshot;
- (void) pollApplicationControls
{
	// does fullscreen / quit / snapshot
	//
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	//
	//  command-key controls
	//
	if (([gameView isCommandDown])&&([[gameView gameController] inFullScreenMode]))
	{
		if (([gameView isCommandDown])&&([gameView isDown:102]))   //  command f
		{
			[[gameView gameController] exitFullScreenMode];
			if (mouse_control_on)
				[universe addMessage:@"Mouse control off." forCount:3.0];
			mouse_control_on = NO;
		}
		//
		if (([gameView isCommandDown])&&([gameView isDown:113]))   //  command q
		{
			[[gameView gameController] pauseFullScreenModeToPerform:@selector(exitApp) onTarget:[gameView gameController]];
		}
	}
	//
	//  snapshot
	//
	if ([gameView isDown:42])   //  '*' key
	{
		if (!taking_snapshot)
		{
			taking_snapshot = YES;
			[gameView snapShot];
		}
	}
	else
	{
		taking_snapshot = NO;
	}
	//
	// FPS display
	//
	if ([gameView isDown:70])   //  'F' key
	{
		if (!f_key_pressed)
			[universe setDisplayFPS:![universe displayFPS]];
		f_key_pressed = YES;
	}
	else
	{
		f_key_pressed = NO;
	}
	//
	// Mouse control
	//
	if ([[gameView gameController] inFullScreenMode])
	{
		if ([gameView isDown:77])   //  'M' key
		{
			if (!m_key_pressed)
			{
				mouse_control_on = !mouse_control_on;
				if (mouse_control_on)
					[universe addMessage:@"Mouse control on." forCount:3.0];
				else
					[universe addMessage:@"Mouse control off." forCount:3.0];
			}
			m_key_pressed = YES;
		}
		else
		{
			m_key_pressed = NO;
		}
	}
}

- (void) pollFlightArrowKeyControls:(double) delta_t
{
	MyOpenGLView	*gameView = (MyOpenGLView *)[universe gameView];
	NSPoint			virtualStick =[gameView virtualJoystickPosition];
	double sensitivity = 2.0;
	virtualStick.x *= sensitivity;
	virtualStick.y *= sensitivity;
	
	rolling = NO;
	if (!mouse_control_on)
	{
		if ([gameView isDown:gvArrowKeyLeft])
		{
			if (flight_roll > 0.0)  flight_roll = 0.0;
			[self decrease_flight_roll:delta_t*roll_delta];
			rolling = YES;
		}
		if ([gameView isDown:gvArrowKeyRight])
		{
			if (flight_roll < 0.0)  flight_roll = 0.0;
			[self increase_flight_roll:delta_t*roll_delta];
			rolling = YES;
		}
	}
	else
	{
		double stick_roll = max_flight_roll * virtualStick.x;
		if (flight_roll < stick_roll)
		{
			[self increase_flight_roll:delta_t*roll_delta];
			if (flight_roll > stick_roll)
				flight_roll = stick_roll;
		}
		if (flight_roll > stick_roll)
		{
			[self decrease_flight_roll:delta_t*roll_delta];
			if (flight_roll < stick_roll)
				flight_roll = stick_roll;
		}
		rolling = (abs(virtualStick.x) > .10);
	}
	if (!rolling)
	{
		if (flight_roll > 0.0)
		{
			if (flight_roll > delta_t)	[self decrease_flight_roll:delta_t];
			else	flight_roll = 0.0;
		}
		if (flight_roll < 0.0)
		{
			if (flight_roll < -delta_t)   [self increase_flight_roll:delta_t];
			else	flight_roll = 0.0;
		}
	}

	pitching = NO;
	if (!mouse_control_on)
	{
		if ([gameView isDown:gvArrowKeyDown])
		{
			if (flight_pitch < 0.0)  flight_pitch = 0.0;
			[self increase_flight_pitch:delta_t*pitch_delta];
			pitching = YES;
		}
		if ([gameView isDown:gvArrowKeyUp])
		{
			if (flight_pitch > 0.0)  flight_pitch = 0.0;
			[self decrease_flight_pitch:delta_t*pitch_delta];
			pitching = YES;
		}
	}
	else
	{
		double stick_pitch = max_flight_pitch * virtualStick.y;
		if (flight_pitch < stick_pitch)
		{
			[self increase_flight_pitch:delta_t*roll_delta];
			if (flight_pitch > stick_pitch)
				flight_pitch = stick_pitch;
		}
		if (flight_pitch > stick_pitch)
		{
			[self decrease_flight_pitch:delta_t*roll_delta];
			if (flight_pitch < stick_pitch)
				flight_pitch = stick_pitch;
		}
		pitching = (abs(virtualStick.x) > .10);
	}
	if (!pitching)
	{
		if (flight_pitch > 0.0)
		{
			if (flight_pitch > delta_t)	[self decrease_flight_pitch:delta_t];
			else	flight_pitch = 0.0;
		}
		if (flight_pitch < 0.0)
		{
			if (flight_pitch < -delta_t)	[self increase_flight_pitch:delta_t];
			else	flight_pitch = 0.0;
		}
	}
}

static BOOL cursor_moving;
static BOOL disc_operation_in_progress;
static BOOL switching_resolution;
static BOOL wait_for_key_up;
static int searchStringLength;
- (void) pollGuiArrowKeyControls:(double) delta_t
{
	MyOpenGLView	*gameView = (MyOpenGLView *)[universe gameView];
	BOOL			moving = NO;
	double			cursor_speed = 10.0;
	GuiDisplayGen*  gui = [universe gui];
	
	[gameView allowStringInput:(gui_screen == GUI_SCREEN_LONG_RANGE_CHART)];

	switch (gui_screen)
	{
		case	GUI_SCREEN_LONG_RANGE_CHART :
			if ([[gameView typedString] length])
			{
				planetSearchString = [gameView typedString];
				//NSLog(@"searching for %@", planetSearchString);
				NSPoint search_coords = [universe findSystemCoordinatesWithPrefix:planetSearchString withGalaxySeed:galaxy_seed];
				if ((search_coords.x > 0.0)&&(search_coords.y > 0.0))
				{
					//NSLog(@"found somewhere");
					moving = ((cursor_coordinates.x != search_coords.x)||(cursor_coordinates.y != search_coords.y));
					//searchStringLength = [[gameView typedString] length];
					cursor_coordinates = search_coords;
				}
				else
				{
					[gameView resetTypedString];
				}
			}
			else
			{
				planetSearchString = nil;
			}
			moving |= (searchStringLength != [[gameView typedString] length]);
			searchStringLength = [[gameView typedString] length];
		case	GUI_SCREEN_SHORT_RANGE_CHART :
			if (status != STATUS_WITCHSPACE_COUNTDOWN)
			{
				if ([gameView isDown:gvArrowKeyLeft])
				{
					[gameView resetTypedString];
					cursor_coordinates.x -= cursor_speed*delta_t;
					if (cursor_coordinates.x < 0.0) cursor_coordinates.x = 0.0;
					moving = YES;
				}
				if ([gameView isDown:gvArrowKeyRight])
				{
					[gameView resetTypedString];
					cursor_coordinates.x += cursor_speed*delta_t;
					if (cursor_coordinates.x > 256.0) cursor_coordinates.x = 256.0;
					moving = YES;
				}
				if ([gameView isDown:gvArrowKeyDown])
				{
					[gameView resetTypedString];
					cursor_coordinates.y += cursor_speed*delta_t*2.0;
					if (cursor_coordinates.y > 256.0) cursor_coordinates.y = 256.0;
					moving = YES;
				}
				if ([gameView isDown:gvArrowKeyUp])
					{
					[gameView resetTypedString];
					cursor_coordinates.y -= cursor_speed*delta_t*2.0;
					if (cursor_coordinates.y < 0.0) cursor_coordinates.y = 0.0;
					moving = YES;
				}
				if ((cursor_moving)&&(!moving))
				{
					target_system_seed = [universe findSystemAtCoords:cursor_coordinates withGalaxySeed:galaxy_seed];
					cursor_coordinates.x = target_system_seed.d;
					cursor_coordinates.y = target_system_seed.b;
					if (gui_screen == GUI_SCREEN_LONG_RANGE_CHART) [self setGuiToLongRangeChartScreen];
					if (gui_screen == GUI_SCREEN_SHORT_RANGE_CHART) [self setGuiToShortRangeChartScreen];
				}
				cursor_moving = moving;
				if ((cursor_moving)&&(gui_screen == GUI_SCREEN_LONG_RANGE_CHART)) [self setGuiToLongRangeChartScreen]; // update graphics
				if (gui_screen == GUI_SCREEN_SHORT_RANGE_CHART) [self setGuiToShortRangeChartScreen]; // update graphics
			}
			break;

		case	GUI_SCREEN_OPTIONS :
			{
				int save_row =		GUI_ROW_OPTIONS_SAVE;
				int load_row =		GUI_ROW_OPTIONS_LOAD;
				//int options_row =   GUI_ROW_OPTIONS_OPTIONS;
				int display_row =   GUI_ROW_OPTIONS_DISPLAY;
				int speech_row =	GUI_ROW_OPTIONS_SPEECH;
				GameController  *controller = [universe gameController];
				NSArray *modes = [controller displayModes];
				
				if ([gameView isDown:gvArrowKeyDown])
				{
					if ([gui setSelectedRow:[gui selectedRow] + 1])
					{
						[gui click];
						[universe guiUpdated];
					}
				}
				if ([gameView isDown:gvArrowKeyUp])
				{
					if ([gui setSelectedRow:[gui selectedRow] - 1])
					{
						[gui click];
						[universe guiUpdated];
					}
				}
				
				if ([gameView isDown:13])   // 'enter' I hope!
				{
					if (([gui selectedRow] == save_row)&&(!disc_operation_in_progress))
					{
						disc_operation_in_progress = YES;
						if ([[universe gameController] inFullScreenMode])
							[[universe gameController] pauseFullScreenModeToPerform:@selector(savePlayer) onTarget:self];
						else
							[self savePlayer];
					}
					if (([gui selectedRow] == load_row)&&(!disc_operation_in_progress))
					{
						disc_operation_in_progress = YES;
						if ([[universe gameController] inFullScreenMode])
							[[universe gameController] pauseFullScreenModeToPerform:@selector(loadPlayer) onTarget:self];
						else
							[self loadPlayer];
					}
				}
				else
				{
					disc_operation_in_progress = NO;
				}
				
				if (([gui selectedRow] == display_row)&&(([gameView isDown:gvArrowKeyRight])||([gameView isDown:gvArrowKeyLeft]))&&(!switching_resolution))
				{
					int direction = ([gameView isDown:gvArrowKeyRight]) ? 1 : -1;
					int displayModeIndex = [controller indexOfCurrentDisplayMode];
					if (displayModeIndex == NSNotFound)
					{
						NSLog(@"***** couldn't find current display mode switching to basic 640x480");
						displayModeIndex = 0;
					}
					else
					{
						displayModeIndex = displayModeIndex + direction;
						if (displayModeIndex < 0)
							displayModeIndex = [modes count] - 1;
						if (displayModeIndex >= [modes count])
							displayModeIndex = 0;
					}
					NSDictionary	*mode = [modes objectAtIndex:displayModeIndex];
					int modeWidth = [[mode objectForKey: (NSString *)kCGDisplayWidth] intValue];
					int modeHeight = [[mode objectForKey: (NSString *)kCGDisplayHeight] intValue];
					int modeRefresh = [[mode objectForKey: (NSString *)kCGDisplayRefreshRate] intValue];
					[controller setDisplayWidth:modeWidth Height:modeHeight Refresh:modeRefresh];
					NSString	*displayModeString = [NSString stringWithFormat:@" Fullscreen: %d x %d at %d Hz ", modeWidth, modeHeight, modeRefresh];
					[gui click];
					{
						GuiDisplayGen* gui = [universe gui];
						int display_row =   GUI_ROW_OPTIONS_DISPLAY;
						[gui setText:displayModeString	forRow:display_row  align:GUI_ALIGN_CENTER];
						[gui generateDisplay];
						[universe guiUpdated];
					}
					switching_resolution = YES;
				}
				if ((![gameView isDown:gvArrowKeyRight])&&(![gameView isDown:gvArrowKeyLeft])&&(![gameView isDown:13]))
					switching_resolution = NO;
				
				if (([gui selectedRow] == speech_row)&&(([gameView isDown:gvArrowKeyRight])||([gameView isDown:gvArrowKeyLeft])))
				{
					GuiDisplayGen* gui = [universe gui];
					speech_on = !speech_on;
					[gui click];
					if (speech_on)
						[gui setText:@" Spoken messages: ON "	forRow:speech_row  align:GUI_ALIGN_CENTER];
					else
						[gui setText:@" Spoken messages: OFF "	forRow:speech_row  align:GUI_ALIGN_CENTER];
					[gui generateDisplay];
					[universe guiUpdated];
				}

			}
			break;
		
		case	GUI_SCREEN_EQUIP_SHIP :
			{
				if ([gameView isDown:gvArrowKeyDown])
				{
					if ([gui setSelectedRow:[gui selectedRow] + 1])
					{
						[gui click];
						[universe guiUpdated];
					}
				}
				if ([gameView isDown:gvArrowKeyUp])
				{
					if ([gui setSelectedRow:[gui selectedRow] - 1])
					{
						[gui click];
						[universe guiUpdated];
					}
				}
				if ([gameView isDown:13])   // 'enter' I hope!
				{
					if (!wait_for_key_up)
					{
						int item = [(NSString *)[gui selectedRowKey] intValue];
						NSString	*item_text = [gui selectedRowText];
						if ([item_text isEqual:FORWARD_FACING_STRING])
							chosen_weapon_facing = WEAPON_FACING_FORWARD;
						if ([item_text isEqual:AFT_FACING_STRING])
							chosen_weapon_facing = WEAPON_FACING_AFT;
						if ([item_text isEqual:PORT_FACING_STRING])
							chosen_weapon_facing = WEAPON_FACING_PORT;
						if ([item_text isEqual:STARBOARD_FACING_STRING])
							chosen_weapon_facing = WEAPON_FACING_STARBOARD;
						//NSLog(@"Try Buying Item %d",item);
						if ([self tryBuyingItem:item])
						{
							if ([buySound isPlaying])
								[buySound stop];
							[buySound play];
						}
						else
						{
							if ([boopSound isPlaying])
								[boopSound stop];
							[boopSound play];
						}
						wait_for_key_up = YES;
					}
				}
				else
				{
					wait_for_key_up = NO;
				}
			}
			break;

		case	GUI_SCREEN_MARKET :
			if (status == STATUS_DOCKED)
			{
				if ([gameView isDown:gvArrowKeyDown])
				{
					if ([gui setSelectedRow:[gui selectedRow] + 1])
					{
						[gui click];
						[universe guiUpdated];
					}
				}
				if ([gameView isDown:gvArrowKeyUp])
				{
					if ([gui setSelectedRow:[gui selectedRow] - 1])
					{
						[gui click];
						[universe guiUpdated];
					}
				}
				if (([gameView isDown:gvArrowKeyRight])||([gameView isDown:gvArrowKeyLeft])||([gameView isDown:13]))
				{
					if ([gameView isDown:gvArrowKeyRight])   // -->
					{
						if (!wait_for_key_up)
						{
							int item = [(NSString *)[gui selectedRowKey] intValue];
							//NSLog(@"Try Buying Commodity %d",item);
							if ([self tryBuyingCommodity:item])
								[self setGuiToMarketScreen];
							else
								[boopSound play];
							wait_for_key_up = YES;
						}
					}
					if ([gameView isDown:gvArrowKeyLeft])   // <--
					{
						if (!wait_for_key_up)
						{
							int item = [(NSString *)[gui selectedRowKey] intValue];
							//NSLog(@"Try Selling Commodity %d",item);
							if ([self trySellingCommodity:item])
								[self setGuiToMarketScreen];
							else
								[boopSound play];
							wait_for_key_up = YES;
						}
					}
					if ([gameView isDown:13])   // 'enter'
					{
						if (!wait_for_key_up)
						{
							int item = [(NSString *)[gui selectedRowKey] intValue];
							int yours =		[(NSNumber *)[(NSArray *)[shipCommodityData objectAtIndex:item] objectAtIndex:1] intValue];
							//NSLog(@"buy/sell all of item %d (you have %d)",item,yours);
							if (yours > 0)  // sell all you can
							{
								int i;
								for (i = 0; i < yours; i++)
									[self trySellingCommodity:item];
								//NSLog(@"... you sold %d.", yours);
								{
									if ([sellSound isPlaying])
										[sellSound stop];
									[sellSound play];
								}
								[self setGuiToMarketScreen];
							}
							else			// buy as much as possible
							{
								int amount_bought = 0;
								while ([self tryBuyingCommodity:item])
									amount_bought++;
								//NSLog(@"... you bought %d.", amount_bought);
								[self setGuiToMarketScreen];
								if (amount_bought == 0)
								{
									if ([boopSound isPlaying])
										[boopSound stop];
									[boopSound play];
								}
								else
								{
									if ([buySound isPlaying])
										[buySound stop];
									[buySound play];
								}
							}
							wait_for_key_up = YES;
						}
					}
				}
				else
				{
					wait_for_key_up = NO;
				}
			}
			break;

	}
	
	//
	// damp any rotations we entered with
	//
	if (flight_roll > 0.0)
	{
		if (flight_roll > delta_t)	[self decrease_flight_roll:delta_t];
		else	flight_roll = 0.0;
	}
	if (flight_roll < 0.0)
	{
		if (flight_roll < -delta_t)   [self increase_flight_roll:delta_t];
		else	flight_roll = 0.0;
	}
	if (flight_pitch > 0.0)
	{
		if (flight_pitch > delta_t)	[self decrease_flight_pitch:delta_t];
		else	flight_pitch = 0.0;
	}
	if (flight_pitch < 0.0)
	{
		if (flight_pitch < -delta_t)	[self increase_flight_pitch:delta_t];
		else	flight_pitch = 0.0;
	}
}

- (void) pollViewControls
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	//
	//  view keys
	//
	if (([gameView isDown:gvFunctionKey1])||([gameView isDown:gvNumberKey1]))
	{
		if ([universe displayText])
		{
			gui_screen = GUI_SCREEN_MAIN;
			[gameView allowStringInput:NO];
		}
		[universe setViewDirection:VIEW_FORWARD];
	}
	if (([gameView isDown:gvFunctionKey2])||([gameView isDown:gvNumberKey2]))
	{
		if ([universe displayText])
		{
			gui_screen = GUI_SCREEN_MAIN;
			[gameView allowStringInput:NO];
		}
		[universe setViewDirection:VIEW_AFT];
	}
	if (([gameView isDown:gvFunctionKey3])||([gameView isDown:gvNumberKey3]))
	{
		if ([universe displayText])
		{
			gui_screen = GUI_SCREEN_MAIN;
			[gameView allowStringInput:NO];
		}
		[universe setViewDirection:VIEW_PORT];
	}
	if (([gameView isDown:gvFunctionKey4])||([gameView isDown:gvNumberKey4]))
	{
		if ([universe displayText])
		{
			gui_screen = GUI_SCREEN_MAIN;
			[gameView allowStringInput:NO];
		}
		[universe setViewDirection:VIEW_STARBOARD];
	}
}

static BOOL switching_chart_screens;
- (void) pollGuiScreenControls
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	//
	//  text displays
	//
	if (([gameView isDown:gvFunctionKey5])||([gameView isDown:gvNumberKey5]))
	{
		if (gui_screen != GUI_SCREEN_STATUS)
		{
			[self setGuiToStatusScreen];
			[self checkScript];
		}
		[universe setDisplayText:YES];
	}
	
	if (([gameView isDown:gvFunctionKey6])||([gameView isDown:gvNumberKey6]))
	{
		if  (!switching_chart_screens)
		{
			switching_chart_screens = YES;
			if (gui_screen == GUI_SCREEN_SHORT_RANGE_CHART)
			{
				[self setGuiToLongRangeChartScreen];
				[universe setDisplayText:YES];
			}
			else
			{
				[self setGuiToShortRangeChartScreen];
				[universe setDisplayText:YES];
			}
		}
	}
	else
	{
		switching_chart_screens = NO;
	}
	
	if (([gameView isDown:gvFunctionKey7])||([gameView isDown:gvNumberKey7]))
	{
		if (gui_screen != GUI_SCREEN_SYSTEM_DATA)
		{
			[self setGuiToSystemDataScreen];
			[self checkScript];
		}
		[universe setDisplayText:YES];
	}
	
	if (status == STATUS_DOCKED)
	{
		BOOL at_main_station =  (docked_station == [universe station]);
		if (at_main_station)
		{
			if ((([gameView isDown:gvFunctionKey2])||([gameView isDown:gvNumberKey2]))&&(gui_screen != GUI_SCREEN_OPTIONS))
			{
				[self setGuiToLoadSaveScreen];
			}
		}
		//
		if ((([gameView isDown:gvFunctionKey3])||([gameView isDown:gvNumberKey3]))&&(gui_screen != GUI_SCREEN_EQUIP_SHIP))
			[self setGuiToEquipShipScreen];
		//
		if (([gameView isDown:gvFunctionKey8])||([gameView isDown:gvNumberKey8]))
		{
			if (gui_screen != GUI_SCREEN_MARKET)
			{
				[self setGuiToMarketScreen];
				[[universe gui] setSelectedRow:GUI_ROW_MARKET_START];
			}
			[self setGuiToMarketScreen];
		}
	}
	else
	{
		if (([gameView isDown:gvFunctionKey8])||([gameView isDown:gvNumberKey8]))
		{
			if (gui_screen != GUI_SCREEN_MARKET)
				[self setGuiToMarketScreen];
			[universe setDisplayText:YES];
		}
	}
}

- (void) pollGameOverControls:(double) delta_t
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	if ([gameView isDown:32])   // look for the spacebar
	{
		[self set_up];
		[universe game_over]; // restart
		//[breakPatternSound play];
	}
}

static BOOL toggling_music;
- (void) pollAutopilotControls:(double) delta_t
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	//
	//  view keys
	//
	[self pollViewControls];
	//
	//  text displays
	//
	[self pollGuiScreenControls];
	//
	if ([universe displayText])
		[self pollGuiArrowKeyControls:delta_t];
	//
	//
	if (([gameView isDown:99])&&(has_docking_computer)&&(![beepSound isPlaying]))   // look for the 'c' key
	{
		[self abortDocking];			// let the station know that you are no longer on approach
		autopilot_engaged = NO;
		status = STATUS_IN_FLIGHT;
		[beepSound play];
		[universe addMessage:@"Autopilot disengaged" forCount:4.5];
		//
		if (docking_music_on)
		{
			StopMovie ([dockingMusic QTMovie]);
		}
	}
	//
	if ([gameView isDown:115])   // look for the 's' key
	{
		if (!toggling_music)
		{
			docking_music_on = !docking_music_on;
			if (docking_music_on)
			{
				if (IsMovieDone ([dockingMusic QTMovie]))
					GoToBeginningOfMovie ([dockingMusic QTMovie]);
				StartMovie ([dockingMusic QTMovie]);
			}
			else
			{
				StopMovie ([dockingMusic QTMovie]);
			}
		}
		toggling_music = YES;
	}
	else
	{
		toggling_music = NO;
	}
}

- (void) pollDockedControls:(double) delta_t
{
	MyOpenGLView  *gameView = (MyOpenGLView *)[universe gameView];
	if (([gameView isDown:gvFunctionKey1])||([gameView isDown:gvNumberKey1]))   // look for the f1 key
	{
		[universe set_up_universe_from_station]; // launch!
		if (!docked_station)
			docked_station = [universe station];
		//NSLog(@"Leaving dock (%@)...%@",docked_station,[docked_station name]);
		[self leaveDock:docked_station];

		[breakPatternSound play];
	}
	//
	//  text displays
	//
	[self pollGuiScreenControls];
	//
	[self pollGuiArrowKeyControls:delta_t];
	//
	//
}

- (void) pollDemoControls:(double) delta_t
{
	MyOpenGLView	*gameView = (MyOpenGLView *)[universe gameView];
	
	switch (gui_screen)
	{
		case	GUI_SCREEN_INTRO1 :
			if (!disc_operation_in_progress)
			{
				if (([gameView isDown:121])||([gameView isDown:89]))	//  'yY'
				{
					if (themeMusic)
					{
						StopMovie ([themeMusic QTMovie]);
						GoToBeginningOfMovie ([themeMusic QTMovie]);
					}
					disc_operation_in_progress = YES;
					if ([[universe gameController] inFullScreenMode])
						[[universe gameController] pauseFullScreenModeToPerform:@selector(loadPlayer) onTarget:self];
					else
						[self loadPlayer];
					[self setStatus:STATUS_DOCKED];
					[self setGuiToStatusScreen];
					[universe setDisplayText:YES];
				}
			}
			if (([gameView isDown:110])||([gameView isDown:78]))	//  'nN'
			{
				[self setGuiToIntro2Screen];
			}
			break;

		case	GUI_SCREEN_INTRO2 :
		case	GUI_SCREEN_MISSION :
			if ([gameView isDown:32])	//  '<space>'
			{
				docked_station = [universe station];
				[self setStatus:STATUS_DOCKED];
				[self setGuiToStatusScreen];
				[universe removeDemoShips];
				[universe setDisplayText:YES];
				if (themeMusic)
				{
					StopMovie ([themeMusic QTMovie]);
					GoToBeginningOfMovie ([themeMusic QTMovie]);
				}
			}
			break;
	}
	
}

- (void) interpretAIMessage:(NSString *)ms
{
		
	if (![ms isEqual:@"UPDATE"])
		NSLog(@"Player entity to respond to message '%@'",ms);
	if ((([ms isEqual:@"TARGET_LOST"])||([ms isEqual:@"TARGET_DESTROYED"]))&&(missile_status != MISSILE_STATUS_ARMED))
	{
		// targetting 'off' here
		missile_status = MISSILE_STATUS_SAFE;
		if (![boopSound isPlaying])
			[boopSound play];
		//[universe displayMessage:@"Target Lost." forCount:2.0];
		[universe addMessage:@"Target Lost." forCount:2.0];
	}
	if ([ms isEqual:@"INCOMING_MISSILE"])
	{
		[warningSound play];
		[universe addMessage:@"Incoming missile." forCount:4.5];
	}
	if ([ms isEqual:@"ENERGY_LOW"])
		[universe addMessage:@"Energy low." forCount:6.0];
	if (([ms isEqual:@"ECM"])&&(![ecmSound isPlaying]))
		[ecmSound play];
	//[beepSound play];	
}

- (BOOL) fireECM
{
	if ([super fireECM])
	{
		ecm_in_operation = YES;
		ecm_start_time = [universe getTime];
		return YES;
	}
	else
		return NO;
}

- (BOOL) fireEnergyBomb
{
	NSArray* targets = [universe getEntitiesWithinRange:SCANNER_MAX_RANGE ofEntity:self];
	if ([targets count] > 0)
	{
		int i;
		for (i = 0; i < [targets count]; i++)
		{
			Entity *e2 = [targets objectAtIndex:i];
			if ([e2 isKindOfClass:[ShipEntity class]])
				[(ShipEntity *)e2 takeEnergyDamage:1000 from:self becauseOf:self];
		}
	}
	[universe addMessage:@"Energy bomb triggered." forCount:4.5];
	[destructionSound play];
	return YES;
}



- (BOOL) fireMainWeapon
{
	int weapon_to_be_fired = 0;
	
	switch ([universe viewDir])
	{
		case VIEW_FORWARD :
			weapon_to_be_fired = forward_weapon;
			break;
		case VIEW_PORT :
			weapon_to_be_fired = port_weapon;
			break;
		case VIEW_STARBOARD :
			weapon_to_be_fired = starboard_weapon;
			break;
		case VIEW_AFT :
			weapon_to_be_fired = aft_weapon;
			break;
	}
	
	if (weapon_temp / PLAYER_MAX_WEAPON_TEMP >= 0.85)
	{
		[universe addMessage:@"Weapon system overheated." forCount:3.0];
		return NO;
	}
	
	if (weapon_to_be_fired == WEAPON_NONE)
		return NO;
	
	switch (weapon_to_be_fired)
	{
		case WEAPON_PLASMA_CANNON :
			weapon_energy =						6.0;
			weapon_energy_per_shot =			6.0;
			weapon_heat_increment_per_shot =	16.0;
			weapon_reload_time =				0.25;
			break;
		case WEAPON_PULSE_LASER :
			weapon_energy =						15.0;
			weapon_energy_per_shot =			1.0;
			weapon_heat_increment_per_shot =	8.0;
			weapon_reload_time =				0.5;
			break;
		case WEAPON_BEAM_LASER :
			weapon_energy =						15.0;
			weapon_energy_per_shot =			1.0;
			weapon_heat_increment_per_shot =	8.0;
			weapon_reload_time =				0.1;
			break;
		case WEAPON_MINING_LASER :
			weapon_energy =						50.0;
			weapon_energy_per_shot =			1.0;
			weapon_heat_increment_per_shot =	8.0;
			weapon_reload_time =				2.5;
			break;
		case WEAPON_MILITARY_LASER :
			weapon_energy =						23.0;
			weapon_energy_per_shot =			1.0;
			weapon_heat_increment_per_shot =	8.0;
			weapon_reload_time =				0.1;
			break;
	}
		
	if (energy <= weapon_energy_per_shot)
	{
		[universe addMessage:@"No energy to weapon system." forCount:3.0];
		return NO;
	}
	
	using_mining_laser = (weapon_to_be_fired == WEAPON_MINING_LASER);
	
	energy -= weapon_energy_per_shot;
	weapon_temp += weapon_heat_increment_per_shot;
	
	//NSLog(@"%@ firing weapon",name);
	switch (weapon_to_be_fired)
	{
		case WEAPON_PLASMA_CANNON :
			[self firePlasmaShot:10.0:1000.0:[NSColor greenColor]];
			return YES;
			break;
			
		case WEAPON_PULSE_LASER :
		case WEAPON_BEAM_LASER :
		case WEAPON_MINING_LASER :
		case WEAPON_MILITARY_LASER :
			[self fireLaserShot];
			return YES;
			break;
	}
	return NO;
}

- (void) takeEnergyDamage:(double) amount from:(Entity *) ent becauseOf:(Entity *) other
{
	Vector  rel_pos;
	double  d_forward;
	
	if (status == STATUS_DEAD)
		return;
	
	[ent retain];
	[other retain];
	rel_pos = [ent getPosition];
	
	rel_pos.x -= position.x;
	rel_pos.y -= position.y;
	rel_pos.z -= position.z;

	d_forward   =   dot_product(rel_pos, v_forward);

	if (damageSound)
		[damageSound play];
		
	if (d_forward >= 0)
	{
		//NSLog(@"hit on FORWARD shields");
		forward_shield -= amount;
		if (forward_shield < 0.0)
		{
			amount = -forward_shield;
			forward_shield = 0.0;
		}
		else
		{
			amount = 0.0;
		}
	}
	else
	{
		//NSLog(@"hit on AFT shields");
		aft_shield -= amount;
		if (aft_shield < 0.0)
		{
			amount = -aft_shield;
			aft_shield = 0.0;
		}
		else
		{
			amount = 0.0;
		}
	}
	
	if (amount > 0.0)
	{
		energy -= amount;
		if (scrapeDamageSound)
		{
			if (![scrapeDamageSound isPlaying])
				[scrapeDamageSound play];
		}	
	}
	
	if (energy <= 0.0)
	{
		if ([other isKindOfClass:[ShipEntity class]])
		{
			ShipEntity* hunter = (ShipEntity *)other;
			[hunter collectBountyFor:self];
			if ([hunter getPrimaryTarget] == (Entity *)self)
			{
				[hunter removeTarget:(Entity *)self];
				[[hunter getAI] message:@"TARGET_DESTROYED"];
			}
		}
		[self getDestroyed];
	}
	
	[ent release];
	[other release];
	//
}

- (void) takeScrapeDamage:(double) amount from:(Entity *) ent
{
	Vector  rel_pos;
	double  d_forward;
	
	if (status == STATUS_DEAD)
		return;
	
	[ent retain];
	rel_pos = [ent getPosition];
	
	rel_pos.x -= position.x;
	rel_pos.y -= position.y;
	rel_pos.z -= position.z;

	d_forward   =   dot_product(rel_pos, v_forward);

	if (scrapeDamageSound)
	{
		if (![scrapeDamageSound isPlaying])
			[scrapeDamageSound play];
	}	
	if (d_forward >= 0)
	{
		forward_shield -= amount;
		if (forward_shield < 0.0)
		{
			amount = -forward_shield;
			forward_shield = 0.0;
		}
		else
		{
			amount = 0.0;
		}
	}
	else
	{
		aft_shield -= amount;
		if (aft_shield < 0.0)
		{
			amount = -aft_shield;
			aft_shield = 0.0;
		}
		else
		{
			amount = 0.0;
		}
	}
	
	energy -= amount;
	if (energy <= 0.0)
	{
		if ([ent isKindOfClass:[ShipEntity class]])
		{
			ShipEntity* hunter = (ShipEntity *)ent;
			[hunter collectBountyFor:self];
			if ([hunter getPrimaryTarget] == (Entity *)self)
			{
				[hunter removeTarget:(Entity *)self];
				[[hunter getAI] message:@"TARGET_DESTROYED"];
			}
		}
		[self getDestroyed];
	}
	//
	[ent release];
}

- (int) launchEscapeCapsule
{
	ShipEntity *doppelganger;
	Vector  vel;
	Vector  origin = position;
	//double  start = 10.0;
	//double  throw_speed = 20.0;
	Quaternion q1 = [self QRotation];
	
	status = STATUS_ESCAPE_SEQUENCE;	// firstly
	has_escape_pod = NO;
	
	q1.w = -q1.w;   // player view is reversed remember!
	
	if (flight_speed < 50.0)
		flight_speed = 50.0;
	
	vel.x = flight_speed * v_forward.x;
	vel.y = flight_speed * v_forward.y;
	vel.z = flight_speed * v_forward.z;
	
	doppelganger = [universe getShip:PLAYER_SHIP_DESC];   // retain count = 1
	[doppelganger setPosition:origin];						// directly below
	[doppelganger setScanClass: CLASS_NEUTRAL];
	[doppelganger setQRotation:q1];
	[doppelganger setVelocity: vel];
	[doppelganger setSpeed: flight_speed];
	[doppelganger setOwner:self];
	[[doppelganger getAI] setStateMachine:@"nullAI.plist"];  // fly straight on
	[doppelganger setStatus: STATUS_IN_FLIGHT];  // necessary to get it going!
	[universe addEntity:doppelganger];
	[doppelganger release]; //release
	
	// set up you
	[self setModel:@"capsule.dat"];				// look right to anyone else (for multiplayer later)
	[universe setViewDirection:VIEW_FORWARD];
	flight_speed = 0.0;
	[cargo removeAllObjects];
	energy = 25;
	[universe addMessage:@"Escape sequence active" forCount:4.5];
	shot_time = 0.0;
	
	return [doppelganger universal_id];
}

- (int) dumpCargo
{
	int result = [super dumpCargo];
	if (result != CARGO_NOT_CARGO)
	{
		[universe addMessage:[NSString stringWithFormat:@"%@ ejected.",[universe nameForCommodity:result]] forCount:3.0];
	}
	return result;
}

- (int) getBounty		// overrides returning 'bounty'
{
	return legal_status;
}
- (int) legal_status
{
	return legal_status;
}

- (void) markAsOffender:(int)offence_value
{
	legal_status |= offence_value;
}

- (void) collectBountyFor:(ShipEntity *)other
{
	int score = [other getBounty];
	if ([[other roles] isEqual:@"police"])   // oops, we shot a copper!
		legal_status |= 64;
	credits += 10.0 * score;
	ship_kills++;
	NSString* bonusMS = [NSString stringWithFormat:@"Bounty\t: %d\nPurse\t: %.1f Cr",score,credits/10.0];
	[universe addMessage:bonusMS forCount:6.0];
	if ((ship_kills % 256) == 0)
		[universe addMessage:@"Right On, Commander!" forCount:6.0];
}

- (void) getDestroyed
{
	NSString* gameOverMS = [NSString stringWithFormat:@"Game Over\n\nScore\t: %.1f Cr",credits/10.0];
	energy = 0.0;
	[universe setDisplayText:NO];
	[universe setViewDirection:VIEW_AFT];
	[self becomeLargeExplosion:4.0];
	[self moveForward:100.0];
	[destructionSound play];
	flight_speed = 160.0;
	status = STATUS_DEAD;
	[universe displayMessage:gameOverMS forCount:30.0];
	shot_time = 0.0;
}

- (void) enterDock:(StationEntity *)station
{
	status = STATUS_DOCKING;
	[universe setDisplayText:NO];
	[universe set_up_break_pattern:position quaternion:q_rotation];
	[breakPatternSound play];
	[station noteDockedShip:self];
	docked_station = station;
}

- (void) docked
{
	NSString* dockedMS = @"Docked with station.";
	status = STATUS_DOCKED;
	[universe setViewDirection:VIEW_DOCKED];
	
	[self setTargetToStation];
	[self setPosition:[[self getPrimaryTarget] getPosition]];
	[self setQRotation:[[self getPrimaryTarget] QRotation]];

	v_forward = vector_forward_from_quaternion(q_rotation);
	v_right = vector_right_from_quaternion(q_rotation);
	v_up = vector_up_from_quaternion(q_rotation);

	q_rotation.w = -q_rotation.w;   // need this as a fix...
	
	flight_roll = 0.0;
	flight_pitch = 0.0;
	flight_speed = 0.0;
	
	if (![docked_station localMarket])
		[docked_station initialiseLocalMarketWithSeed:system_seed andRandomFactor:market_rnd];

	[self unloadCargoPods];
	
	[self setGuiToStatusScreen];
	[universe setDisplayText:YES];
	[universe displayMessage:dockedMS forCount:3.0];
	
	if (docking_music_on)
	{
		StopMovie ([dockingMusic QTMovie]);
	}

}

- (void) leaveDock:(StationEntity *)station
{
	legal_status |= [universe legal_status_of_manifest:shipCommodityData];  // 'leaving with those guns were you sir?'
	[self loadCargoPods];
	
	[station autoDockShipsOnApproach];  // clear the way
	
	Vector launchPos = [station getPosition];
	Vector stat_f = vector_forward_from_quaternion([station QRotation]);
	launchPos.x += 500.0*stat_f.x;
	launchPos.y += 500.0*stat_f.y;
	launchPos.z += 500.0*stat_f.z;
    position = launchPos;
	
	q_rotation = [station QRotation];
	q_rotation.w = -q_rotation.w;   // need this as a fix...
	
	flight_roll = -[station flight_roll];
	flight_pitch = 0.0;
	flight_speed = max_flight_speed * 0.5;
	status = STATUS_LAUNCHING;
	[universe setDisplayText:NO];
	gui_screen = GUI_SCREEN_MAIN;
	[universe set_up_break_pattern:position quaternion:q_rotation];
	[breakPatternSound play];
	docked_station = nil;
}

- (void) enterGalacticWitchspace
{	
	status = STATUS_ENTERING_WITCHSPACE;
	[universe setDisplayText:NO];
	[universe removeAllEntitiesExceptPlayer:NO];
	
	has_galactic_hyperdrive = NO;								// fuel cost to target system
	
	galaxy_number++;
	galaxy_number &= 7;
	
	galaxy_seed.a = rotate_byte_left(galaxy_seed.a);
	galaxy_seed.b = rotate_byte_left(galaxy_seed.b);
	galaxy_seed.c = rotate_byte_left(galaxy_seed.c);
	galaxy_seed.d = rotate_byte_left(galaxy_seed.d);
	galaxy_seed.e = rotate_byte_left(galaxy_seed.e);
	galaxy_seed.f = rotate_byte_left(galaxy_seed.f);
	
	system_seed = [universe findSystemAtCoords:NSMakePoint(0x60, 0x60) withGalaxySeed:galaxy_seed];
	target_system_seed = system_seed;
	
	[universe setSystemTo:system_seed];
	galaxy_coordinates.x = system_seed.d;
	galaxy_coordinates.y = system_seed.b;
    srand([[NSDate date] timeIntervalSince1970]);	// seed randomiser by time
	market_rnd = rand() & 255;						// random factor for market values is reset
	legal_status = 0;
	[universe set_up_universe_from_witchspace];
}

- (void) enterWitchspace
{
	//int			dx = target_system_seed.d - galaxy_coordinates.x;
	//int			dy = target_system_seed.b - galaxy_coordinates.y;
	//double		distance = 4.0*sqrt(dx*dx + dy*dy*0.25); 
	double		distance = distanceBetweenPlanetPositions(target_system_seed.d,target_system_seed.b,galaxy_coordinates.x,galaxy_coordinates.y); 
	
	status = STATUS_ENTERING_WITCHSPACE;
	[universe setDisplayText:NO];
	[universe removeAllEntitiesExceptPlayer:NO];
	//
	//  perform any check here for forced witchspace encounters
	//
	BOOL misjump = ((flight_pitch == max_flight_pitch)||(rand() % 256 > 253));
	
	fuel -= 10.0 * distance;								// fuel cost to target system
	if (!misjump)
	{	
		[universe setSystemTo:target_system_seed];
		system_seed = target_system_seed;
		galaxy_coordinates.x = system_seed.d;
		galaxy_coordinates.y = system_seed.b;
		srand([[NSDate date] timeIntervalSince1970]);	// seed randomiser by time
		market_rnd = rand() & 255;						// random factor for market values is reset
		//
		[universe set_up_universe_from_witchspace];
	}
	else
	{
		//NSLog(@"---> Witchspace misjump");
		[ecmSound play];
		[universe set_up_universe_from_misjump];
	}
}

- (void) leaveWitchspace
{
	Vector		pos = [universe getWitchspaceExitPosition];
	Quaternion  q_rtn = [universe getWitchspaceExitRotation];
	position = pos;
	q_rotation = q_rtn;
	flight_roll = 0.0;
	flight_pitch = 0.0;
	flight_speed = max_flight_speed * 0.25;
	status = STATUS_EXITING_WITCHSPACE;
	gui_screen = GUI_SCREEN_MAIN;
	[universe setDisplayText:NO];
	[universe set_up_break_pattern:position quaternion:q_rotation];
	[breakPatternSound play];
	legal_status /= 2;  // 'another day, another system'
}

- (void) performDocking
{
	[self abortDocking];			// let the station know that you are no longer on approach
	autopilot_engaged = NO;
	status = STATUS_IN_FLIGHT;
}

///////////////////////////////////////

- (void) savePlayer
{
	NSSavePanel *sp;
	int runResult;

	/* create or get the shared instance of NSSavePanel */
	sp = [NSSavePanel savePanel];

	/* set up new attributes */
	[sp setRequiredFileType:@"oolite-save"];

	/* display the NSSavePanel */
	
	runResult = [sp runModalForDirectory:nil file:player_name];

	/* if successful, save file under designated name */
	if (runResult == NSOKButton)
	{
		NSArray*	path_components = [[sp filename] pathComponents];
		NSString*   new_name = [[path_components objectAtIndex:[path_components count]-1] stringByDeletingPathExtension]; 
		
		//NSLog(@"Attempting to save to %@",[sp filename]);
		//NSLog(@"Will set Commander's name to %@", new_name);
		
		if (player_name)	[player_name release];
		player_name = [new_name retain];
		
		if (![[self commanderDataDictionary] writeToFile:[sp filename] atomically:YES])
		{
			NSBeep();
			NSLog(@"***** ERROR: Save failed!");
		}
	}
	[self setGuiToStatusScreen];
}

- (void) loadPlayer
{
    int result;
    NSArray *fileTypes = [NSArray arrayWithObject:@"oolite-save"];
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];

    [oPanel setAllowsMultipleSelection:NO];
    result = [oPanel runModalForDirectory:nil file:nil types:fileTypes];
    if (result == NSOKButton)
		[self loadPlayerFromFile:[oPanel filename]];
}

- (void) loadPlayerFromFile:(NSString *)fileToOpen
{
	BOOL loadedOK = NO;
	if (fileToOpen)
	{
		NSDictionary *fileDic = [NSDictionary dictionaryWithContentsOfFile:fileToOpen];
		if (fileDic)
		{
			[self setCommanderDataFromDictionary:fileDic];
			loadedOK = YES;
			//[universe removeAllEntitiesExceptPlayer:NO];	// get rid of current entities
		}
	}
	if (!loadedOK)
		NSBeep();

	[universe removeAllEntitiesExceptPlayer:NO];
	[universe set_up_space];
	
	status = STATUS_DOCKED;
	[universe setViewDirection:VIEW_DOCKED];
	
	docked_station = [universe station];
	[self setPosition:[docked_station getPosition]];
	[self setQRotation:[docked_station QRotation]];
	v_forward = vector_forward_from_quaternion(q_rotation);
	v_right = vector_right_from_quaternion(q_rotation);
	v_up = vector_up_from_quaternion(q_rotation);

	q_rotation.w = -q_rotation.w;   // need this as a fix...
	
	flight_roll = 0.0;
	flight_pitch = 0.0;
	flight_speed = 0.0;
	
	if (![docked_station localMarket])
		[docked_station initialiseLocalMarketWithSeed:system_seed andRandomFactor:market_rnd];

	[self setGuiToStatusScreen];
}

- (void) changePlayerName
{
}


///////////////////////////////////////

- (void) setGuiToStatusScreen
{
	NSDictionary*   descriptions = [universe descriptions];
	NSString*		systemName;
	NSString*		targetSystemName;
	
	system_seed =			[universe findSystemAtCoords:galaxy_coordinates withGalaxySeed:galaxy_seed];
	target_system_seed =	[universe findSystemAtCoords:cursor_coordinates withGalaxySeed:galaxy_seed];
	
	systemName =	[universe generateSystemName:system_seed];
			
	targetSystemName =	[universe generateSystemName:target_system_seed];
	
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		int equip_row = 10;

		int legal_index = 0;
		if (legal_status != 0)
			legal_index = (legal_status <= 50) ? 1 : 2;
		int rating = 0;
		int kills = 8;  // Mostly Harmless starts at 8
		while (kills <= ship_kills)
		{
			rating ++;
			kills *= 2;
		}
		NSString*   legal_desc = (NSString *)[(NSArray *)[descriptions objectForKey:@"legal_status"] objectAtIndex:legal_index];
		NSString*   rating_desc = (NSString *)[(NSArray *)[descriptions objectForKey:@"rating"] objectAtIndex:rating];
		NSString*   alert_desc = (NSString *)[(NSArray *)[descriptions objectForKey:@"condition"] objectAtIndex:[self scanAlertCondition]];
		[gui clear];
		[gui setTitle:[NSString stringWithFormat:@"Commander %@",   player_name]];
		//
		[gui setText:[NSString stringWithFormat:@"Present System:\t\t%@",	systemName]			forRow:1];
		[gui setText:[NSString stringWithFormat:@"Hyperspace System:\t%@",  targetSystemName]   forRow:2];
		[gui setText:[NSString stringWithFormat:@"Condition:\t\t\t%@",		alert_desc]			forRow:3];
		[gui setText:[NSString stringWithFormat:@"Fuel:\t\t%.1f Light Years",   fuel/10.0]		forRow:4];
		[gui setText:[NSString stringWithFormat:@"Cash:\t\t%.1f Cr",			credits/10.0]	forRow:5];
		[gui setText:[NSString stringWithFormat:@"Legal Status:\t%@",		legal_desc]			forRow:6];
		[gui setText:[NSString stringWithFormat:@"Rating:\t\t%@",			rating_desc]		forRow:7];
		//
		[gui setText:@"Equipment:"  forRow:9];
		
		if (max_cargo > 20)				[gui setText:@"\tLarge Cargo Bay"  forRow:equip_row++];
		if (has_escape_pod)				[gui setText:@"\tEscape Pod"  forRow:equip_row++];
		if (has_scoop)					[gui setText:@"\tCargo/Fuel Scoops"  forRow:equip_row++];
		if (has_ecm)					[gui setText:@"\tE.C.M. System"  forRow:equip_row++];
		if (has_energy_bomb)			[gui setText:@"\tEnergy Bomb"  forRow:equip_row++];
		if (has_energy_unit)
		{
			switch (energy_unit)
			{
				case ENERGY_UNIT_NORMAL :
					[gui setText:@"\tExtra Energy Unit"  forRow:equip_row++];
					break;
				case ENERGY_UNIT_NAVAL :
					[gui setText:@"\tNaval Energy Unit"  forRow:equip_row++];
					break;
			}
		}
		if (has_docking_computer)		[gui setText:@"\tDocking Computers"  forRow:equip_row++];
		if (has_galactic_hyperdrive)	[gui setText:@"\tGalactic Hyperdrive"  forRow:equip_row++];
		if (forward_weapon > 0)
			[gui setText:[NSString stringWithFormat:@"\tForward %@",(NSString *)[(NSArray *)[descriptions objectForKey:@"weapon_name"] objectAtIndex:forward_weapon]]  forRow:equip_row++];
		if (aft_weapon > 0)
			[gui setText:[NSString stringWithFormat:@"\tAft %@",(NSString *)[(NSArray *)[descriptions objectForKey:@"weapon_name"] objectAtIndex:aft_weapon]]  forRow:equip_row++];
		if (starboard_weapon > 0)
			[gui setText:[NSString stringWithFormat:@"\tStarboard %@",(NSString *)[(NSArray *)[descriptions objectForKey:@"weapon_name"] objectAtIndex:starboard_weapon]]  forRow:equip_row++];
		if (port_weapon > 0)
			[gui setText:[NSString stringWithFormat:@"\tPort %@",(NSString *)[(NSArray *)[descriptions objectForKey:@"weapon_name"] objectAtIndex:port_weapon]]  forRow:equip_row++];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	//NSLog(@"gui_screen = GUI_SCREEN_STATUS");
	if (lastTextKey)
	{
		[lastTextKey release];
		lastTextKey = nil;
	}
	
	gui_screen = GUI_SCREEN_STATUS;
}

- (void) setGuiToSystemDataScreen
{
	NSDictionary*   targetSystemData;
	NSString*		targetSystemName;
	NSDictionary*   descriptions = [universe descriptions];
	
	target_system_seed =	[universe findSystemAtCoords:cursor_coordinates withGalaxySeed:galaxy_seed];
	
	targetSystemData =		[[universe generateSystemData:target_system_seed] retain];  // retained
	targetSystemName =		[[universe generateSystemName:target_system_seed] retain];  // retained
	
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		
		int government =	[(NSNumber *)[targetSystemData objectForKey:@"government"] intValue];
		int economy =		[(NSNumber *)[targetSystemData objectForKey:@"economy"] intValue];
		int techlevel =		[(NSNumber *)[targetSystemData objectForKey:@"techlevel"] intValue];
		int population =	[(NSNumber *)[targetSystemData objectForKey:@"population"] intValue];
		int productivity =	[(NSNumber *)[targetSystemData objectForKey:@"productivity"] intValue];
		int radius =		[(NSNumber *)[targetSystemData objectForKey:@"radius"] intValue];
		
		NSString*   government_desc =   (NSString *)[(NSArray *)[descriptions objectForKey:@"government"] objectAtIndex:government];
		NSString*   economy_desc =		(NSString *)[(NSArray *)[descriptions objectForKey:@"economy"] objectAtIndex:economy];
		NSString*   inhabitants =		(NSString *)[targetSystemData objectForKey:@"inhabitants"];
		NSString*   system_desc =		(NSString *)[targetSystemData objectForKey:@"description"];

		[gui clear];
		[gui setTitle:[NSString stringWithFormat:@"Data on %@",   targetSystemName]];
		//
		[gui setText:[NSString stringWithFormat:@"Economy:\t%@",				economy_desc]		forRow:1];
		//
		[gui setText:[NSString stringWithFormat:@"Government:\t%@",				government_desc]	forRow:3];
		//
		[gui setText:[NSString stringWithFormat:@"Tech Level:\t%d",				techlevel + 1]		forRow:5];
		//
		[gui setText:[NSString stringWithFormat:@"Population:\t%.1f Billion",	population/10.0]	forRow:7];
		//
		[gui setText:[NSString stringWithFormat:@"\t(%@)",						inhabitants]		forRow:9];
		//
		[gui setText:[NSString stringWithFormat:@"Gross Productivity:\t%5d M Cr.",	productivity]	forRow:11];
		//
		[gui setText:[NSString stringWithFormat:@"Average radius:\t%5d km",			radius]			forRow:13];
		//
		int i = [gui addLongText:system_desc startingAtRow:15 align:GUI_ALIGN_LEFT];
		missionTextRow = i;
		for (i-- ; i > 14 ; i--)
			[gui setColor:[NSColor greenColor] forRow:i];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	//NSLog(@"gui_screen = GUI_SCREEN_SYSTEM_DATA for system %@ at (%d,%d)",targetSystemName,target_system_seed.d,target_system_seed.b);
	
	gui_screen = GUI_SCREEN_SYSTEM_DATA;

	if (lastTextKey)
	{
		[lastTextKey release];
		lastTextKey = nil;
	}
	
	[targetSystemData release]; // released
	[targetSystemName release]; // released
}

- (void) setGuiToLongRangeChartScreen
{
	NSString*   targetSystemName;
	//int			dx = target_system_seed.d - galaxy_coordinates.x;
	//int			dy = target_system_seed.b - galaxy_coordinates.y;
	//double		distance = 4.0*sqrt(dx*dx + dy*dy*0.25); 
	double		distance = distanceBetweenPlanetPositions(target_system_seed.d,target_system_seed.b,galaxy_coordinates.x,galaxy_coordinates.y); 
	
	target_system_seed =	[universe findSystemAtCoords:cursor_coordinates withGalaxySeed:galaxy_seed];
	targetSystemName =		[[universe generateSystemName:target_system_seed] retain];  // retained
	
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		NSImage*	drawImage;
		int			i;
		NSPoint		star;
		
		Random_Seed g_seed = galaxy_seed;
		double		hscale = GUI_PIXEL_WIDTH / 256.0;
		double		vscale = -1.0 * GUI_PIXEL_HEIGHT / 512.0;
		double		hoffset = 0.0;
		double		voffset = GUI_PIXEL_HEIGHT - GUI_PIXEL_TITLE_HEIGHT - 5;
		
		[gui clear];
		[gui setTitle:[NSString stringWithFormat:@"Galactic Chart %d",   galaxy_number+1]];
		//
		[gui setText:targetSystemName														forRow:17];
		[gui setText:[NSString stringWithFormat:@"Distance:\t%.1f Light Years", distance]   forRow:18];
		//
		if (planetSearchString)
			[gui setText:[NSString stringWithFormat:@"Find planet: %@|", [planetSearchString capitalizedString]]  forRow:16];
		else
			[gui setText:@"Find planet: |"  forRow:16];
		[gui setColor:[NSColor cyanColor] forRow:16];
		//
		drawImage = [[gui generateDisplay] retain];
		
		[drawImage lockFocus];
		[[NSColor clearColor] set];
		//
		//NSMutableString *stardump = [NSMutableString stringWithString:@"\nStar List\n"];
		//
		// draw stars
		[[NSColor whiteColor] set];
		for (i = 0; i < 256; i++)
		{
			star.x = g_seed.d * hscale + hoffset;
			star.y = g_seed.b * vscale + voffset;

			NSRectFill(NSMakeRect(star.x,star.y,1.0,1.0));
			
			if ((g_seed.e | 0x50) < 0x90)
				NSRectFill(NSMakeRect(star.x + 1.0,star.y,1.0,1.0));
			
			//[stardump appendFormat:@"System %d, %@ (%d,%d)\n", i, [universe generateSystemName:g_seed], g_seed.d, g_seed.b];
			
			rotate_seed(&g_seed);
			rotate_seed(&g_seed);
			rotate_seed(&g_seed);
			rotate_seed(&g_seed);
		}
		
		//NSLog(@"%@",stardump);
		
		// draw bottom divider
		[[NSColor lightGrayColor] set];
		NSRectFill(NSMakeRect(0,voffset + 256.0*vscale - 2,GUI_PIXEL_WIDTH,2));
		
		// draw fuel range circle
		[[NSColor greenColor] set];
		{
			double fu = fuel/4.0;
			NSPoint cu = NSMakePoint(hscale*galaxy_coordinates.x+hoffset,vscale*galaxy_coordinates.y+voffset);
			NSRect limits = NSMakeRect(cu.x-hscale*fu,cu.y-vscale*fu*2.0,fu*2.0*hscale,fu*4.0*vscale);
			[[NSBezierPath bezierPathWithOvalInRect:limits] stroke];
			// draw cross-hairs over current location
			NSRectFill(NSMakeRect(cu.x-1.0,cu.y-14.0,2.0,28.0));
			NSRectFill(NSMakeRect(cu.x-14.0,cu.y-1.0,28.0,2.0));
		}
		
		// draw cross hairs over current location
		[[NSColor redColor] set];
		{
			NSPoint cu = NSMakePoint(hscale*cursor_coordinates.x+hoffset,vscale*cursor_coordinates.y+voffset);
			NSRectFill(NSMakeRect(cu.x-1.0,cu.y-7.0,2.0,14.0));
			NSRectFill(NSMakeRect(cu.x-7.0,cu.y-1.0,14.0,2.0));
		}
		
		//
		[drawImage unlockFocus];
		//
		[universe guiUpdated];
	}
	/* ends */
	
	//NSLog(@"gui_screen = GUI_SCREEN_LONG_RANGE_CHART");
	
	gui_screen = GUI_SCREEN_LONG_RANGE_CHART;
	
	
	
	[targetSystemName release]; // released
}

- (void) setGuiToShortRangeChartScreen
{
	NSString*   targetSystemName;
	//int			dx = target_system_seed.d - galaxy_coordinates.x;
	//int			dy = target_system_seed.b - galaxy_coordinates.y;
	//double		distance = 4.0*sqrt(dx*dx + dy*dy*0.25); 
	double		distance = distanceBetweenPlanetPositions(target_system_seed.d,target_system_seed.b,galaxy_coordinates.x,galaxy_coordinates.y); 
	
	target_system_seed =	[universe findSystemAtCoords:cursor_coordinates withGalaxySeed:galaxy_seed];
	targetSystemName =		[[universe generateSystemName:target_system_seed] retain];  // retained
	
	//NSLog(@"found %@ at (%d, %d)",targetSystemName,(int)cursor_coordinates.x,(int)cursor_coordinates.y);
	
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		NSImage*	drawImage;
		int			i;
		NSPoint		star;
		
		Random_Seed g_seed = galaxy_seed;
		double		hcenter = GUI_PIXEL_WIDTH/2.0;
		double		vcenter = 160.0;
		double		hscale = 4.0 * GUI_PIXEL_WIDTH / 256.0;
		double		vscale = -4.0 * GUI_PIXEL_HEIGHT / 512.0;
		double		hoffset = hcenter - galaxy_coordinates.x*hscale;
		double		voffset = GUI_PIXEL_HEIGHT - GUI_PIXEL_TITLE_HEIGHT - 5 - vcenter - galaxy_coordinates.y*vscale;
		
		if ((abs(cursor_coordinates.x-galaxy_coordinates.x)>=20)||(abs(cursor_coordinates.y-galaxy_coordinates.y)>=38))
			cursor_coordinates = galaxy_coordinates;	// home
		
		[gui clear];
		[gui setTitle:@"Short Range Chart"];
		//
		[gui setText:targetSystemName														forRow:19];
		[gui setText:[NSString stringWithFormat:@"Distance:\t%.1f Light Years", distance]   forRow:20];
		//
		drawImage = [[gui generateDisplay] retain];
		
		[drawImage lockFocus];
		[[NSColor clearColor] set];
		//
		// draw stars
		[[NSColor colorWithCalibratedRed:1.0 green:1.0 blue:0.75 alpha:1.0] set];
		clear_carry_flag();
		for (i = 0; i < 256; i++)
		{
			int dx, dy;
			int blob_size = 2.0 * ((g_seed.e & 1) + 2 + get_carry_flag());
			
			//[[NSColor colorWithCalibratedHue:(g_seed.f & 15)/16.0 saturation:0.5 brightness:1.0 alpha:1.0] set];
			
			star.x = g_seed.d * hscale + hoffset;
			star.y = g_seed.b * vscale + voffset;
			
			dx = abs(galaxy_coordinates.x - g_seed.d);
			dy = abs(galaxy_coordinates.y - g_seed.b);
			
			if ((dx < 20)&&(dy < 38))
			{
				NSString*   pName = [universe generateSystemName:g_seed];
				NSRect blobRect = NSMakeRect(star.x-blob_size, star.y-blob_size, blob_size*2.0, blob_size*2.0);
				[[NSBezierPath bezierPathWithOvalInRect:blobRect] fill];
				[pName drawAtPoint:NSMakePoint(star.x+blob_size, star.y+blob_size) withAttributes:[NSMutableDictionary dictionaryWithObjectsAndKeys:
							[NSFont boldSystemFontOfSize:0.75*GUI_PIXEL_ROW_HEIGHT], NSFontAttributeName,
							[NSColor yellowColor], NSForegroundColorAttributeName, NULL]];
				
				//if ((g_seed.e | 0x50) < 0x90)
				//	NSRectFill(NSMakeRect(star.x + 1.0,star.y,1.0,1.0));
			}
			rotate_seed(&g_seed);
			rotate_seed(&g_seed);
			rotate_seed(&g_seed);
			rotate_seed(&g_seed);
		}
		
		// draw fuel range circle
		[[NSColor greenColor] set];
		{
			double fu = fuel/4.0;
			NSPoint cu = NSMakePoint(hscale*galaxy_coordinates.x+hoffset,vscale*galaxy_coordinates.y+voffset);
			NSRect limits = NSMakeRect(cu.x-hscale*fu,cu.y-vscale*fu*2.0,fu*2.0*hscale,fu*4.0*vscale);
			[[NSBezierPath bezierPathWithOvalInRect:limits] stroke];
			// draw cross-hairs over current location
			NSRectFill(NSMakeRect(cu.x-1.0,cu.y-14.0,2.0,28.0));
			NSRectFill(NSMakeRect(cu.x-14.0,cu.y-1.0,28.0,2.0));
		}
		
		// draw cross hairs over cursor
		[[NSColor redColor] set];
		{
			NSPoint cu = NSMakePoint(hscale*cursor_coordinates.x+hoffset,vscale*cursor_coordinates.y+voffset);
			NSRectFill(NSMakeRect(cu.x-1.0,cu.y-7.0,2.0,14.0));
			NSRectFill(NSMakeRect(cu.x-7.0,cu.y-1.0,14.0,2.0));
		}
		
		//
		[drawImage unlockFocus];
		//
		[universe guiUpdated];
	}
	/* ends */
	
	//NSLog(@"gui_screen = GUI_SCREEN_LONG_RANGE_CHART");
	
	gui_screen = GUI_SCREEN_SHORT_RANGE_CHART;
	
	[targetSystemName release]; // released
}

- (void) setGuiToLoadSaveScreen
{
	GameController  *controller = [universe gameController];
	int displayModeIndex = [controller indexOfCurrentDisplayMode];
	if (displayModeIndex == NSNotFound)
	{
		NSLog(@"***** couldn't find current display mode switching to basic 640x480");
		displayModeIndex = 0;
	}
	NSDictionary	*mode = [(NSArray *)[controller displayModes] objectAtIndex:displayModeIndex];
	int modeWidth = [[mode objectForKey: (NSString *)kCGDisplayWidth] intValue];
	int modeHeight = [[mode objectForKey: (NSString *)kCGDisplayHeight] intValue];
	int modeRefresh = [[mode objectForKey: (NSString *)kCGDisplayRefreshRate] intValue];
	
	NSString	*displayModeString = [NSString stringWithFormat:@" Fullscreen: %d x %d at %d Hz ", modeWidth, modeHeight, modeRefresh];
		
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		int save_row =		GUI_ROW_OPTIONS_SAVE;
		int load_row =		GUI_ROW_OPTIONS_LOAD;
		int options_row =   GUI_ROW_OPTIONS_OPTIONS;
		int display_row =   GUI_ROW_OPTIONS_DISPLAY;
		int speech_row =	GUI_ROW_OPTIONS_SPEECH;
		
		[gui clear];
		[gui setTitle:[NSString stringWithFormat:@"Commander %@",   player_name]];
		//
		[gui setText:@" Save Commander "  forRow:save_row		align:GUI_ALIGN_CENTER];
		[gui setText:@" Load Commander "  forRow:load_row		align:GUI_ALIGN_CENTER];
		[gui setText:@" Game Options "	forRow:options_row  align:GUI_ALIGN_CENTER];
		[gui setText:displayModeString	forRow:display_row  align:GUI_ALIGN_CENTER];
		if (speech_on)
			[gui setText:@" Spoken messages: ON "	forRow:speech_row  align:GUI_ALIGN_CENTER];
		else
			[gui setText:@" Spoken messages: OFF "	forRow:speech_row  align:GUI_ALIGN_CENTER];
		[gui setSelectableRange:NSMakeRange(save_row,1 + speech_row - save_row)];
		[gui setSelectedRow:save_row];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	//NSLog(@"gui_screen = GUI_SCREEN_OPTIONS");
	
	gui_screen = GUI_SCREEN_OPTIONS;
}

- (void) setGuiToEquipShipScreen
{
	NSArray*	equipdata = [universe equipmentdata];
		
	double price_factor = 1.0;
	int techlevel =		[(NSNumber *)[[universe generateSystemData:system_seed] objectForKey:@"techlevel"] intValue];
	
	if (docked_station)
	{
		price_factor = [docked_station equipment_price_factor];
		if ([docked_station equivalent_tech_level] != NSNotFound)
		{
			//NSLog(@"DEBUG setting equivalent tech level...");
			techlevel = [docked_station equivalent_tech_level];
		}	
		//NSLog(@"DEBUG docked_station name '%@' with price_factor %.1f and equivalent_tech_level %d",[docked_station name],[docked_station equipment_price_factor],[docked_station equivalent_tech_level]);
	}
	
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		
		int start_row =		GUI_ROW_EQUIPMENT_START;
		int row = start_row;
		int i;
		int tab_stops[GUI_MAX_COLUMNS]; 
		
		[gui clear];
		[gui setTitle:@"Ship Outfitting"];
		//
		[gui setText:[NSString stringWithFormat:@"Cash:\t%.1f Cr.", 0.1*credits]  forRow: GUI_ROW_EQUIPMENT_CASH];
		//
		tab_stops[0] = 0;
		tab_stops[1] = 320;
		//
		[gui setTabStops:tab_stops];
		//
		for (i = 0; i < [equipdata count]; i++)
		{
			int			min_techlevel   = [(NSNumber *)[(NSArray *)[equipdata objectAtIndex:i] objectAtIndex:0] intValue];
			int			price_per_unit  = [(NSNumber *)[(NSArray *)[equipdata objectAtIndex:i] objectAtIndex:1] intValue];
			NSString*   desc			= (NSString *)[(NSArray *)[equipdata objectAtIndex:i] objectAtIndex:2];
			NSString*   eq_type			= (NSString *)[(NSArray *)[equipdata objectAtIndex:i] objectAtIndex:3];
			double		price			= ([eq_type isEqual:@"EQ_FUEL"]) ? ((70 - fuel) * price_per_unit) : (price_per_unit) ;
			
			price *= price_factor;  // increased prices at some stations
			
			if (([eq_type isEqual:@"EQ_MISSILE"])&&(missiles > 3)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_ECM"])&&(has_ecm)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_FUEL_SCOOPS"])&&(has_scoop)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_ESCAPE_POD"])&&(has_escape_pod)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_ENERGY_BOMB"])&&(has_energy_bomb)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_ENERGY_UNIT"])&&(has_energy_unit)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_DOCK_COMP"])&&(has_docking_computer)) price = 0.0;
			
			if (([eq_type isEqual:@"EQ_GAL_DRIVE"])&&(has_galactic_hyperdrive)) price = 0.0;
			
			if ([eq_type isEqual:@"EQ_CARGO_BAY"])
			{
				int standard_max_cargo = 20;
				if (max_cargo > standard_max_cargo)
					price = 0.0;
			}
			
			if ((price > 0.0)&&(techlevel >= min_techlevel))
			{
				NSMutableArray*		row_info = [NSMutableArray arrayWithCapacity:2];
				[gui setKey:[NSString stringWithFormat:@"%d",i] forRow:row];			// save the index of the item as the key for the row
				[row_info addObject:[NSString stringWithFormat:@" %@ ",desc]];
				[row_info addObject:[NSString stringWithFormat:@" %.1f ",0.1*price]];
				[gui setArray:(NSArray *)row_info forRow:row++];
			}
		}
		//
		[gui setSelectableRange:NSMakeRange(start_row,row - start_row)];
		[gui setSelectedRow:start_row];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	//NSLog(@"gui_screen = GUI_SCREEN_EQUIP_SHIP");
	
	chosen_weapon_facing = WEAPON_FACING_NONE;
	gui_screen = GUI_SCREEN_EQUIP_SHIP;
}

- (void) setGuiToIntro1Screen
{
	GuiDisplayGen* gui = [universe gui];

	// GUI stuff
	{
		[gui clear];
		[gui setTitle:@"Oolite"];
		//
		[gui setText:@"by Giles Williams \x00a9 2003,2004" forRow:17 align:GUI_ALIGN_CENTER];
		[gui setColor:[NSColor whiteColor] forRow:17];
		//
		[gui setText:@"Oolite Theme Music by No Sleep \x00a9 2004" forRow:19 align:GUI_ALIGN_CENTER];
		[gui setColor:[NSColor grayColor] forRow:19];
		//
		[gui setText:@"Load Previous Commander (Y/N)?" forRow:21 align:GUI_ALIGN_CENTER];
		[gui setColor:[NSColor yellowColor] forRow:21];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	[universe set_up_intro1];
		
	if (gui)
		gui_screen = GUI_SCREEN_INTRO1;
		
	if (themeMusic)
	{
		GoToBeginningOfMovie ([themeMusic QTMovie]);
		StartMovie ([themeMusic QTMovie]);
	}
}

- (void) setGuiToIntro2Screen
{
	GuiDisplayGen* gui = [universe gui];

	// GUI stuff
	{
		[gui clear];
		[gui setTitle:@"Oolite"];
		//
		[gui setText:@"Press Space Commander" forRow:21 align:GUI_ALIGN_CENTER];
		[gui setColor:[NSColor yellowColor] forRow:21];
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	/* ends */
	
	[universe set_up_intro2];
		
	if (gui)
		gui_screen = GUI_SCREEN_INTRO2;
}

- (BOOL) tryBuyingItem:(int) index
{
	NSArray*	equipdata = [universe equipmentdata];
	int			price_per_unit  = [(NSNumber *)[(NSArray *)[equipdata objectAtIndex:index] objectAtIndex:1] intValue];
	NSString*   eq_type			= (NSString *)[(NSArray *)[equipdata objectAtIndex:index] objectAtIndex:3];
	double		price			= ([eq_type isEqual:@"EQ_FUEL"]) ? ((70 - fuel) * price_per_unit) : (price_per_unit) ;
	double		price_factor	= 1.0;

	if (docked_station)
	{
		price_factor = [docked_station equipment_price_factor];
	}
	
	price *= price_factor;  // increased prices at some stations
	
	if (([eq_type hasPrefix:@"EQ_WEAPON"])&&(chosen_weapon_facing == WEAPON_FACING_NONE))
	{
		// expand the gui rows with Forward, Aft, Port and Starboard choices.
		// the row will have the same key value as before
		// but consist of the string '\t$facing' where $facing is Forward, Aft, Port or Starboard
		GuiDisplayGen*		gui = [universe gui];
		int					row = [gui selectedRow];				// save for after resetting
		NSMutableArray*		row_info = [NSMutableArray arrayWithCapacity:2];
		NSRange				range = [gui selectableRange];
		//int					row = index + GUI_ROW_EQUIPMENT_START;
		int					i;
		
		//NSLog(@"index = %d, selectedRow = %d, index + GRES = %d",index,[gui selectedRow],index + GUI_ROW_EQUIPMENT_START);
		
		[self setGuiToEquipShipScreen];											// reset
		range = [gui selectableRange];
		for (i = range.location; i < range.location + range.length; i++)		// determine row with selected index
		{
			[gui setSelectedRow:i];
			if ([[gui selectedRowKey] intValue] == index)
				row = i;
		}
		range.length += 4;
		[gui setSelectableRange:range];
		[gui setSelectedRow:row];												// from determined value
		for (i = GUI_MAX_ROWS; i >= row+5 ; i--)
		{
			id	old_row_info = [gui objectForRow:i-4];
			if (old_row_info)
				[gui setArray:old_row_info forRow:i];
		}
		[gui setText:[gui selectedRowText] forRow:row];							// set it as text rather than an array eliminating the price
		[gui setColor:[NSColor greenColor] forRow:row];							// color the selection
		row++;																	// skip forward one row
		[gui setSelectedRow:row];												// new selection is 'forward' facing
		[row_info addObject:FORWARD_FACING_STRING];
		[row_info addObject:[NSString stringWithFormat:@" %.1f ",0.1*price]];
		[gui setKey:[NSString stringWithFormat:@"%d",index] forRow:row];
		[gui setColor:[NSColor greenColor] forRow:row];
		[gui setArray:[NSArray arrayWithArray:row_info] forRow:row++];
		[row_info replaceObjectAtIndex:0 withObject:AFT_FACING_STRING];
		[row_info replaceObjectAtIndex:1 withObject:[NSString stringWithFormat:@" %.1f ",0.1*price]];
		[gui setKey:[NSString stringWithFormat:@"%d",index] forRow:row];
		[gui setColor:[NSColor greenColor] forRow:row];
		[gui setArray:[NSArray arrayWithArray:row_info] forRow:row++];
		[row_info replaceObjectAtIndex:0 withObject:PORT_FACING_STRING];
		[row_info replaceObjectAtIndex:1 withObject:[NSString stringWithFormat:@" %.1f ",0.1*price]];
		[gui setKey:[NSString stringWithFormat:@"%d",index] forRow:row];
		[gui setColor:[NSColor greenColor] forRow:row];
		[gui setArray:[NSArray arrayWithArray:row_info] forRow:row++];
		[row_info replaceObjectAtIndex:0 withObject:STARBOARD_FACING_STRING];
		[row_info replaceObjectAtIndex:1 withObject:[NSString stringWithFormat:@" %.1f ",0.1*price]];
		[gui setKey:[NSString stringWithFormat:@"%d",index] forRow:row];
		[gui setColor:[NSColor greenColor] forRow:row];
		[gui setArray:[NSArray arrayWithArray:row_info] forRow:row++];
		[gui generateDisplay];
		[universe guiUpdated];
		return YES;
	}

	if (price > credits)
	{
		[self setGuiToEquipShipScreen];
		return NO;
	}
	
	if (([eq_type hasPrefix:@"EQ_WEAPON"])&&(chosen_weapon_facing != WEAPON_FACING_NONE))
	{
		int chosen_weapon = WEAPON_NONE;
		int current_weapon = WEAPON_NONE;
		
		if ([eq_type isEqual:@"EQ_WEAPON_TWIN_PLASMA_CANNON"])
			chosen_weapon = WEAPON_PLASMA_CANNON;
		if ([eq_type isEqual:@"EQ_WEAPON_PULSE_LASER"])
			chosen_weapon = WEAPON_PULSE_LASER;
		if ([eq_type isEqual:@"EQ_WEAPON_BEAM_LASER"])
			chosen_weapon = WEAPON_BEAM_LASER;
		if ([eq_type isEqual:@"EQ_WEAPON_MINING_LASER"])
			chosen_weapon = WEAPON_MINING_LASER;
		if ([eq_type isEqual:@"EQ_WEAPON_MILITARY_LASER"])
			chosen_weapon = WEAPON_MILITARY_LASER;
			
		switch (chosen_weapon_facing)
		{
			case WEAPON_FACING_FORWARD :
				current_weapon = forward_weapon;
				forward_weapon = chosen_weapon;
				break;
			case WEAPON_FACING_AFT :
				current_weapon = aft_weapon;
				aft_weapon = chosen_weapon;
				break;
			case WEAPON_FACING_PORT :
				current_weapon = port_weapon;
				port_weapon = chosen_weapon;
				break;
			case WEAPON_FACING_STARBOARD :
				current_weapon = starboard_weapon;
				starboard_weapon = chosen_weapon;
				break;
		}
		
		credits -= price;
		
		// refund here for current_weapon;
		switch (current_weapon)
		{
			case WEAPON_PLASMA_CANNON :
				credits += [universe getPriceForWeaponSystemWithKey:@"EQ_WEAPON_TWIN_PLASMA_CANNON"];
				break;
			case WEAPON_PULSE_LASER :
				credits += [universe getPriceForWeaponSystemWithKey:@"EQ_WEAPON_PULSE_LASER"];
				break;
			case WEAPON_BEAM_LASER :
				credits += [universe getPriceForWeaponSystemWithKey:@"EQ_WEAPON_BEAM_LASER"];
				break;
			case WEAPON_MINING_LASER :
				credits += [universe getPriceForWeaponSystemWithKey:@"EQ_WEAPON_MINING_LASER"];
				break;
			case WEAPON_MILITARY_LASER :
					credits += [universe getPriceForWeaponSystemWithKey:@"EQ_WEAPON_MILITARY_LASER"];
				break;
			case WEAPON_NONE :
				break;
		}
			
		[self setGuiToEquipShipScreen];
		return YES;
	}
	
	if (([eq_type isEqual:@"EQ_MISSILE"])&&(missiles > 3))
		return NO;	
	
	if ([eq_type isEqual:@"EQ_FUEL"])
	{
		fuel = 70;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}
	
	if ([eq_type isEqual:@"EQ_MISSILE"])
	{
		missiles++;
		credits -= price;
		[self setGuiToEquipShipScreen];
		missile_status = (missiles > 0) ? MISSILE_STATUS_SAFE : MISSILE_STATUS_NONE;
		return YES;
	}

	if ([eq_type isEqual:@"EQ_CARGO_BAY"])
	{
		// find the standard cargo limit for your craft first
		int standard_max_cargo = 20;
		if (max_cargo <= standard_max_cargo)
		{
			max_cargo += 15;
			credits -= price;
			[self setGuiToEquipShipScreen];
			return YES;
		}
	}

	if (([eq_type isEqual:@"EQ_ECM"]) && (!has_ecm))
	{
		has_ecm = YES;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	if (([eq_type isEqual:@"EQ_FUEL_SCOOPS"]) && (!has_scoop))
	{
		has_scoop = YES;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	if (([eq_type isEqual:@"EQ_ESCAPE_POD"]) && (!has_escape_pod))
	{
		has_escape_pod = YES;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	if (([eq_type isEqual:@"EQ_ENERGY_BOMB"]) && (!has_energy_bomb))
	{
		has_energy_bomb = YES;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	if (([eq_type isEqual:@"EQ_ENERGY_UNIT"]) && (!has_energy_unit))
	{
		has_energy_unit = YES;
		energy_unit = ENERGY_UNIT_NORMAL;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	if (([eq_type isEqual:@"EQ_DOCK_COMP"]) && (!has_docking_computer))
	{
		has_docking_computer = YES;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	if (([eq_type isEqual:@"EQ_GAL_DRIVE"]) && (!has_galactic_hyperdrive))
	{
		has_galactic_hyperdrive = YES;
		credits -= price;
		[self setGuiToEquipShipScreen];
		return YES;
	}

	return NO;
}

- (void) setGuiToMarketScreen
{
	NSMutableArray*		row_info = [NSMutableArray arrayWithCapacity:5];
	NSArray*			localMarket;
	
	if (status == STATUS_DOCKED)
	{
		if (docked_station == nil)
			docked_station = [universe station];
		if ([docked_station localMarket])
			localMarket = [docked_station localMarket];
		else
			localMarket = [docked_station initialiseLocalMarketWithSeed:system_seed andRandomFactor:market_rnd];
	}
	else
	{
		if ([[universe station] localMarket])
			localMarket = [[universe station] localMarket];
		else
			localMarket = [[universe station] initialiseLocalMarketWithSeed:system_seed andRandomFactor:market_rnd];
	}
	
	// GUI stuff
	{
		GuiDisplayGen* gui = [universe gui];
		
		int start_row =		GUI_ROW_MARKET_START;
		int row = start_row;
		int i;
		int tab_stops[GUI_MAX_COLUMNS]; 
		int n_commodities = [localMarket count];
		int in_hold[n_commodities];
		
		if (status == STATUS_DOCKED)
		{
			for (i = 0; i < n_commodities; i++)
				in_hold[i] = [(NSNumber *)[(NSArray *)[shipCommodityData objectAtIndex:i] objectAtIndex:MARKET_QUANTITY] intValue];
		}
		else
		{
			for (i = 0; i < n_commodities; i++)
				in_hold[i] = 0;
			for (i = 0; i < [cargo count]; i++)
			{
				ShipEntity *container = (ShipEntity *)[cargo objectAtIndex:i];
				in_hold[[container getCommodityType]] += [container getCommodityAmount];
			}
		}
		
		[gui clear];
		[gui setTitle:[NSString stringWithFormat:@"%@ Commodity Market",[universe generateSystemName:system_seed]]];
		//
		tab_stops[0] = 0;
		tab_stops[1] = 192;
		tab_stops[2] = 288;
		tab_stops[3] = 384;
		//
		[gui setTabStops:tab_stops];
		//
		[row_info addObject:@"Commodity:"];
		[row_info addObject:@"Price:"];
		[row_info addObject:@"For sale:"];
		[row_info addObject:@"In hold:"];
		//
		[gui setColor:[NSColor greenColor] forRow:GUI_ROW_MARKET_KEY];
		[gui setArray:row_info forRow:GUI_ROW_MARKET_KEY];
		//
		current_cargo = 0;  // for calculating remaining hold space
		//
		for (i = 0; i < n_commodities; i++)
		{
			NSMutableArray* row_data = [[NSMutableArray alloc] initWithCapacity:5];
			NSString* desc = [NSString stringWithFormat:@" %@ ",(NSString *)[(NSArray *)[localMarket objectAtIndex:i] objectAtIndex:MARKET_NAME]];
			int available_units = [(NSNumber *)[(NSArray *)[localMarket objectAtIndex:i] objectAtIndex:MARKET_QUANTITY] intValue];
			int units_in_hold = in_hold[i];
			int price_per_unit = [(NSNumber *)[(NSArray *)[localMarket objectAtIndex:i] objectAtIndex:MARKET_PRICE] intValue];
			int unit = [(NSNumber *)[(NSArray *)[localMarket objectAtIndex:i] objectAtIndex:MARKET_UNITS] intValue];
			NSString* available = (available_units > 0) ? [NSString stringWithFormat:@"%d",available_units] : @"-";
			NSString* price = [NSString stringWithFormat:@" %.1f ",0.1*price_per_unit];
			NSString* owned = (units_in_hold > 0) ? [NSString stringWithFormat:@"%d",units_in_hold] : @"-";
			NSString* units = @"t";
			if (unit == UNITS_KILOGRAMS)  units = @"kg";
			if (unit == UNITS_GRAMS)  units = @"g"; 
			NSString* units_available = [NSString stringWithFormat:@" %@ %@ ",available,units];
			NSString* units_owned = [NSString stringWithFormat:@" %@ %@ ",owned,units];
			
			if (unit == UNITS_TONS)
				current_cargo += units_in_hold;
			
			[row_data addObject:desc]; // description of commodity
			[row_data addObject:price]; // price
			[row_data addObject:units_available]; // amount on station
			[row_data addObject:units_owned]; // amount in hold
			
			[gui setKey:[NSString stringWithFormat:@"%d",i] forRow:row];
			[gui setArray:row_data forRow:row++];
		}
		//
		if ([cargo count] > 0)
			current_cargo = ([cargo count] <= max_cargo) ? [cargo count] : max_cargo;  // actually count the containers and things (may be > max_cargo)
		//
		[gui setText:[NSString stringWithFormat:@"Cash:\t%.1f Cr.\t\tLoad %d of %d t.", 0.1*credits, current_cargo, max_cargo]  forRow: GUI_ROW_MARKET_CASH];
		//
		if (status == STATUS_DOCKED)	// can only buy or sell in dock
		{
			[gui setSelectableRange:NSMakeRange(start_row,row - start_row)];
			if (([gui selectedRow] < start_row)||([gui selectedRow] >=row))
				[gui setSelectedRow:start_row];
		}
		else
		{
			[gui setNoSelectedRow];
		}
		//
		[gui generateDisplay];
		[universe guiUpdated];
	}
	
	gui_screen = GUI_SCREEN_MARKET;
}

- (int) gui_screen
{
	return gui_screen;
}

- (BOOL) tryBuyingCommodity:(int) index
{
	NSMutableArray*			localMarket;
	if (status == STATUS_DOCKED)
	{
		if (docked_station == nil)
			docked_station = [universe station];
		if ([docked_station localMarket])
			localMarket = [docked_station localMarket];
		else
			localMarket = [docked_station initialiseLocalMarketWithSeed:system_seed andRandomFactor:market_rnd];
	}
	else
	{
		return NO; // can't buy if not docked
	}
	NSArray *commodityArray = (NSArray *)[localMarket objectAtIndex:index];
	int available_units =   [(NSNumber *)[commodityArray objectAtIndex:MARKET_QUANTITY] intValue];
	int price_per_unit =	[(NSNumber *)[commodityArray objectAtIndex:MARKET_PRICE] intValue];
	int unit =				[(NSNumber *)[commodityArray objectAtIndex:MARKET_UNITS] intValue];
	if ((available_units == 0)||(price_per_unit > credits)||((unit == 0)&&(current_cargo >= max_cargo)))		return NO;
	{
		NSMutableArray* manifest =  [NSMutableArray arrayWithArray:shipCommodityData];
		NSMutableArray* manifest_commodity =	[NSMutableArray arrayWithArray:(NSArray *)[manifest objectAtIndex:index]];
		NSMutableArray* market_commodity =		[NSMutableArray arrayWithArray:(NSArray *)[localMarket objectAtIndex:index]];
		int manifest_quantity = [(NSNumber *)[manifest_commodity objectAtIndex:MARKET_QUANTITY] intValue];
		int market_quantity =   [(NSNumber *)[market_commodity objectAtIndex:MARKET_QUANTITY] intValue];
		manifest_quantity++;
		market_quantity--;
		credits -= price_per_unit;
		if (unit == 0)
			current_cargo++;
		[manifest_commodity replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:manifest_quantity]];
		[market_commodity replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:market_quantity]];
		[manifest replaceObjectAtIndex:index withObject:[NSArray arrayWithArray:manifest_commodity]];
		[localMarket replaceObjectAtIndex:index withObject:[NSArray arrayWithArray:market_commodity]];

		[shipCommodityData release];
		shipCommodityData = [[NSArray arrayWithArray:manifest] retain];
	}
	return YES;
}

- (BOOL) trySellingCommodity:(int) index
{
	NSMutableArray*			localMarket;
	if (status == STATUS_DOCKED)
	{
		if (docked_station == nil)
			docked_station = [universe station];
		if ([docked_station localMarket])
			localMarket = [docked_station localMarket];
		else
			localMarket = [docked_station initialiseLocalMarketWithSeed:system_seed andRandomFactor:market_rnd];
	}
	else
	{
		return NO; // can't sell if not docked
	}
	int available_units = [(NSNumber *)[(NSArray *)[shipCommodityData objectAtIndex:index] objectAtIndex:MARKET_QUANTITY] intValue];
	int price_per_unit = [(NSNumber *)[(NSArray *)[localMarket objectAtIndex:index] objectAtIndex:MARKET_PRICE] intValue];
	if (available_units == 0)		return NO;
	{
		NSMutableArray* manifest =  [NSMutableArray arrayWithArray:shipCommodityData];
		NSMutableArray* manifest_commodity =	[NSMutableArray arrayWithArray:(NSArray *)[manifest objectAtIndex:index]];
		NSMutableArray* market_commodity =		[NSMutableArray arrayWithArray:(NSArray *)[localMarket objectAtIndex:index]];
		int manifest_quantity = [(NSNumber *)[manifest_commodity objectAtIndex:MARKET_QUANTITY] intValue];
		int market_quantity =   [(NSNumber *)[market_commodity objectAtIndex:MARKET_QUANTITY] intValue];
		manifest_quantity--;
		market_quantity++;
		credits += price_per_unit;
		[manifest_commodity replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:manifest_quantity]];
		[market_commodity replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:market_quantity]];
		[manifest replaceObjectAtIndex:index withObject:[NSArray arrayWithArray:manifest_commodity]];
		[localMarket replaceObjectAtIndex:index withObject:[NSArray arrayWithArray:market_commodity]];
		[shipCommodityData release];
		shipCommodityData = [[NSArray arrayWithArray:manifest] retain];
	}
	return YES;
}

- (BOOL) isMining
{
	return using_mining_laser;
}

- (BOOL) speech_on
{
	return speech_on;
}

@end
