//
//  RingEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "RingEntity.h"
#import "entities.h"

#import "Universe.h"


@implementation RingEntity

- (void) setLifetime:(double) amount
{
	lifetime = amount;
}

- (id) init
{    
	self = [super init];

	[self setModel:@"ring.dat"];
	
	lifetime = 50.0;
	status = STATUS_EFFECT;
	
	velocity.x = 0.0;
	velocity.y = 0.0;
	velocity.z = 1.0;
	
    return self;
}

- (void) update:(double) delta_t
{	
	[super update:delta_t];
			
    {
		double movement = RING_SPEED * delta_t;
		position.x -= movement * velocity.x; // swap out for setting a velocity vector
		position.y -= movement * velocity.y; // swap out for setting a velocity vector
		position.z -= movement * velocity.z; // swap out for setting a velocity vector
		lifetime -= movement;
		if (lifetime < 0.0)
		{
			//NSLog(@"removing ring %@ movement %.3f delta_t %.3f",self,movement,delta_t);
			[universe removeEntity:self];
		}
    }
}

- (void) drawEntity:(BOOL) immediate :(BOOL) translucent
{
    // roll out each face and vertex in turn
    //
    int fi,vi;
	glShadeModel(GL_SMOOTH);
	glDisable(GL_LIGHTING);			// face culling
    //
	if ((translucent)||(immediate))
	{
		if (basefile)
		{
			//NSLog(@"Drawing ring : %@ n_faces %d, n_vertices %d", basefile, n_faces, n_vertices);
			if (immediate)
			{
				glBegin(GL_TRIANGLES);
				for (fi = 0; fi < n_faces; fi++)
				{
					GLfloat amb_diff1[] = { 1.0, 0.0, 0.0, 0.5};
					GLfloat amb_diff2[] = { 0.0, 0.0, 1.0, 0.25};
									
					glNormal3f(faces[fi].normal.x, faces[fi].normal.y, faces[fi].normal.z);
					
					for (vi = 0; vi < faces[fi].n_verts; vi++)
					{
						int v = faces[fi].vertex[vi];
						if (vertices[v].z < -20.0)
						{
							// color for inner edge of ring
							glColor4fv(amb_diff1);
						}
						else
						{
							// color for outer edge of ring
							glColor4fv(amb_diff2);
						}
						glVertex3f(vertices[v].x, vertices[v].y, vertices[v].z);
					}
				}
				glEnd();
			}
			else
			{
				if (displayListName != 0)
					glCallList(displayListName);
				else
					[self generateDisplayList];
			}
		}
	}
	glEnable(GL_LIGHTING);			// face culling
}

- (BOOL) canCollide
{
	return NO;
}


@end
