//
//  ShipEntity.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <Foundation/Foundation.h>

#import "Entity.h"

#define CONDITION_IDLE			0
#define CONDITION_TRACK_TARGET	1
#define CONDITION_FLY_TO_TARGET	2
#define CONDITION_HANDS_OFF		3
#define CONDITION_TUMBLE		4

#define CONDITION_ATTACK_TARGET				100
#define CONDITION_ATTACK_FLY_TO_TARGET		101
#define CONDITION_ATTACK_FLY_FROM_TARGET	102

#define CONDITION_FLY_RANGE_FROM_DESTINATION	200
#define CONDITION_FLY_TO_DESTINATION			201
#define CONDITION_FLY_FROM_DESTINATION			202
#define CONDITION_FACE_DESTINATION				203

#define CONDITION_COLLECT_TARGET			300
#define CONDITION_INTERCEPT_TARGET			350

#define CONDITION_FLEE_TARGET				400

#define CONDITION_MISSILE_FLY_TO_TARGET		901

#define MAX_TARGETS				24
#define RAIDER_MAX_CARGO					5
#define MERCHANTMAN_MAX_CARGO				125

#define LAUNCH_DELAY					2.0

#define WEAPON_NONE						0
#define WEAPON_PLASMA_CANNON			1
#define WEAPON_PULSE_LASER				2
#define WEAPON_BEAM_LASER				3
#define WEAPON_MINING_LASER				4
#define WEAPON_MILITARY_LASER			5

#define CARGO_NOT_CARGO					-1
#define CARGO_SLAVES					3
#define CARGO_MINERALS					12
#define CARGO_ALLOY						9
#define CARGO_THARGOID					16
#define CARGO_RANDOM					100

#define CARGO_FLAG_NONE					400
#define CARGO_FLAG_FULL_PLENTIFUL		501
#define CARGO_FLAG_FULL_SCARCE			502
#define CARGO_FLAG_CANISTERS			600

#define PIRATES_PREFER_PLAYER			YES

#define AEGIS_NONE						0
#define AEGIS_CLOSE_TO_PLANET			1
#define AEGIS_IN_DOCKING_RANGE			2


@class StationEntity, ParticleEntity, PlanetEntity, AI;

@interface ShipEntity : Entity {
	
	//scripting
	NSMutableArray *launch_actions;
	NSMutableArray *death_actions;
	
	NSArray *sub_entities;
	NSColor *laser_color;
	
	// per ship-type variables
	//
	double  max_flight_speed;		// top speed			(160.0 for player)  (200.0 for fast raider)
	double  max_flight_roll;		// maximum roll rate	(2.0 for player)	(3.0 for fast raider)	
	double  max_flight_pitch;		// maximum pitch rate   (1.0 for player)	(1.5 for fast raider)
	
	double  thrust;					// acceleration
	
	BOOL	has_ecm;				// anti-missile system
	BOOL	has_scoop;				// fuel/cargo scoops
	BOOL	has_escape_pod;			// escape pod
	
	int		likely_cargo;			// likely amount of cargo (for merchantmen, this is what is spilled as loot)
	int		max_cargo;				// capacity of cargo hold
	int		bounty;					// bounty (if any)
	int		cargo_type;				// if this is scooped, this is indicates contents
	int		cargo_flag;				// indicates contents for merchantmen
	
	double  energy_recharge_rate;   // recharge rate for energy banks
	
	int		forward_weapon_type;	// type of forward weapon (allows lasers, plasma cannon, others)
	double  weapon_energy;			// energy used/delivered by weapon
	double  weapon_offset_x;		// if weapon is in twin configuration this is the distance from the centerline to the weapon
	
	int		missiles;				// number of on-board missiles
	
	AI*		shipAI;					// ship's AI system
	
	NSString*   name;				// descriptive name
	NSString*   roles;				// names fo roles a ship can take, eg. trader, hunter, police, pirate, scavenger &c.
	
	// AI stuff
	//
	Vector		jink;				// x and y set factors for offsetting a pursuing ship's position
	Vector		destination;		// for flying to/from a set point
	Vector		coordinates;		// for flying to/from a set point
	int			primaryTarget;		// for combat or rendezvous
	int			primaryAggressor;   // recorded after an attack
	int			found_target;		// from scans
	double		desired_range;		// range to which to journey/scan
	double		desired_speed;		// speed at which to travel
	double		launch_time;		// time at which launched
	
	int		condition;					// ship's behavioural state
	
	// derived variables
	//
	double  shot_time;					// time elapsed since last shot was fired
	double  weapon_recharge_rate;		// time between shots
	int		shot_counter;				// number of shots fired
	double  cargo_dump_time;			// time cargo was last dumped
	
	NSMutableArray* cargo;				// cargo containers go in here
	
	int commodity_type;					// type of commodity in a container
	int commodity_amount;				// 1 if unit is TONNES (0), possibly more if precious metals KILOGRAMS (1)
										// or gem stones GRAMS (2) 
	
	// navigation
	//
	Vector v_forward, v_up, v_right;	// unit vectors derived from the direction faced
		
	double flight_speed;				// current speed
	double flight_roll;					// current roll rate
	double flight_pitch;				// current pitch rate
	
	BOOL	pitching_over;				// set to YES if executing a sharp loop

	BOOL	within_station_aegis;		// set to YES when within the station's protective zone
	
	BOOL	reportAImessages;			// normally NO, suppressing AI message reporting
	
	BOOL	being_mined;				// normally NO, set to Yes when fired on by mining laser

}

- (void) reinit;

- (id) initWithDictionary:(NSDictionary *) dict;
- (void) setUpShipFromDictionary:(NSDictionary *) dict;

- (void) addExhaust:(ParticleEntity *) exhaust;
- (void) addExhaustAt:(Vector) ex_position withScale:(Vector) ex_scale;

- (void) applyThrust:(double) delta_t;

- (NSString *) name;
- (NSString *) roles;

- (BOOL) reportAImessages;
- (void) setReportAImessages:(BOOL) yn;

- (int) checkForAegis;
- (BOOL) within_station_aegis;

- (void) setAI:(AI *) ai;
- (AI *) getAI;

- (void) setRoll:(double) amount;
- (void) setPitch:(double) amount;

- (void) setThrust:(double) amount;

- (void) setBounty:(int) amount;
- (int) getBounty;
- (int) legal_status;

- (void) setCommodity:(int) co_type andAmount:(int) co_amount;
- (int) getCommodityType;
- (int) getCommodityAmount;

- (int) getMaxCargo;
- (int) getCargoType;
- (void) setCargo:(NSArray *) some_cargo;

- (int) cargoFlag;
- (void) setCargoFlag:(int) flag;

- (void) setSpeed:(double) amount;

- (void) increase_flight_speed:(double) delta;
- (void) decrease_flight_speed:(double) delta;
- (void) increase_flight_roll:(double) delta;
- (void) decrease_flight_roll:(double) delta;
- (void) increase_flight_pitch:(double) delta;
- (void) decrease_flight_pitch:(double) delta;

- (double) flight_roll;
- (double) flight_pitch;
- (double) flight_speed;
- (double) speed_factor;

- (int) damage;

- (void) becomeExplosion;
- (void) becomeLargeExplosion:(double) factor;

- (void) collectBountyFor:(ShipEntity *)other;

/*-----------------------------------------

	AI piloting methods

-----------------------------------------*/

- (void) setPrimaryAggressor:(Entity *) targetEntity;
- (void) addTarget:(Entity *) targetEntity;
- (void) removeTarget:(Entity *) targetEntity;
- (Entity *) getPrimaryTarget;

- (void) setCondition:(int) cond;

- (double) rangeToDestination;
- (double) trackDestination:(double) delta_t :(BOOL) retreat;

- (double) trackPrimaryTarget:(double) delta_t :(BOOL) retreat;
- (double) rangeToPrimaryTarget;
- (BOOL) onTarget;

- (BOOL) fireMainWeapon:(double) range;
- (void) setLaserColor:(NSColor *) color;
- (BOOL) fireLaserShot;
- (BOOL) firePlasmaShot:(double) offset :(double) speed :(NSColor *) color;
- (BOOL) fireMissile;
- (BOOL) fireTharglet;
- (BOOL) fireECM;
- (int) launchEscapeCapsule;
- (int) dumpCargo;

- (void) manageCollisions;
- (BOOL) collideWithShip:(ShipEntity *)other;
- (void) scoopUp:(ShipEntity *)other;
- (BOOL) beCollidedBy:(ShipEntity *)other;
- (void) takeScrapeDamage:(double) amount from:(Entity *) ent;

- (void) enterDock:(StationEntity *)station;
- (void) leaveDock:(StationEntity *)station;

- (void) enterWitchspace;
- (void) leaveWitchspace;

- (void) markAsOffender:(int)offence_value;

/*-----------------------------------------

	methods from AI

-----------------------------------------*/

- (void) pauseAI:(NSString *)intervalString;

- (void) setDestinationToCurrentLocation;

- (void) setDesiredRangeTo:(NSString *)rangeString;

- (void) setSpeedTo:(NSString *)speedString;

- (void) setSpeedFactorTo:(NSString *)speedString;

- (void) performFlyToRangeFromDestination;

- (void) performIdle;

- (void) setTargetToPrimaryAggressor;

- (void) performAttack;

- (void) scanForMerchantmen;

- (void) scanForLoot;

- (void) scanForRandomLoot;

- (void) setTargetToFoundTarget;

- (void) checkForFullHold;

- (void) performCollect;

- (void) performIntercept;

- (void) performFlee;

- (void) requestDockingCoordinates;

- (void) getWitchspaceEntryCoordinates;

- (void) abortDocking;

- (void) setDestinationFromCoordinates;

- (void) performDocking;

- (void) performFaceDestination;

- (void) setTargetToStation;

- (void) performTumble;

- (void) fightOrFleeMissile;

// new
- (PlanetEntity *) findNearestPlanet;
- (PlanetEntity *) findNearestLargeBody;

- (void) setCourseToPlanet;

- (void) setTakeOffFromPlanet;

- (void) landOnPlanet;

- (void) setAITo:(NSString *)aiString;

- (void) checkTargetLegalStatus;

- (void) exitAI;

- (void) setDestinationToTarget;

- (void) checkCourseToDestination;

- (void) scanForOffenders;

- (void) setCourseToWitchpoint;

- (void) performHyperSpaceExit;

- (void) broadcastDistressMessage;
- (void) acceptDistressMessageFrom:(ShipEntity *)other;

- (void) broadcastThargoidDestroyed;
- (void) scanForThargoid;
- (void) scanForNonThargoid;


@end
