//
//  ShipEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "ShipEntity.h"
#import "entities.h"

#import "vector.h"
#import "Universe.h"

#import "AI.h"


@implementation ShipEntity

- (id) init
{    
    self = [super init];
	//
	// scripting
	launch_actions = [[NSMutableArray alloc] initWithCapacity:4];
	death_actions = [[NSMutableArray alloc] initWithCapacity:4];
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
	//
	v_forward   = vector_forward_from_quaternion(q_rotation);
	v_up		= vector_up_from_quaternion(q_rotation);
	v_right		= vector_right_from_quaternion(q_rotation);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
	//
	max_flight_speed = 200.0;
	max_flight_roll = 3.0;
	max_flight_pitch = 1.5;
	//
	flight_speed = 0.0;
	flight_roll = 0.0;
	flight_pitch = 0.0;
	//
	thrust = 80.0;
	//
	pitching_over = NO;
	//
	energy = 100.0;
	max_energy = 100.0;
	energy_recharge_rate = 2.0;
	//
	forward_weapon_type = WEAPON_PLASMA_CANNON;
	weapon_energy = 8.0;
	weapon_recharge_rate = 1.0/3.0;
	weapon_offset_x = 10.0;
	shot_time = 0.0;
	//
	launch_time = 0.0;
	//
	cargo_dump_time = 0.0;
	//
	missiles = 3;
	has_ecm = NO;
	has_scoop = NO;
	has_escape_pod = NO;
	//
	bounty = 0;
	//
	primaryTarget = NO_TARGET;
	//
	condition = CONDITION_IDLE;
	//
	shipAI = [[AI alloc] init]; // alloc retains
	[shipAI setOwner:self];
	[shipAI setState:@"GLOBAL"];
	//
	max_cargo = RAIDER_MAX_CARGO;
	likely_cargo = 0;
	cargo_type = 0;
	cargo = [[NSMutableArray alloc] initWithCapacity:max_cargo]; // alloc retains;
	cargo_flag = CARGO_FLAG_NONE;
	//
	[self setOwner:self];
	//
	reportAImessages = NO;
	//
	sub_entities = nil;
	//
	laser_color = [[NSColor redColor] retain];
	//
    return self;
}

- (void) dealloc
{
	if (shipAI)				[shipAI release];
	if (cargo)				[cargo release];
	if (name)				[name release];
	if (roles)				[roles release];
	if (sub_entities)		[sub_entities release];
	if (laser_color)		[laser_color release];
	//scripting
	if (launch_actions)		[launch_actions release];
	if (death_actions)		[death_actions release];
	[super dealloc];
}

- (void) setUniverse:(Universe *)univ
{
    if (univ)
    {
        if (universe)	[universe release];
        universe = [univ retain];
    }
	else
	{
        if (universe)	[universe release];
        universe = nil;
    }
	
	if (sub_entities != nil)
	{
		int i;
		for (i = 0; i < [sub_entities count]; i++)
		{
			[(Entity *)[sub_entities objectAtIndex:i] setUniverse:univ];
		}
	}
}



- (void) reinit
{
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
	//
	v_forward   = vector_forward_from_quaternion(q_rotation);
	v_up		= vector_up_from_quaternion(q_rotation);
	v_right		= vector_right_from_quaternion(q_rotation);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
	//
	zero_distance = SCANNER_MAX_RANGE2 * 2.0;   // beyond scanner range to avoid the momentary blip
	//
	max_flight_speed = 200.0;
	max_flight_roll = 3.0;
	max_flight_pitch = 1.5;
	//
	flight_speed = 0.0;
	flight_roll = 0.0;
	flight_pitch = 0.0;
	//
	thrust = 80.0;
	//
	pitching_over = NO;
	//
	energy = 100.0;
	max_energy = 100.0;
	energy_recharge_rate = 2.0;
	//
	forward_weapon_type = WEAPON_PLASMA_CANNON;
	weapon_energy = 8.0;
	weapon_recharge_rate = 1.0/3.0;
	weapon_offset_x = 10.0;
	shot_time = 0.0;
	//
	launch_time = 0.0;
	//
	cargo_dump_time = 0.0;
	//
	missiles = 3;
	has_ecm = NO;
	has_scoop = NO;
	has_escape_pod = NO;
	//
	bounty = 0;
	//
	primaryTarget = NO_TARGET;
	//
	condition = CONDITION_IDLE;
	//
	if (!shipAI)
		shipAI = [[AI alloc] init]; // alloc retains
	[shipAI setOwner:self];
	[shipAI setState:@"GLOBAL"];
	//
	max_cargo = RAIDER_MAX_CARGO;
	likely_cargo = 0;
	cargo_type = 0;
	cargo_flag = CARGO_FLAG_NONE;
	if (!cargo)
		cargo = [[NSMutableArray alloc] initWithCapacity:max_cargo]; // alloc retains;
	[cargo removeAllObjects];
	//
	[self setOwner:self];
	//
	reportAImessages = NO;
	//
	if (sub_entities) [sub_entities release];
	sub_entities = nil;
	//
}

- (id) initWithDictionary:(NSDictionary *) dict
{
	self = [super init];
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
	//
	v_forward   = vector_forward_from_quaternion(q_rotation);
	v_up		= vector_up_from_quaternion(q_rotation);
	v_right		= vector_right_from_quaternion(q_rotation);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
	//
	flight_speed = 0.0;
	flight_roll = 0.0;
	flight_pitch = 0.0;
	//
	pitching_over = NO;
	//
	primaryTarget = NO_TARGET;
	//
	condition = CONDITION_IDLE;
	//
	[self setUpShipFromDictionary:dict];
	//
	reportAImessages = NO;
	//
	return self;
}


- (void) setUpShipFromDictionary:(NSDictionary *) dict
{
    NSString*   cargo_type_string;
    NSString*   ai_type_string;
    NSString*   weapon_type_string;
	
	// reset all settings
	[self reinit];
	
	//
	// set things from dictionary from here out
	//
	[self setValue:[dict objectForKey:@"max_flight_speed"]  forKey:@"max_flight_speed"];
	[self setValue:[dict objectForKey:@"max_flight_roll"]   forKey:@"max_flight_roll"];
	[self setValue:[dict objectForKey:@"max_flight_pitch"]  forKey:@"max_flight_pitch"];
	//
	[self setValue:[dict objectForKey:@"thrust"]  forKey:@"thrust"];
	//
	[self setValue:[dict objectForKey:@"max_energy"]			forKey:@"max_energy"];
	[self setValue:[dict objectForKey:@"energy_recharge_rate"]  forKey:@"energy_recharge_rate"];
	energy = max_energy;
	//
	[self setValue:[dict objectForKey:@"weapon_offset_x"]   forKey:@"weapon_offset_x"];
	weapon_type_string = (NSString *)[dict objectForKey:@"forward_weapon_type"];
	if ([weapon_type_string isEqual:@"WEAPON_PULSE_LASER"])
	{
		forward_weapon_type = WEAPON_PULSE_LASER;
		weapon_recharge_rate = 1.0/3.0;
	}
	if ([weapon_type_string isEqual:@"WEAPON_BEAM_LASER"])
	{
		forward_weapon_type = WEAPON_BEAM_LASER;
		weapon_recharge_rate = 1.0/4.0;
	}
	if ([weapon_type_string isEqual:@"WEAPON_MINING_LASER"])
	{
		forward_weapon_type = WEAPON_MINING_LASER;
		weapon_recharge_rate = 1.0/2.0;
	}
	if ([weapon_type_string isEqual:@"WEAPON_MILITARY_LASER"])
	{
		forward_weapon_type = WEAPON_MILITARY_LASER;
		weapon_recharge_rate = 1.0/5.0;
	}
	if ([weapon_type_string isEqual:@"WEAPON_PLASMA_CANNON"])
	{
		forward_weapon_type = WEAPON_PLASMA_CANNON;
		weapon_recharge_rate = 1.0/3.0;
	}
	if ([weapon_type_string isEqual:@"WEAPON_NONE"])
	{
		forward_weapon_type = WEAPON_NONE;
		weapon_recharge_rate = 0.0;
	}
	//
	[self setValue:[dict objectForKey:@"weapon_energy"]  forKey:@"weapon_energy"];
	//
	[self setValue:[dict objectForKey:@"missiles"]			forKey:@"missiles"];
	[self setValue:[dict objectForKey:@"has_ecm"]			forKey:@"has_ecm"];
	[self setValue:[dict objectForKey:@"has_scoop"]			forKey:@"has_scoop"];
	[self setValue:[dict objectForKey:@"has_escape_pod"]	forKey:@"has_escape_pod"];
	//
	[self setValue:[dict objectForKey:@"bounty"]	forKey:@"bounty"];
	//
	if (shipAI)
		[shipAI autorelease];
	shipAI = [[AI alloc] init]; // alloc retains
	[shipAI setOwner:self];
	ai_type_string = (NSString *)[dict objectForKey:@"ai_type"];
	if ([ai_type_string isEqual:@"AI_PLAYER"])
		[shipAI setStateMachine:@"dockingAI.plist"];
	if ([ai_type_string isEqual:@"AI_PIRATE"])
		[shipAI setStateMachine:@"pirateAI.plist"];
	if ([ai_type_string isEqual:@"AI_NONE"])
		[shipAI setStateMachine:@"nullAI.plist"];
	if ([ai_type_string isEqual:@"AI_TUMBLE"])
		[shipAI setStateMachine:@"dumbAI.plist"];
	if ([ai_type_string isEqual:@"AI_HOME"])
		[shipAI setStateMachine:@"homeAI.plist"];
	if ([ai_type_string isEqual:@"AI_MISSILE"])
		[shipAI setStateMachine:@"missileAI.plist"];
	if ([ai_type_string isEqual:@"AI_STATION"])
		[shipAI setStateMachine:@"stationAI.plist"];
	if ([ai_type_string isEqual:@"AI_THARGLET"])
		[shipAI setStateMachine:@"thargletAI.plist"];
	if ([ai_type_string isEqual:@"AI_THARGOID"])
		[shipAI setStateMachine:@"thargoidAI.plist"];
	if ([ai_type_string isEqual:@"AI_HERMIT"])
		[shipAI setStateMachine:@"rockHermitAI.plist"];
	[shipAI setState:@"GLOBAL"];
	//
	[self setValue:[dict objectForKey:@"max_cargo"]   forKey:@"max_cargo"];
	[self setValue:[dict objectForKey:@"likely_cargo"] forKey:@"likely_cargo"];
	//
	cargo_type_string = (NSString *)[dict objectForKey:@"cargo_type"];
	if ([cargo_type_string isEqual:@"CARGO_THARGOID"])
		cargo_type = CARGO_THARGOID;
	if ([cargo_type_string isEqual:@"CARGO_ALLOY"])
		cargo_type = CARGO_ALLOY;
	if ([cargo_type_string isEqual:@"CARGO_MINERALS"])
		cargo_type = CARGO_MINERALS;
	if ([cargo_type_string isEqual:@"CARGO_SLAVES"])
		cargo_type = CARGO_SLAVES;
	if ([cargo_type_string isEqual:@"CARGO_NOT_CARGO"])
		cargo_type = CARGO_NOT_CARGO;
	if ([cargo_type_string isEqual:@"CARGO_RANDOM"])
		cargo_type = CARGO_RANDOM;
	if (cargo)
		[cargo autorelease];
	cargo = [[NSMutableArray alloc] initWithCapacity:max_cargo]; // alloc retains;
	//
	// A HACK!! - must do this before the model is set
	if ([dict objectForKey:@"smooth"])
		is_smooth_shaded = YES;
	else
		is_smooth_shaded = NO;
	//
	[self setModel:(NSString *)[dict objectForKey:@"model"]];
	//
	[self setValue:[dict objectForKey:@"name"] forKey:@"name"];
	//
	[self setValue:[dict objectForKey:@"roles"] forKey:@"roles"];
	//
	[self setOwner:self];
	//
	if ([dict objectForKey:@"exhaust"])
	{
		int i;
		NSArray *plumes = (NSArray *)[dict objectForKey:@"exhaust"];
		for (i = 0; i < [plumes count]; i++)
		{
			ParticleEntity *exhaust = [[ParticleEntity alloc] initExhaustFromShip:self details:(NSString *)[plumes objectAtIndex:i]];
			[self addExhaust:exhaust];
			[exhaust release];
		}
	}		
	//	
	if ([dict objectForKey:@"laser_color"])
	{
		NSString *laser_color_string = (NSString *)[dict objectForKey:@"laser_color"];
		SEL color_selector = NSSelectorFromString(laser_color_string);
		if ([NSColor respondsToSelector:color_selector])
		{
			id  color_thing = [NSColor performSelector:color_selector];
			if ([color_thing isKindOfClass:[NSColor class]])
				[self setLaserColor:(NSColor *)color_thing];
		}
	}   
	else
		[self setLaserColor:[NSColor redColor]];
	//
	// scan class
	if ([dict objectForKey:@"scanClass"])
	{
		NSString *s_class= (NSString *)[dict objectForKey:@"scanClass"];
		
		//NSLog(@"----- initialising ship with scan class '%@'",s_class);
		
		scan_class = CLASS_NEUTRAL;
		if ([s_class isEqual:@"CLASS_STATION"])
			scan_class = CLASS_STATION;
		if ([s_class isEqual:@"CLASS_THARGOID"])
			scan_class = CLASS_THARGOID;
		if ([s_class isEqual:@"CLASS_TARGET"])
			scan_class = CLASS_TARGET;
		if ([s_class isEqual:@"CLASS_CARGO"])
			scan_class = CLASS_CARGO;
		if ([s_class isEqual:@"CLASS_POLICE"])
			scan_class = CLASS_POLICE;
		if ([s_class isEqual:@"CLASS_BUOY"])
			scan_class = CLASS_BUOY;
		if ([s_class isEqual:@"CLASS_NO_DRAW"])
			scan_class = CLASS_NO_DRAW;
		if ([s_class isEqual:@"CLASS_NEUTRAL"])
			scan_class = CLASS_NEUTRAL;
	}
	//
	// scripting	
	if ([dict objectForKey:@"launch_actions"])
		[launch_actions addObjectsFromArray:(NSArray *)[dict objectForKey:@"launch_actions"]];
	if ([dict objectForKey:@"death_actions"])
		[death_actions addObjectsFromArray:(NSArray *)[dict objectForKey:@"death_actions"]];
	//
	if ([self isKindOfClass:[StationEntity class]])
	{
		if ([dict objectForKey:@"port_radius"])   // this gets set for rock-hermits and other specials, otherwise it's 500m
			[self setValue:[dict objectForKey:@"port_radius"] forKey:@"port_radius"];
	}
		
	//if ([self isSmoothShaded])
	//	//NSLog(@"%@ is smooth shaded", name);

}
//////////////////////////////////////////////

- (BOOL) canCollide
{
	if (status == STATUS_DEMO)
		return NO;
	if (([self scanClass] == CLASS_MISSILE)&&(shot_time < 0.25)) // not yet fused
		return NO;
	return YES;
}

- (BOOL) checkCloseCollisionWith:(Entity *)other
{
	if ([other isKindOfClass:[ShipEntity class]])
	{
		// check bounding boxes ...
		//
		// get bounding box relative to this ship's orientation
		BoundingBox arbb = [other findBoundingBoxRelativeTo:self InVectors: v_right: v_up: v_forward];
		
		// construct 6 rectangles based on the sides of the possibly overlapping bounding boxes
		NSRect  ship_x_rect = NSMakeRect(boundingBox.min_z, boundingBox.min_y, boundingBox.max_z - boundingBox.min_z, boundingBox.max_y - boundingBox.min_y);
		NSRect  ship_y_rect = NSMakeRect(boundingBox.min_x, boundingBox.min_z, boundingBox.max_x - boundingBox.min_x, boundingBox.max_z - boundingBox.min_z);
		NSRect  ship_z_rect = NSMakeRect(boundingBox.min_x, boundingBox.min_y, boundingBox.max_x - boundingBox.min_x, boundingBox.max_y - boundingBox.min_y);

		NSRect  other_x_rect = NSMakeRect(arbb.min_z, arbb.min_y, arbb.max_z - arbb.min_z, arbb.max_y - arbb.min_y);
		NSRect  other_y_rect = NSMakeRect(arbb.min_x, arbb.min_z, arbb.max_x - arbb.min_x, arbb.max_z - arbb.min_z);
		NSRect  other_z_rect = NSMakeRect(arbb.min_x, arbb.min_y, arbb.max_x - arbb.min_x, arbb.max_y - arbb.min_y);

		return (NSIntersectsRect(ship_x_rect,other_x_rect))&&(NSIntersectsRect(ship_y_rect,other_y_rect))&&(NSIntersectsRect(ship_z_rect,other_z_rect));
	}
	return YES;
}

- (void) update:(double) delta_t
{
	double confidenceFactor;
	int missile_chance = 1 + (rand() % (int)( 32 * 0.1 / delta_t));
	double hurt_factor = 16 * pow(energy/max_energy, 4.0);
	[super update:delta_t];
	
	// update time between shots
	//
	shot_time +=delta_t;

	// deal with collisions
	//
	[self manageCollisions];
	
	// check outside factors
	//
	[self checkForAegis];   // is a station nearby??
	
	//scripting
	if ((status == STATUS_IN_FLIGHT)&&([launch_actions count]))
	{
		int i;
		for (i = 0; i < [launch_actions count]; i++)
			[(PlayerEntity *)[universe entityZero] scriptAction:(NSString *)[launch_actions objectAtIndex:i]];
		[launch_actions removeAllObjects];
	}
	
	// behaviours according to status and condition
    //
	if (status == STATUS_LAUNCHING)
	{
		if ([universe getTime] > launch_time + LAUNCH_DELAY)		// move for while before thinking
			status = STATUS_IN_FLIGHT;
		else
		{
			// ignore condition just keep moving...
			[self applyRoll:delta_t*flight_roll andClimb:delta_t*flight_pitch];
			[self applyThrust:delta_t];
			if (energy < max_energy)
			{
				energy += energy_recharge_rate * delta_t;
				if (energy > max_energy)
				{
					energy = max_energy;
					[shipAI message:@"ENERGY_FULL"];
				}
			}
			return;
		}
	}
	//
	if (status == STATUS_DEMO)
    {
        [self applyRoll:delta_t*flight_roll andClimb:delta_t*flight_pitch];
		position.x += delta_t*velocity.x;
		position.y += delta_t*velocity.y;
		position.z += delta_t*velocity.z;
		if (position.z <= collision_radius * 3.6)
		{
			position.z = collision_radius * 3.6;
			velocity.z = 0.0;
		}
		return;
    }
	else
	{
		double  range = [self rangeToPrimaryTarget];
		double  distance = [self rangeToDestination];
		
		if (primaryTarget != NO_TARGET)
		{
			// check if the target is still with us...
			if (![universe entityForUniversalID:primaryTarget]) // oops! It's no longer a parrot, it has ceased to be, it has joined the choir invisible...
			{
				//NSLog(@"---> %@ %d has lost its target.", name, universal_id);
				[shipAI message:@"TARGET_LOST"];
				primaryTarget = NO_TARGET;
			}
		}
				
		switch (condition)
		{
			case CONDITION_TUMBLE :
			case CONDITION_IDLE :
				break;

			case CONDITION_TRACK_TARGET :
				[self trackPrimaryTarget:delta_t:NO];
				break;
				
			case CONDITION_INTERCEPT_TARGET :
			case CONDITION_COLLECT_TARGET :
				if (condition == CONDITION_INTERCEPT_TARGET)
				{
					desired_speed = max_flight_speed;
					if (range < desired_range)
						[shipAI reactToMessage:@"DESIRED_RANGE_ACHIEVED"];
				}
				else
				{
					double target_speed = [[self getPrimaryTarget] getVelocityAsSpeed];
					double eta = range / (flight_speed - target_speed);
					if ((eta < 3.0)&&(flight_speed > max_flight_speed * 0.05))
						desired_speed = flight_speed * 0.75;   // cut speed to a minimum of 5 % speed
					else
						desired_speed = max_flight_speed;
					if (desired_speed < target_speed)
					{
						desired_speed += target_speed;
						if (target_speed > max_flight_speed)
							[shipAI reactToMessage:@"TARGET_LOST"];
					}
				}
				[self trackPrimaryTarget:delta_t:NO];
				break;
				
			case CONDITION_ATTACK_TARGET :
				desired_speed = max_flight_speed;
				if (range < 350.0)
					condition = CONDITION_ATTACK_FLY_FROM_TARGET;
				else
					condition = CONDITION_ATTACK_FLY_TO_TARGET;
				break;
				
			case CONDITION_ATTACK_FLY_TO_TARGET :
				if (range < 350.0)
				{
					jink.x = (rand() % 256) - 128.0;
					jink.y = (rand() % 256) - 128.0;
					jink.z = 1000.0;
					condition = CONDITION_ATTACK_FLY_FROM_TARGET;
					desired_speed = max_flight_speed;
				}
				else
				{
					if (range > SCANNER_MAX_RANGE)
					{
						condition = CONDITION_IDLE;
						[shipAI message:@"TARGET_LOST"];
					}
				}
				[self trackPrimaryTarget:delta_t:NO];
				
				// trying this out
				if (missiles > missile_chance * hurt_factor)
				{
					//NSLog(@"]==> firing missile : missiles %d, missile_chance %d, hurt_factor %.3f", missiles, missile_chance, hurt_factor);
					[self fireMissile];
				}
				[self fireMainWeapon:range];	// if necessary
				break;
				
			case CONDITION_ATTACK_FLY_FROM_TARGET :
				if (range > 2500.0 + 15.0 * jink.x)
				{
					jink.x = 0.0;
					jink.y = 0.0;
					jink.z = 0.0;
					condition = CONDITION_ATTACK_FLY_TO_TARGET;
					desired_speed = max_flight_speed;
				}
				[self trackPrimaryTarget:delta_t:YES];
				break;
				
			case CONDITION_FLEE_TARGET :
				if (range > SCANNER_MAX_RANGE)
				{
					[shipAI message:@"REACHED_SAFETY"];
				}
				else
				{
					desired_speed = max_flight_speed;
				}
				[self trackPrimaryTarget:delta_t:YES];
				break;
				
			case CONDITION_FLY_RANGE_FROM_DESTINATION :
				if (distance < desired_range)
					condition = CONDITION_FLY_FROM_DESTINATION;
				else
					condition = CONDITION_FLY_TO_DESTINATION;
				break;
				
			case CONDITION_FACE_DESTINATION :
				desired_speed = 0.0;
				confidenceFactor = [self trackDestination:delta_t:NO];
				if (confidenceFactor > 0.99)
				{
					// desired facing achieved
					[shipAI message:@"FACING_DESTINATION"];
					condition = CONDITION_IDLE;
				}
				break;
				
			case CONDITION_FLY_TO_DESTINATION :
				if (distance < desired_range)
				{
					// desired range achieved
					[shipAI message:@"DESIRED_RANGE_ACHIEVED"];
					condition = CONDITION_IDLE;
					desired_speed = 0.0;
				}
				else
				{
					double eta = (distance - desired_range) / flight_speed;
					if ((eta < 2.0)&&(flight_speed > max_flight_speed * 0.10))
						desired_speed = flight_speed * 0.50;   // cut speed to a minimum of 10 % speed
				}
				[self trackDestination:delta_t:NO];
				break;
				
			case CONDITION_FLY_FROM_DESTINATION :
				if (distance > desired_range)
				{
					// desired range achieved
					[shipAI message:@"DESIRED_RANGE_ACHIEVED"];
					condition = CONDITION_IDLE;
					desired_speed = 0.0;
				}
				else
				{
					double eta = (desired_range - distance) / flight_speed;
					desired_speed = max_flight_speed;
					if ((eta < 1.0)&&(flight_speed > max_flight_speed*0.25))
						desired_speed = flight_speed * 0.5;   // cut speed to a minimum of 1/4 speed
				}
				[self trackDestination:delta_t:YES];
				break;
				
		}
		//
		// in every case...
		//
		[self applyRoll:delta_t*flight_roll andClimb:delta_t*flight_pitch];
		[self applyThrust:delta_t];
		if (energy < max_energy)
		{
			energy += energy_recharge_rate * delta_t;
			if (energy > max_energy)
			{
				energy = max_energy;
				[shipAI message:@"ENERGY_FULL"];
			}
		}
    }
}


- (void) drawEntity:(BOOL) immediate :(BOOL) translucent
{
	if (zero_distance > no_draw_distance)   return; // TOO FAR AWAY

	if ([universe breakPatternHide])   return; // DON'T DRAW

	//debug
	//NSLog(@"drawing ship (%@ %d)", name, universal_id);

	[super drawEntity:immediate:translucent];

	//if (sub_entities)
	if ((sub_entities)&&(status != STATUS_DEMO))
	{
		int i;
		for (i = 0; i < [sub_entities count]; i++)
		{
			Entity  *se = (Entity *)[sub_entities objectAtIndex:i];
			[se setOwner:self]; // refresh ownership
			[se drawEntity:immediate:translucent];
		}
	}
}

- (void) addExhaust:(ParticleEntity *)exhaust
{
	if (!exhaust)
		return;
	if (sub_entities == nil)
		sub_entities = [[NSArray arrayWithObject:exhaust] retain];
	else
	{
		NSMutableArray *temp = [NSMutableArray arrayWithArray:sub_entities];
		[temp addObject:exhaust];
		[sub_entities release];
		sub_entities = [[NSArray arrayWithArray:temp] retain];
	}
}

- (void) addExhaustAt:(Vector) ex_position withScale:(Vector) ex_scale
{
	ParticleEntity *exhaust = [[ParticleEntity alloc] initExhaustFromShip:self offsetVector:ex_position scaleVector:ex_scale];
	[self addExhaust:exhaust];
}
	

- (void) applyThrust:(double) delta_t
{
	position.x += delta_t*velocity.x;
	position.y += delta_t*velocity.y;
	position.z += delta_t*velocity.z;
	
	if (velocity.x > 0.0)
		velocity.x -= (velocity.x > delta_t*thrust) ? delta_t*thrust : velocity.x;
	if (velocity.y > 0.0)
		velocity.y -= (velocity.y > delta_t*thrust) ? delta_t*thrust : velocity.y;
	if (velocity.z > 0.0)
		velocity.z -= (velocity.z > delta_t*thrust) ? delta_t*thrust : velocity.z;
	if (velocity.x < 0.0)
		velocity.x -= (velocity.x < -delta_t*thrust) ? -delta_t*thrust : velocity.x;
	if (velocity.y < 0.0)
		velocity.y -= (velocity.y < -delta_t*thrust) ? -delta_t*thrust : velocity.y;
	if (velocity.z < 0.0)
		velocity.z -= (velocity.z < -delta_t*thrust) ? -delta_t*thrust : velocity.z;

	if (condition == CONDITION_TUMBLE)  return; //testing

	if (flight_speed > desired_speed)
	{
		[self decrease_flight_speed:delta_t*thrust];
		if (flight_speed < desired_speed)   flight_speed = desired_speed;
	}
	if (flight_speed < desired_speed)
	{
		[self increase_flight_speed:delta_t*thrust];
		if (flight_speed > desired_speed)   flight_speed = desired_speed;
	}
	[self moveForward:delta_t*flight_speed];
}

- (void) applyRoll:(GLfloat) roll1 andClimb:(GLfloat) climb1
{	
	Quaternion q1;
	
	quaternion_set_identity(&q1);
	
	quaternion_rotate_about_z( &q1, -roll1);
	quaternion_rotate_about_x( &q1, -climb1);
	
	q_rotation = quaternion_multiply( q1, q_rotation);
	
    quaternion_normalise(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
	
	v_forward   = vector_forward_from_quaternion(q_rotation);
	v_up		= vector_up_from_quaternion(q_rotation);
	v_right		= vector_right_from_quaternion(q_rotation);
}

- (NSString *) name
{
	return name;
}

- (NSString *) roles
{
	return roles;
}

- (BOOL) reportAImessages
{
	return reportAImessages;
}

- (void) setReportAImessages:(BOOL) yn
{
	reportAImessages = yn;
}

- (int) checkForAegis
{
	// checkk planet
	Vector p1 = [[universe planet] getPosition];
	double cr = [[universe planet] collisionRadius];
	int result = AEGIS_NONE;
	p1.x -= position.x;	p1.y -= position.y;	p1.z -= position.z;
	double d2 = p1.x*p1.x + p1.y*p1.y + p1.z*p1.z - cr * cr * 9.0; // 3x radius of planet
	if (d2 < 0.0)
		result = AEGIS_CLOSE_TO_PLANET;
	// check station
	p1 = [[universe station] getPosition];
	p1.x -= position.x;	p1.y -= position.y;	p1.z -= position.z;
	d2 = p1.x*p1.x + p1.y*p1.y + p1.z*p1.z - SCANNER_MAX_RANGE2*4.0; // double scanner range
	if (d2 < 0.0)
		result = AEGIS_IN_DOCKING_RANGE;
	within_station_aegis = (d2 < 0.0);
	return result;
}

- (BOOL) within_station_aegis
{
	return within_station_aegis;
}

- (void) setStatus:(int) stat
{
	status = stat;
	if ((status == STATUS_LAUNCHING)&&(universe))
		launch_time = [universe getTime];
}

- (void) setAI:(AI *) ai
{
	if (shipAI) [shipAI release];
	shipAI = [ai retain];
}

- (AI *) getAI
{
	return shipAI;
}

- (void) setRoll:(double) amount
{
	flight_roll = amount * PI / 2.0;
}
	
- (void) setPitch:(double) amount
{
	flight_pitch = amount * PI / 2.0;
}


- (void) setThrust:(double) amount
{
	thrust = amount;
}


- (void) setBounty:(int) amount
{
	bounty = amount;
}

- (int) getBounty
{
	return bounty;
}

- (int) legal_status
{
	if ([roles isEqual:@"asteroid"])
		return 0;
	if ([roles isEqual:@"boulder"])
		return 0;
	if ([roles isEqual:@"splinter"])
		return 0;
	return bounty;
}

- (void) setCommodity:(int) co_type andAmount:(int) co_amount;
{
	commodity_type = co_type;
	commodity_amount = co_amount;
}
- (int) getCommodityType
{
	return commodity_type;
}
- (int) getCommodityAmount
{
	return commodity_amount;
}

- (int) getMaxCargo
{
	return max_cargo;
}

- (int) getCargoType
{
	return cargo_type;
}

- (void) setCargo:(NSArray *) some_cargo
{
	[cargo removeAllObjects];
	[cargo addObjectsFromArray:some_cargo];
}

- (int) cargoFlag
{
	return cargo_flag;
}

- (void) setCargoFlag:(int) flag
{
	cargo_flag = flag;
}

- (void) setSpeed:(double) amount
{
	flight_speed = amount;
}

- (void) increase_flight_speed:(double) delta
{
	if (flight_speed < max_flight_speed)
		flight_speed += delta;
	else
		flight_speed = max_flight_speed;
}

- (void) decrease_flight_speed:(double) delta
{
	if (flight_speed > -max_flight_speed)
		flight_speed -= delta;
	else
		flight_speed = -max_flight_speed;
}


- (void) increase_flight_roll:(double) delta
{
	if (flight_roll < max_flight_roll)
		flight_roll += delta;
	if (flight_roll > max_flight_roll)
		flight_roll = max_flight_roll;
}

- (void) decrease_flight_roll:(double) delta
{
	if (flight_roll > -max_flight_roll)
		flight_roll -= delta;
	if (flight_roll < -max_flight_roll)
		flight_roll = -max_flight_roll;
}


- (void) increase_flight_pitch:(double) delta
{
	if (flight_pitch < max_flight_pitch)
		flight_pitch += delta;
	if (flight_pitch > max_flight_pitch)
		flight_pitch = max_flight_pitch;
}


- (void) decrease_flight_pitch:(double) delta
{
	if (flight_pitch > -max_flight_pitch)
		flight_pitch -= delta;
	if (flight_pitch < -max_flight_pitch)
		flight_pitch = -max_flight_pitch;
}

- (double) flight_roll
{
	return flight_roll;
}
- (double) flight_pitch
{
	return flight_pitch;
}
- (double) flight_speed
{
	return flight_speed;
}
- (double) speed_factor
{
	if (max_flight_speed <= 0.0)
		return 0.0;
	return flight_speed / max_flight_speed;
}

- (int) damage
{
	return (int)(100 - (100 * energy / max_energy));
}


- (void) dealEnergyDamageWithinDesiredRange
{
	NSArray* targets = [universe getEntitiesWithinRange:desired_range ofEntity:self];
	if ([targets count] > 0)
	{
		int i;
		for (i = 0; i < [targets count]; i++)
		{
			Entity *e2 = [targets objectAtIndex:i];
			Vector p2 = [e2 getPosition];
			p2.x -= position.x;	p2.y -= position.y;	p2.z -= position.z;
			//double cr = [e2 collisionRadius];
			double d2 = p2.x*p2.x + p2.y*p2.y + p2.z*p2.z;
			double damage = weapon_energy*desired_range/d2;
			[e2 takeEnergyDamage:damage from:self becauseOf:[self owner]];
			//if ([e2 isKindOfClass:[ShipEntity class]])
			//	//NSLog(@"Doing %.1f damage to %@ %d",damage,[(ShipEntity *)e2 name],[(ShipEntity *)e2 universal_id]);
		}
	}
}

- (void) takeEnergyDamage:(double) amount from:(Entity *) ent becauseOf:(Entity *) other
{
	if (status == STATUS_DEAD)  // it's too late for this one!
		return;
	//
	energy -= amount;
	being_mined = NO;
	// if the other entity is a ship note it as an aggressor
	if ([other isKindOfClass:[ShipEntity class]])
	{
		ShipEntity* hunter = (ShipEntity *)other;
		//if (primaryAggressor) [primaryAggressor release]; // note primaryAggressor is assumed to be retained elsewhere
		primaryAggressor = [hunter universal_id];
		[shipAI reactToMessage:@"ATTACKED"];			// note use the reactToMessage: method NOT the think-delayed message: method
		
		// if I'm a copper and you're not, then mark the other as an offender!
		if (([roles isEqual:@"police"])&&(![[hunter roles] isEqual:@"police"]))
			[hunter markAsOffender:64];
		
		if ([other isKindOfClass:[PlayerEntity class]])
			being_mined = [(PlayerEntity *)other isMining];
	}
	// die if I'm out of energy
	if (energy <= 0.0)
	{
		[self becomeExplosion];
		if ([other isKindOfClass:[ShipEntity class]])
		{
			ShipEntity* hunter = (ShipEntity *)other;
			[hunter collectBountyFor:self];
			if ([hunter getPrimaryTarget] == (Entity *)self)
			{
				[hunter removeTarget:(Entity *)self];
				[[hunter getAI] message:@"TARGET_DESTROYED"];
			}
		}
	}
	// warn if I'm low on energy
	if (energy < max_energy *0.25)
		[shipAI reactToMessage:@"ENERGY_LOW"];
	if ((energy < max_energy *0.125)&&(has_escape_pod)&&(rand() % 100 < 25))  // 25% chance we get to an escape pod
	{
		has_escape_pod = NO;
		//NSLog(@"Escape Pod launched");
		[shipAI setStateMachine:@"nullAI.plist"];
		[shipAI setState:@"GLOBAL"];
		[self launchEscapeCapsule];
		[self setScanClass: CLASS_NEUTRAL];			// we're unmanned now!
	}
}

- (void) becomeExplosion
{
	ParticleEntity  *fragment;
	int i;
	Vector v;
	Quaternion q;
	int speed_low = 200;
	int speed_high = 800;
	int n_fragments = 32;
	int n_alloys = floor((boundingBox.max_z - boundingBox.min_z) / 50.0);
	
	if (status == STATUS_DEAD)
		return;
	status = STATUS_DEAD;
	//scripting
	if ([death_actions count])
	{
		int i;
		for (i = 0; i < [death_actions count]; i++)
			[(PlayerEntity *)[universe entityZero] scriptAction:(NSString *)[death_actions objectAtIndex:i]];
		[death_actions removeAllObjects];
	}
	
	
	if ([roles isEqual:@"thargoid"])
		[self broadcastThargoidDestroyed];
	
	if (collision_radius > 49.9) // big!
	{
		// quick test of hyperring
		ParticleEntity *ring = [[ParticleEntity alloc] initHyperringFromShip:self]; // retained
		[universe addEntity:ring];
		[ring release];
	}
	
	for (i = 0 ; i < n_fragments; i++)
	{
		int speed = (rand() % (speed_high - speed_low)) + speed_low;
		double duration = 0.5 + (rand() % 256) / 256.0;
		v.x = (rand() % speed) - speed / 2;
		v.y = (rand() % speed) - speed / 2;
		v.z = (rand() % speed) - speed / 2;
		//NSLog(@"In ShipEntity.becomeExplosion we need to make a bang!");
		fragment = [[ParticleEntity alloc] init];	// alloc retains!
		[fragment setPosition:position]; // here
		[fragment setScanClass: CLASS_NO_DRAW];
		[fragment setVelocity: v];
		[fragment setDuration: duration];
		[fragment setCollisionRadius: 0.0];
		[fragment setEnergy: 0.0];
		[fragment setParticleType: PARTICLE_SHOT_GREEN_PLASMA];
		[fragment setColor:[NSColor yellowColor]];
		[universe addEntity:fragment];
		[fragment release];							//release
	}
	for (i = 0 ; i < n_fragments / 4; i++)
	{
		int speed = ((rand() % (speed_high - speed_low)) + speed_low) /8;
		double duration = 0.5 + (rand() % 256) / 256.0;
		v.x = (rand() % speed) - speed / 2;
		v.y = (rand() % speed) - speed / 2;
		v.z = (rand() % speed) - speed / 2;
		//NSLog(@"In ShipEntity.becomeExplosion we need to make a bang!");
		fragment = [[ParticleEntity alloc] init];	// alloc retains!
		[fragment setPosition:position]; // here
		[fragment setScanClass: CLASS_NO_DRAW];
		[fragment setVelocity: v];
		[fragment setDuration: duration];
		[fragment setCollisionRadius: 0.0];
		[fragment setEnergy: 0.0];
		[fragment setParticleType: PARTICLE_EXPLOSION];
		[fragment setColor:[NSColor yellowColor]];
		[universe addEntity:fragment];
		[fragment release];							//release
	}
	
	// we need to throw out cargo at this point.
	NSArray *jetsam = nil;  // this will contain the stuff to get thrown out
	int cargo_chance = 10;
	if (cargo_flag == CARGO_FLAG_FULL_PLENTIFUL)
	{
		int cargo_to_go = max_cargo * cargo_chance / 100;
		while (cargo_to_go > 15)
			cargo_to_go = rand() % cargo_to_go;
		//NSLog(@"explosion in %@ %d will launch %d pieces of cargo (max_cargo = %d)", name, universal_id, cargo_to_go, max_cargo);
		jetsam = [[universe getContainersOfPlentifulGoods:cargo_to_go] retain];
		cargo_chance = 100;  //  chance of any given piece of cargo surviving decompression
	}
	if (cargo_flag == CARGO_FLAG_FULL_SCARCE)
	{
		int cargo_to_go = max_cargo * cargo_chance / 100;
		while (cargo_to_go > 15)
			cargo_to_go = rand() % cargo_to_go;
		//NSLog(@"explosion in %@ %d will launch %d pieces of cargo (max_cargo = %d)", name, universal_id, cargo_to_go, max_cargo);
		jetsam = [[universe getContainersOfScarceGoods:cargo_to_go] retain];
		cargo_chance = 100;  //  chance of any given piece of cargo surviving decompression
	}
	if (cargo_flag == CARGO_FLAG_CANISTERS)
	{
		jetsam = [[NSArray arrayWithArray:cargo] retain];   // what the ship is carrying
		[cargo removeAllObjects];   // dispense with it!
	}
	if (jetsam)
	{
		for (i = 0; i < [jetsam count]; i++)
		{
			if (rand() % 100 < cargo_chance)  //  chance of any given piece of cargo surviving decompression
			{
				ShipEntity* container = [[jetsam objectAtIndex:i] retain];
				Vector  rpos = position;
				rpos.x += (rand() % 7) - 3;
				rpos.y += (rand() % 7) - 3;
				rpos.z += (rand() % 7) - 3;
				[container setPosition:rpos];
				v.x = 0.1 *((rand() % speed_low) - speed_low / 2);
				v.y = 0.1 *((rand() % speed_low) - speed_low / 2);
				v.z = 0.1 *((rand() % speed_low) - speed_low / 2);
				[container setVelocity:v];
				quaternion_set_random(&q);
				[container setQRotation:q];
				[container setStatus:STATUS_IN_FLIGHT];
				[container setScanClass: CLASS_CARGO];
				[universe addEntity:container];
				[[container getAI] setState:@"GLOBAL"];
				//NSLog(@"Launched %@ %d with %@",[container name], [container universal_id], [universe describeCommodity:[container getCommodityType] amount:[container getCommodityAmount]]);
				[container release];
			}
		}
		[jetsam release];   // done with this now!
	}

	//
	//  Throw out rocks and alloys to be scooped up
	//
	if ([roles isEqual:@"asteroid"])
	{
		if ((being_mined)||(rand() % 100 < 20))
		{
			// if hit by a mining laser, break up into 2..6 boulders
			int n_rocks = 2 + (rand() % 5);
			//NSLog(@"Throwing %d boulders", n_rocks);
			for (i = 0; i < n_rocks; i++)
			{
				ShipEntity* rock = [universe getShipWithRole:@"boulder"];   // retain count = 1
				Vector  rpos = position;
				int  r_speed = 20.0 * [(NSNumber *)[rock valueForKey:@"max_flight_speed"] doubleValue];
				int cr = 3 * [rock collisionRadius];
				rpos.x += (rand() % cr) - cr/2;
				rpos.y += (rand() % cr) - cr/2;
				rpos.z += (rand() % cr) - cr/2;
				[rock setPosition:rpos];
				v.x = 0.1 *((rand() % r_speed) - r_speed / 2);
				v.y = 0.1 *((rand() % r_speed) - r_speed / 2);
				v.z = 0.1 *((rand() % r_speed) - r_speed / 2);
				[rock setVelocity:v];
				quaternion_set_random(&q);
				[rock setQRotation:q];
				[rock setStatus:STATUS_IN_FLIGHT];
				[rock setScanClass: CLASS_CARGO];
				[universe addEntity:rock];
				[[rock getAI] setState:@"GLOBAL"];
				[rock release];
			}
		}
		[universe removeEntity:self];
		return; // don't do anything more
	}
	
	if ([roles isEqual:@"boulder"])
	{
		if ((being_mined)||(rand() % 100 < 20))
		{
			// if hit by a mining laser, break up into 2..6 splinters
			int n_rocks = 2 + (rand() % 5);
			//NSLog(@"Throwing %d splinters", n_rocks);
			for (i = 0; i < n_rocks; i++)
			{
				ShipEntity* rock = [universe getShipWithRole:@"splinter"];   // retain count = 1
				Vector  rpos = position;
				int  r_speed = 20.0 * [(NSNumber *)[rock valueForKey:@"max_flight_speed"] doubleValue];
				int cr = 3 * [rock collisionRadius];
				rpos.x += (rand() % cr) - cr/2;
				rpos.y += (rand() % cr) - cr/2;
				rpos.z += (rand() % cr) - cr/2;
				[rock setPosition:rpos];
				v.x = 0.1 *((rand() % r_speed) - r_speed / 2);
				v.y = 0.1 *((rand() % r_speed) - r_speed / 2);
				v.z = 0.1 *((rand() % r_speed) - r_speed / 2);
				[rock setVelocity:v];
				quaternion_set_random(&q);
				[rock setQRotation:q];
				[rock setStatus:STATUS_IN_FLIGHT];
				[rock setScanClass: CLASS_CARGO];
				[universe addEntity:rock];
				[[rock getAI] setState:@"GLOBAL"];
				[rock release];
			}
		}
		[universe removeEntity:self];
		return; // don't do anything more
	}

	//NSLog(@"Throwing %d pieces of alloy", n_alloys);
	for (i = 0; i < n_alloys; i++)
	{
		ShipEntity* plate = [universe getShip:@"alloy"];   // retain count = 1
		Vector  rpos = position;
		rpos.x += (rand() % 7) - 3;
		rpos.y += (rand() % 7) - 3;
		rpos.z += (rand() % 7) - 3;
		[plate setPosition:rpos];
		v.x = 0.1 *((rand() % speed_low) - speed_low / 2);
		v.y = 0.1 *((rand() % speed_low) - speed_low / 2);
		v.z = 0.1 *((rand() % speed_low) - speed_low / 2);
		[plate setVelocity:v];
		quaternion_set_random(&q);
		[plate setQRotation:q];
		[plate setStatus:STATUS_TEST];
		[plate setScanClass: CLASS_CARGO];
		[plate setCommodity:9 andAmount:1];
		[universe addEntity:plate];
		[[plate getAI] setState:@"GLOBAL"];
		[plate release];
	}
	//
	if (![self isKindOfClass:[PlayerEntity class]])
		[universe removeEntity:self];
}

- (void) becomeLargeExplosion:(double) factor
{
	ParticleEntity  *fragment;
	int i;
	Vector v;
	int speed_low = 200;
	int speed_high = 800;
	int n_fragments = 32*factor;
	int n_cargo = (rand() % (likely_cargo + 1));
	
	if (status == STATUS_DEAD)
		return;
	
	status = STATUS_DEAD;
	//scripting
	if ([death_actions count])
	{
		int i;
		for (i = 0; i < [death_actions count]; i++)
			[(PlayerEntity *)[universe entityZero] scriptAction:(NSString *)[death_actions objectAtIndex:i]];
		[death_actions removeAllObjects];
	}
	
	for (i = 0 ; i < n_fragments; i++)
	{
		int speed = (rand() % (speed_high - speed_low)) + speed_low;
		double duration = factor*(0.5 + (rand() % 256) / 256.0);
		v.x = (rand() % speed) - speed / 2;
		v.y = (rand() % speed) - speed / 2;
		v.z = (rand() % speed) - speed / 2;
		//NSLog(@"In ShipEntity.becomeExplosion we need to make a bang!");
		fragment = [[ParticleEntity alloc] init];	// alloc retains!
		[fragment setPosition:position]; // here
		[fragment setScanClass: CLASS_NO_DRAW];
		[fragment setVelocity: v];
		[fragment setDuration: duration];
		[fragment setCollisionRadius: 0.0];
		[fragment setEnergy: 0.0];
		[fragment setParticleType: PARTICLE_SHOT_GREEN_PLASMA];
		[fragment setColor:[NSColor yellowColor]];
		[universe addEntity:fragment];
		[fragment release];							//release
	}
	for (i = 0 ; i < n_fragments / 4; i++)
	{
		int speed = ((rand() % (speed_high - speed_low)) + speed_low) /8;
		double duration = factor*(0.5 + (rand() % 256) / 256.0);
		v.x = (rand() % speed) - speed / 2;
		v.y = (rand() % speed) - speed / 2;
		v.z = (rand() % speed) - speed / 2;
		//NSLog(@"In ShipEntity.becomeExplosion we need to make a bang!");
		fragment = [[ParticleEntity alloc] init];	// alloc retains!
		[fragment setPosition:position]; // here
		[fragment setScanClass: CLASS_NO_DRAW];
		[fragment setVelocity: v];
		[fragment setDuration: duration];
		[fragment setCollisionRadius: 0.0];
		[fragment setEnergy: 0.0];
		[fragment setParticleType: PARTICLE_EXPLOSION];
		[fragment setColor:[NSColor yellowColor]];
		[universe addEntity:fragment];
		[fragment release];							//release
	}
	
	// we need to throw out cargo at this point.
	int cargo_chance = 10;
	if (cargo_flag == CARGO_FLAG_FULL_PLENTIFUL)
	{
		int cargo_to_go = max_cargo / 10;
		while (cargo_to_go > 15)
			cargo_to_go = rand() % cargo_to_go;
		//NSLog(@"explosion in %@ %d will launch %d pieces of cargo (max_cargo = %d)", name, universal_id, cargo_to_go, max_cargo);
		[self setCargo:[universe getContainersOfPlentifulGoods:cargo_to_go]];
		cargo_chance = 100;
	}
	if (cargo_flag == CARGO_FLAG_FULL_SCARCE)
	{
		int cargo_to_go = max_cargo / 10;
		while (cargo_to_go > 15)
			cargo_to_go = rand() % cargo_to_go;
		//NSLog(@"explosion in %@ %d will launch %d pieces of cargo (max_cargo = %d)", name, universal_id, cargo_to_go, max_cargo);
		[self setCargo:[universe getContainersOfScarceGoods:cargo_to_go]];
		cargo_chance = 100;
	}
	while ([cargo count] > 0)
	{
		if (rand() % 100 < cargo_chance)  //  10% chance of any given piece of cargo surviving decompression
		{
			ShipEntity* container = [[cargo objectAtIndex:0] retain];
			Vector  rpos = position;
			rpos.x += (rand() % 7) - 3;
			rpos.y += (rand() % 7) - 3;
			rpos.z += (rand() % 7) - 3;
			[container setStatus:STATUS_TEST];
			[container setPosition:rpos];
			[container setScanClass: CLASS_CARGO];
			[universe addEntity:container];
			[[container getAI] setState:@"GLOBAL"];
			[container release];
			if (n_cargo > 0)
				n_cargo--;  // count down extra cargo
		}
		[cargo removeObjectAtIndex:0];
	}
	//
		
	if (![self isKindOfClass:[PlayerEntity class]])
		[universe removeEntity:self];
}

- (void) collectBountyFor:(ShipEntity *)other
{
	if ([roles isEqual:@"pirate"])
		bounty += [other getBounty];
}

/*-----------------------------------------

	AI piloting methods

-----------------------------------------*/

- (void) setPrimaryAggressor:(Entity *) targetEntity
{
	if (targetEntity)
		primaryAggressor = [targetEntity universal_id];
}

- (void) addTarget:(Entity *) targetEntity
{
	if (targetEntity)
		primaryTarget = [targetEntity universal_id];
}

- (void) removeTarget:(Entity *) targetEntity
{
	if (primaryTarget != NO_TARGET)
		[shipAI reactToMessage:@"TARGET_LOST"];
	primaryTarget = NO_TARGET;
}

- (Entity *) getPrimaryTarget
{
	return [universe entityForUniversalID:primaryTarget];
}

- (void) setCondition:(int) cond
{
	condition = cond;
}

- (double) trackPrimaryTarget:(double) delta_t :(BOOL) retreat
{
	Vector  relativePosition;	
	GLfloat  d_forward, d_up, d_right;
	Entity  *target;
	
	double  damping = 0.5 * delta_t;
	double  rate2 = 4.0 * delta_t;
	double  rate1 = 2.0 * delta_t;
	
	double stick_roll = 0.0;	//desired roll and pitch
	double stick_pitch = 0.0;
	
	double pitch_pitch = 0.0;
	double roll_roll = 0.0;
	
	double reverse = 1.0;
	
	target = [self getPrimaryTarget];
	
	if (target == nil)   // leave now!
		return 0.0;
	
	if (retreat)
		reverse = -1.0;
	
	relativePosition = [target getPosition];
	relativePosition.x -= position.x;
	relativePosition.y -= position.y;
	relativePosition.z -= position.z;
	
	//jink if retreating
	if (retreat)
	{
		Quaternion q = [target QRotation];
		Vector vx = vector_right_from_quaternion(q);
		Vector vy = vector_up_from_quaternion(q);
		Vector vz = vector_forward_from_quaternion(q);
		relativePosition.x += jink.x * vx.x + jink.y * vy.x + jink.z * vz.x;
		relativePosition.y += jink.x * vx.y + jink.y * vy.y + jink.z * vz.y;
		relativePosition.z += jink.x * vx.z + jink.y * vy.z + jink.z * vz.z;
	}
	
	relativePosition = unit_vector(&relativePosition);
	
	d_right		=   dot_product(relativePosition, v_right);
	d_up		=   dot_product(relativePosition, v_up);
	d_forward   =   dot_product(relativePosition, v_forward);

	// begin rule-of-thumb manoeuvres
	
	stick_roll = 0.0;
	
	if (pitching_over)
		pitching_over = (stick_pitch != 0.0);
	
	if ((d_forward*reverse < -0.85) && (!pitching_over))
	{
		pitching_over = YES;
		pitch_pitch = 1.0;
		//NSLog(@"Pitch - over");
		if (d_up >= 0)
			stick_pitch = -max_flight_pitch * pitch_pitch * reverse;
		if (d_up < 0)
			stick_pitch = max_flight_pitch * pitch_pitch * reverse;
	}
	
			
	if (pitching_over)
	{
		pitching_over = (d_forward * reverse < 0.5);
	}
	else
	{
		stick_pitch = 0.0;
		pitch_pitch = d_up;
		if (pitch_pitch < 0.0)
			pitch_pitch = -pitch_pitch;
		roll_roll = d_right;
		if (roll_roll < 0.0)
			roll_roll = -roll_roll;
		
		if (retreat)
			pitch_pitch = 1.0 - pitch_pitch;
				
		if (d_up > 0.0)
			stick_pitch = -max_flight_pitch * pitch_pitch * reverse;
		if (d_up < 0.0)
			stick_pitch = max_flight_pitch * pitch_pitch * reverse;

		if (d_right > 0.0)
			stick_roll = -max_flight_roll * roll_roll;
		if (d_right < 0.0)
			stick_roll = max_flight_roll * roll_roll;
	}
	
	// end rule-of-thumb manoeuvres

	// apply stick movement limits
	if (flight_roll < stick_roll - rate1)
		stick_roll = flight_roll + rate1;
	if (flight_roll > stick_roll + rate1)
		stick_roll = flight_roll - rate1;
	if (flight_pitch < stick_pitch - rate2)
		stick_pitch = flight_pitch + rate2;
	if (flight_pitch > stick_pitch + rate2)
		stick_pitch = flight_pitch - rate2;
	flight_roll = stick_roll;
	flight_pitch = stick_pitch;
			
	// apply damping
	if (flight_roll < -damping)
		flight_roll += damping;
	if (flight_roll > damping)
		flight_roll -= damping;
	if (flight_pitch < -damping)
		flight_pitch += damping;
	if (flight_pitch > damping)
		flight_pitch -= damping;

	//
	//  return target confidence 0.0 .. 1.0
	//
	if (d_forward < 0.0)
		return 0.0;
	return d_forward;
}

- (double) trackDestination:(double) delta_t :(BOOL) retreat
{
	Vector  relativePosition;	
	GLfloat  d_forward, d_up, d_right;
	
	BOOL	we_are_docking = NO;
	
	double  damping = 0.5 * delta_t;
	double  rate2 = 4.0 * delta_t;
	double  rate1 = 2.0 * delta_t;
	
	double stick_roll = 0.0;	//desired roll and pitch
	double stick_pitch = 0.0;
		
	double reverse = 1.0;
	
	double min_d = 0.004;
	
	if (retreat)
		reverse = -reverse;

	if ([self isKindOfClass:[PlayerEntity class]])
		reverse = -reverse;
		
	relativePosition = destination;
	relativePosition.x -= position.x;
	relativePosition.y -= position.y;
	relativePosition.z -= position.z;
	
	relativePosition = unit_vector(&relativePosition);
	
	d_right		=   dot_product(relativePosition, v_right);
	d_up		=   dot_product(relativePosition, v_up);
	d_forward   =   dot_product(relativePosition, v_forward);

	// begin rule-of-thumb manoeuvres
	
	if (d_forward < -0.99)  // hack to avoid just flying away from the destination
	{
		d_up = min_d * 2.0;
	}
	
	{
		stick_pitch = 0.0;
		stick_roll = 0.0;
			
		if (d_up >= 0.0)
		{
			if (d_right > min_d)
				stick_roll = stick_roll - max_flight_roll * reverse * 0.25;  //roll_roll * reverse;
			if (d_right < -min_d)
				stick_roll = stick_roll + max_flight_roll * reverse * 0.25; //roll_roll * reverse;
		}
		else
		{
			if (d_right > min_d)
				stick_roll = stick_roll + max_flight_roll * reverse * 0.25;  //roll_roll * reverse;
			if (d_right < -min_d)
				stick_roll = stick_roll - max_flight_roll * reverse * 0.25; //roll_roll * reverse;
		}
		
		if (stick_roll == 0.0)
		{
			if (d_up > min_d)
				stick_pitch = -max_flight_pitch * reverse * 0.25;  //pitch_pitch * reverse;
			if (d_up < -min_d)
				stick_pitch = max_flight_pitch * reverse * 0.25;  //pitch_pitch * reverse;
		}

	}
	
	if ([[self getPrimaryTarget] isKindOfClass:[StationEntity class]])
	{
		/* we are docking and need to consider the rotation of the space station */
		Vector up_vec = [(StationEntity *)[self getPrimaryTarget] upVector];
		double rot_up = dot_product(up_vec, v_up);
		double station_roll = [(StationEntity *)[self getPrimaryTarget] flight_roll];
		//NSLog(@"docking rot_up %.3f",rot_up);
		if (rot_up < 0.0)
			rot_up = -rot_up;
		if ((stick_pitch == 0.0)&&(stick_roll == 0.0))
		{
			if (rot_up > 0.75)
				stick_roll = station_roll;
			if (rot_up < 0.25)
				stick_roll = 0.0;
		}
		we_are_docking = YES;
	}
	
	// end rule-of-thumb manoeuvres

	// apply stick movement limits
	if (flight_roll < stick_roll - rate1)
		stick_roll = flight_roll + rate1;
	if (flight_roll > stick_roll + rate1)
		stick_roll = flight_roll - rate1;
	if (flight_pitch < stick_pitch - rate2)
		stick_pitch = flight_pitch + rate2;
	if (flight_pitch > stick_pitch + rate2)
		stick_pitch = flight_pitch - rate2;
	flight_roll = stick_roll;
	flight_pitch = stick_pitch;
			
	// apply damping
	if (flight_roll < -damping)
		flight_roll += damping;
	if (flight_roll > damping)
		flight_roll -= damping;
	if (flight_pitch < -damping)
		flight_pitch += damping;
	if (flight_pitch > damping)
		flight_pitch -= damping;

	if (d_forward < 0.0)
		return 0.0;
	return d_forward;
}

- (double) rangeToDestination
{
	double dist;
	Vector delta = destination;
	delta.x -= position.x;
	delta.y -= position.y;
	delta.z -= position.z;
	dist = sqrt(delta.x*delta.x + delta.y*delta.y + delta.z*delta.z);
	return dist;
}

- (double) rangeToPrimaryTarget
{
	double dist;
	Vector delta;
	Entity  *target = [self getPrimaryTarget];
	if (target == nil)   // leave now!
		return 0.0;
	delta = [target getPosition];
	delta.x -= position.x;
	delta.y -= position.y;
	delta.z -= position.z;
	dist = sqrt(delta.x*delta.x + delta.y*delta.y + delta.z*delta.z);
	dist -= [target collisionRadius];
	dist -= collision_radius;
	return dist;
}

- (BOOL) onTarget
{
	double dist, radius;
	Vector rel_pos, delta;
	Entity  *target = [self getPrimaryTarget];
	if (target == nil)   // leave now!
		return NO;
	if ([target getStatus] == STATUS_DEAD)
		return NO;
	rel_pos = [target getPosition];
	rel_pos.x -= position.x;
	rel_pos.y -= position.y;
	rel_pos.z -= position.z;
	delta.z = dot_product(rel_pos, v_forward);
	if (delta.z < 0.0)
		return NO;
	delta.x = dot_product(rel_pos, v_right);
	delta.y = dot_product(rel_pos, v_up);
	//NSLog(@"Targetting (%3.3f,%3.3f)",delta.x,delta.y);
	dist = delta.x*delta.x + delta.y*delta.y;
	radius = [target collisionRadius] * 2.0;		//  fudge factor 
	return (dist < radius * radius);
}

- (BOOL) fireMainWeapon:(double) range
{
	//NSLog(@"%@ firing weapon type:%d shot_time:%.1f weapon_recharge_rate:%.1f, onTarget:%d", name, forward_weapon_type, shot_time, weapon_recharge_rate, [self onTarget]);
	if ((shot_time < weapon_recharge_rate)||(![self onTarget]))
		return NO;
	//NSLog(@"%@ firing weapon",name);
	switch (forward_weapon_type)
	{
		case WEAPON_PLASMA_CANNON :
			if (range > 3750)
				return NO;
			[self firePlasmaShot:weapon_offset_x:1500.0:[NSColor yellowColor]];
			[self firePlasmaShot:weapon_offset_x:1500.0:[NSColor yellowColor]];
			return YES;
			break;
			
		case WEAPON_PULSE_LASER :
		case WEAPON_BEAM_LASER :
		case WEAPON_MINING_LASER :
		case WEAPON_MILITARY_LASER :
			if (range > 7500)
				return NO;
			[self fireLaserShot];
			return YES;
			break;
			
	}
	return NO;
}

- (void) setLaserColor:(NSColor *) color
{
	if (color)
	{
		[laser_color release];
		laser_color = [color retain];
	}
}

- (BOOL) fireLaserShot
{
	int				ent_hit = NO_TARGET;
	ParticleEntity  *shot;
	int				direction = VIEW_FORWARD;
	double			range_limit = PARTICLE_LASER_RANGE_LIMIT;
	Vector  vel;
	vel.x = v_forward.x * flight_speed;
	vel.y = v_forward.y * flight_speed;
	vel.z = v_forward.z * flight_speed;

	if ([self isKindOfClass:[PlayerEntity class]])		// only the player has weapons on other facings
		direction = [universe viewDir];					// set the weapon facing here

	ent_hit = [universe getFirstEntityHitByLaserFromEntity:self inView:direction];

	shot = [[ParticleEntity alloc] initLaserFromShip:self view:direction];	// alloc retains!
	[shot setColor:laser_color];
	[shot setScanClass: CLASS_NO_DRAW];
	[shot setVelocity: vel];
	if (ent_hit != NO_TARGET)
	{
		Entity *victim = [universe entityForUniversalID:ent_hit];
		if (([victim isKindOfClass:[ShipEntity class]])&&([victim getZeroDistance] < range_limit))
		{
			Vector rel = [victim getPosition];
			Vector p1 = [shot getPosition];
			double dist;
			[(ShipEntity *)victim takeEnergyDamage:weapon_energy from:self becauseOf:self];	// a very palpable hit
			rel.x -= p1.x; rel.y -= p1.y; rel.z -= p1.z;
			dist = sqrt(rel.x*rel.x+rel.y*rel.y+rel.z*rel.z);
			[shot setCollisionRadius:dist];
		}
	}
	[universe addEntity:shot];
	[shot release]; //release
	
	shot_time = 0.0;
	
	return YES;
}

- (BOOL) firePlasmaShot:(double) offset :(double) speed :(NSColor *) color
{
	ParticleEntity *shot;
	Vector  vel, rt;
	Vector  origin = position;
	double  start = collision_radius + 0.5;
	
	speed += flight_speed;

	if (++shot_counter % 2)
		offset = -offset;
	
	vel = v_forward;
	rt = v_right;
	
	if ([self isKindOfClass:[PlayerEntity class]])					// player can fire into multiple views!
	{
		switch ([universe viewDir])
		{
			case VIEW_AFT :
				vel = v_forward;
				vel.x = -vel.x; vel.y = -vel.y; vel.z = -vel.z; // reverse
				rt = v_right;
				rt.x = -rt.x;   rt.y = -rt.y;   rt.z = -rt.z; // reverse
				break;
			case VIEW_STARBOARD :
				vel = v_right;
				rt = v_forward;
				rt.x = -rt.x;   rt.y = -rt.y;   rt.z = -rt.z; // reverse
				break;
			case VIEW_PORT :
				vel = v_right;
				vel.x = -vel.x; vel.y = -vel.y; vel.z = -vel.z; // reverse
				rt = v_forward;
				break;
		}
	}
	
	origin.x += vel.x * start;
	origin.y += vel.y * start;
	origin.z += vel.z * start;
	
	origin.x += rt.x * offset;
	origin.y += rt.y * offset;
	origin.z += rt.z * offset;
	
	vel.x *= speed;
	vel.y *= speed;
	vel.z *= speed;
	
	shot = [[ParticleEntity alloc] init];	// alloc retains!
	[shot setPosition:origin]; // directly ahead
	[shot setScanClass: CLASS_NO_DRAW];
	[shot setVelocity: vel];
	[shot setDuration: 5.0];
	[shot setCollisionRadius: 2.0];
	[shot setEnergy: weapon_energy];
	[shot setParticleType: PARTICLE_SHOT_GREEN_PLASMA];
	[shot setColor:color];
	[shot setOwner:self];
	[universe addEntity:shot];
	[shot release]; //release
	
	shot_time = 0.0;
	
	return YES;
}

- (BOOL) fireMissile
{
	ShipEntity *missile;
	Vector  vel;
	Vector  origin = position;
	Vector  start;
	start.x = 0.0;						// in the middle
	start.y = boundingBox.min_y - 4.0;	// 4m below bounding box
	start.z = boundingBox.max_z + 1.0;	// 1m ahead of bounding box
	double  throw_speed = 250.0;
	Quaternion q1 = [self QRotation];
	Entity  *target = [self getPrimaryTarget];
	
	if ((missiles <= 0)||(target == nil))
		return NO;
		
	if ([roles isEqual:@"thargoid"])
		return [self fireTharglet];
	
	missiles--;
	
	if ([self isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   // player view is reversed remember!
		
	vel.x = (flight_speed + throw_speed) * v_forward.x;
	vel.y = (flight_speed + throw_speed) * v_forward.y;
	vel.z = (flight_speed + throw_speed) * v_forward.z;
	
	origin.x = position.x + v_right.x * start.x + v_up.x * start.y + v_forward.x * start.z;
	origin.y = position.y + v_right.y * start.x + v_up.y * start.y + v_forward.y * start.z;
	origin.z = position.z + v_right.z * start.x + v_up.z * start.y + v_forward.z * start.z;
	
	//vel.x *= throw_speed;	vel.y *= throw_speed;	vel.z *= throw_speed;
	
	missile = [universe getShip:@"missile"];   // retain count = 1
	[missile setPosition:origin];						// directly below
	[missile setScanClass: CLASS_MISSILE];
	[missile addTarget:target];
	[missile setQRotation:q1];
	[missile setStatus: STATUS_IN_FLIGHT];  // necessary to get it going!
	[missile setVelocity: vel];
	[missile setSpeed:150.0];
	[missile setOwner:self];
	//[missile setReportAImessages:YES];
	[universe addEntity:missile];
	//NSLog(@"Missile collision radius is %.1f",[missile collisionRadius]);
	[missile release]; //release
	
	[(ShipEntity *)target setPrimaryAggressor:self];
	[[(ShipEntity *)target getAI] reactToMessage:@"INCOMING_MISSILE"];
	
	return YES;
}

- (BOOL) fireTharglet
{
	ShipEntity *tharglet;
	Vector  vel;
	Vector  origin = position;
	Vector  start;
	start.x = 0.0;						// in the middle
	start.y = boundingBox.min_y - 10.0;	// 10m below bounding box
	start.z = 0.0;	// 1m ahead of bounding box
	double  throw_speed = 250.0;
	Quaternion q1 = [self QRotation];
	Entity  *target = [self getPrimaryTarget];
	
	if ((missiles <= 0)||(target == nil))
		return NO;
	
	missiles--;
	
	if ([self isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   // player view is reversed remember!
		
	vel.x = (flight_speed + throw_speed) * v_forward.x;
	vel.y = (flight_speed + throw_speed) * v_forward.y;
	vel.z = (flight_speed + throw_speed) * v_forward.z;
	
	origin.x = position.x + v_right.x * start.x + v_up.x * start.y + v_forward.x * start.z;
	origin.y = position.y + v_right.y * start.x + v_up.y * start.y + v_forward.y * start.z;
	origin.z = position.z + v_right.z * start.x + v_up.z * start.y + v_forward.z * start.z;
		
	tharglet = [universe getShip:@"tharglet"];   // retain count = 1
	[tharglet setPosition:origin];						// directly below
	[tharglet setScanClass: CLASS_THARGOID];
	[tharglet addTarget:target];
	[tharglet setQRotation:q1];
	[tharglet setStatus: STATUS_IN_FLIGHT];  // necessary to get it going!
	[tharglet setVelocity: vel];
	[tharglet setSpeed:150.0];
	[tharglet setOwner:self];
	[tharglet setReportAImessages:YES]; // debug
	[universe addEntity:tharglet];
	//NSLog(@"tharglet collision radius is %.1f",[tharglet collisionRadius]);
	[tharglet release]; //release
	
	return YES;
}

- (BOOL) fireECM
{
	if (!has_ecm)
		return NO;
	else
	{
		ParticleEntity  *ecmDevice = [[ParticleEntity alloc] initECMMineFromShip:self]; // retained
		[universe addEntity:ecmDevice];
		[ecmDevice release];
	}
	return YES;
}

- (int) launchEscapeCapsule
{
	ShipEntity *pod;
	Vector  vel;
	Vector  origin = position;
	double  start = boundingBox.min_y - 10.0;
	double  throw_speed = 20.0;
	int co_type, co_amount;
	Quaternion q1 = [self QRotation];
	
	if ([self isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   // player view is reversed remember!
		
	vel.x = (-v_forward.x) * throw_speed + flight_speed * v_forward.x;
	vel.y = (-v_forward.y) * throw_speed + flight_speed * v_forward.y;
	vel.z = (-v_forward.z) * throw_speed + flight_speed * v_forward.z;
	
	origin.x += vel.x * start / throw_speed;
	origin.y += vel.y * start / throw_speed;
	origin.z += vel.z * start / throw_speed;
	
	pod = [universe getShip:@"escape-capsule"];   // retain count = 1
	[pod setPosition:origin];						// directly below
	[pod setScanClass: CLASS_CARGO];
	co_type = [universe commodityForName:@"Slaves"];
	co_amount = 1;
	[pod setCommodity:co_type andAmount:co_amount];
	[pod setQRotation:q1];
	[pod setVelocity: vel];
	[pod setOwner:self];
	[[pod getAI] setState:@"GLOBAL"];   // set the AI going
	[pod setStatus: STATUS_IN_FLIGHT];  // necessary to get it going!
	[universe addEntity:pod];
	[pod release]; //release
		
	return [pod universal_id];
}

- (int) dumpCargo
{
	int result = CARGO_NOT_CARGO;
	if (([cargo count] > 0)&&([universe getTime] - cargo_dump_time > 0.5))  // space them 0.5s or 10m apart
	{
		ShipEntity* jetto = [[cargo objectAtIndex:0] retain];
		double  start = collision_radius + [jetto collisionRadius];
		double  eject_speed = -20.0;
		Quaternion  random_direction;
		Vector  vel;
		Vector  rpos = position;
		double random_roll =	((rand() % 1024) - 512.0)/1024.0;  //  -0.5 to +0.5
		double random_pitch =   ((rand() % 1024) - 512.0)/1024.0;  //  -0.5 to +0.5
		quaternion_set_random(&random_direction);
		rpos.x -= v_forward.x * start;
		rpos.y -= v_forward.y * start;
		rpos.z -= v_forward.z * start;
		vel.x = v_forward.x * flight_speed;
		vel.y = v_forward.y * flight_speed;
		vel.z = v_forward.z * flight_speed;

		vel.x += v_forward.x * eject_speed;
		vel.y += v_forward.y * eject_speed;
		vel.z += v_forward.z * eject_speed;
		eject_speed *= 0.1 * ((rand() % 10) - 5);   //  -0.5x .. +0.5x
		vel.x += v_up.x * eject_speed;
		vel.y += v_up.y * eject_speed;
		vel.z += v_up.z * eject_speed;
		eject_speed *= 0.1 * ((rand() % 10) - 5);   //  -0.25x .. +0.25x
		vel.x += v_right.x * eject_speed;
		vel.y += v_right.y * eject_speed;
		vel.z += v_right.z * eject_speed;
		
		result = [jetto getCommodityType];
		[jetto setStatus:STATUS_TEST];
		[jetto setPosition:rpos];
		[jetto setQRotation:random_direction];
		[jetto setRoll:random_roll];
		[jetto setPitch:random_pitch];
		[jetto setVelocity:vel];
		[jetto setScanClass: CLASS_CARGO];
		[jetto setStatus: STATUS_IN_FLIGHT];
		[universe addEntity:jetto];
		[[jetto getAI] setState:@"GLOBAL"];
		[jetto release];
		[cargo removeObjectAtIndex:0];
		cargo_dump_time = [universe getTime];
	}
	return result;
}

- (void) manageCollisions
{
	// deal with collisions
	//
	Entity*		ent;
	ShipEntity* other_ship;

	while ([collidingEntities count] > 0)
	{
		ent = (Entity *)[collidingEntities objectAtIndex:0];
		[collidingEntities removeObjectAtIndex:0];
		if ([ent isKindOfClass:[ShipEntity class]])
		{
			other_ship = (ShipEntity *)ent;
			[self collideWithShip:other_ship];
		}
		if ([ent isKindOfClass:[PlanetEntity class]])
		{
			if ([self isKindOfClass:[PlayerEntity class]])
			{
				[(PlayerEntity *)self getDestroyed];
				return;
			}
			[self becomeExplosion];
		}
	}	
}

- (BOOL) collideWithShip:(ShipEntity *)other
{
	Vector  loc, fd;
	double  min_dist = collision_radius + [other collisionRadius];
	double  inc1, dam1;
	
	fd = v_forward;
	
	loc = [other getPosition];
	loc.x -= position.x;	loc.y -= position.y;	loc.z -= position.z;
	
	loc = unit_vector(&loc);
	
	inc1 = (v_forward.x*loc.x)+(v_forward.y*loc.y)+(v_forward.z*loc.z);
	
	if ((inc1 > 0)&&([other scanClass] == CLASS_CARGO)&&(has_scoop)&&([cargo count] < max_cargo)&&([other getCargoType] != CARGO_NOT_CARGO))
	{
		double inc2 = (v_up.x*loc.x)+(v_up.y*loc.y)+(v_up.z*loc.z);
		if ((inc2 < 0.0)||(![self isKindOfClass:[PlayerEntity class]])) // ie now in forward-lower quadrant or we're not the player
		{
			[self scoopUp:other];
			return NO;
		}
	}
	
	if (inc1 < 0.0)
		inc1 = -inc1;
		
	dam1 = inc1 * flight_speed * 0.25;   //  10 times what it was before!
	
	// reverse the collision in case this is cargo colliding with scoops
	if (![other beCollidedBy:self])
		return NO;

	[self takeScrapeDamage:dam1 from:other];
	[other takeScrapeDamage:dam1 from:self];
		
	position = [other getPosition];
	position.x -= (loc.x * min_dist);
	position.y -= (loc.y * min_dist);
	position.z -= (loc.z * min_dist);
	
	velocity.x -= (loc.x * min_dist)*0.1;
	velocity.y -= (loc.y * min_dist)*0.1;
	velocity.z -= (loc.z * min_dist)*0.1;
	
	return YES;
}

- (void) scoopUp:(ShipEntity *)other
{
	int		co_type,co_amount;
	switch ([other getCargoType])
	{
		case	CARGO_RANDOM :
			co_type = [other getCommodityType];
			co_amount = [other getCommodityAmount];
			break;
		case	CARGO_SLAVES :
			co_amount = 1;
			co_type = [universe commodityForName:@"Slaves"];
			if (co_type == NSNotFound)  // No 'Slaves' in this game, get something else instead...
			{
				co_type = [universe getRandomCommodity];
				co_amount = [universe getRandomAmountOfCommodity:co_type];
			}
			break;
		case	CARGO_ALLOY :
			co_amount = 1;
			co_type = [universe commodityForName:@"Alloys"];
			break;
		case	CARGO_MINERALS :
			co_amount = 1;
			co_type = [universe commodityForName:@"Minerals"];
			break;
		case	CARGO_THARGOID :
			co_amount = 1;
			co_type = [universe commodityForName:@"Alien Items"];
			break;
	}
	[other setCommodity:co_type andAmount:co_amount];   // belt and braces setting this!
	if (cargo_flag !=CARGO_FLAG_CANISTERS)
		cargo_flag = CARGO_FLAG_CANISTERS;
	//NSLog(@"---> %@ %d scooped %@", name, universal_id, [universe describeCommodity:co_type amount:co_amount]);
	if ([self isKindOfClass:[PlayerEntity class]])
		[universe addMessage:[universe describeCommodity:co_type amount:co_amount] forCount:4.5];
	[cargo addObject:other];
	[other setStatus:STATUS_IN_HOLD];		// prevents entity from being recycled!
	[[other collisionArray] removeObject:self]; // so it can't be scooped twice!
	[universe removeEntity:other];
	[shipAI message:@"CARGO_SCOOPED"];
	if ([cargo count] == max_cargo)
		[shipAI message:@"HOLD_FULL"];
}


- (BOOL) beCollidedBy:(ShipEntity *)other
{
	Vector  loc, fd;
	double  min_dist = collision_radius + [other collisionRadius];
	double  inc1, dam1;
	
	fd = v_forward;
	
	loc = [other getPosition];
	loc.x -= position.x;	loc.y -= position.y;	loc.z -= position.z;
	
	loc = unit_vector(&loc);
	
	inc1 = (v_forward.x*loc.x)+(v_forward.y*loc.y)+(v_forward.z*loc.z);
	
	if ((inc1 > 0)&&([other scanClass] == CLASS_CARGO)&&(has_scoop)&&([cargo count] < max_cargo)&&([other getCargoType] != CARGO_NOT_CARGO))
	{
		double inc2 = (v_up.x*loc.x)+(v_up.y*loc.y)+(v_up.z*loc.z);
		if ((inc2 < 0.0)||(![self isKindOfClass:[PlayerEntity class]])) // ie now in forward-lower quadrant or we're not the player
		{
			[self scoopUp:other];
			return NO;
		}
	}
	
	if (inc1 < 0.0)
		inc1 = -inc1;
	
	dam1 = inc1 * flight_speed * 0.25;
	
	[self takeScrapeDamage:dam1 from:other];
	[other takeScrapeDamage:dam1 from:self];
	
	position = [other getPosition];
	position.x -= (loc.x * min_dist);
	position.y -= (loc.y * min_dist);
	position.z -= (loc.z * min_dist);
	
	velocity.x -= (loc.x * min_dist)*0.1;
	velocity.y -= (loc.y * min_dist)*0.1;
	velocity.z -= (loc.z * min_dist)*0.1;
	
	return YES;
}

- (void) takeScrapeDamage:(double) amount from:(Entity *) ent
{
	if (status == STATUS_DEAD)  // it's too late for this one!
		return;
	//
	energy -= amount;
	// oops we hit too hard!!!
	if (energy <= 0.0)
	{
		being_mined = YES;  // same as using a mining laser
		[self becomeExplosion];
		if ([ent isKindOfClass:[ShipEntity class]])
		{
			ShipEntity* hunter = (ShipEntity *)ent;
			[hunter collectBountyFor:self];
			if ([hunter getPrimaryTarget] == (Entity *)self)
			{
				[hunter removeTarget:(Entity *)self];
				[[hunter getAI] message:@"TARGET_DESTROYED"];
			}
		}
	}
	// warn if I'm low on energy
	if (energy < max_energy *0.25)
		[shipAI message:@"ENERGY_LOW"];
}

- (void) enterDock:(StationEntity *)station
{
	[shipAI message:@"DOCKED"];
	[station noteDockedShip:self];
	[universe removeEntity:self];
}

- (void) leaveDock:(StationEntity *)station
{
	Vector launchPos = [station getPosition];
	Vector stat_f = vector_forward_from_quaternion([station QRotation]);
	launchPos.x += 500.0*stat_f.x;
	launchPos.y += 500.0*stat_f.y;
	launchPos.z += 500.0*stat_f.z;
    position = launchPos;
	q_rotation = [station QRotation];
	flight_roll = [station flight_roll];
	flight_pitch = 0.0;
	flight_speed = max_flight_speed * 0.5;
	status = STATUS_LAUNCHING;
	[shipAI message:@"LAUNCHED"];
	[universe addEntity:self];
}

- (void) enterWitchspace
{
	// witchspace entry effects here
	ParticleEntity *ring1 = [[ParticleEntity alloc] initHyperringFromShip:self]; // retained
	[universe addEntity:ring1];
	[ring1 release];
	ParticleEntity *ring2 = [[ParticleEntity alloc] initHyperringFromShip:self]; // retained
	[ring2 setSize:NSMakeSize([ring2 size].width * -2.5 ,[ring2 size].height * -2.0 )]; // shrinking!
	[universe addEntity:ring2];
	[ring2 release];
	
	[shipAI message:@"ENTERED_WITCHSPACE"];
	
	[universe witchspaceShipWithRole:roles];   // add a new ship like this one leaving!
	
	[universe removeEntity:self];
}

- (void) leaveWitchspace
{
	Vector		pos = [universe getWitchspaceExitPosition];
	Quaternion  q_rtn = [universe getWitchspaceExitRotation];
	position = pos;
	position.x += SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5); // randomise exit position
	position.y += SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
	position.z += SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
	q_rotation = q_rtn;
	flight_roll = 0.0;
	flight_pitch = 0.0;
	flight_speed = max_flight_speed * 0.25;
	status = STATUS_LAUNCHING;
	[shipAI message:@"EXITED_WITCHSPACE"];
	[universe addEntity:self];
	
	//NSLog(@"DEBUG Ship: %@ %d exiting witchspace now!", name, universal_id);
	
	// witchspace exit effects here
	ParticleEntity *ring1 = [[ParticleEntity alloc] initHyperringFromShip:self]; // retained
	[universe addEntity:ring1];
	[ring1 release];
	ParticleEntity *ring2 = [[ParticleEntity alloc] initHyperringFromShip:self]; // retained
	[ring2 setSize:NSMakeSize([ring2 size].width * -2.5 ,[ring2 size].height * -2.0 )]; // shrinking!
	[universe addEntity:ring2];
	[ring2 release];
}

- (void) markAsOffender:(int)offence_value
{
	bounty |= offence_value;
}

/*-----------------------------------------

	methods from AI

-----------------------------------------*/

- (void) pauseAI:(NSString *)intervalString
{
	[shipAI setNextThinkTime:[universe getTime] + [intervalString doubleValue]];
}

- (void) setDestinationToCurrentLocation
{
	destination = position;
	destination.x += (rand() % 100)*0.01 - 0.5;		// randomly add a .5m variance
	destination.y += (rand() % 100)*0.01 - 0.5;
	destination.z += (rand() % 100)*0.01 - 0.5;
}

- (void) setDesiredRangeTo:(NSString *)rangeString
{
	desired_range = [rangeString doubleValue];
}

- (void) performFlyToRangeFromDestination
{
	//NSLog(@"ShipEntity.performFlyToRangeFromDestination NOT YET IMPLEMENTED");
	condition = CONDITION_FLY_RANGE_FROM_DESTINATION;
}

- (void) setSpeedTo:(NSString *)speedString
{
	desired_speed = [speedString doubleValue];
}

- (void) setSpeedFactorTo:(NSString *)speedString
{
	desired_speed = max_flight_speed * [speedString doubleValue];
}

- (void) performIdle
{
	condition = CONDITION_IDLE;
}

- (void) setTargetToPrimaryAggressor
{
	if ([universe entityForUniversalID:primaryAggressor])
		[self addTarget:[universe entityForUniversalID:primaryAggressor]];
}

- (void) performAttack
{
	condition = CONDITION_ATTACK_TARGET;
}

- (void) scanForMerchantmen
{
	/*-- Locates the nearest merchantman in range --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	double found_d2 = desired_range*desired_range;
	found_target = NO_TARGET;
	for (i = 0; i < [entList count] ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if (([thing isKindOfClass:[ShipEntity class]])&&(thing != (Entity *)self))
		{
			ShipEntity* ship = (ShipEntity *)thing;
			if ((([[ship roles] isEqual:@"trader"])||([ship isKindOfClass:[PlayerEntity class]]))&&([ship getStatus] != STATUS_DEAD))
			{
				double d2;
				Vector delta = [ship getPosition];
				delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
				d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
				if (([ship isKindOfClass:[PlayerEntity class]])&&(PIRATES_PREFER_PLAYER)&&(d2*d2 < desired_range))
					d2 = 0.0;
				if (d2 < found_d2)
				{
					found_d2 = d2;
					found_target = [ship universal_id];
				}
			}
		}
	}
	[entList release];
	if (found_target != NO_TARGET)
		[shipAI message:@"TARGET_FOUND"];
	else
		[shipAI message:@"NOTHING_FOUND"];
}

- (void) scanForLoot
{
	/*-- Locates the nearest debris in range --*/
	if ((![self isKindOfClass:[StationEntity class]])&&(!has_scoop))
	{
		[shipAI message:@"NOTHING_FOUND"];		//can't collect loot if you have no scoop!
		return;
	}
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	double found_d2 = desired_range*desired_range;
	found_target = NO_TARGET;
	for (i = 0; i < [entList count]; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			ShipEntity *other = (ShipEntity *)thing;
			if (([other scanClass] == CLASS_CARGO)&&([other getCargoType] != CARGO_NOT_CARGO))
			{
				double d2;
				Vector delta = [thing getPosition];
				delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
				d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
				if (d2 < found_d2)
				{
					found_d2 = d2;
					found_target = [thing universal_id];
				}
			}
		}
	}
	[entList release];
	if (found_target != NO_TARGET)
		[shipAI message:@"TARGET_FOUND"];
	else
		[shipAI message:@"NOTHING_FOUND"];
}

- (void) scanForRandomLoot
{
	/*-- Locates the all debris in range and chooses a piece at random from the first sixteen found --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int thing_uids_found[16];
	int i, things_found;
	double found_d2 = desired_range*desired_range;
	found_target = NO_TARGET;
	if ((![self isKindOfClass:[StationEntity class]])&&(!has_scoop))
	{
		[shipAI message:@"NOTHING_FOUND"];		//can't collect loot if you have no scoop!
		return;
	}
	things_found = 0;
	for (i = 0; (i < [entList count])&&(things_found < 16) ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			ShipEntity *other = (ShipEntity *)thing;
			if (([other scanClass] == CLASS_CARGO)&&([other getCargoType] != CARGO_NOT_CARGO))
			{
				double d2;
				Vector delta = [thing getPosition];
				delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
				d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
				if (d2 < found_d2)
				{
					found_target = [thing universal_id];
					thing_uids_found[things_found++] = [thing universal_id];
				}
			}
		}
	}
	[entList release];
	if (found_target != NO_TARGET)
	{
		found_target = thing_uids_found[rand() % things_found];
		[shipAI message:@"TARGET_FOUND"];
	}
	else
		[shipAI message:@"NOTHING_FOUND"];
}

- (void) setTargetToFoundTarget
{
	if ([universe entityForUniversalID:found_target])
		[self addTarget:[universe entityForUniversalID:found_target]];
}

- (void) checkForFullHold
{
	if ([cargo count] >= max_cargo)
		[shipAI message:@"HOLD_FULL"];
}

- (void) performCollect
{
	condition = CONDITION_COLLECT_TARGET;
}

- (void) performIntercept
{
	condition = CONDITION_INTERCEPT_TARGET;
}

- (void) performFlee
{
	condition = CONDITION_FLEE_TARGET;
}

- (void) requestDockingCoordinates
{
	/*- requests coordinates from the nearest station it can find -*/
	NSArray* entList = [[universe getAllEntities] retain];
	StationEntity* station =  nil;
	Vector  p1 = position;
	double nearest2 = SCANNER_MAX_RANGE2 * 1000000.0; // 1000x scanner range (25600 km), squared.
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[StationEntity class]])
		{
			Vector p2 = [thing getPosition];
			p2.x -= p1.x;   p2.y -= p1.y; p2.z -= p1.z;
			double range2 = (p2.x * p2.x + p2.y * p2.y + p2.z * p2.z);
			if (range2 < nearest2)
			{
				station = (StationEntity *)thing;
				nearest2 = range2;
			}
		}
	}
	[entList release];
	if (station)
	{
		//NSLog(@"Station '%@' %@ with universal_id %d",[station name],station,[station universal_id]);
		coordinates = [station nextDockingCoordinatesForShip:self];
	}
	else
	{
		[shipAI message:@"NO_STATION_FOUND"];
	}
}

- (void) getWitchspaceEntryCoordinates
{
	/*- calculates coordinates from the nearest station it can find, or just fly 10s forward -*/
	NSArray* entList = [[universe getAllEntities] retain];
	StationEntity* station =  nil;
	Vector  p1 = position;
	double nearest2 = SCANNER_MAX_RANGE2 * 1000000.0; // 1000x scanner range (25600 km), squared.
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[StationEntity class]])
		{
			Vector p2 = [thing getPosition];
			p2.x -= p1.x;   p2.y -= p1.y; p2.z -= p1.z;
			double range2 = (p2.x * p2.x + p2.y * p2.y + p2.z * p2.z);
			if (range2 < nearest2)
			{
				station = (StationEntity *)thing;
				nearest2 = range2;
			}
		}
	}
	[entList release];
	if (station)
	{
		coordinates = [station getPosition];
		Vector  vr = vector_right_from_quaternion([station QRotation]);
		coordinates.x += 10000 * vr.x;  // 10km from station
		coordinates.y += 10000 * vr.y;
		coordinates.z += 10000 * vr.z;
	}
	else
	{
		coordinates = position;
		coordinates.x += v_forward.x * max_flight_speed * 10.0;
		coordinates.y += v_forward.y * max_flight_speed * 10.0;
		coordinates.z += v_forward.z * max_flight_speed * 10.0;
	}
	//[shipAI message:@"OKAY"];
}

- (void) abortDocking
{
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[StationEntity class]])
			[(StationEntity *)thing abortDockingForShip:self];
	}
	[entList release];
}

- (void) setDestinationFromCoordinates
{
	destination = coordinates;
}

- (void) performDocking
{
	//NSLog(@"ShipEntity.performDocking NOT IMPLEMENTED!");
}

- (void) performFaceDestination
{
	condition = CONDITION_FACE_DESTINATION;
}

- (void) setTargetToStation
{
	/*- selects the nearest station it can find -*/
	NSArray* entList = [[universe getAllEntities] retain];
	StationEntity* station =  nil;
	Vector  p1 = position;
	double nearest2 = SCANNER_MAX_RANGE2 * 1000000.0; // 1000x scanner range (25600 km), squared.
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[StationEntity class]])
		{
			Vector p2 = [thing getPosition];
			p2.x -= p1.x;   p2.y -= p1.y; p2.z -= p1.z;
			double range2 = (p2.x * p2.x + p2.y * p2.y + p2.z * p2.z);
			if (range2 < nearest2)
			{
				station = (StationEntity *)thing;
				nearest2 = range2;
			}
		}
	}
	[entList release];
	if (station)
		primaryTarget = [station universal_id];
}

- (void) performTumble
{
	flight_roll = max_flight_roll*((rand() % 2000) - 1000)/1000.0;
	flight_pitch = max_flight_pitch*((rand() % 2000) - 1000)/1000.0;
	velocity.x = (rand() % 1000) - 500;
	velocity.y = (rand() % 1000) - 500;
	velocity.z = (rand() % 1000) - 500;
	velocity = unit_vector(&velocity);
	velocity.x *= max_flight_speed;
	velocity.y *= max_flight_speed;
	velocity.z *= max_flight_speed;
	
	condition = CONDITION_TUMBLE;
}

- (void) fightOrFleeMissile
{
	NSArray* entList = [[universe getAllEntities] retain];
	ShipEntity* missile =  nil;
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing scanClass] == CLASS_MISSILE)
		{
			if ([(ShipEntity *)thing getPrimaryTarget] == self)
				missile = (ShipEntity *)thing;
		}
	}
	[entList release];
	
	//NSLog(@"---> %@ %d targetting the Missile %d", name, universal_id, [missile universal_id]);
	[self addTarget:missile];
	
	if ((missile)&&(has_ecm))
	{
		// use the ECM and battle on
		//NSLog(@"---> and firing ecm!");
		[self fireECM];
		return;
	}
	
	if (missile)
	{
		//NSLog(@"---> and running away!");
		jink.x = 0.0;
		jink.y = 0.0;
		jink.z = 1000.0;
		[self performFlee];
	}
}

// new

- (PlanetEntity *) findNearestPlanet
{
	/*- selects the nearest planet it can find -*/
	NSArray			*entList = [[universe getAllEntities] retain];
	PlanetEntity	*the_planet =  nil;
	Vector  p1 = position;
	double nearest2 = SCANNER_MAX_RANGE2 * 10000000000.0; // 100 000x scanner range (2 560 000 km), squared.
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity  *thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[PlanetEntity class]])
		{
			if ([(PlanetEntity *)thing getPlanetType] == PLANET_TYPE_GREEN)
			{
				Vector p2 = [thing getPosition];
				p2.x -= p1.x;   p2.y -= p1.y; p2.z -= p1.z;
				double range2 = (p2.x * p2.x + p2.y * p2.y + p2.z * p2.z);
				if ((!the_planet)||(range2 < nearest2))
				{
					the_planet = (PlanetEntity *)thing;
					nearest2 = range2;
				}
			}
		}
	}
	[entList release];
	return the_planet;
}

- (PlanetEntity *) findNearestLargeBody
{
	/*- selects the nearest planet it can find -*/
	NSArray			*entList = [[universe getAllEntities] retain];
	PlanetEntity	*the_planet =  nil;
	Vector  p1 = position;
	double nearest2 = SCANNER_MAX_RANGE2 * 10000000000.0; // 100 000x scanner range (2 560 000 km), squared.
	int i;
	for (i = 0; i < [entList count]; i++)
	{
		Entity  *thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[PlanetEntity class]])
		{
			Vector p2 = [thing getPosition];
			p2.x -= p1.x;   p2.y -= p1.y; p2.z -= p1.z;
			double range2 = (p2.x * p2.x + p2.y * p2.y + p2.z * p2.z);
			if ((!the_planet)||(range2 < nearest2))
			{
				the_planet = (PlanetEntity *)thing;
				nearest2 = range2;
			}
		}
	}
	[entList release];
	return the_planet;
}

- (void) setCourseToPlanet
{
	/*- selects the nearest planet it can find -*/
	PlanetEntity	*the_planet =  [self findNearestPlanet];
	if (the_planet)
	{
		destination = [the_planet getPosition];
		desired_range = [the_planet collisionRadius] + 100.0;   // 100m from the surface
	}
}

- (void) setTakeOffFromPlanet
{
	/*- selects the nearest planet it can find -*/
	PlanetEntity	*the_planet =  [self findNearestPlanet];
	if (the_planet)
	{
		destination = [the_planet getPosition];
		desired_range = [the_planet collisionRadius] + 10000.0;   // 10km from the surface
	}
	else
		NSLog(@"***** Ackk!! planet not found!!!");
}

- (void) landOnPlanet
{
	/*- selects the nearest planet it can find -*/
	PlanetEntity	*the_planet =  [self findNearestPlanet];
	if (the_planet)
	{
		[the_planet welcomeShuttle:self];   // 10km from the surface
	}
	[shipAI message:@"LANDED_ON_PLANET"];
	[universe removeEntity:self];
}

- (void) setAITo:(NSString *)aiString
{
	[[self getAI] setStateMachine:aiString];
}

- (void) checkTargetLegalStatus
{
	ShipEntity  *other_ship = (ShipEntity *)[universe entityForUniversalID:primaryTarget];
	if (!other_ship)
	{
		[shipAI message:@"NO_TARGET"];
		return;
	}
	else
	{
		int ls = [other_ship legal_status];
		if (ls > 50)
		{
			[shipAI message:@"TARGET_FUGITIVE"];
			return;
		}
		if (ls > 0)
		{
			[shipAI message:@"TARGET_OFFENDER"];
			return;
		}
		[shipAI message:@"TARGET_CLEAN"];
	}
}

- (void) exitAI
{
	[shipAI exitStateMachine];
}

- (void) setDestinationToTarget
{
	destination = [[universe entityForUniversalID:primaryTarget] getPosition];
}

- (void) checkCourseToDestination
{
	if ([universe isVectorClearFromEntity:self toDistance:desired_range fromPoint:destination])
		[shipAI message:@"COURSE_OK"];
	else
	{
		destination = [universe getSafeVectorFromEntity:self toDistance:desired_range fromPoint:destination];
		[shipAI message:@"WAYPOINT_SET"];
	}
}

- (void) scanForOffenders
{
	/*-- Locates all the ships in range and compares their legal status or bounty against rand() % 64 - chooses the worst offender --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int i, ships_found, worst_legal_factor;
	double found_d2 = desired_range*desired_range;
	found_target = NO_TARGET;
	ships_found = 0;
	worst_legal_factor = 0;
	for (i = 0; (i < [entList count])&&(ships_found < 16) ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if (([thing isKindOfClass:[ShipEntity class]])&&([thing scanClass] != CLASS_CARGO))
		{
			double d2;
			int legal_factor = [(ShipEntity *)thing legal_status];
			int random_factor = rand() % 256;   // 25% chance of spotting a fugitive in 15s
			Vector delta = [thing getPosition];
			delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
			d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
			if ((d2 < found_d2)&&(random_factor < legal_factor)&&(legal_factor > worst_legal_factor))
			{
				found_target = [thing universal_id];
				worst_legal_factor = legal_factor;
			}
		}
	}
	[entList release];
	if (found_target != NO_TARGET)
		[shipAI message:@"TARGET_FOUND"];
	else
		[shipAI message:@"NOTHING_FOUND"];
}

- (void) setCourseToWitchpoint
{
	if (universe)
	{
		destination = [universe getWitchspaceExitPosition];
		desired_range = 10000.0;   // 10km away
	}
}

- (void) performHyperSpaceExit
{
	[self enterWitchspace];
}

- (void) broadcastDistressMessage
{
	/*-- Locates all the stations, bounty hunters and police ships in range and tells them that you are under attack --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	double d2;
	double found_d2 = SCANNER_MAX_RANGE2;
	found_target = NO_TARGET;
	for (i = 0; i < [entList count] ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			Vector delta = [thing getPosition];
			delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
			d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
			if (d2 < found_d2)
			{
				ShipEntity  *ship = (ShipEntity *)thing;
				// tell it! //
				if ([ship isKindOfClass:[StationEntity class]])
					[ship acceptDistressMessageFrom:self];
				if ([[ship roles] isEqual:@"police"])
					[ship acceptDistressMessageFrom:self];
				if ([[ship roles] isEqual:@"hunter"])
					[ship acceptDistressMessageFrom:self];
			}
		}
	}
	[entList release];
}

- (void) acceptDistressMessageFrom:(ShipEntity *)other
{
	found_target = [[other getPrimaryTarget] universal_id];
	switch (condition)
	{
		case CONDITION_ATTACK_TARGET :
		case CONDITION_ATTACK_FLY_TO_TARGET :
		case CONDITION_ATTACK_FLY_FROM_TARGET :
			// busy - ignore the request
			break;
			
		case CONDITION_FLEE_TARGET :
			// scared - ignore the request;
			break;
			
		default :
			//NSLog(@"%@ %d responding to distress message from %@ %d", name, universal_id, [other name], [other universal_id]);
			if ([roles isEqual:@"police"])
				[(ShipEntity *)[universe entityForUniversalID:found_target] markAsOffender:8];  // you have been warned!!
			[shipAI reactToMessage:@"ACCEPT_DISTRESS_CALL"];
			break;
	}
}

- (void) broadcastThargoidDestroyed
{
	/*-- Locates all tharglets in range and tells them you've gone --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	double d2;
	double found_d2 = SCANNER_MAX_RANGE2;
	for (i = 0; i < [entList count] ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			Vector delta = [thing getPosition];
			delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
			d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
			if (d2 < found_d2)
			{
				ShipEntity  *ship = (ShipEntity *)thing;
				// tell it! //
				if ([[ship roles] isEqual:@"tharglet"])
					[[ship getAI] message:@"THARGOID_DESTROYED"];
			}
		}
	}
	[entList release];
}

- (void) scanForThargoid
{
	/*-- Locates all the thargoid warships in range and chooses the nearest --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	double found_d2 = desired_range*desired_range;
	found_target = NO_TARGET;
	for (i = 0; i < [entList count] ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			double d2;
			ShipEntity *ship = (ShipEntity *)thing;
			Vector delta = [thing getPosition];
			delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
			d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
			if (([[ship roles] isEqual:@"thargoid"])&&(d2 < found_d2))
			{
				found_target = [thing universal_id];
				found_d2 = d2;
			}
		}
	}
	[entList release];
	if (found_target != NO_TARGET)
		[shipAI message:@"TARGET_FOUND"];
	else
	{
		if ([roles isEqual:@"tharglet"])
		{
			// now we're just a bunch of alien artefacts!
			scan_class = CLASS_CARGO;
			reportAImessages = NO;
			[shipAI setStateMachine:@"nullAI.plist"];
		}
		[shipAI message:@"NOTHING_FOUND"];
	}
}

- (void) scanForNonThargoid
{
	/*-- Locates all the non thargoid ships in range and chooses the nearest --*/
	NSArray* entList = [[universe getAllEntities] retain];
	int i;
	double found_d2 = desired_range*desired_range;
	found_target = NO_TARGET;
	for (i = 0; i < [entList count] ; i++)
	{
		Entity* thing = (Entity *)[entList objectAtIndex:i];
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			double d2;
			NSString *shiproles = [(ShipEntity *)thing roles];
			Vector delta = [thing getPosition];
			delta.x -= position.x;  delta.y -= position.y;  delta.z -= position.z;
			d2 = delta.x*delta.x + delta.y*delta.y + delta.z*delta.z;
			if (([thing scanClass] != CLASS_CARGO)&&(![shiproles hasPrefix:@"tharg"])&&(d2 < found_d2))
			{
				found_target = [thing universal_id];
				if ([thing isKindOfClass:[PlayerEntity class]]) d2 = 0.0;   // prefer the player
				found_d2 = d2;
			}
		}
	}
	[entList release];
	if (found_target != NO_TARGET)
		[shipAI message:@"TARGET_FOUND"];
	else
		[shipAI message:@"NOTHING_FOUND"];
}



@end
