//
//  SkyEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#define SKY_SCALE	2000.0

#import "Entity.h"
#import "SkyEntity.h"

#import "vector.h"
#import "Universe.h"
#import "TextureStore.h"
#import "MyOpenGLView.h"

@implementation SkyEntity

- (id) init
{    
    self = [super init];
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
    //
    n_vertices = 0;
    n_faces = 0;
	delta = 0.0;
    //
    displayListName = 0;
    //
    status = STATUS_EFFECT;
	sky_type = SKY_BILLBOARDS;
	//
	int i;
	float h1 = (rand() % 1024)/1024.0;
	float h2 = h1 + 1.0 / (1.0 + (rand() % 5));
	while (h2 > 1.0)
		h2 -= 1.0;
	NSColor *col1 = [NSColor colorWithCalibratedHue:h1 saturation:(rand() % 1024)/1024.0 brightness:0.5 +(rand() % 1024)/2048.0 alpha:1.0];
	NSColor *col2 = [NSColor colorWithCalibratedHue:h2 saturation:0.5 +(rand() % 1024)/2048.0 brightness:0.5 +(rand() % 1024)/2048.0 alpha:1.0];
	//
	sky_color = [[col2 blendedColorWithFraction:0.5 ofColor:col1] retain];
	//
	// init stars
	//
	for (i = 0; i < SKY_MAX_STARS; i++)
	{
		NSColor *col3 = [col1 blendedColorWithFraction:(rand() % 1024)/1024.0 ofColor:col2];
		star_color[0][i] = [col3 redComponent];
		star_color[1][i] = [col3 greenComponent];
		star_color[2][i] = [col3 blueComponent];
		star_color[3][i] = 1.0;
		Quaternion q;
		quaternion_set_random(&q);
		Vector vi = vector_right_from_quaternion(q);
		Vector vj = vector_up_from_quaternion(q);
		Vector vk = vector_forward_from_quaternion(q);
		star_vector[i] = vk;
		star_vector[i].x *= 500.0;
		star_vector[i].y *= 500.0;
		star_vector[i].z *= 500.0;
		double p_size = 1 + (rand() % 6);
		star_quad[i][0] = star_vector[i];
		star_quad[i][1].x = star_quad[i][0].x + p_size * vj.x;
		star_quad[i][1].y = star_quad[i][0].y + p_size * vj.y;
		star_quad[i][1].z = star_quad[i][0].z + p_size * vj.z;
		star_quad[i][2].x = star_quad[i][1].x + p_size * vi.x;
		star_quad[i][2].y = star_quad[i][1].y + p_size * vi.y;
		star_quad[i][2].z = star_quad[i][1].z + p_size * vi.z;
		star_quad[i][3].x = star_quad[i][0].x + p_size * vi.x;
		star_quad[i][3].y = star_quad[i][0].y + p_size * vi.y;
		star_quad[i][3].z = star_quad[i][0].z + p_size * vi.z;
		int j;
		for (j = 0; j < 4; j++)
		{
			star_quad[i][j].x -= 0.5 * p_size * (vi.x + vj.x);
			star_quad[i][j].y -= 0.5 * p_size * (vi.y + vj.y);
			star_quad[i][j].z -= 0.5 * p_size * (vi.z + vj.z);
		}
	}
	star_textureName = 0;
	//
    return self;
}

- (id) initAsWitchspace
{
    self = [super init];
    //
    quaternion_set_identity(&q_rotation);
    quaternion_into_gl_matrix(q_rotation, rotMatrix);
    //
    position.x = 0.0;
    position.y = 0.0;
    position.z = 0.0;
    //
    n_vertices = 0;
    n_faces = 0;
	delta = 0.0;
    //
    displayListName = 0;
    //
    status = STATUS_EFFECT;
	sky_type = SKY_BILLBOARDS;
    //
	int i;
	NSColor *col1 = [NSColor colorWithCalibratedRed:0.0 green:1.0 blue:0.5 alpha:1.0];
	NSColor *col2 = [NSColor colorWithCalibratedRed:0.0 green:1.0 blue:0.0 alpha:1.0];
	//
	sky_color = [[col2 blendedColorWithFraction:0.5 ofColor:col1] retain];
	//
	// init stars
	//
	for (i = 0; i < SKY_MAX_STARS; i++)
	{
		NSColor *col3 = [col1 blendedColorWithFraction:(rand() % 1024)/1024.0 ofColor:col2];
		star_color[0][i] = [col3 redComponent];
		star_color[1][i] = [col3 greenComponent];
		star_color[2][i] = [col3 blueComponent];
		star_color[3][i] = 1.0;
		Quaternion q;
		quaternion_set_random(&q);
		Vector vi = vector_right_from_quaternion(q);
		Vector vj = vector_up_from_quaternion(q);
		Vector vk = vector_forward_from_quaternion(q);
		star_vector[i] = vk;
		star_vector[i].x *= 500.0;
		star_vector[i].y *= 500.0;
		star_vector[i].z *= 500.0;
		double p_size = 1 + (rand() % 6);
		star_quad[i][0] = star_vector[i];
		star_quad[i][1].x = star_quad[i][0].x + p_size * vj.x;
		star_quad[i][1].y = star_quad[i][0].y + p_size * vj.y;
		star_quad[i][1].z = star_quad[i][0].z + p_size * vj.z;
		star_quad[i][2].x = star_quad[i][1].x + p_size * vi.x;
		star_quad[i][2].y = star_quad[i][1].y + p_size * vi.y;
		star_quad[i][2].z = star_quad[i][1].z + p_size * vi.z;
		star_quad[i][3].x = star_quad[i][0].x + p_size * vi.x;
		star_quad[i][3].y = star_quad[i][0].y + p_size * vi.y;
		star_quad[i][3].z = star_quad[i][0].z + p_size * vi.z;
		int j;
		for (j = 0; j < 4; j++)
		{
			star_quad[i][j].x -= 0.5 * p_size * (vi.x + vj.x);
			star_quad[i][j].y -= 0.5 * p_size * (vi.y + vj.y);
			star_quad[i][j].z -= 0.5 * p_size * (vi.z + vj.z);
		}
	}
	star_textureName = 0;
	//
    return self;
}

- (void) dealloc
{
	if (sky_color)  [sky_color release];
	[super dealloc];
}

- (NSColor *) sky_color
{
	return sky_color;
}

- (void) update:(double) delta_t
{
	zero_distance = MAX_CLEAR_DEPTH*MAX_CLEAR_DEPTH;
	position = [[universe entityZero] getPosition];
}

- (BOOL) canCollide
{
	return NO;
}

- (void) drawEntity:(BOOL) immediate :(BOOL) translucent
{
    // roll out each face and vertex in turn
    //
    int vi;
	double  star_size = [(MyOpenGLView *)[universe gameView] viewSize].width / 512.0;
    
	if ([universe breakPatternHide])   return; // DON'T DRAW

    //
    if (!translucent)
	{
		// disapply lighting
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);	// don't read the depth buffer
		glDepthMask(GL_FALSE);	// don't write to depth buffer
		glDisable(GL_CULL_FACE);			// face culling
		//
		glShadeModel(GL_SMOOTH);	// smoothing for color values...
			
		if (immediate)
		{
			switch (sky_type)
			{
				case SKY_POINTS :
					glEnable(GL_POINT_SMOOTH);
					glBegin(GL_POINTS);
					for (vi = 0; vi < SKY_MAX_STARS; vi++)
					{
						int p_size = 1 + floor((rand() % 4) * star_size);
						glPointSize(p_size);
						glColor4f( star_color[0][vi], star_color[1][vi], star_color[2][vi], star_color[3][vi]);
						glVertex3f( star_vector[vi].x, star_vector[vi].y, star_vector[vi].z );
					}
					glEnd();
					glDisable(GL_POINT_SMOOTH);
					break;
					
				case SKY_BILLBOARDS :	
					if ((star_textureName == 0)&&(universe))
						star_textureName = [[universe textureStore] getTextureNameFor:@"star64.png"];
					glEnable(GL_TEXTURE_2D);
					glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					glBindTexture(GL_TEXTURE_2D, star_textureName);
					glBegin(GL_QUADS);
					for (vi = 0; vi < SKY_MAX_STARS; vi++)
					{            
						glColor4f( star_color[0][vi], star_color[1][vi], star_color[2][vi], star_color[3][vi]);
						glTexCoord2i(0,0);
						glVertex3f( star_quad[vi][0].x, star_quad[vi][0].y, star_quad[vi][0].z);
						glTexCoord2i(1,0);
						glVertex3f( star_quad[vi][1].x, star_quad[vi][1].y, star_quad[vi][1].z);
						glTexCoord2i(1,1);
						glVertex3f( star_quad[vi][2].x, star_quad[vi][2].y, star_quad[vi][2].z);
						glTexCoord2i(0,1);
						glVertex3f( star_quad[vi][3].x, star_quad[vi][3].y, star_quad[vi][3].z);
					}
					glEnd();
					glDisable(GL_TEXTURE_2D);
					break;
			}
		}
		else
		{
			if (displayListName != 0)
				glCallList(displayListName);
			else
			{
				[self initialiseTextures];
				[self generateDisplayList];
				//
			}
		}
		// reapply lighting &c
		glEnable(GL_CULL_FACE);			// face culling
		glEnable(GL_LIGHTING);
		glEnable(GL_DEPTH_TEST);	// read the depth buffer
		glDepthMask(GL_TRUE);	// restore write to depth buffer
	}
}





@end
