//
//  StationEntity.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <Foundation/Foundation.h>

#include "entities.h"

#define STATION_ALERT_LEVEL_GREEN		0
#define STATION_ALERT_LEVEL_YELLOW		1
#define STATION_ALERT_LEVEL_RED			2

#define STATION_MAX_POLICE				8

#define STATION_DELAY_BETWEEN_LAUNCHES  6.0

@interface StationEntity : ShipEntity {
	
	NSMutableDictionary		*shipsOnApproach;
	NSMutableArray			*launchQueue;
	double					last_launch_time;
	double					approach_spacing;
	int						alert_level;
	int						police_launched;
	
	int						equivalent_tech_level;
	double					equipment_price_factor;

	double					port_radius;
	
	BOOL					no_docking_while_launching;
	
	int						planet;
	
	NSMutableArray			*localMarket;
	
	int						docked_shuttles;
	double					last_shuttle_launch_time;
	double					shuttle_launch_interval;
	
	int						docked_traders;
	double					last_trader_launch_time;
	double					trader_launch_interval;
	
}

- (NSMutableArray *) localMarket;
- (NSMutableArray *) initialiseLocalMarketWithSeed: (Random_Seed) s_seed andRandomFactor: (int) random_factor;

- (int) equivalent_tech_level;
- (double) equipment_price_factor;

- (void) setPlanet:(PlanetEntity *)planet_entity;

- (PlanetEntity *) planet;

- (void) sanityCheckShipsOnApproach;

- (void) autoDockShipsOnApproach;

- (Vector) nextDockingCoordinatesForShip:(ShipEntity *) ship;

- (Vector) upVector;

- (void) clear;


- (void) abortAllDockings;

- (void) abortDockingForShip:(ShipEntity *) ship;

- (void) addShipToLaunchQueue:(ShipEntity *) ship;

- (int) countShipsInLaunchQueueWithRole:(NSString *) a_role;

- (void) launchShip:(ShipEntity *) ship;

- (void) noteDockedShip:(ShipEntity *) ship;

////////////////////////////////////////////////////////////// AI methods...

- (void) increaseAlertLevel;

- (void) decreaseAlertLevel;

- (void) launchPolice;

- (void) launchDefenseShip;

- (void) launchScavenger;

- (void) launchShuttle;

- (void) launchTrader;

@end
