//
//  StationEntity.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import "StationEntity.h"
#import "entities.h"

#import "AI.h"


@implementation StationEntity

- (void) acceptDistressMessageFrom:(ShipEntity *)other
{
	if (self != [universe station])
	{
		//NSLog(@"DEBUG acceptDistressMessageFrom rejected from sub-station '%@'", name);
		return;
	}
	
	//NSLog(@"DEBUG %@ %d responding to distress message from %@ %d", name, universal_id, [other name], [other universal_id]);
	{
		int police_target = [[other getPrimaryTarget] universal_id];
		[(ShipEntity *)[universe entityForUniversalID:police_target] markAsOffender:8];
		if (police_launched < STATION_MAX_POLICE)
		{
			ShipEntity  *police_ship;
			if (![universe entityForUniversalID:police_target])
			{
				[shipAI message:@"TARGET_LOST"];
				return;
			}
				
			//NSLog(@"DEBUG Launching Police Ship to intercept %@",[universe entityForUniversalID:police_target]);
				
			police_ship = [universe getShipWithRole:@"police"];   // retain count = 1
			[police_ship addTarget:[universe entityForUniversalID:police_target]];
			[police_ship setScanClass: CLASS_POLICE];
			
			//[police_ship setReportAImessages:YES]; // debug
			
			[[police_ship getAI] setStateMachine:@"policeInterceptAI.plist"];
			[self addShipToLaunchQueue:police_ship];
			[police_ship release];
			police_launched++;
		}
		no_docking_while_launching = YES;
		[self abortAllDockings];
	}
}

- (int) equivalent_tech_level
{
	return equivalent_tech_level;
}

- (double) equipment_price_factor
{
	return equipment_price_factor;
}


- (NSMutableArray *) localMarket
{
	return localMarket;
}

- (NSMutableArray *) initialiseLocalMarketWithSeed: (Random_Seed) s_seed andRandomFactor: (int) random_factor
{
	//NSLog(@"///// Initialising local market for station %@ with roles %@",self,[self roles]);
	int economy = [(NSNumber *)[[universe generateSystemData:s_seed] objectForKey:@"economy"] intValue];
	if (localMarket)
		[localMarket release];
	localMarket = [[NSMutableArray alloc] initWithArray:[universe commodityDataForEconomy:economy andStation:self andRandomFactor:random_factor]];
	return localMarket;
}

- (void) setUniverse:(Universe *)univ
{
    if (univ)
    {
        if (universe)	[universe release];
        universe = [univ retain];
    }
	else
	{
        if (universe)	[universe release];
        universe = nil;
		[localMarket release];
		localMarket = nil;
    }
}

- (void) setPlanet:(PlanetEntity *)planet_entity
{
	if (planet_entity)
		planet = [planet_entity universal_id];
	else
		planet = NO_TARGET;
}

- (PlanetEntity *) planet
{
	return (PlanetEntity *)[universe entityForUniversalID:planet];
}

- (void) sanityCheckShipsOnApproach
{
	int i;
	NSArray*	ships = [shipsOnApproach allKeys];
	for (i = 0; i < [ships count]; i++)
	{
		int sid = [(NSString *)[ships objectAtIndex:i] intValue];
		if ((sid == NO_TARGET)||(![universe entityForUniversalID:sid]))
			[shipsOnApproach removeObjectForKey:[ships objectAtIndex:i]];
	}
	if ([shipsOnApproach count] == 0)
		last_launch_time = [universe getTime];
}

- (void) abortAllDockings
{
	int i;
	NSArray*	ships = [shipsOnApproach allKeys];
	for (i = 0; i < [ships count]; i++)
	{
		int sid = [(NSString *)[ships objectAtIndex:i] intValue];
		if ([universe entityForUniversalID:sid])
			[[(ShipEntity *)[universe entityForUniversalID:sid] getAI] message:@"DOCKING_ABORTED"];
		[shipsOnApproach removeObjectForKey:[ships objectAtIndex:i]];
	}
	last_launch_time = [universe getTime];
}


- (void) autoDockShipsOnApproach
{
	int i;
	NSArray*	ships = [shipsOnApproach allKeys];
	for (i = 0; i < [ships count]; i++)
	{
		int sid = [(NSString *)[ships objectAtIndex:i] intValue];
		if ([universe entityForUniversalID:sid])
			[(ShipEntity *)[universe entityForUniversalID:sid] enterDock:self];
		[shipsOnApproach removeObjectForKey:[ships objectAtIndex:i]];
	}
}


- (Vector) nextDockingCoordinatesForShip:(ShipEntity *) ship
{
	Vector		coords;
	NSString*   shipID = [NSString stringWithFormat:@"%d",[ship universal_id]];
	
	if (no_docking_while_launching)
	{
		[[ship getAI] message:@"TRY_AGAIN_LATER"];
		return [ship getPosition];  // hold position
	}
	
	if (![shipsOnApproach objectForKey:shipID])
	{
		NSMutableArray*		coordinatesStack =  [NSMutableArray arrayWithCapacity:3];
		NSString*			speedMessage = @"SLOW";
		double offset = 6 * port_radius;
		double offset2 = 6 * port_radius + approach_spacing;
		approach_spacing += port_radius;  // space out incoming ships by 500m
		while (offset >= 0.0)
		{
			NSMutableDictionary*	nextCoords =	[NSMutableDictionary dictionaryWithCapacity:3];
			Vector coords = [self getPosition];
			coords.x += offset*v_forward.x;
			coords.y += offset*v_forward.y;
			coords.z += offset*v_forward.z;
			coords.x += offset2*v_up.x;
			coords.y += offset2*v_up.y;
			coords.z += offset2*v_up.z;
			offset2 = 0;	// just the first one is offset to one side
			//NSLog(@"docking coordinates [%d] = (%.2f, %.2f, %.2f)", 3-i, coords.x, coords.y, coords.z);
			[nextCoords setObject:[NSString stringWithFormat:@"%f",coords.x] forKey:@"x"];
			[nextCoords setObject:[NSString stringWithFormat:@"%f",coords.y] forKey:@"y"];
			[nextCoords setObject:[NSString stringWithFormat:@"%f",coords.z] forKey:@"z"];
			[nextCoords setObject:[NSString stringWithFormat:@"%@",speedMessage] forKey:@"speed"];
			[coordinatesStack addObject:nextCoords];
			if (offset > 3 * port_radius)
				offset -= port_radius;
			else
			{
				offset -= port_radius;
				speedMessage = @"DEAD_SLOW";
			}	
		}
		//NSLog(@"Coordinates stack = %@",[coordinatesStack description]);
		[shipsOnApproach setObject:coordinatesStack forKey:shipID];
	}
	//NSLog(@"shipsOnApproach = %@",[[shipsOnApproach allKeys] description]);
	if ([shipsOnApproach objectForKey:shipID])
	{
		NSMutableArray* coordinatesStack = (NSMutableArray *)[shipsOnApproach objectForKey:shipID];
		//NSLog(@"coordinatesStack = %@", [coordinatesStack description]);
		if ([coordinatesStack count] == 0)
		{
			[[ship getAI] message:@"DOCKED"];
			// set last launch time to avoid clashes with outgoing ships
			last_launch_time = [universe getTime];
			[shipsOnApproach removeObjectForKey:shipID];
			//NSLog(@"shipsOnApproach = %@",[[shipsOnApproach allKeys] description]);
		}
		else
		{
			NSDictionary* nextCoords = (NSDictionary *)[coordinatesStack objectAtIndex:0];
			NSString* speedMessage = (NSString *)[nextCoords objectForKey:@"speed"];
			//NSLog(@"nextCoords = %@", [nextCoords description]);
			coords.x = [(NSString *)[nextCoords objectForKey:@"x"] doubleValue];
			coords.y = [(NSString *)[nextCoords objectForKey:@"y"] doubleValue];
			coords.z = [(NSString *)[nextCoords objectForKey:@"z"] doubleValue];
			//NSLog(@"Next docking coordinates ---> (%.2f, %.2f, %.2f)", coords.x, coords.y, coords.z);
			if ([speedMessage isEqual:@"DEAD_SLOW"])
				[[ship getAI] message:@"APPROACH_STATION"];
			else
				[[ship getAI] message:@"APPROACH_COORDINATES"];
			[coordinatesStack removeObjectAtIndex:0];
			if ([coordinatesStack count] == 0)
			{
				// set last launch time to avoid clashes with outgoing ships
				last_launch_time = [universe getTime];
				[shipsOnApproach removeObjectForKey:shipID];
				//NSLog(@"shipsOnApproach = %@",[[shipsOnApproach allKeys] description]);
			}
		}
	}
	
	return coords;
}

- (void) abortDockingForShip:(ShipEntity *) ship
{
	NSString*   shipID = [NSString stringWithFormat:@"%d",[ship universal_id]];
	if ([universe entityForUniversalID:[ship universal_id]])
		[[(ShipEntity *)[universe entityForUniversalID:[ship universal_id]] getAI] message:@"DOCKING_ABORTED"];
	if ([shipsOnApproach objectForKey:shipID])
		[shipsOnApproach removeObjectForKey:shipID];
}

- (Vector) upVector
{
	return v_right; // because the slot is horizontal dammit
}

//////////////////////////////////////////////// from superclass

- (id) init
{
	self = [super init];
	
	shipsOnApproach = [[NSMutableDictionary alloc] initWithCapacity:5]; // alloc retains
	launchQueue = [[NSMutableArray alloc] initWithCapacity:16]; // retained
	alert_level = 0;
	police_launched = 0;
	last_launch_time = 0.0;
	no_docking_while_launching = NO;
	
	localMarket = nil;
	
	// local specials
	equivalent_tech_level = NSNotFound;
	equipment_price_factor = 1.0;
	port_radius = 500.0;
	
	docked_shuttles = rand() % 4;   // 0..3;
	last_shuttle_launch_time = 0.0;
	shuttle_launch_interval = 15.0 * 60.0;  // every 15 minutes
	last_shuttle_launch_time = - (rand() % 60) * shuttle_launch_interval / 60.0;
	
	docked_traders = 1 + (rand() % 3);   // 1..3;
	last_trader_launch_time = 0.0;
	trader_launch_interval = ((rand() % 10) + 20.0) * 60.0;  // every few minutes
	last_trader_launch_time = 60.0 - trader_launch_interval; // in one minute's time
	
	return self;
}

- (id) initWithDictionary:(NSDictionary *) dict
{
	port_radius = 500.0;	// may be overwritten by [super initWithDictionary:dict]
	
	self = [super initWithDictionary:dict];
	
	//NSLog(@"DEBUG setting up station '%@' from dict:%@",name,[dict description]);
	
	if ([dict objectForKey:@"equivalent_tech_level"])
		equivalent_tech_level = [(NSNumber *)[dict objectForKey:@"equivalent_tech_level"] intValue];
	else
		equivalent_tech_level = NSNotFound;
	
	if ([dict objectForKey:@"equipment_price_factor"])
		equipment_price_factor = [(NSNumber *)[dict objectForKey:@"equipment_price_factor"] doubleValue];
	else
		equipment_price_factor = 1.0;
	
	
	shipsOnApproach = [[NSMutableDictionary alloc] initWithCapacity:16]; // alloc retains
	launchQueue = [[NSMutableArray alloc] initWithCapacity:16]; // retained
	alert_level = 0;
	police_launched = 0;
	last_launch_time = 0.0;
	
	localMarket = nil;
	
	docked_shuttles = rand() % 4;   // 0..3;
	last_shuttle_launch_time = 0.0;
	shuttle_launch_interval = 15.0 * 60.0;  // every 15 minutes
	last_shuttle_launch_time = - (rand() % 60) * shuttle_launch_interval / 60.0;
	
	docked_traders = 1 + (rand() % 3);   // 1..3;
	last_trader_launch_time = 0.0;
	trader_launch_interval = ((rand() % 10) + 20.0) * 60.0;  // every few minutes
	last_trader_launch_time = 60.0 - trader_launch_interval; // in one minute's time
	
	return self;
}

- (void) setUpShipFromDictionary:(NSDictionary *) dict
{
	[super setUpShipFromDictionary:dict];
	
	//NSLog(@"DEBUG setting up station '%@' from dict:%@",name,[dict description]);
	
	if ([dict objectForKey:@"equivalent_tech_level"])
		equivalent_tech_level = [(NSNumber *)[dict objectForKey:@"equivalent_tech_level"] intValue];
	else
		equivalent_tech_level = NSNotFound;
	
	if ([dict objectForKey:@"equipment_price_factor"])
		equipment_price_factor = [(NSNumber *)[dict objectForKey:@"equipment_price_factor"] doubleValue];
	else
		equipment_price_factor = 1.0;
}

- (void) dealloc
{
	if (shipsOnApproach)	[shipsOnApproach release];
	if (launchQueue)		[launchQueue release];
	
	if (localMarket)		[localMarket release];
    [super dealloc];
}

- (BOOL) checkCloseCollisionWith:(Entity *)other
{
	//
	//  check if other is within docking corridor
	//
	//NSLog(@"Checking Station CloseContact...");
	//
	//if ([roles isEqual:@"rockhermit"])
	//	return [super checkCloseCollisionWith:other];   // can't dock with rockhermit!
	//
	if ([other isKindOfClass:[ShipEntity class]])
	{
		Vector rel_pos, delta;
		ShipEntity* ship =  (ShipEntity *) other;
		double radius =		[ship collisionRadius];
		//NSLog(@"Checking docking corridor...");
		rel_pos = [ship getPosition];
		rel_pos.x -= position.x;
		rel_pos.y -= position.y;
		rel_pos.z -= position.z;
		delta.x = dot_product(rel_pos, v_right);
		delta.y = dot_product(rel_pos, v_up);
		delta.z = dot_product(rel_pos, v_forward);

		if ((delta.x > boundingBox.max_x + radius)||(delta.x < boundingBox.min_x - radius))
			return NO;
		if ((delta.y > boundingBox.max_y + radius)||(delta.y < boundingBox.min_y - radius))
			return NO;
		if ((delta.z > boundingBox.max_z + radius)||(delta.z < boundingBox.min_z - radius))
			return NO;
		//
		// within bounding box at this point
		//
		//NSLog(@"Docking corridor placement (%3.2f,%3.2f,%3.2f)", delta.x, delta.y, delta.z);
		if  ((delta.x > -32.0)&&(delta.x < 32.0)
			&&(delta.y > -96.0)&&(delta.y < 96.0)
			&&(delta.z > 0.0))
		{
			//NSLog(@"In the lane!");
			if (delta.z + radius < port_radius)
				[ship enterDock:self];
			return NO;
		}
	}
	return YES;
}

- (void) update:(double) delta_t
{
	BOOL isRockHermit = [roles isEqual:@"rockhermit"];
	
	[super update:delta_t];
	if (([launchQueue count] > 0)&&([shipsOnApproach count] == 0)&&([universe getTime] > last_launch_time + STATION_DELAY_BETWEEN_LAUNCHES))
	{
		[self launchShip:(ShipEntity *)[launchQueue objectAtIndex:0]];
		[launchQueue removeObjectAtIndex:0];
	}
	if (([launchQueue count] == 0)&&(no_docking_while_launching))
		no_docking_while_launching = NO;	// launching complete
	if (approach_spacing > 0.0)
	{
		approach_spacing -= delta_t * 5.0;
		if (approach_spacing < 0.0)   approach_spacing = 0.0;
	}
	if ((docked_shuttles > 0)&&(!isRockHermit))
	{
		if ([universe getTime] > last_shuttle_launch_time + shuttle_launch_interval)
		{
			[self launchShuttle];
			docked_shuttles--;
			last_shuttle_launch_time = [universe getTime];
		}
	}
	if ((docked_traders > 0)&&(!isRockHermit))
	{
		if ([universe getTime] > last_trader_launch_time + trader_launch_interval)
		{
			//NSLog(@"O ---> %d docked traders  at %.1f (%.1f and every %.1f)", docked_traders, [universe getTime], last_trader_launch_time + trader_launch_interval, trader_launch_interval);
			[self launchTrader];
			docked_traders--;
			last_trader_launch_time = [universe getTime];
		}
	}
}

- (void) clear
{
	if (launchQueue)
		[launchQueue removeAllObjects];
	if (shipsOnApproach)
		[shipsOnApproach removeAllObjects];
}

- (void) addShipToLaunchQueue:(ShipEntity *) ship
{
	[self sanityCheckShipsOnApproach];
	if (!launchQueue)
		launchQueue = [[NSMutableArray alloc] initWithCapacity:16]; // retained
	if (ship)
		[launchQueue addObject:ship];
}

- (int) countShipsInLaunchQueueWithRole:(NSString *) a_role
{
	if ([launchQueue count] == 0)
		return 0;
	int i;
	int result = 0;
	for (i = 0; i < [launchQueue count]; i++)
	{
		if ([[(ShipEntity *)[launchQueue objectAtIndex:i] roles] isEqual:a_role])
			result++;
	}
	return result;
}

- (void) launchShip:(ShipEntity *) ship
{
	Vector launchPos = position;
	double dist = port_radius;	// + [ship collisionRadius];
	Quaternion q1 = q_rotation;
	quaternion_rotate_about_axis(&q1,v_forward,PI*0.5);  // to account for the slot being at 90 degrees to vertical
	launchPos.x += dist * v_forward.x;
	launchPos.y += dist * v_forward.y;
	launchPos.z += dist * v_forward.z;
    [ship setPosition:launchPos];
	[ship setQRotation:q1];
	[ship setRoll:flight_roll];
	[ship setPitch:0.0];
	[ship setSpeed:40.0];
	[ship setStatus:STATUS_LAUNCHING];
	[[ship getAI] reactToMessage:@"pauseAI: 2.0"]; // pause while launching
	[universe addEntity:ship];
	last_launch_time = [universe getTime];
}

- (void) noteDockedShip:(ShipEntity *) ship
{
	// set last launch time to avoid clashes with outgoing ships
	last_launch_time = [universe getTime];
	if ([[ship roles] isEqual:@"shuttle"])
		docked_shuttles++;
	if ([[ship roles] isEqual:@"trader"])
		docked_traders++;
	if ([[ship roles] isEqual:@"police"])
		police_launched--;
	if ([[ship roles] isEqual:@"hermit-ship"])
		police_launched--;
}


- (BOOL) collideWithShip:(ShipEntity *)other
{
	[self abortAllDockings];
	return [super collideWithShip:other];
}

- (BOOL) beCollidedBy:(ShipEntity *)other
{
	[self abortAllDockings];
	return [super beCollidedBy:other];
}

//////////////////////////////////////////////// extra AI routines


- (void) increaseAlertLevel
{
	switch (alert_level)
	{
		case STATION_ALERT_LEVEL_GREEN :
			alert_level = STATION_ALERT_LEVEL_YELLOW;
			[shipAI reactToMessage:@"YELLOW_ALERT"];
			break;
		case STATION_ALERT_LEVEL_YELLOW :
			alert_level = STATION_ALERT_LEVEL_RED;
			[shipAI reactToMessage:@"RED_ALERT"];
			break;
	}
}

- (void) decreaseAlertLevel
{
	switch (alert_level)
	{
		case STATION_ALERT_LEVEL_RED :
			alert_level = STATION_ALERT_LEVEL_YELLOW;
			[shipAI reactToMessage:@"CONDITION_YELLOW"];
			break;
		case STATION_ALERT_LEVEL_YELLOW :
			alert_level = STATION_ALERT_LEVEL_GREEN;
			[shipAI reactToMessage:@"CONDITION_GREEN"];
			break;
	}
}

- (void) launchPolice
{
	int police_target = primaryTarget;
	int i;
	for (i = 0; (i < 4)&&(police_launched < STATION_MAX_POLICE) ; i++)
	{
		ShipEntity  *police_ship;
		if (![universe entityForUniversalID:police_target])
		{
			[shipAI message:@"TARGET_LOST"];
			return;
		}
			
		//NSLog(@"Launching Police Ship to intercept %@",[universe entityForUniversalID:police_target]);
			
		police_ship = [universe getShipWithRole:@"police"];   // retain count = 1
		[police_ship addTarget:[universe entityForUniversalID:police_target]];
		[police_ship setScanClass: CLASS_POLICE];
		
		//[police_ship setReportAImessages:YES]; // debug
		
		[[police_ship getAI] setStateMachine:@"policeInterceptAI.plist"];
		[self addShipToLaunchQueue:police_ship];
		[police_ship release];
		police_launched++;
	}
	no_docking_while_launching = YES;
	[self abortAllDockings];
}

- (void) launchDefenseShip
{
	int defense_target = primaryTarget;
	//int n_ships = [universe countShipsWithRole:@"hermit-ship"] + [self countShipsInLaunchQueueWithRole:@"hermit-ship"];
	ShipEntity  *defense_ship;
	
	if (police_launched >= 3)   // shuttles are to rockhermits what police ships are to stations
		return;
	
	if (![universe entityForUniversalID:defense_target])
	{
		[shipAI message:@"TARGET_LOST"];
		return;
	}
		
	//NSLog(@"Launching defense Ship to intercept %@",[(ShipEntity *)[universe entityForUniversalID:defense_target] name]);
	
	police_launched++;
	
	defense_ship = [universe getShipWithRole:@"hermit-ship"];   // retain count = 1
	[defense_ship addTarget:[universe entityForUniversalID:defense_target]];
	[defense_ship setScanClass: CLASS_NEUTRAL];
	
	[defense_ship setReportAImessages:YES]; // debug
	
	[[defense_ship getAI] setStateMachine:@"policeInterceptAI.plist"];
	[self addShipToLaunchQueue:defense_ship];
	[defense_ship release];
	no_docking_while_launching = YES;
	[self abortAllDockings];
}

- (void) launchScavenger
{
	ShipEntity  *scavenger_ship;
	
	int		scavs = [universe countShipsWithRole:@"scavenger" inRange:SCANNER_MAX_RANGE ofEntity:self] + [self countShipsInLaunchQueueWithRole:@"scavenger"];
	if (scavs > 3)
		return;
		
	//NSLog(@"Launching Scavenger");
		
	scavenger_ship = [universe getShipWithRole:@"scavenger"];   // retain count = 1
	[scavenger_ship setScanClass: CLASS_NEUTRAL];

	//[scavenger_ship setReportAImessages:YES]; // debug

	[[scavenger_ship getAI] setStateMachine:@"scavengerAI.plist"];
	[self addShipToLaunchQueue:scavenger_ship];
	[scavenger_ship release];
}

- (void) launchShuttle
{
	ShipEntity  *shuttle_ship;
		
	shuttle_ship = [universe getShipWithRole:@"shuttle"];   // retain count = 1
	[shuttle_ship setScanClass: CLASS_NEUTRAL];
	
	[shuttle_ship setCargoFlag:CARGO_FLAG_FULL_SCARCE];

	//[shuttle_ship setReportAImessages:YES]; // debug

	[[shuttle_ship getAI] setStateMachine:@"fallingShuttleAI.plist"];
	[self addShipToLaunchQueue:shuttle_ship];

	//NSLog(@"%@ Prepping shuttle: %@ %d for launch.", [self name], [shuttle_ship name], [shuttle_ship universal_id]);
	
	[shuttle_ship release];
}

- (void) launchTrader
{
	ShipEntity  *trader_ship;
		
	trader_ship = [universe getShipWithRole:@"trader"];   // retain count = 1
	[trader_ship setScanClass: CLASS_NEUTRAL];
	
	[trader_ship setCargoFlag:CARGO_FLAG_FULL_PLENTIFUL];

	[trader_ship setReportAImessages:YES]; // debug

	[[trader_ship getAI] setStateMachine:@"exitingTraderAI.plist"];
	[self addShipToLaunchQueue:trader_ship];

	//NSLog(@"%@ Prepping trader: %@ %d for launch.", [self name], [trader_ship name], [trader_ship universal_id]);
	
	[trader_ship release];
}



@end
