//
//  TextureStore.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <Cocoa/Cocoa.h>
#import <OpenGL/gl.h>

#import "ResourceManager.h"

#import "TextureStore.h"


@implementation TextureStore

- (id) init
{
    self = [super init];
    //
    textureDictionary = [[NSMutableDictionary dictionaryWithCapacity:5] retain];
    //
    return self;
}

- (void) dealloc
{
    if (textureDictionary) [textureDictionary release];
    //
    [super dealloc];
}

- (GLuint) getTextureNameFor:(NSString *)filename
{
    NSBitmapImageRep	*bitmapImageRep;
    NSRect		textureRect = NSMakeRect(0.0,0.0,4.0,4.0);
    NSImage		*texImage, *image;
    NSSize		imageSize;
    NSData		*textureData;
    GLuint		texName;
	
	int			n_planes;
	int			n_bytes;
    
    if (![textureDictionary objectForKey:filename])
    {
        NSMutableDictionary*	texProps = [NSMutableDictionary dictionaryWithCapacity:3];  // autoreleased
        
        texImage = [ResourceManager imageNamed:filename inFolder:@"Textures"];
        if (!texImage)
        {
            NSLog(@"***** Couldn't find texture : %@", filename);
                return 0;
        }
        
        imageSize = [texImage size];
    
        while (textureRect.size.width < imageSize.width)
            textureRect.size.width *= 2.0;
        while (textureRect.size.height < imageSize.height)
            textureRect.size.height *= 2.0;
        
        textureRect.origin= NSMakePoint(0.0,0.0);
    
        //  NSLog(@"textureSize = %f %f",textureRect.size.width,textureRect.size.height);
    
        image = [[NSImage alloc] initWithSize:textureRect.size]; // is retained
        
        // draw the texImage into an image of an appropriate size
        //
        [image lockFocus];
        
		[[NSColor clearColor] set];
        NSRectFill(textureRect);
        
		[texImage drawAtPoint:NSMakePoint(0.0,0.0) fromRect:NSMakeRect(0.0,0.0,imageSize.width,imageSize.height) operation:NSCompositeSourceOver fraction:1.0];
        bitmapImageRep = [[NSBitmapImageRep alloc] initWithFocusedViewRect:textureRect];// is retained
        
		//  NSLog(@"TextureStore %@ texture has %d planes, %d bytes per plane", filename, [bitmapImageRep numberOfPlanes], [bitmapImageRep bytesPerPlane]);
		
		n_bytes = [bitmapImageRep bytesPerPlane];
		n_planes = 3;
		if (n_bytes > textureRect.size.width*textureRect.size.height*3)
			n_planes = 4;
		
		[image unlockFocus];
    
        textureData = [[NSData dataWithBytes:[bitmapImageRep bitmapData] length:(int)(textureRect.size.width*textureRect.size.height*n_planes)] retain];
                
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        glGenTextures(1, &texName);			// get a new unique texture name
        glBindTexture(GL_TEXTURE_2D, texName);	// initialise it
    
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	// adjust this
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	// adjust this
    
        if (n_planes == 4)
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureRect.size.width, textureRect.size.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, [textureData bytes]);
		else
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureRect.size.width, textureRect.size.height, 0, GL_RGB, GL_UNSIGNED_BYTE, [textureData bytes]);
    
        // add to dictionary
        //
        [texProps setObject:textureData forKey:@"textureData"];
        [texProps setObject:[NSNumber numberWithInt:texName] forKey:@"texName"];
        [texProps setObject:[NSNumber numberWithInt:textureRect.size.width] forKey:@"width"];
        [texProps setObject:[NSNumber numberWithInt:textureRect.size.height] forKey:@"height"];

        [textureDictionary setObject:texProps forKey:filename];
        
        [image autorelease]; // is released
        
        [bitmapImageRep autorelease];// is released
        
		[textureData autorelease];// is released (retain count has been incremented by adding it to the texProps dictionary) 
    
    }
    else
    {
        texName = (GLuint)[(NSNumber *)[[textureDictionary objectForKey:filename] objectForKey:@"texName"] intValue];
    }
    return texName;
}

- (NSSize) getSizeOfTexture:(NSString *)filename
{
    NSSize size = NSMakeSize(0.0, 0.0);	// zero size
    if ([textureDictionary objectForKey:filename])
    {
        size.width = [(NSNumber *)[[textureDictionary objectForKey:filename] objectForKey:@"width"] intValue];
        size.height = [(NSNumber *)[[textureDictionary objectForKey:filename] objectForKey:@"height"] intValue];
    }
    return size;
}

- (void) reloadTextures
{
	[textureDictionary removeAllObjects];
	return;
}

@end
