//
//  Universe.h
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#define SCANNER_CENTRE_X	0
#define SCANNER_CENTRE_Y	-180
#define SCANNER_SCALE		256

#define COMPASS_IMAGE			@"compass.png"
#define COMPASS_CENTRE_X		132
#define COMPASS_CENTRE_Y		-216
#define COMPASS_SIZE			64
#define COMPASS_HALF_SIZE		32
#define COMPASS_REDDOT_IMAGE	@"reddot.png"
#define COMPASS_GREENDOT_IMAGE  @"greendot.png"
#define COMPASS_DOT_SIZE		16
#define COMPASS_HALF_DOT_SIZE	8

#define TEXT_MESSAGE_Y			-144

#define AEGIS_CENTRE_X		-132
#define AEGIS_CENTRE_Y		-216

#define SCANNER_IMAGE			@"scanner.png"
#define SCANNER_IMAGE_WIDTH		512
#define SCANNER_IMAGE_HEIGHT	128

#define CROSSHAIR_IMAGE			@"crosshair.png"
#define CROSSHAIR_IMAGE_WIDTH   64
#define CROSSHAIR_IMAGE_HEIGHT	64
#define CROSSHAIR_SIZE			32.0

#define VIEW_FORWARD			0
#define VIEW_AFT				1
#define VIEW_PORT				2
#define VIEW_STARBOARD			3
#define VIEW_NONE				-1
#define VIEW_DOCKED				100
#define VIEW_BREAK_PATTERN		200

#define UNITS_TONS			0
#define UNITS_KILOGRAMS		1
#define UNITS_GRAMS			2

#define DEMO_NO_DEMO		0
#define DEMO_FLY_IN			101
#define DEMO_SHOW_THING		102
#define DEMO_FLY_OUT		103

#define MARKET_NAME					0
#define MARKET_QUANTITY				1
#define MARKET_PRICE				2
#define MARKET_BASE_PRICE			3
#define MARKET_ECO_ADJUST_PRICE		4
#define MARKET_ECO_ADJUST_QUANTITY  5
#define MARKET_BASE_QUANTITY		6
#define MARKET_MASK_PRICE			7
#define MARKET_MASK_QUANTITY		8
#define MARKET_UNITS				9

#define MAX_MESSAGES		5


#import <OpenGL/gl.h>
#import <OpenGL/glu.h>

#import <Cocoa/Cocoa.h>

#import "entities.h"
#import "GuiDisplayGen.h"

@class TextureStore, OpenGLSprite, GameController;

@interface Universe : NSObject
{
		NSView			*gameView;
        TextureStore	*textureStore;
        
		NSSpeechSynthesizer *speechSynthesizer;
		
		NSMutableDictionary *universal_id_dictionary;
		int					next_universal_id;
		
		NSLock				*recycleLock;
		NSMutableDictionary *entityRecyclePool;

		NSMutableDictionary *preloadedDataFiles;

		NSMutableArray	*entities;
        NSMutableArray	*entsInDrawOrder;
				
		int				station;
		int				sun;
		int				planet;
		
        double			messageDisplayDuration;
        double			fpsDisplayDuration;
        NSString		*currentMessage;
		OpenGLSprite	*messageSprite;
		OpenGLSprite	*msSpriteArray[MAX_MESSAGES];
		double			msDuration[MAX_MESSAGES];
        
		GuiDisplayGen   *gui;
		OpenGLSprite	*textDisplaySprite;
		BOOL			displayText;
        
        OpenGLSprite	*fpsSprite;
		BOOL			displayFPS;
		
        OpenGLSprite	*crosshairSprite;
        OpenGLSprite	*scannerSprite;
        OpenGLSprite	*compassSprite;
        OpenGLSprite	*greendotSprite;
        OpenGLSprite	*reddotSprite;
        OpenGLSprite	*legend1Sprite;
        OpenGLSprite	*legend2Sprite;
        OpenGLSprite	*aegisSprite;
		int				viewDirection;
		
		double			universal_time;
		//double			universe_think_time;
		double			ai_think_time;
		
		double			demo_stage_time;
		int				demo_stage;
		int				demo_ship_index;
		NSArray			*demo_ships;
		
		GLfloat			sun_center_position[4];
		
		BOOL			dumpCollisionInfo;
		
		NSDictionary	*shipdata;			// holds data on all ships available, loaded at initialisation
		
		NSDictionary	*commoditylists;	// holds data on commodities for various types of station, loaded at initialisation
		NSArray			*commoditydata;		// holds data on commodities extracted from commoditylists
		
		NSDictionary	*illegal_goods;		// holds the legal penalty for illicit commodities, loaded at initialisation
		NSDictionary	*descriptions;		// holds descriptive text for lots of stuff, loaded at initialisation
		NSDictionary	*missiontext;		// holds descriptive text for missions, loaded at initialisation
		NSArray			*equipmentdata;		// holds data on available equipment, loaded at initialisation
		
		Random_Seed		galaxy_seed;
		Random_Seed		system_seed;
		Random_Seed		target_system_seed;
		
		int				breakPatternCounter;
}

- (id) init;
- (void) dealloc;

- (void) initSprites;

- (void) set_up_universe_from_station;
- (void) set_up_universe_from_witchspace;
- (void) set_up_universe_from_misjump;
- (void) set_up_witchspace;
- (void) set_up_space;
- (void) populateSpaceFromHyperPoint:(Vector) h1_pos toPlanetPosition:(Vector) p1_pos andSunPosition:(Vector) s1_pos;
- (void) addShipWithRole:(NSString *) desc nearRouteOneAt:(double) route_fraction;
- (void) witchspaceShipWithRole:(NSString *) desc;
- (void) set_up_break_pattern:(Vector) pos quaternion:(Quaternion) q;
- (void) game_over;

- (void) set_up_intro1;
- (void) set_up_intro2;

- (StationEntity *) station;
- (PlanetEntity *) planet;
- (PlanetEntity *) sun;

- (BOOL) breakPatternOver;
- (BOOL) breakPatternHide;

- (id) recycleOrDiscard:(Entity *) entity;
- (Entity *) recycledOrNew:(NSString *) classname;

- (NSMutableDictionary *) preloadedDataFiles;

- (ShipEntity *) getShipWithRole:(NSString *) desc;
- (ShipEntity *) getShip:(NSString *) desc;
- (NSDictionary *) getDictionaryForShip:(NSString *) desc;

- (int) getPriceForWeaponSystemWithKey:(NSString *)weapon_key;

- (int) legal_status_of_manifest:(NSArray *)manifest;

- (NSArray *) getContainersOfPlentifulGoods:(int) how_many;
- (NSArray *) getContainersOfScarceGoods:(int) how_many;

- (int) getRandomCommodity;
- (int) getRandomAmountOfCommodity:(int) co_type;

- (int) commodityForName:(NSString *) co_name;
- (NSString *) nameForCommodity:(int) co_type;
- (int) unitsForCommodity:(int) co_type;
- (NSString *) describeCommodity:(int) co_type amount:(int) co_amount;

- (void) setGameView:(NSView *)view;
- (NSView *) gameView;
- (GameController *) gameController;

- (TextureStore *) textureStore;
- (void) resetTextureStore;

- (void) drawFromEntity:(int) n;
- (void) drawScannerFromEntity:(int) n;
- (void) drawDialsForEntity:(int) n;
- (void) drawIndicatorAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz andValue:(double) amount;
- (void) drawBarAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz andValue:(double) amount;
- (void) drawSurroundAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz;
- (void) drawMissileIconAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz;
- (void) drawCrosshairs;
- (void) drawMessage;

- (Entity *) entityZero;

- (Entity *) entityForUniversalID:(int)u_id;
- (BOOL) addEntity:(Entity *) entity;
- (BOOL) removeEntity:(Entity *) entity;
- (void) removeAllEntitiesExceptPlayer:(BOOL) restore;
- (void) removeDemoShips;
- (NSArray *) getAllEntities;

- (BOOL) isVectorClearFromEntity:(Entity *) e1 toDistance:(double)dist fromPoint:(Vector) p2;
- (Vector) getSafeVectorFromEntity:(Entity *) e1 toDistance:(double)dist fromPoint:(Vector) p2;

- (NSArray *) getLaserLineEntitiesForEntity:(Entity *) e1 inView:(int) viewdir;
- (int) getFirstEntityHitByLaserFromEntity:(Entity *) e1;
- (int) getFirstEntityHitByLaserFromEntity:(Entity *) e1 inView:(int) viewdir;
- (int) getFirstEntityTargettedFromEntity:(Entity *) e1 inView:(int) viewdir;

- (NSArray *) getEntitiesWithinRange:(double) range1 ofEntity:(Entity *) e1;
- (int) countShipsWithRole:(NSString *) desc inRange:(double) range1 ofEntity:(Entity *)e1;
- (int) countShipsWithRole:(NSString *) desc;
- (void) sendShipsWithRole:(NSString *) desc messageToAI:(NSString *) ms;

- (double) getTime;

- (void) findCollisions;
- (void) dumpCollisions;

- (void) setViewDirection:(int) vd;
- (int) viewDir;

- (void) displayMessage:(NSString *) text forCount:(int) count;
- (void) addMessage:(NSString *) text forCount:(int) count;

- (void) update:(double) delta_t;

//- (void) think;

///////////////////////////////////////

- (void) setSystemTo:(Random_Seed) s_seed;

- (NSDictionary *) descriptions;
- (NSDictionary *) missiontext;

- (NSDictionary *) generateSystemData:(Random_Seed) system_seed;
- (NSString *) generateSystemName:(Random_Seed) system_seed;
- (NSString *) generateSystemInhabitants:(Random_Seed) s_seed;
- (Random_Seed) findSystemAtCoords:(NSPoint) coords withGalaxySeed:(Random_Seed) gal_seed;
- (int) findSystemNumberAtCoords:(NSPoint) coords withGalaxySeed:(Random_Seed) gal_seed;
- (NSPoint) findSystemCoordinatesWithPrefix:(NSString *) p_fix withGalaxySeed:(Random_Seed) gal_seed;

- (NSArray *) equipmentdata;
- (NSDictionary *) commoditylists;
- (NSArray *) commoditydata;

- (BOOL) generateEconomicDataWithEconomy:(int) economy andRandomFactor:(int) random_factor;
- (NSArray *) commodityDataForEconomy:(int) economy andStation:(StationEntity *)some_station andRandomFactor:(int) random_factor;

- (NSString *) generateSystemDescription:(Random_Seed) s_seed;
- (NSString *) expandDescription:(NSString *) desc forSystem:(Random_Seed)s_seed;
- (NSString *) getRandomDigrams;

- (Vector) getWitchspaceExitPosition;
- (Quaternion) getWitchspaceExitRotation;

///////////////////////////////////////

- (GuiDisplayGen *) gui;
- (void) guiUpdated;
- (void) setDisplayText:(BOOL) value;
- (BOOL) displayText;

- (void) setDisplayFPS:(BOOL) value;
- (BOOL) displayFPS;


@end
