//
//  Universe.m
/*
 *
 *  Oolite
 *
 *  Created by Giles Williams on Sat Apr 03 2004.
 *  Copyright (c) 2004 for aegidian.org. All rights reserved.
 *

Copyright (c) 2004, Giles C Williams
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

•   Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

•   Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

•   Neither the name of aegidian.org nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#import <OpenGL/gl.h>
#import <OpenGL/glu.h>

#import "Universe.h"

#import "entities.h"

#import "MyOpenGLView.h"
#import "ResourceManager.h"
#import "TextureStore.h"
#import "OpenGLSprite.h"
#import "AI.h"

#import "GuiDisplayGen.h"

#define MAX_NUMBER_OF_ENTITIES				200
#define MAX_NUMBER_OF_SOLAR_SYSTEM_ENTITIES 20


@implementation Universe

- (id) init
{	
    PlayerEntity	*player;
	
	self = [super init];
	// set up the universal entity data store
	if (![Entity dataStore])
		[Entity setDataStore:self];
	//
	speechSynthesizer   = [[NSSpeechSynthesizer alloc] initWithVoice:[NSSpeechSynthesizer defaultVoice]];
	//
 	dumpCollisionInfo = NO;
	next_universal_id = 100;	// start arbitrarily above zero
	universal_id_dictionary =   [[NSMutableDictionary dictionaryWithCapacity:MAX_NUMBER_OF_ENTITIES] retain];
	//
	preloadedDataFiles =   [[NSMutableDictionary dictionaryWithCapacity:16] retain];
	//
	entityRecyclePool =			[[NSMutableDictionary dictionaryWithCapacity:MAX_NUMBER_OF_ENTITIES] retain];
	recycleLock =				[[NSLock alloc] init];
	//
    entities =				[[NSMutableArray arrayWithCapacity:MAX_NUMBER_OF_ENTITIES] retain];
    entsInDrawOrder =		[[NSMutableArray arrayWithCapacity:MAX_NUMBER_OF_ENTITIES] retain];
	//
	sun_center_position[0] = 4000000.0;
	sun_center_position[1] = 0.0;
	sun_center_position[2] = 0.0;
	sun_center_position[3] = 1.0;
    //
	textureStore = [[TextureStore alloc] init];	// alloc retains
	//
	scannerSprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:SCANNER_IMAGE inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, SCANNER_IMAGE_WIDTH, SCANNER_IMAGE_HEIGHT)
											size:NSMakeSize(SCANNER_IMAGE_WIDTH, SCANNER_IMAGE_HEIGHT)];	// alloc retains
	compassSprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:COMPASS_IMAGE inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, COMPASS_SIZE, COMPASS_SIZE)
											size:NSMakeSize(COMPASS_HALF_SIZE, COMPASS_HALF_SIZE)];	// alloc retains
	reddotSprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:COMPASS_REDDOT_IMAGE inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, COMPASS_DOT_SIZE, COMPASS_DOT_SIZE)
											size:NSMakeSize(COMPASS_HALF_DOT_SIZE, COMPASS_HALF_DOT_SIZE)];	// alloc retains
	greendotSprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:COMPASS_GREENDOT_IMAGE inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, COMPASS_DOT_SIZE, COMPASS_DOT_SIZE)
											size:NSMakeSize(COMPASS_HALF_DOT_SIZE, COMPASS_HALF_DOT_SIZE)];	// alloc retains
	legend1Sprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:@"legend-left.png" inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, 32, 100)
											size:NSMakeSize(32, 100)];	// alloc retains
	legend2Sprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:@"legend-right.png" inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, 32, 100)
											size:NSMakeSize(32, 100)];	// alloc retains
	aegisSprite = [[OpenGLSprite alloc]   initWithImage:[ResourceManager imageNamed:@"aegis.png" inFolder:@"Images"]
											cropRectangle:NSMakeRect(0, 0, 32, 32)
											size:NSMakeSize(32, 32)];	// alloc retains
	//
    gui = [[GuiDisplayGen alloc] init]; // alloc retains
    displayText = NO;
	//
	displayFPS = NO;
	//
    messageSprite = [[OpenGLSprite alloc] initWithText:@""];	// alloc retains
    messageDisplayDuration = 1.0;
	//
    fpsSprite = [[OpenGLSprite alloc] initWithText:@"FPS ---"];	// alloc retains
    fpsDisplayDuration = 0.0;
	universal_time = 0.0;
	ai_think_time = AI_THINK_INTERVAL;				// one eighth of a second
	//
	shipdata = [[ResourceManager dictionaryFromFilesNamed:@"shipdata.plist" inFolder:@"Config" andMerge:YES] retain];
	//
	commoditylists = [(NSDictionary *)[ResourceManager dictionaryFromFilesNamed:@"commodities.plist" inFolder:@"Config" andMerge:YES] retain];
	commoditydata = [[NSArray arrayWithArray:(NSArray *)[commoditylists objectForKey:@"default"]] retain];
	//
	illegal_goods = [[ResourceManager dictionaryFromFilesNamed:@"illegal_goods.plist" inFolder:@"Config" andMerge:YES] retain];
	//
	descriptions = [[ResourceManager dictionaryFromFilesNamed:@"descriptions.plist" inFolder:@"Config" andMerge:YES] retain];
	//
	missiontext = [[ResourceManager dictionaryFromFilesNamed:@"missiontext.plist" inFolder:@"Config" andMerge:YES] retain];
	//
	equipmentdata = [[ResourceManager arrayFromFilesNamed:@"equipment.plist" inFolder:@"Config" andMerge:YES] retain];
	//
	demo_ships = [[ResourceManager arrayFromFilesNamed:@"demoships.plist" inFolder:@"Config" andMerge:YES] retain];
	demo_ship_index = 0;
	//
	breakPatternCounter = 0;
	//
	station = NO_TARGET;
	planet = NO_TARGET;
	sun = NO_TARGET;
	//
	player = [[PlayerEntity alloc] init];	// alloc retains!
	[self addEntity:player];
	[player set_up];

	[player setStatus:STATUS_DEMO];
	
	galaxy_seed = [player galaxy_seed];
	system_seed = [self findSystemAtCoords:[player galaxy_coordinates] withGalaxySeed:galaxy_seed];
	
	// test
	[self set_up_space];
	
	
	//NSLog(@"Well whaddayaknow - we're at %@", [self generateSystemName:system_seed]);


	[player release];
	//
	[self setViewDirection:VIEW_DOCKED];
	//
	
	//NSLog(@"UNIVERSE INIT station %d, planet %d, sun %d",station,planet,sun);
	
    return self;
}

- (void) dealloc
{
    if (currentMessage)			[currentMessage release];
    if (gui)					[gui release];
    if (messageSprite)			[messageSprite release];
    if (fpsSprite)				[fpsSprite release];
    if (scannerSprite)			[scannerSprite release];
    if (compassSprite)			[compassSprite release];
    if (greendotSprite)			[greendotSprite release];
    if (reddotSprite)			[reddotSprite release];
    if (legend1Sprite)			[legend1Sprite release];
    if (legend2Sprite)			[legend2Sprite release];
    if (aegisSprite)			[aegisSprite release];
    if (textureStore)			[textureStore release];
    if (universal_id_dictionary)	[universal_id_dictionary release];
    if (preloadedDataFiles)			[preloadedDataFiles release];
	
    if (entityRecyclePool)		[entityRecyclePool release];
    if (recycleLock)			[recycleLock release];
	
    if (entities)				[entities release];
    if (entsInDrawOrder)		[entsInDrawOrder release];
    if (shipdata)				[shipdata release];
	
    if (commoditylists)			[commoditylists release];
    if (commoditydata)			[commoditydata release];
	
    if (illegal_goods)			[illegal_goods release];
    if (descriptions)			[descriptions release];
    if (missiontext)			[missiontext release];
	if (equipmentdata)			[equipmentdata release];
    if (demo_ships)				[demo_ships release];
    if (gameView)				[gameView release];
	
	if (speechSynthesizer)		[speechSynthesizer release];

    [super dealloc];
}

- (void) initSprites
{
    if (scannerSprite)			[scannerSprite  autorelease];
    if (compassSprite)			[compassSprite  autorelease];
    if (greendotSprite)			[greendotSprite autorelease];
    if (reddotSprite)			[reddotSprite   autorelease];
    if (legend1Sprite)			[legend1Sprite  autorelease];
    if (legend2Sprite)			[legend2Sprite  autorelease];
    if (aegisSprite)			[aegisSprite	autorelease];
	//
	scannerSprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:SCANNER_IMAGE]
											cropRectangle:NSMakeRect(0, 0, SCANNER_IMAGE_WIDTH, SCANNER_IMAGE_HEIGHT)
											size:NSMakeSize(SCANNER_IMAGE_WIDTH, SCANNER_IMAGE_HEIGHT)];	// alloc retains
	compassSprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:COMPASS_IMAGE]
											cropRectangle:NSMakeRect(0, 0, COMPASS_SIZE, COMPASS_SIZE)
											size:NSMakeSize(COMPASS_HALF_SIZE, COMPASS_HALF_SIZE)];	// alloc retains
	reddotSprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:COMPASS_REDDOT_IMAGE]
											cropRectangle:NSMakeRect(0, 0, COMPASS_DOT_SIZE, COMPASS_DOT_SIZE)
											size:NSMakeSize(COMPASS_HALF_DOT_SIZE, COMPASS_HALF_DOT_SIZE)];	// alloc retains
	greendotSprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:COMPASS_GREENDOT_IMAGE]
											cropRectangle:NSMakeRect(0, 0, COMPASS_DOT_SIZE, COMPASS_DOT_SIZE)
											size:NSMakeSize(COMPASS_HALF_DOT_SIZE, COMPASS_HALF_DOT_SIZE)];	// alloc retains
	legend1Sprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:@"legend-left.png"]
											cropRectangle:NSMakeRect(0, 0, 32, 100)
											size:NSMakeSize(32, 100)];	// alloc retains
	legend2Sprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:@"legend-right.png"]
											cropRectangle:NSMakeRect(0, 0, 32, 100)
											size:NSMakeSize(32, 100)];	// alloc retains
	aegisSprite = [[OpenGLSprite alloc]   initWithImage:[NSImage imageNamed:@"aegis.png"]
											cropRectangle:NSMakeRect(0, 0, 32, 32)
											size:NSMakeSize(32, 32)];	// alloc retains
}

- (void) set_up_universe_from_station
{
	//NSLog(@"UNIVERSE set_up_universe_from_station station %d, planet %d, sun %d",station,planet,sun);
	if (station == NO_TARGET)
	{
		//NSLog(@"----> FIRST LAUNCH!");
		[self set_up_space];	// first launch
	}
	station = [[self station] universal_id];
	planet = [[self planet] universal_id];
	sun = [[self sun] universal_id];
	
	[self setViewDirection:VIEW_FORWARD];
	displayText = NO;
}

- (void) set_up_universe_from_witchspace
{
    PlayerEntity		*player;

    //
	// check the player is still around!
    //
	if ([entities count] == 0)
	{
		/*- the player ship -*/
		player = [[PlayerEntity alloc] init];	// alloc retains!
		//
		[self addEntity:player];
		//
		/*--*/
	}
	else
	{
		player = [(PlayerEntity *)[self entityZero] retain];	// retained here
	}
	//

	[self set_up_space];
	
	[player leaveWitchspace];
	[player release];											// released here
	
	[self setViewDirection:VIEW_FORWARD];
	//
    //
	/* test stuff */
	displayText = NO;
	/* ends */
}

- (void) set_up_universe_from_misjump
{
    PlayerEntity		*player;

    //
	// check the player is still around!
    //
	if ([entities count] == 0)
	{
		/*- the player ship -*/
		player = [[PlayerEntity alloc] init];	// alloc retains!
		//
		[self addEntity:player];
		//
		/*--*/
	}
	else
	{
		player = [(PlayerEntity *)[self entityZero] retain];	// retained here
	}
	//

	[self set_up_witchspace];
	
	[player leaveWitchspace];
	[player release];											// released here
	
	[self setViewDirection:VIEW_FORWARD];
	//
    //
	/* test stuff */
	displayText = NO;
	/* ends */
}

- (void) set_up_witchspace
{
	// new system is hyper-centric : witchspace exit point is origin

    Entity				*thing;
	
	Quaternion			randomQ;
	
	//
	// fixed entities (part of the graphics system really) come first...
	//
	
	/*- the sky backdrop -*/
	thing = [[SkyEntity alloc] initAsWitchspace];	// alloc retains!
	[thing setScanClass: CLASS_NO_DRAW];
	quaternion_set_random(&randomQ);
	[thing setQRotation:randomQ];
	[self addEntity:thing]; // [entities addObject:thing];
	[thing release];
	/*--*/
	
	/*- the dust particle system -*/
	thing = [[DustEntity alloc] init];	// alloc retains!
	[thing setScanClass: CLASS_NO_DRAW];
	[self addEntity:thing]; // [entities addObject:thing];
	[thing release];
	/*--*/
	
	sun = NO_TARGET;
	station = NO_TARGET;
	planet = NO_TARGET;
	sun_center_position[0] = 0.0;
	sun_center_position[1] = 0.0;
	sun_center_position[2] = 0.0;
	sun_center_position[3] = 1.0;
	
	srand([[NSDate date] timeIntervalSince1970]);   // reset randomiser with current time
	
	//
	// actual thargoids and tharglets next...
	//
	int n_thargs = 1 + (rand() % 4);
	if (n_thargs < 1)
		n_thargs = 1;   // just to be sure
	int i;
	for (i = 0; i < n_thargs; i++)
	{
		ShipEntity  *thargoid = [self getShipWithRole:@"thargoid"]; // is retained
		Vector		tharg_pos = [self getWitchspaceExitPosition];
		tharg_pos.x *= 2.0;	tharg_pos.y *= 2.0;	tharg_pos.z *= 2.0;
		[thargoid setPosition:[self getWitchspaceExitPosition]];
		[thargoid setQRotation:[self getWitchspaceExitRotation]];
		[thargoid setScanClass: CLASS_THARGOID];
		[thargoid setBounty:100];
		[thargoid setStatus:STATUS_IN_FLIGHT];
		[thargoid setReportAImessages:YES];
		[self addEntity:thargoid];
		[thargoid release];
	}
	
}

- (void) set_up_space
{
	// new system is hyper-centric : witchspace exit point is origin

    Entity				*thing;
    ShipEntity			*nav_buoy;
    StationEntity		*a_station;
    PlanetEntity		*a_sun;
    PlanetEntity		*a_planet;
	
	Vector				stationPos;
	double				stationRoll;
	
	Quaternion			randomQ;

	Vector				vf;

	NSDictionary		*systeminfo = [self generateSystemData:system_seed];
	int					techlevel = [(NSNumber *)[systeminfo objectForKey:@"techlevel"] intValue];
	NSString			*stationDesc;
	NSColor				*bgcolor;
	
	// set the system seed for random number generation
	seed_for_planet_description(system_seed);
	srand(gen_rnd_number()+gen_rnd_number()+gen_rnd_number());  // seed the standard random number generator
	
	//
	// fixed entities (part of the graphics system really) come first...
	//
	
	/*- the sky backdrop -*/
	thing = [[SkyEntity alloc] init];	// alloc retains!
	[thing setScanClass: CLASS_NO_DRAW];
	quaternion_set_random(&randomQ);
	[thing setQRotation:randomQ];
	[self addEntity:thing]; // [entities addObject:thing];
	bgcolor = [(SkyEntity *)thing sky_color];
	[thing release];
	/*--*/
	
	/*- the dust particle system -*/
	thing = [[DustEntity alloc] init];	// alloc retains!
	[thing setScanClass: CLASS_NO_DRAW];
	[self addEntity:thing]; // [entities addObject:thing];
	[(DustEntity *)thing setDustColor:[bgcolor blendedColorWithFraction:0.5 ofColor:[NSColor whiteColor]]]; 
	[thing release];
	/*--*/
	
	//
	// actual entities next...
	//
	
	// set the system seed for random number generation
	seed_for_planet_description(system_seed);
	srand(gen_rnd_number()+gen_rnd_number()+gen_rnd_number());  // seed the standard random number generator
	
	/*- space planet -*/
	a_planet = [[PlanetEntity alloc] initWithSeed: system_seed fromUniverse: self];	// alloc retains!
	double planet_radius = [a_planet getRadius];
	
	[a_planet setPlanetType:PLANET_TYPE_GREEN];
	[a_planet setStatus:STATUS_TEST];
	[a_planet setPosition:0.0:0.0:12.0 * planet_radius]; // 12 pr (planet radii)  ahead
	[a_planet setScanClass: CLASS_NO_DRAW];
	[a_planet setEnergy:  1000000.0];
	[self addEntity:a_planet]; // [entities addObject:a_planet];
	planet = [a_planet universal_id];
	/*--*/

	/*- space sun -*/
	double		sun_distance = 20.0 * planet_radius;
	double		sun_radius = 2.5 * planet_radius;
	Quaternion  q_sun;
	Vector		sunPos = [a_planet getPosition];
	quaternion_set_random(&q_sun);
	// set up planet's direction in space so it gets a proper day
	[a_planet setQRotation:q_sun];
	
	vf = vector_right_from_quaternion(q_sun);
	sunPos.x -= sun_distance * vf.x;	// back off from the planet by 20 pr
	sunPos.y -= sun_distance * vf.y;
	sunPos.z -= sun_distance * vf.z;
	
	a_sun = [[PlanetEntity alloc] init];	// alloc retains!
	[a_sun setPlanetType:PLANET_TYPE_SUN];
	[a_sun setStatus:STATUS_TEST];
	[a_sun setPosition:sunPos];
	sun_center_position[0] = sunPos.x;
	sun_center_position[1] = sunPos.y;
	sun_center_position[2] = sunPos.z;
	sun_center_position[3] = 1.0;
	[a_sun setRadius:sun_radius];			// 2.5 pr
	[a_sun setScanClass: CLASS_NO_DRAW];
	[a_sun setEnergy:  1000000.0];
	[self addEntity:a_sun];					// [entities addObject:a_sun];
	sun = [a_sun universal_id];
	/*--*/
	
	/*- space station -*/
	stationPos = [a_planet getPosition];
	double  station_orbit = 2.0 * planet_radius;
	Quaternion  q_station;
	quaternion_set_random(&q_station);
	vf = vector_forward_from_quaternion(q_station);
	stationPos.x -= station_orbit * vf.x;					// back away from the planet
	stationPos.y -= station_orbit * vf.y;
	stationPos.z -= station_orbit * vf.z;
	stationRoll =   0.4;
	stationDesc = @"coriolis";
	if (techlevel > 10)
	{
		if (system_seed.f & 0x03)   // 3 out of 4 get this type
			stationDesc = @"dodecahedron";
		else
			stationDesc = @"icosahedron";
	}
	//NSLog(@"* INFO *\t>>\tAdding %@ station for TL %d", stationDesc, techlevel);
	a_station = (StationEntity *)[self getShipWithRole:stationDesc];			   // retain count = 1
	[a_station setStatus:STATUS_TEST];
	[a_station setQRotation: q_station];
	[a_station setPosition:stationPos];
	[a_station setRoll: stationRoll ];
	[a_station setPitch: 0.0];
	[a_station setScanClass: CLASS_STATION];
	[a_station setPlanet:(PlanetEntity *)[self entityForUniversalID:planet]];
	[self addEntity:a_station];
	station = [a_station universal_id];
	
	/*- nav beacon -*/
	double buoy_distance = 10000.0;					// distance from station entrance
	nav_buoy = [self getShipWithRole:@"buoy"];  // retain count = 1
	[nav_buoy setStatus:STATUS_TEST];
	[nav_buoy setRoll: 0.2];
	[nav_buoy setPitch: 0.1];
	[nav_buoy setPosition:stationPos.x+buoy_distance*vf.x:stationPos.y+buoy_distance*vf.y:stationPos.z+buoy_distance*vf.z];
	[nav_buoy setScanClass: CLASS_BUOY];
	[self addEntity:nav_buoy]; // [entities addObject:nav_buoy];
	/*--*/
	
	srand([[NSDate date] timeIntervalSince1970]);   // reset randomiser with current time
	[self populateSpaceFromHyperPoint:[self getWitchspaceExitPosition] toPlanetPosition:[a_planet getPosition] andSunPosition: [a_sun getPosition]];
	
	[a_sun release];
	[nav_buoy release];
	[a_station release];
	[a_planet release];
	
	//station = [[self station] universal_id];
	//planet = [[self planet] universal_id];
	//sun = [[self sun] universal_id];
	//NSLog(@"UNIVERSE SET_UP_SPACE station %d, planet %d, sun %d",station,planet,sun);
}

- (void) populateSpaceFromHyperPoint:(Vector) h1_pos toPlanetPosition:(Vector) p1_pos andSunPosition:(Vector) s1_pos
{
	int i, r;
	NSDictionary		*systeminfo = [self generateSystemData:system_seed];
	int government =	[(NSNumber *)[systeminfo objectForKey:@"government"] intValue]; // 0 .. 7 (0 anarchic .. 7 most stable)
	int economy =		[(NSNumber *)[systeminfo objectForKey:@"economy"] intValue];	// 0 .. 7 (0 richest .. 7 poorest)
	int thargoidChance = (system_seed.e < 127) ? 10 : 3; // if Human Colonials live here, there's a greater % chance the Thargoids will attack!
	Vector  lastPiratePosition;
	int		wolfPackCounter = 0;
	
	NSLog(@"Populating a system with economy %d, and government %d", economy, government);

	// traders
	int trading_parties = (9 - economy);			// 2 .. 9
	if (government == 0) trading_parties *= 1.25;	// 25% more trade where there are no laws!
	if (trading_parties > 0)
		trading_parties = 1 + trading_parties * (randf()+randf());   // randomize 0..2
	while (trading_parties > 15)
		trading_parties = 1 + (rand() % trading_parties);   // reduce
	
	NSLog(@"... adding %d trading vessels", trading_parties);
	
	// pirates
	int raiding_parties = (8 - government) * trading_parties / 3;
	if (raiding_parties > 0)
		raiding_parties =  raiding_parties * (randf()+randf());   // randomize
	while (raiding_parties > 21)
		raiding_parties = 7 + (rand() % raiding_parties);   // reduce
	
	NSLog(@"... adding %d pirate vessels", raiding_parties);

	// bounty-hunters and the law
	int hunting_parties = (1 + government) * trading_parties / 8;
	if (government == 0) hunting_parties *= 1.25;   // 25% more bounty hunters in an anarchy
	if (hunting_parties > 0)
		hunting_parties = hunting_parties * (randf()+randf());   // randomize
	while (hunting_parties > 15)
		hunting_parties = 5 + (rand() % hunting_parties);   // reduce
	
	//debug
	if (hunting_parties < 1)
		hunting_parties = 1;
	
	NSLog(@"... adding %d law/bounty-hunter vessels", hunting_parties);

	int thargoid_parties = 0;
	while ((rand() % 100) < thargoidChance)
		thargoid_parties++;

	NSLog(@"... adding %d Thargoid warships", thargoid_parties);
	
	int rock_clusters = rand() % 3;
	if (trading_parties + raiding_parties + hunting_parties < 10)
		rock_clusters += 1 + (rand() % 3);

	NSLog(@"... adding %d asteroid clusters", rock_clusters);

	int total_clicks = trading_parties + raiding_parties + hunting_parties + thargoid_parties + rock_clusters;
	
	NSLog(@"... for a total of %d ships", total_clicks);
	
	Vector  v_route1 = p1_pos;
	v_route1.x -= h1_pos.x;	v_route1.y -= h1_pos.y;	v_route1.z -= h1_pos.z;
	double d_route1 = sqrt(v_route1.x*v_route1.x + v_route1.y*v_route1.y + v_route1.z*v_route1.z) - 60000.0; // -60km to avoid planet
	v_route1 = unit_vector(&v_route1);
	
	// add the traders to route1 (witchspace exit to space-station / planet)
	for (i = 0; i < trading_parties; i++)
	{
		ShipEntity  *trader_ship;
		Vector		launch_pos = h1_pos;
		if (total_clicks < 3)   total_clicks = 3;
		r = 2 + (rand() % (total_clicks - 2));  // find an empty slot
		double ship_location = d_route1 * r / total_clicks;
		launch_pos.x += ship_location * v_route1.x + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		launch_pos.y += ship_location * v_route1.y + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		launch_pos.z += ship_location * v_route1.z + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		trader_ship = [self getShipWithRole:@"trader"];   // retain count = 1
		[trader_ship setScanClass: CLASS_NEUTRAL];
		[trader_ship setPosition:launch_pos];
		[trader_ship setBounty:0];
		[trader_ship setCargoFlag:CARGO_FLAG_FULL_SCARCE];
		[trader_ship setStatus:STATUS_IN_FLIGHT];
		
		//[trader_ship setReportAImessages: (i == 0) ? YES:NO ]; // debug

		[self addEntity:trader_ship];
		[[trader_ship getAI] setStateMachine:@"route1traderAI.plist"];	// must happen after adding to the universe!
		[trader_ship release];
	}
	
	// add the raiders to route1 (witchspace exit to space-station / planet)
	for (i = 0; i < raiding_parties; i++)
	{
		ShipEntity  *pirate_ship;
		Vector		launch_pos = h1_pos;
		if ((i > 0)&&((rand() % 5) > wolfPackCounter))
		{
			// use last position
			launch_pos = lastPiratePosition;
			launch_pos.x += SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5)*0.1; // pack them closer together
			launch_pos.y += SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5)*0.1;
			launch_pos.z += SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5)*0.1;
			wolfPackCounter++;
		}
		else
		{
			// random position along route1
			if (total_clicks < 3)   total_clicks = 3;
			r = 2 + (rand() % (total_clicks - 2));  // find an empty slot
			double ship_location = d_route1 * r / total_clicks;
			launch_pos.x += ship_location * v_route1.x + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
			launch_pos.y += ship_location * v_route1.y + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
			launch_pos.z += ship_location * v_route1.z + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
			lastPiratePosition = launch_pos;
			wolfPackCounter = 0;
		}
		pirate_ship = [self getShipWithRole:@"pirate"];   // retain count = 1
		[pirate_ship setScanClass: CLASS_NEUTRAL];
		[pirate_ship setPosition:launch_pos];
		[pirate_ship setStatus:STATUS_IN_FLIGHT];
		[pirate_ship setBounty:20 + government + wolfPackCounter + (rand() % 5)];
		
		//[pirate_ship setReportAImessages: (i == 0) ? YES:NO ]; // debug

		[self addEntity:pirate_ship];
		[[pirate_ship getAI] setStateMachine:@"pirateAI.plist"];	// must happen after adding to the universe!
		[pirate_ship release];
	}
	
	// add the hunters and police ships to route1 (witchspace exit to space-station / planet)
	for (i = 0; i < hunting_parties; i++)
	{
		ShipEntity  *hunter_ship;
		Vector		launch_pos = h1_pos;
		// random position along route1
		if (total_clicks < 3)   total_clicks = 3;
		r = 2 + (rand() % (total_clicks - 2));  // find an empty slot
		double ship_location = d_route1 * r / total_clicks;
		launch_pos.x += ship_location * v_route1.x + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		launch_pos.y += ship_location * v_route1.y + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		launch_pos.z += ship_location * v_route1.z + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		if ((rand() % 8) < government)
		{
			hunter_ship = [self getShipWithRole:@"police"];   // retain count = 1
			[hunter_ship setScanClass: CLASS_POLICE];
		}
		else
		{
			hunter_ship = [self getShipWithRole:@"hunter"];   // retain count = 1
			[hunter_ship setScanClass: CLASS_NEUTRAL];
		}
		[hunter_ship setPosition:launch_pos];
		[hunter_ship setStatus:STATUS_IN_FLIGHT];
		[hunter_ship setBounty:0];
		
		//[hunter_ship setReportAImessages: (i == 0) ? YES:NO ]; // debug

		[self addEntity:hunter_ship];
		[[hunter_ship getAI] setStateMachine:@"route1patrolAI.plist"];	// must happen after adding to the universe!
		[hunter_ship release];
	}
	
	// add the thargoids to route1 (witchspace exit to space-station / planet) clustered together
	if (total_clicks < 3)   total_clicks = 3;
	r = 2 + (rand() % (total_clicks - 2));  // find an empty slot
	double thargoid_location = d_route1 * r / total_clicks;
	for (i = 0; i < thargoid_parties; i++)
	{
		ShipEntity  *thargoid_ship;
		Vector		launch_pos;
		launch_pos.x = h1_pos.x + thargoid_location * v_route1.x + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		launch_pos.y = h1_pos.y + thargoid_location * v_route1.y + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		launch_pos.z = h1_pos.z + thargoid_location * v_route1.z + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
		thargoid_ship = [self getShipWithRole:@"thargoid"];   // retain count = 1
		[thargoid_ship setScanClass: CLASS_THARGOID];
		[thargoid_ship setPosition:launch_pos];
		[thargoid_ship setBounty:100];
		[thargoid_ship setStatus:STATUS_IN_FLIGHT];
		[self addEntity:thargoid_ship];
		[[thargoid_ship getAI] setState:@"GLOBAL"];
		[thargoid_ship release];
	}
	
	// add the asteroids to route1 (witchspace exit to space-station / planet) clustered together in a preset location.
	// set the system seed for random number generation
	int total_rocks = 0;
	seed_for_planet_description(system_seed);
	srand(gen_rnd_number()+gen_rnd_number()+gen_rnd_number());  // seed the standard random number generator
	if (total_clicks < 3)   total_clicks = 3;
	for (i = 0; i < rock_clusters; i++)
	{
		int j;
		int cluster_size = 1 + (rand() % 6) + (rand() % 6);
		r = 2 + (rand() % (total_clicks - 2));  // find an empty slot
		double asteroid_location = d_route1 * r / total_clicks;
		for (j = 0; j < cluster_size; j++)
		{
			ShipEntity  *asteroid;
			Vector		launch_pos;
			launch_pos.x = h1_pos.x + asteroid_location * v_route1.x + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
			launch_pos.y = h1_pos.y + asteroid_location * v_route1.y + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
			launch_pos.z = h1_pos.z + asteroid_location * v_route1.z + SCANNER_MAX_RANGE*((rand() % 256)/256.0 - 0.5);
			asteroid = [self getShipWithRole:@"asteroid"];   // retain count = 1
			[asteroid setScanClass: CLASS_CARGO];
			[asteroid setPosition:launch_pos];
			[asteroid setBounty:10];
			[asteroid setStatus:STATUS_IN_FLIGHT];
			[self addEntity:asteroid];
			[[asteroid getAI] setState:@"GLOBAL"];
			[asteroid release];
			total_rocks++;
		}
		// rock-hermit : chance is related to the number of asteroids
		// hermits are placed near to other asteroids for obvious reasons
		if ((rand() % 31) <= cluster_size)
		{
			//debug
			NSLog(@"DEBUG ... adding rock-hermit");
			StationEntity  *hermit;
			Vector		launch_pos;
			launch_pos.x = h1_pos.x + asteroid_location * v_route1.x + 0.5 * SCANNER_MAX_RANGE * ((rand() % 256)/256.0 - 0.5);
			launch_pos.y = h1_pos.y + asteroid_location * v_route1.y + 0.5 * SCANNER_MAX_RANGE * ((rand() % 256)/256.0 - 0.5);
			launch_pos.z = h1_pos.z + asteroid_location * v_route1.z + 0.5 * SCANNER_MAX_RANGE * ((rand() % 256)/256.0 - 0.5);
			hermit = (StationEntity *)[self getShipWithRole:@"rockhermit"];   // retain count = 1
			[hermit setScanClass: CLASS_CARGO];
			[hermit setPosition:launch_pos];
			[hermit setStatus:STATUS_IN_FLIGHT];
			[self addEntity:hermit];
			[[hermit getAI] setState:@"GLOBAL"];
			[hermit release];
		}
	}
	
		
}

- (void) addShipWithRole:(NSString *) desc nearRouteOneAt:(double) route_fraction
{
	// adds a ship within scanner range of a point on route 1
	//
	Vector  h1_pos = [self getWitchspaceExitPosition];
	Vector  launch_pos = [[self station] getPosition];
	//NSLog(@"h1( %.1f, %.1f, %.1f) launch_pos( %.1f, %.1f, %.1f)", h1_pos.x, h1_pos.y, h1_pos.z, launch_pos.x, launch_pos.y, launch_pos.z);
	launch_pos.x -= h1_pos.x;		launch_pos.y -= h1_pos.y;		launch_pos.z -= h1_pos.z;
	//NSLog(@"launch_pos - h1_pos =( %.1f, %.1f, %.1f)", launch_pos.x, launch_pos.y, launch_pos.z);
	launch_pos.x *= route_fraction; launch_pos.y *= route_fraction; launch_pos.z *= route_fraction;
	//NSLog(@"*route_fraction <%.3f> =( %.1f, %.1f, %.1f)", route_fraction, launch_pos.x, launch_pos.y, launch_pos.z);
	launch_pos.x += h1_pos.x;		launch_pos.y += h1_pos.y;		launch_pos.z += h1_pos.z;
	//NSLog(@"+ h1_pos =( %.1f, %.1f, %.1f)", launch_pos.x, launch_pos.y, launch_pos.z);
	launch_pos.x += SCANNER_MAX_RANGE*(randf() - randf());
	launch_pos.y += SCANNER_MAX_RANGE*(randf() - randf());
	launch_pos.z += SCANNER_MAX_RANGE*(randf() - randf());
	//
	ShipEntity  *ship;
	ship = [self getShipWithRole:desc];   // retain count = 1
	[ship setPosition:launch_pos];
	[self addEntity:ship];
	[[ship getAI] setState:@"GLOBAL"];	// must happen after adding to the universe!
	[ship release];
	//
}

- (void) witchspaceShipWithRole:(NSString *) desc
{
	// adds a ship exiting witchspace (corollary of when ships leave the system)
	ShipEntity  *ship;
	ship = [self getShipWithRole:desc];   // retain count = 1
	//[ship setScanClass: CLASS_NEUTRAL];
	if ([desc isEqual:@"trader"])
	{
		[ship setCargoFlag: CARGO_FLAG_FULL_SCARCE];
		[[ship getAI] setStateMachine:@"route1traderAI.plist"];
	}
	[ship setUniverse:self];
	[ship leaveWitchspace];				// gets added to the universe here!
	[[ship getAI] setState:@"GLOBAL"];	// must happen after adding to the universe!
	[ship release];
}

- (void) set_up_break_pattern:(Vector) pos quaternion:(Quaternion) q
{
	int				i;
	RingEntity*		ring;
	
	q.w = -q.w;		// reverse the quaternion because this is from the player's viewpoint
	
	Vector			v = vector_forward_from_quaternion(q);
		
	for (i = 1; i < 11; i++)
	{
		ring = (RingEntity *)[self recycledOrNew:@"RingEntity"];	// alloc retains!
		[ring setPosition:pos.x+v.x*i*50.0:pos.y+v.y*i*50.0:pos.z+v.z*i*50.0]; // ahead of the player
		[ring setQRotation:q];
		[ring setVelocity:v];
		[ring setLifetime:i*50.0];
		[ring setScanClass: CLASS_NO_DRAW];
		[self addEntity:ring]; // [entities addObject:ring];
		breakPatternCounter++;
		[ring release];
    }
}

- (void) game_over
{
	PlayerEntity*   player = (PlayerEntity *)[[self entityZero] retain];
	//
	[self removeAllEntitiesExceptPlayer:NO];	// don't want to restore afterwards
	//
	[player set_up];						//reset the player
	//
	[player setStatus:STATUS_DOCKED];
	[player setGuiToStatusScreen];
	[self setViewDirection:VIEW_DOCKED];
	displayText = YES;
	//
	[player release];    
	//
}

- (void) set_up_intro1
{
	ShipEntity		*ship;
	Quaternion		q2;
	q2.x = 0.0;   q2.y = 0.0;   q2.z = 0.0; q2.w = 1.0;
	quaternion_rotate_about_y(&q2,PI);
	//
	// in status demo : draw ships and display text
	//
	//[self removeAllEntitiesExceptPlayer:NO];
	[[self entityZero] setStatus:STATUS_DEMO];
	displayText = YES;
	//
	/*- cobra -*/
	ship = [self getShip:PLAYER_SHIP_DESC];   // retain count = 1
	[ship setStatus:STATUS_DEMO];
	[ship setQRotation:q2];
	[ship setPosition:0.0:0.0: 3.6 * [ship collisionRadius]];  // 250m ahead
	
	//NSLog(@"demo ship %@ has collision radius %.1f 250.0/cr = %.1f", [ship name], [ship collisionRadius], 250.0/[ship collisionRadius]);
	
	[ship setScanClass: CLASS_NO_DRAW];
	[ship setRoll:PI/5.0];
	[ship setPitch:PI/10.0];
	[[ship getAI] setStateMachine:@"nullAI.plist"];
	[self addEntity:ship];
	[ship release];
	//
	[self setViewDirection:VIEW_DOCKED];
	displayText = YES;
	//
	//
}

- (void) set_up_intro2
{
	ShipEntity		*ship;
	Quaternion		q2;
	q2.x = 0.0;   q2.y = 0.0;   q2.z = 0.0; q2.w = 1.0;
	quaternion_rotate_about_y(&q2,PI);
	//
	// in status demo draw ships and display text
	//
	//[self removeAllEntitiesExceptPlayer:NO];
	[self removeDemoShips];
	[[self entityZero] setStatus:STATUS_DEMO];
	displayText = YES;
	//
	/*- demo ships -*/
	demo_ship_index = 0;
	ship = [self getShip:[demo_ships objectAtIndex:0]];   // retain count = 1
	[ship setStatus:STATUS_DEMO];
	[ship setQRotation:q2];
	[ship setPosition:0.0:0.0: 3.6 * [ship collisionRadius]];
	
	//NSLog(@"demo ship %@ has collision radius %.1f 250.0/cr = %.1f", [ship name], [ship collisionRadius], 250.0/[ship collisionRadius]);
	
	[ship setScanClass: CLASS_NO_DRAW];
	[ship setRoll:PI/5.0];
	[ship setPitch:PI/10.0];
	[[ship getAI] setStateMachine:@"nullAI.plist"];
	[self addEntity:ship];
	
	[gui setText:[ship name] forRow:19 align:GUI_ALIGN_CENTER];
	[gui setColor:[NSColor whiteColor] forRow:19];
	[gui generateDisplay];
	[self guiUpdated];
	
	[ship release];
	//
	[self setViewDirection:VIEW_DOCKED];
	displayText = YES;
	//
	demo_stage = DEMO_SHOW_THING;
	demo_stage_time = universal_time + 3.0;
	//
}

- (StationEntity *) station
{
	if (![self entityForUniversalID:station])
	{
		int i;
		station = NO_TARGET;
		NSArray* entList = [NSArray arrayWithArray:entities];
		for (i = 0; ((i < [entList count])&&(station == NO_TARGET)) ; i++)
		{
			Entity* thing = (Entity *)[entList objectAtIndex:i];
			if (([thing scanClass] == CLASS_STATION)&&([thing isKindOfClass:[StationEntity class]]))
				station = [thing universal_id];
		}
	}
	
	return (StationEntity *)[self entityForUniversalID:station];
}

- (PlanetEntity *) planet
{
	if (![self entityForUniversalID:planet])
	{
		int i;
		planet = NO_TARGET;
		NSArray* entList = [NSArray arrayWithArray:entities];
		for (i = 0; ((i < [entList count])&&(planet == NO_TARGET)) ; i++)
		{
			Entity* thing = (Entity *)[entList objectAtIndex:i];
			if ([thing isKindOfClass:[PlanetEntity class]])
			{
				if ([(PlanetEntity *)thing getPlanetType] == PLANET_TYPE_GREEN)
					planet = [thing universal_id];
			}
		}
	}
	return (PlanetEntity *)[self entityForUniversalID:planet];
}

- (PlanetEntity *) sun
{
	if (![self entityForUniversalID:sun])
	{
		int i;
		sun = NO_TARGET;
		NSArray* entList = [NSArray arrayWithArray:entities];
		for (i = 0; ((i < [entList count])&&(sun == NO_TARGET)) ; i++)
		{
			Entity* thing = (Entity *)[entList objectAtIndex:i];
			if ([thing isKindOfClass:[PlanetEntity class]])
			{
				if ([(PlanetEntity *)thing getPlanetType] == PLANET_TYPE_SUN)
					sun = [thing universal_id];
			}
		}
	}
	return (PlanetEntity *)[self entityForUniversalID:sun];
}

- (BOOL) breakPatternOver
{
	return (breakPatternCounter == 0);
}

- (BOOL) breakPatternHide
{
	return ((breakPatternCounter > 5)||([[self entityZero] getStatus] == STATUS_DOCKING));
}

- (id) recycleOrDiscard:(Entity *) entity
{
	NSMutableArray  *entlist;
	NSString		*classname = [entity className];
	
	if ((![classname isEqual:@"ShipEntity"])&&(![classname isEqual:@"StationEntity"])&&(![classname isEqual:@"RingEntity"]))
		return entity;
	
	if ([entity getStatus] == STATUS_IN_HOLD)
		return entity;  // don't recycle scooped objects
	
	//NSLog(@"Considering a used %@ with retainCount:%d for recycling",classname,[entity retainCount]);
	if (classname)
	{
		[recycleLock lock];
		if (![entityRecyclePool objectForKey:classname])
			[entityRecyclePool setObject:[NSMutableArray arrayWithCapacity:100] forKey:classname];   // add a new array
		entlist = (NSMutableArray *)[entityRecyclePool objectForKey:classname];
		
		[entity setScanClass: CLASS_NO_DRAW];   //  housekeeping, keeps glitches from appearing on scanner
		if ([entity isKindOfClass:[ShipEntity class]])
			[[(ShipEntity *)entity getAI] setOwner:nil];					//  save ai misreporting
		
		if ([entlist count] < 100)		//  keep only up to 100 of each thing
			[entlist addObject:entity]; // add the entity to the array
		[recycleLock unlock];
	}
	return entity;																	// pass through
}

- (Entity *) recycledOrNew:(NSString *) classname
{
	Entity			*entity = nil;
	NSMutableArray  *entlist;
	if (classname)
	{
		if ([entityRecyclePool objectForKey:classname])
		{
			[recycleLock lock];
			entlist = (NSMutableArray *)[entityRecyclePool objectForKey:classname];
			if ([entlist count] > 0)
			{
				//NSLog(@"Recycling a used %@ from %@",classname,[entlist description]);
				entity = [[entlist objectAtIndex:0] retain];
				[entlist removeObjectAtIndex:0];
				//NSLog(@"Recycling a used %@",classname);
			}
			[recycleLock unlock];
		}
	}
	if (!entity)
	{
		Class   required_class = [[NSBundle mainBundle] classNamed:classname];
		entity = [[required_class alloc] init];
		[entity setUniverse:self];  // ensures access to preloaded data
	}
	return entity;
}

- (NSMutableDictionary *) preloadedDataFiles
{
	return preloadedDataFiles;
}

- (ShipEntity *) getShipWithRole:(NSString *) desc
{
	int i, found;
	ShipEntity		*ship;
	NSMutableArray  *foundShips = [NSMutableArray arrayWithCapacity:16];
	NSArray			*shipKeys = [shipdata allKeys];
	for (i = 0; i < [shipKeys count]; i++)
	{
		NSDictionary	*shipDict = (NSDictionary *)[shipdata objectForKey:[shipKeys objectAtIndex:i]];
		NSString		*shipRoles = (NSString *)[shipDict objectForKey:@"roles"];
		if ([shipRoles rangeOfString:desc].location != NSNotFound)
			[foundShips addObject:[shipKeys objectAtIndex:i]];
	}
	found = [foundShips count];
	if (found == 0)
		return nil;
	i = rand() % found;
	ship = [self getShip:(NSString *)[foundShips objectAtIndex:i]];
	[ship setValue:desc forKey:@"roles"];								// set it's roles to this one particular chosen role
	return ship;
}


- (ShipEntity *) getShip:(NSString *) desc
{
	NSDictionary	*shipDict;
	ShipEntity		*ship;
	shipDict = [self getDictionaryForShip:desc];
	NSString	*shipRoles = (NSString *)[shipDict objectForKey:@"roles"];
	BOOL		isStation = ([shipRoles rangeOfString:@"station"].location != NSNotFound);
	if (!shipDict)		return nil;
	if (isStation)
		ship = (StationEntity *)[self recycledOrNew:@"StationEntity"];
	else
		ship = (ShipEntity *)[self recycledOrNew:@"ShipEntity"];

	[ship setUpShipFromDictionary:shipDict];
	return ship;   // retain count = 1
}

- (NSDictionary *) getDictionaryForShip:(NSString *) desc
{
	if (![shipdata objectForKey:desc])
	{
		NSLog(@"***** Universe couldn't find a dictionary for a ship with description '%@'",desc);
		return nil;
	}
	else
		return [NSDictionary dictionaryWithDictionary:(NSDictionary *)[shipdata objectForKey:desc]];	// is autoreleased
}

- (int) getPriceForWeaponSystemWithKey:(NSString *)weapon_key
{
	int i;
	int price = 0;
	for (i = 0; ((i < [equipmentdata count])&&(price == 0)) ; i++)
	{
		int			price_per_unit  = [(NSNumber *)[(NSArray *)[equipmentdata objectAtIndex:i] objectAtIndex:1] intValue];
		NSString*   eq_type			= (NSString *)[(NSArray *)[equipmentdata objectAtIndex:i] objectAtIndex:3];
		if ([eq_type isEqual:weapon_key])
			price = price_per_unit;
	}
	return price;
}


- (int) legal_status_of_manifest:(NSArray *)manifest
{
	int i;
	int penalty = 0;
	for (i = 0 ; i < [manifest count] ; i++)
	{
		NSString *commodity = (NSString *)[(NSArray *)[manifest objectAtIndex:i] objectAtIndex:MARKET_NAME];
		int amount = [(NSNumber *)[(NSArray *)[manifest objectAtIndex:i] objectAtIndex:MARKET_QUANTITY] intValue];
		if ((amount > 0)&&([illegal_goods objectForKey:commodity]))
			penalty += amount * [(NSNumber *)[illegal_goods objectForKey:commodity] intValue];
	}
	return penalty;
}

- (NSArray *) getContainersOfPlentifulGoods:(int) how_many
{
	// build list of goods allocating 0..100 for each based on how
	// much of each quantity there is. Use a ratio of n x 100/64
	NSMutableArray  *accumulator = [NSMutableArray arrayWithCapacity:how_many];
	int quantities[[commoditydata count]];
	int total_quantity = 0;
	int i;
	for (i = 0; i < [commoditydata count]; i++)
	{
		int q = [(NSNumber *)[(NSArray *)[commoditydata objectAtIndex:i] objectAtIndex:MARKET_QUANTITY] intValue];
		if (q < 0)  q = 0;
		if (q > 64) q = 64;
		q *= 100;   q/= 64;
		quantities[i] = q;
		total_quantity += q;
	}
	// quantities is now used to determine which good get into the containers
	for (i = 0; i < how_many; i++)
	{
		ShipEntity* container = [self getShip:@"barrel"];   // retain count = 1
		int co_type, co_amount, qr;
		
		// select a random point in the histogram
		qr = rand() % total_quantity;
		co_type = 0;
		while (qr > 0)
		{
			qr -= quantities[co_type++];
		}
		co_type--;
		
		co_amount = [self getRandomAmountOfCommodity:co_type];
		
		//NSLog(@"... loading with plentiful %@",[self describeCommodity:co_type amount:co_amount]);
		
		// into the barrel it goes...
		[container setUniverse:self];
		[container setScanClass: CLASS_CARGO];
		[container setCommodity:co_type andAmount:co_amount];
		[accumulator addObject:container];
		[container release];
	}
	return [NSArray arrayWithArray:accumulator];	
}

- (NSArray *) getContainersOfScarceGoods:(int) how_many
{
	// build list of goods allocating 0..100 for each based on how
	// much of each quantity there is. Use a ratio of (64 - n) x 100/64
	NSMutableArray  *accumulator = [NSMutableArray arrayWithCapacity:how_many];
	int quantities[[commoditydata count]];
	int total_quantity = 0;
	int i;
	for (i = 0; i < [commoditydata count]; i++)
	{
		int q = 64 - [(NSNumber *)[(NSArray *)[commoditydata objectAtIndex:i] objectAtIndex:MARKET_QUANTITY] intValue];
		if (q < 0)  q = 0;
		if (q > 64) q = 64;
		q *= 100;   q/= 64;
		quantities[i] = q;
		total_quantity += q;
	}
	// quantities is now used to determine which good get into the containers
	for (i = 0; i < how_many; i++)
	{
		ShipEntity* container = [self getShip:@"barrel"];   // retain count = 1
		int co_type, co_amount, qr;
		
		// select a random point in the histogram
		qr = rand() % total_quantity;
		co_type = 0;
		while (qr > 0)
		{
			qr -= quantities[co_type++];
		}
		co_type--;
		
		co_amount = [self getRandomAmountOfCommodity:co_type];
		
		//NSLog(@"... loading with scarce %@",[self describeCommodity:co_type amount:co_amount]);
		
		[container setUniverse:self];
		[container setScanClass: CLASS_CARGO];
		[container setCommodity:co_type andAmount:co_amount];
		[accumulator addObject:container];
		[container release];
	}
	return [NSArray arrayWithArray:accumulator];	
}

- (int) getRandomCommodity
{
	int cd = rand() % [commoditydata count];
	return cd;
}

- (int) getRandomAmountOfCommodity:(int) co_type
{
	int units;
	if ((co_type < 0)||(co_type >= [commoditydata count]))
		return 0;
	units = [[[commoditydata objectAtIndex:co_type] objectAtIndex:MARKET_UNITS] intValue];
	switch (units)
	{
		case 0 :	// TONNES
			return 1;
			break;
		case 1 :	// KILOGRAMS
			return 1 + (rand() % 6) + (rand() % 6) + (rand() % 6);
			break;
		case 2 :	// GRAMS
			return 1 + (rand() % 6) + (rand() % 6);
			break;
	}
	return 1;
}

- (int) commodityForName:(NSString *) co_name
{
	int i;
	for (i = 0; i < [commoditydata count]; i++)
	{
		if ([co_name isEqual:[[commoditydata objectAtIndex:i] objectAtIndex:MARKET_NAME]])
			return i;
	}
	return NSNotFound;
}

- (NSString *) nameForCommodity:(int) co_type
{
	if ((co_type < 0)||(co_type >= [commoditydata count]))
		return @"";
	return [NSString stringWithFormat:@"%@",[[commoditydata objectAtIndex:co_type] objectAtIndex:MARKET_NAME]];
}

- (int) unitsForCommodity:(int) co_type
{
	if ((co_type < 0)||(co_type >= [commoditydata count]))
		return NSNotFound;
	return [[[commoditydata objectAtIndex:co_type] objectAtIndex:MARKET_UNITS] intValue];
}



- (NSString *) describeCommodity:(int) co_type amount:(int) co_amount
{
	int units;
	NSString	*desc2, *desc3;
	if ((co_type < 0)||(co_type >= [commoditydata count])||(co_amount == 0))
		return @"";
	units = [[[commoditydata objectAtIndex:co_type] objectAtIndex:MARKET_UNITS] intValue];
	switch (units)
	{
		case UNITS_TONS :	// TONNES
			desc2 = @"ton";
			break;
		case UNITS_KILOGRAMS :	// KILOGRAMS
			desc2 = @"kilogram";
			break;
		case UNITS_GRAMS :	// GRAMS
			desc2 = @"gram";
			break;
	}
	if (co_amount > 1)
		desc2 = [NSString stringWithFormat:@"%@s",desc2];
	desc3 = [[commoditydata objectAtIndex:co_type] objectAtIndex:MARKET_NAME];
	
	return [NSString stringWithFormat:@"%d %@ %@",co_amount,desc2,desc3];
}

////////////////////////////////////////////////////

- (void) setGameView:(NSView *)view
{
    if (gameView)	[gameView release];
    gameView = view;
    [gameView retain];
}

- (NSView *) gameView
{
    return gameView;
}

- (GameController *) gameController
{
	return [(MyOpenGLView *)gameView gameController];
}


- (TextureStore *) textureStore
{
    return textureStore;
}
- (void) resetTextureStore
{
//	if (textureStore) [textureStore release];
//	textureStore = [[TextureStore alloc] init]; // retained
	
	//NSLog(@"universe resetTextureStore");
	
	[textureStore reloadTextures];

	[self initSprites];
	[gui generateDisplay];
	[gui updateGui];
	
	int i;
	NSArray *entlist = [NSArray arrayWithArray:entities];
	for (i = 0; i < [entlist count]; i++)
	{
		[(Entity *)[entlist objectAtIndex:i] initialiseTextures];
		[(Entity *)[entlist objectAtIndex:i] regenerateDisplayList];
	}
}


- (void) drawFromEntity:(int) n
{
    int i, v_status;
	Vector	position, obj_position, view_dir;
	Matrix rotMatrix;
	//
	// use a non-mutable copy so this can't be changed under us.
	//
	NSArray  *entityList = [[NSArray alloc] initWithArray:entities];	// alloc retains
	//
	Entity	*viewthing = nil;
	Entity	*drawthing = nil;
    
    position.x = 0.0;	position.y = 0.0;	position.z = 0.0;
	set_matrix_identity(rotMatrix);

    if (n < [entityList count])
    {
        viewthing = [entityList objectAtIndex:n];
    }
    
    if (viewthing)
    {
        position = [viewthing getPosition];
		gl_matrix_into_matrix([viewthing rotationMatrix], rotMatrix);
		v_status = [viewthing getStatus];
    }
	
	// make a drawing order
	//
	[entsInDrawOrder setArray:entityList];
	[entsInDrawOrder removeObject:viewthing];
	[entsInDrawOrder sortUsingSelector:@selector(compareZeroDistance:)];
	//
	[entityList release];   // we're done with this now.
	
	//NSLog(@"Drawing from [%f,%f,%f]", position.x, position.y, position.z);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);			// face culling
	glDepthMask(GL_TRUE);	// restore write to depth buffer

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();	// reset matrix                         
    
    gluLookAt(0.0, 0.0, 0.0,	0.0, 0.0, 1.0,	0.0, 1.0, 0.0);
	
	// HACK BUSTED
	glScalef(   -1.0,  1.0,	1.0);   // flip left and right
	
    glPushMatrix(); // save this flat viewpoint

	switch (viewDirection)
	{
		case VIEW_FORWARD :
			view_dir.x = 0.0;   view_dir.y = 0.0;   view_dir.z = -1.0;
			break;
		case VIEW_AFT :
			view_dir.x = 0.0;   view_dir.y = 0.0;   view_dir.z = 1.0;
			break;
		case VIEW_PORT :
			view_dir.x = 1.0;   view_dir.y = 0.0;   view_dir.z = 0.0;
			break;
		case VIEW_STARBOARD :
			view_dir.x = -1.0;   view_dir.y = 0.0;   view_dir.z = 0.0;
			break;
		
		case VIEW_DOCKED :
		case VIEW_BREAK_PATTERN :
		default :
			view_dir.x = 0.0;   view_dir.y = 0.0;   view_dir.z = -1.0;
			break;
	}

    gluLookAt(0.0, 0.0, 0.0,	view_dir.x, view_dir.y, view_dir.z,	0.0, 1.0, 0.0);
	
	
	if ((!displayText) || (v_status == STATUS_DEMO))
	{
		//
		// rotate the view
		glMultMatrixf([viewthing rotationMatrix]);
		// translate the view
		glTranslatef(-position.x,-position.y,-position.z);
		
		// set lighting
		glLightfv(GL_LIGHT1, GL_POSITION, sun_center_position);
		
		//
		//		DRAW ALL THE OPAQUE ENTITIES
		//
		for (i = 0; i < [entsInDrawOrder count]; i++)
		{
			int d_status;
			drawthing = (Entity *)[entsInDrawOrder objectAtIndex:i];
			d_status = [drawthing getStatus];
			
			if (((d_status == STATUS_DEMO)&&(v_status == STATUS_DEMO)) || ((d_status != STATUS_DEMO)&&(v_status != STATUS_DEMO)))
			{
				glPushMatrix();
				obj_position = [drawthing getPosition];
				//translate the object
				glTranslatef(obj_position.x,obj_position.y,obj_position.z);
				//rotate the object
				glMultMatrixf([drawthing rotationMatrix]);
				// draw the thing
				[drawthing drawEntity:NO:NO];
				glPopMatrix();
			}
		}
		
		//
		//		DRAW ALL THE TRANSLUCENT entsInDrawOrder
		//
		glDepthMask(GL_FALSE);				// don't write to depth buffer
		for (i = 0; i < [entsInDrawOrder count]; i++)
		{
			int d_status;
			drawthing = [entsInDrawOrder objectAtIndex:i];
			d_status = [drawthing getStatus];
			
			if (((d_status == STATUS_DEMO)&&(v_status == STATUS_DEMO)) || ((d_status != STATUS_DEMO)&&(v_status != STATUS_DEMO)))
			{
				glPushMatrix();
				obj_position = [drawthing getPosition];
				//translate the object
				glTranslatef(obj_position.x,obj_position.y,obj_position.z);
				//rotate the object
				glMultMatrixf([drawthing rotationMatrix]);
				// draw the thing
				[drawthing drawEntity:NO:YES];
				glPopMatrix();
			}
		}
		glDepthMask(GL_TRUE);	// restore write to depth buffer
	}
	
	glPopMatrix(); //restore saved flat viewpoint
    
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);			// face culling
	glDepthMask(GL_FALSE);				// don't write to depth buffer
	//
	if ((v_status != STATUS_DEAD)&&(v_status != STATUS_ESCAPE_SEQUENCE))
	{
		[self drawScannerFromEntity:n];
		double  line_size = [(MyOpenGLView *)gameView viewSize].width / 1024.0;
		if (line_size < 1.0)
			line_size = 1.0;
		glLineWidth(line_size);
		[self drawDialsForEntity:n];
		if (!displayText)
			[self drawCrosshairs];
	}
	[self drawMessage];
	
	glFlush();	// don't wait around for drawing to complete
}

static BOOL hostiles;
- (void) drawScannerFromEntity:(int) n
{	
	int i;
    int scanner_cx = SCANNER_CENTRE_X;
	int scanner_cy = SCANNER_CENTRE_Y;
	int scanner_scale = SCANNER_SCALE;
	double max_scanner_range2 = scanner_scale*scanner_scale*10000.0;
	BOOL	isHostile = NO;
	BOOL	foundHostiles = NO;
	
	Vector	position, relativePosition;
	Matrix rotMatrix;
	int flash = (int)(universal_time * 4);
	flash &= 1;
	//
	// use a non-mutable copy so this can't be changed under us.
	//
	NSArray  *entityList = [[NSArray alloc] initWithArray:entities];	// alloc retains
	//
	Entity	*viewthing = nil;
	Entity	*drawthing = nil;
	GLfloat cargo_color[4] = {0.9, 0.9, 0.9, 1.0};		// gray
	GLfloat hostile_color[4] = {1.0, 0.25, 0.0, 1.0};	// red/orange
	GLfloat neutral_color[4] = {1.0, 1.0, 0.0, 1.0};	// yellow
	GLfloat friendly_color[4] = {0.0, 1.0, 0.0, 1.0};   // green
	GLfloat missile_color[4] = {0.0, 1.0, 1.0, 1.0};	// cyan
	GLfloat police_color1[4] = {0.5, 0.0, 1.0, 1.0};		// purpley-blue
	GLfloat police_color2[4] = {1.0, 0.0, 0.5, 1.0};		// purpley-red
	GLfloat col[4];								// can be manipulated
    
    position.x = 0.0;	position.y = 0.0;	position.z = 0.0;
	set_matrix_identity(rotMatrix);

    if (n < [entityList count])
    {
        viewthing = [entityList objectAtIndex:n];
    }
    
    if (viewthing)
    {
        position = [viewthing getPosition];
		gl_matrix_into_matrix([viewthing rotationMatrix], rotMatrix);
    }
	
	[scannerSprite blitCentredToX:scanner_cx Y:scanner_cy Z:640.0 Alpha:0.50];
	//
	if (viewDirection != VIEW_DOCKED)
	{
		for (i = 0; i < [entityList count]; i++)  // scanner lollypops
		{
			if (i != n)
			{
				drawthing = (Entity *)[entityList objectAtIndex:i];
				if ([drawthing scanClass] != CLASS_NO_DRAW)
				{
					GLfloat x1,y1,y2,z1;
					double upscale = 1.25;
					
					if ([drawthing getZeroDistance] > max_scanner_range2)
						continue;
					
					relativePosition = [drawthing getPosition];
					// translate the view
					relativePosition.x -= position.x;   relativePosition.y -= position.y;   relativePosition.z -= position.z;
					// rotate the view
					mult_vector(&relativePosition, rotMatrix);
					// scale the view
					relativePosition.x *= upscale/scanner_scale;	relativePosition.y *= upscale/scanner_scale;	relativePosition.z *= upscale/scanner_scale;
					
					x1 = relativePosition.x;
					y1 = relativePosition.z / 4.0;
					y2 = y1 + relativePosition.y / 2.0;
					
					isHostile = NO;
					if ([drawthing isKindOfClass:[ShipEntity class]])
					{
						isHostile = (([(ShipEntity *)drawthing getPrimaryTarget] == viewthing)&&([(ShipEntity *)drawthing getZeroDistance] < 100000000.0));
					}
					
					// position the scanner
					x1 += scanner_cx;   y1 += scanner_cy;   y2 += scanner_cy;   z1 = 640.0;
					switch ([drawthing scanClass])
					{
						case CLASS_CARGO :
							col[0] = cargo_color[0];	col[1] = cargo_color[1];	col[2] = cargo_color[2];	col[3] = cargo_color[3];
							break;
						case CLASS_THARGOID :
							if (flash)
							{
								col[0] = friendly_color[0];	col[1] = friendly_color[1];	col[2] = friendly_color[2];	col[3] = friendly_color[3];
							}
							else
							{
								col[0] = hostile_color[0];	col[1] = hostile_color[1];	col[2] = hostile_color[2];	col[3] = hostile_color[3];
							}
							foundHostiles = YES;
							break;
						case CLASS_MISSILE :
							col[0] = missile_color[0];	col[1] = missile_color[1];	col[2] = missile_color[2];	col[3] = missile_color[3];
							break;
						case CLASS_STATION :
							col[0] = friendly_color[0];	col[1] = friendly_color[1];	col[2] = friendly_color[2];	col[3] = friendly_color[3];
							break;
						case CLASS_BUOY :
							if (flash)
							{
								col[0] = neutral_color[0];	col[1] = neutral_color[1];	col[2] = neutral_color[2];	col[3] = neutral_color[3];
							}
							else
							{
								col[0] = friendly_color[0];	col[1] = friendly_color[1];	col[2] = friendly_color[2];	col[3] = friendly_color[3];
							}
							break;
						case CLASS_POLICE :
							if ((isHostile)&&(flash))
							{
								col[0] = police_color2[0];	col[1] = police_color2[1];	col[2] = police_color2[2];	col[3] = police_color2[3];
							}
							else
							{
								col[0] = police_color1[0];	col[1] = police_color1[1];	col[2] = police_color1[2];	col[3] = police_color1[3];
							}
							break;
						default :
							if (isHostile)
							{
								col[0] = hostile_color[0];	col[1] = hostile_color[1];	col[2] = hostile_color[2];	col[3] = hostile_color[3];
								foundHostiles = YES;
							}
							else
							{
								col[0] = neutral_color[0];	col[1] = neutral_color[1];	col[2] = neutral_color[2];	col[3] = neutral_color[3];
							}
							break;
					}
					glBegin(GL_QUADS);
					glColor4fv(col);
					glVertex3f( x1-3, y2, z1);	glVertex3f( x1+2, y2, z1);	glVertex3f( x1+2, y2+3, z1);	glVertex3f( x1-3, y2+3, z1);	
					col[3] *= 0.3333; // one third the alpha
					glColor4fv(col);
					glVertex3f( x1, y1, z1);	glVertex3f( x1+2, y1, z1);	glVertex3f( x1+2, y2, z1);	glVertex3f( x1, y2, z1);
					glEnd();
				}
			}
		}
		//
		if ((foundHostiles)&&(!hostiles))
		{
			//NSLog(@"FOUND HOSTILES");
			[viewthing warnAboutHostiles];  // there are now hostiles on scope, warn the scanner
			hostiles = YES;
		}
		if ((!foundHostiles)&&(hostiles))
		{
			//NSLog(@"LOST HOSTILES");
			hostiles = NO;					// there are now no hostiles on scope, relax
		}
	}
	
	// draw the compass
	//
	[compassSprite blitCentredToX:COMPASS_CENTRE_X Y:COMPASS_CENTRE_Y Z:640.0 Alpha:1.0];
	//
	if ((viewDirection != VIEW_DOCKED)&&(sun != NO_TARGET))
	{
		relativePosition = [[self entityForUniversalID:planet] getPosition];
		if ([(ShipEntity *)viewthing checkForAegis] != AEGIS_NONE)
			relativePosition = [[self entityForUniversalID:station] getPosition];
		if ([(ShipEntity *)viewthing checkForAegis] == AEGIS_IN_DOCKING_RANGE)
			[aegisSprite blitCentredToX:AEGIS_CENTRE_X Y:AEGIS_CENTRE_Y Z:640.0 Alpha:1.0];
		// translate the view
		relativePosition.x -= position.x;   relativePosition.y -= position.y;   relativePosition.z -= position.z;
		// rotate the view
		mult_vector(&relativePosition, rotMatrix);
		relativePosition = unit_vector(&relativePosition);
		relativePosition.x *= COMPASS_HALF_SIZE * 0.4;
		relativePosition.y *= COMPASS_HALF_SIZE * 0.4;
		relativePosition.x += COMPASS_CENTRE_X;
		relativePosition.y += COMPASS_CENTRE_Y;
		if (relativePosition.z >= 0)
			[greendotSprite blitCentredToX:relativePosition.x Y:relativePosition.y Z:640.0 Alpha:1.0];
		else
			[reddotSprite blitCentredToX:relativePosition.x Y:relativePosition.y Z:640.0 Alpha:1.0];
	}
	
	//
	[entityList release];   // we're done with this now.

}

- (void) drawDialsForEntity:(int) n
{
	GLfloat red_color[4] = {1.0, 0.0, 0.0, 1.0};
	GLfloat redplus_color[4] = {1.0, 0.0, 0.5, 1.0};
	GLfloat yellow_color[4] = {1.0, 1.0, 0.0, 1.0};
	GLfloat green_color[4] = {0.0, 1.0, 0.0, 1.0};
	PlayerEntity	*viewthing = nil;
	int z1 = 640;
	
	int flash = (int)(universal_time * 4);
	
	int speed_dial_cx = 200;
	int speed_dial_cy = -145;
	NSSize speed_dial_size = NSMakeSize(80,8);
	
	int roll_dial_cx = 200;
	int roll_dial_cy = -160;
	NSSize roll_dial_size = NSMakeSize(80,8);
	
	int pitch_dial_cx = 200;
	int pitch_dial_cy = -170;
	NSSize pitch_dial_size = NSMakeSize(80,8);
	
	int energy_dial_cx = 200;
	int energy_dial_cy = -205;
	NSSize energy_dial_size = NSMakeSize(80,48);
	
	int forward_shield_dial_cx = -200;
	int forward_shield_dial_cy = -146;
	NSSize forward_shield_dial_size = NSMakeSize(80,10);
	
	int aft_shield_dial_cx = -200;
	int aft_shield_dial_cy = -162;
	NSSize aft_shield_dial_size = NSMakeSize(80,10);
	
	int weapon_temp_dial_cx = -200;
	int weapon_temp_dial_cy = -194;
	NSSize aux_dial_size = NSMakeSize(80,40);
	NSSize aux1_dial_size = NSMakeSize(80,8);
	
	NSSize missile_icon_size = NSMakeSize(12,12);
	int missile_dial_cy = -224;
	int missile_dial_cx = -228;
	int missile_dial_spacing = 16;
	
	flash &= 1;
	
	if (n < [entities count])
    {
        viewthing = (PlayerEntity	*)[entities objectAtIndex:n];
    }
    
    if (viewthing)
    {
		// Left Hand Dials
		[legend1Sprite blitCentredToX:-256.0 Y:-184 Z:z1 Alpha:1.0];
		
		// draw forward_shield surround
		glColor4fv(green_color);
		[self drawSurroundAt:forward_shield_dial_cx:forward_shield_dial_cy:z1 ofSize:forward_shield_dial_size];
		// draw forward_shield bar
		glColor4fv(green_color);
		if ([viewthing dial_forward_shield] < .80)
			glColor4fv(yellow_color);
		if ([viewthing dial_forward_shield] < .25)
			glColor4fv(red_color);
		[self   drawBarAt:forward_shield_dial_cx:forward_shield_dial_cy:z1
				ofSize:forward_shield_dial_size
				andValue:[viewthing dial_forward_shield]];

		// draw aft_shield surround
		glColor4fv(green_color);
		[self drawSurroundAt:aft_shield_dial_cx:aft_shield_dial_cy:z1 ofSize:aft_shield_dial_size];
		// draw aft_shield bar
		glColor4fv(green_color);
		if ([viewthing dial_aft_shield] < .80)
			glColor4fv(yellow_color);
		if ([viewthing dial_aft_shield] < .25)
			glColor4fv(red_color);
		[self   drawBarAt:aft_shield_dial_cx:aft_shield_dial_cy:z1
				ofSize:aft_shield_dial_size
				andValue:[viewthing dial_aft_shield]];

		// draw auxilliary dials surround
		glColor4fv(yellow_color);
		[self drawSurroundAt:weapon_temp_dial_cx:weapon_temp_dial_cy:z1 ofSize:aux_dial_size];
		// draw fuel bar
		glColor4fv(yellow_color);
		[self drawBarAt:weapon_temp_dial_cx:weapon_temp_dial_cy + 15 :z1 ofSize:aux1_dial_size andValue:[viewthing dial_fuel]];
		// draw cabin_temp bar
		glColor4fv(green_color);
		if ([viewthing dial_cabin_temp] > .25)
			glColor4fv(yellow_color);
		if ([viewthing dial_cabin_temp] > .80)
			glColor4fv(red_color);
		if ((flash)&&([viewthing dial_cabin_temp] > .90))
			glColor4fv(redplus_color);
		[self drawBarAt:weapon_temp_dial_cx:weapon_temp_dial_cy + 5 :z1 ofSize:aux1_dial_size andValue:[viewthing dial_cabin_temp]];
		// draw weapon_temp bar
		glColor4fv(green_color);
		if ([viewthing dial_weapon_temp] > .25)
			glColor4fv(yellow_color);
		if ([viewthing dial_weapon_temp] > .80)
			glColor4fv(red_color);
		[self drawBarAt:weapon_temp_dial_cx:weapon_temp_dial_cy - 5 :z1 ofSize:aux1_dial_size andValue:[viewthing dial_weapon_temp]];
		// draw altitude bar
		glColor4fv(green_color);
		if ([viewthing dial_altitude] < .75)
			glColor4fv(yellow_color);
		if ([viewthing dial_altitude] < .25)
			glColor4fv(red_color);
		if ((flash)&&([viewthing dial_altitude] < .10))
			glColor4fv(redplus_color);
		[self drawBarAt:weapon_temp_dial_cx:weapon_temp_dial_cy - 15 :z1 ofSize:aux1_dial_size andValue:[viewthing dial_altitude]];

		// missiles
		{
			int i;
			for (i = 0; i < [viewthing dial_missiles]; i++)
			{
				if (i == 0)
				{
					switch ([viewthing dial_missile_status])
					{
						case MISSILE_STATUS_SAFE :
							glColor4fv(green_color);	break;
						case MISSILE_STATUS_ARMED :
							glColor4fv(yellow_color);	break;
						case MISSILE_STATUS_TARGET_LOCKED :
							glColor4fv(red_color);	break;
					}
				}
				else
				{
					glColor4fv(green_color);
				}
				glBegin(GL_POLYGON);
				[self drawMissileIconAt:missile_dial_cx + i * missile_dial_spacing :missile_dial_cy:z1 ofSize:missile_icon_size];
				glEnd();
				glColor4fv(green_color);
				glBegin(GL_LINE_LOOP);
				[self drawMissileIconAt:missile_dial_cx + i * missile_dial_spacing :missile_dial_cy:z1 ofSize:missile_icon_size];
				glEnd();
			}
		}

		// Right Hand dials
		[legend2Sprite blitCentredToX:256.0 Y:-184 Z:z1 Alpha:1.0];

		// draw speed surround
		glColor4fv(green_color);
		[self drawSurroundAt:speed_dial_cx:speed_dial_cy:z1 ofSize:speed_dial_size];
		// draw speed bar
		if ([viewthing dial_speed] > .25)
			glColor4fv(yellow_color);
		if ([viewthing dial_speed] > .80)
			glColor4fv(red_color);
		[self drawBarAt:speed_dial_cx:speed_dial_cy:z1 ofSize:speed_dial_size andValue:[viewthing dial_speed]];

		// draw ROLL surround
		glColor4fv(green_color);
		[self drawSurroundAt:roll_dial_cx:roll_dial_cy:z1 ofSize:roll_dial_size];
		// draw roll bar
		glColor4fv(yellow_color);
		[self drawIndicatorAt:roll_dial_cx:roll_dial_cy:z1 ofSize:roll_dial_size andValue:[viewthing dial_roll]];
		
		// draw PITCH surround
		glColor4fv(green_color);
		[self drawSurroundAt:pitch_dial_cx:pitch_dial_cy:z1 ofSize:pitch_dial_size];
		// draw pitch bar
		glColor4fv(yellow_color);
		[self drawIndicatorAt:pitch_dial_cx:pitch_dial_cy:z1 ofSize:pitch_dial_size andValue:[viewthing dial_pitch]];
		
		// draw energy surround
		glColor4fv(yellow_color);
		[self drawSurroundAt:energy_dial_cx:energy_dial_cy:z1 ofSize:energy_dial_size];
		// draw energy bars
		glColor4fv(yellow_color);
		{
			int qy = energy_dial_size.height / 4;
			NSSize dial_size = NSMakeSize(energy_dial_size.width,qy - 2);
			int cy = energy_dial_cy - 3 * qy / 2;
			double energy = [viewthing dial_energy]*4.0;
			int i;
			for (i = 0; i < 4; i++)
			{
				if (energy > 1.0)
					[self   drawBarAt:energy_dial_cx:cy:z1 ofSize:dial_size andValue:1.0];
				if ((energy > 0.0)&&(energy <= 1.0))
					[self   drawBarAt:energy_dial_cx:cy:z1 ofSize:dial_size andValue:energy];
				energy -= 1.0;
				cy += qy;
			}
		}
		
	}
}

- (void) drawIndicatorAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz andValue:(double) amount
{
	int dial_oy =   y - siz.height/2;
	int position =  x + amount * siz.width / 2;
	glBegin(GL_QUADS);
		glVertex3i( position, dial_oy, z);
		glVertex3i( position+2, y, z);
		glVertex3i( position, dial_oy+siz.height, z);
		glVertex3i( position-2, y, z);
	glEnd();
}

- (void) drawBarAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz andValue:(double) amount
{
	int dial_ox =   x - siz.width/2;
	int dial_oy =   y - siz.height/2;
	int position =  dial_ox + amount * siz.width;
	
	glBegin(GL_QUADS);
		glVertex3i( dial_ox, dial_oy, z);
		glVertex3i( position, dial_oy, z);
		glVertex3i( position, dial_oy+siz.height, z);
		glVertex3i( dial_ox, dial_oy+siz.height, z);
	glEnd();
}

- (void) drawSurroundAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz
{
	int dial_ox = x - siz.width/2;
	int dial_oy = y - siz.height/2;

	glBegin(GL_LINE_LOOP);
		glVertex3i( dial_ox-2, dial_oy-2, z);
		glVertex3i( dial_ox+siz.width+2, dial_oy-2, z);
		glVertex3i( dial_ox+siz.width+2, dial_oy+siz.height+2, z);
		glVertex3i( dial_ox-2, dial_oy+siz.height+2, z);
	glEnd();
}

- (void) drawMissileIconAt:(int) x:(int) y:(int) z ofSize:(NSSize) siz
{
	int ox = x - siz.width / 2.0;
	int oy = y - siz.height / 2.0;
	int w = siz.width / 4.0;
	int h = siz.height / 4.0; 

	glVertex3i( ox, oy + 3 * h, z);
	glVertex3i( ox + 2 * w, oy, z);
	glVertex3i( ox + w, oy, z);
	glVertex3i( ox + w, oy - 2 * h, z);
	glVertex3i( ox - w, oy - 2 * h, z);
	glVertex3i( ox - w, oy, z);
	glVertex3i( ox - 2 * w, oy, z);
}

- (void) drawCrosshairs
{
    PlayerEntity*   playerShip = (PlayerEntity *)[self entityZero];
	
	if (([playerShip getStatus] == STATUS_IN_FLIGHT)||([playerShip getStatus] == STATUS_WITCHSPACE_COUNTDOWN))
	{	
		//debug mouse position
		/*--
		NSPoint vjp = [(MyOpenGLView *)gameView virtualJoystickPosition];
		GLfloat vjpx = vjp.x * CROSSHAIR_SIZE;
		GLfloat vjpy = vjp.y * CROSSHAIR_SIZE;
		--*/
		
		GLfloat cx0 = CROSSHAIR_SIZE;
		GLfloat cx1 = CROSSHAIR_SIZE / 2.0;
		GLfloat z1 = 640.0;
		GLfloat cx_col0[4] = { 0.0, 1.0, 0.0, 0.25};
		GLfloat cx_col1[4] = { 0.0, 1.0, 0.0, 0.75};
		glEnable(GL_LINE_SMOOTH);									// alpha blending
		glLineWidth(2.0);
		glBegin(GL_LINES);
		glColor4fv(cx_col0);	glVertex3f(0.0, cx0, z1);   glColor4fv(cx_col1);	glVertex3f(0.0, cx1, z1);
		glColor4fv(cx_col0);	glVertex3f(0.0, -cx0, z1);   glColor4fv(cx_col1);	glVertex3f(0.0, -cx1, z1);
		glColor4fv(cx_col0);	glVertex3f(cx0, 0.0, z1);   glColor4fv(cx_col1);	glVertex3f(cx1, 0.0, z1);
		glColor4fv(cx_col0);	glVertex3f(-cx0, 0.0, z1);   glColor4fv(cx_col1);	glVertex3f(-cx1, 0.0, z1);
		glEnd();
		glLineWidth(1.0);
		
		//DEBUG mouse position
		/*--
		glColor4fv(cx_col1);
		glBegin(GL_LINES);
		glVertex3f(-cx1, vjpy, z1);	glVertex3f(cx1, vjpy, z1);
		glVertex3f(vjpx, -cx1, z1);	glVertex3f(vjpx, cx1, z1);
		glEnd();
		--*/
	}
}

- (void) drawMessage
{
    PlayerEntity*   playerShip = (PlayerEntity *)[self entityZero];

	int ms = 0;
	float  ms_y = TEXT_MESSAGE_Y;
	while ((ms < MAX_MESSAGES)&&(msDuration[ms] > 0.0))
	{
		ms_y += [msSpriteArray[ms] size].height;
		[msSpriteArray[ms] blitCentredToX:0.0 Y:ms_y Z:640.0 Alpha:msDuration[ms]];
		ms++;
	}
	if (displayFPS)
	{
		if ((fpsDisplayDuration > 0.0)&&([playerShip getStatus] != STATUS_DOCKED))
		{
			[fpsSprite blitCentredToX:-290.0 Y:220.0 Z:640.0 Alpha:fpsDisplayDuration];
		}
	}
	if (displayText)
	{
		[[gui openGLSprite] blitCentredToX:0.0 Y:0.0 Z:640.0 Alpha:1.0];
	}
	
}

- (Entity *) entityZero
{
	return [entities objectAtIndex:0];
}


- (Entity *) entityForUniversalID:(int)u_id
{
	if (u_id == NO_TARGET)
		return nil;
	else
	{
		NSString *idkey = [NSString stringWithFormat:@"%d",u_id];
		return [universal_id_dictionary objectForKey:idkey];
	}
}

- (BOOL) addEntity:(Entity *) entity
{
	if (entity)
	{
		if (![entity isKindOfClass:[ParticleEntity class]])
		{
			next_universal_id++;						// increment keeps idkeys unique
			while (next_universal_id == NO_TARGET)		// these are the null values - avoid them!
				next_universal_id++;
			NSString *idkey = [NSString stringWithFormat:@"%d",next_universal_id];
			[entity setUniversal_id:next_universal_id];
			[universal_id_dictionary setObject:entity forKey:idkey];
			if ([entity isKindOfClass:[ShipEntity class]])
			{
				[[(ShipEntity *)entity getAI] setOwner:(ShipEntity *)entity];
				[[(ShipEntity *)entity getAI] setState:@"GLOBAL"];
			}
		}
		else
			[entity setUniversal_id:NO_TARGET];
		[entity setUniverse:self];
		[entities addObject:entity];
		
		//NSLog(@"++(%@)\n%@", entity, [entities description]);
		
		return YES;
	}
	return NO;
}

- (BOOL) removeEntity:(Entity *) entity
{
	if (entity)
	{
		// remove from the reference dictionary
		NSString *idkey = [NSString stringWithFormat:@"%d",[entity universal_id]];
		[universal_id_dictionary removeObjectForKey:idkey];
		[entity setUniversal_id:NO_TARGET];
		[entity setUniverse:nil];
		
		// remove from the definitive list
		if ([entities containsObject:entity])
		{
			if ([entity isKindOfClass:[RingEntity class]])
				breakPatternCounter--;

			[entities removeObject:[self recycleOrDiscard:entity]];
			
			//NSLog(@"--(%@)\n%@", entity, [entities description]);
			
			return YES;
		}
	}
	return NO;
}

- (void) removeAllEntitiesExceptPlayer:(BOOL) restore
{
	if (![[entities objectAtIndex:0] isKindOfClass:[PlayerEntity class]])
	{
		NSLog(@"***** First entity is not the player in Universe.removeAllEntitiesExceptPlayer - exiting.");
		exit(1);
	}
	
	while ([entities count] > 1)
	{
		Entity* ent = [entities objectAtIndex:1];
		if (![ent isKindOfClass:[PlayerEntity class]])
		{
			if ([ent isKindOfClass:[StationEntity class]])  // clear out queues
				[(StationEntity *)ent clear];
			
			[self removeEntity:ent];
		}
	}	
}

- (void) removeDemoShips
{
	int i;
	NSArray *entlist = [NSArray arrayWithArray:entities];
	if ([entlist count] <= 1)
		return;
	for (i = 1; i < [entlist count]; i++)
	{
		Entity* ent = (Entity *)[entlist objectAtIndex:i];
		if ([ent getStatus] == STATUS_DEMO)
			[self removeEntity:ent];
	}
}

- (NSArray *) getAllEntities
{
	return [NSArray arrayWithArray:entities];
}

- (BOOL) isVectorClearFromEntity:(Entity *) e1 toDistance:(double)dist fromPoint:(Vector) p2
{
	NSArray *entlist =  [NSArray arrayWithArray:entities];
	BOOL	result =		YES;
	Vector  f1;
	int i;
	Vector p1 = [e1 getPosition];
	Vector v1 = p2;
	v1.x -= p1.x;   v1.y -= p1.y;   v1.z -= p1.z;   // vector from entity to p2
	
	double  nearest = sqrt(v1.x*v1.x + v1.y*v1.y + v1.z*v1.z) - dist;  // length of vector
	
	if (nearest < dist)
		return YES;			// within range already!
	
	//NSLog(@"Distance being considered is %.1f", nearest);

	f1 = unit_vector(&v1);   // unit vector in direction of p2 from p1
	
	//NSLog(@"f1 = (%.1f,%.1f,%.1f)",f1.x,f1.y,f1.z);
	
	for (i = 0; (i < [entlist count])&&(result) ; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide]))
		{
			Vector epos = [e2 getPosition];
			epos.x -= p1.x;	epos.y -= p1.y;	epos.z -= p1.z; // epos now holds vector from p1 to this entities position
			
			double d_forward = dot_product(epos,f1);	// distance along f1 which is nearest to e2's position
			
			if ((d_forward > 0)&&(d_forward < nearest))
			{
				double cr = 1.10 * ([e2 collisionRadius] + [e1 collisionRadius]); //  10% safety margin
				
				//NSLog(@"Entity %@ being considered, d_forward is %.1f", e2, d_forward);
				
				Vector p0 = [e1 getPosition];
				p0.x += d_forward * f1.x;	p0.y += d_forward * f1.y;	p0.z += d_forward * f1.z;
				// p0 holds nearest point on current course to center of incident object
				
				//NSLog(@"... p0 = (%.1f, %.1f, %.1f)", p0.x, p0.y, p0.z);
				
				Vector epos = [e2 getPosition];
				p0.x -= epos.x;	p0.y -= epos.y;	p0.z -= epos.z;
				// compare with center of incident object
				
				double  dist2 = p0.x * p0.x + p0.y * p0.y + p0.z * p0.z;
				
				//NSLog(@"... which is %.1f from entity.",sqrt(dist2));
				
				if (dist2 < cr*cr)
				{
					//NSLog(@"... which IS incident = against radius %.1f", cr);
					result = NO;
				}
			}
		}
	}
	return result;
}


- (Vector) getSafeVectorFromEntity:(Entity *) e1 toDistance:(double)dist fromPoint:(Vector) p2
{
	//
	// heuristic three
	//
	NSArray *entlist = [NSArray arrayWithArray:entities];
	Vector  f1;
	Vector  result = p2;
	int i;
	Vector p1 = [e1 getPosition];
	Vector v1 = p2;
	v1.x -= p1.x;   v1.y -= p1.y;   v1.z -= p1.z;   // vector from entity to p2
	
	double  nearest = sqrt(v1.x*v1.x + v1.y*v1.y + v1.z*v1.z) - dist;  // length of vector

	f1 = unit_vector(&v1);   // unit vector in direction of p2 from p1
		
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide]))
		{
			Vector epos = [e2 getPosition];
			epos.x -= p1.x;	epos.y -= p1.y;	epos.z -= p1.z;
			double d_forward = dot_product(epos,f1);
			if ((d_forward > 0)&&(d_forward < nearest))
			{
				double cr = 1.10 * ([e2 collisionRadius] + [e1 collisionRadius]); //  10% safety margin
					
				Vector p0 = [e1 getPosition];
				p0.x += d_forward * f1.x;	p0.y += d_forward * f1.y;	p0.z += d_forward * f1.z;
				// p0 holds nearest point on current course to center of incident object
								
				Vector epos = [e2 getPosition];
				p0.x -= epos.x;	p0.y -= epos.y;	p0.z -= epos.z;
				// compare with center of incident object
				
				double  dist2 = p0.x * p0.x + p0.y * p0.y + p0.z * p0.z;
								
				if (dist2 < cr*cr)
				{
					result = [e2 getPosition];			// center of incident object
					nearest = d_forward;
					
					if (dist2 == 0.0)
					{
						// ie. we're on a line through the object's center !
						// jitter the position somewhat!
						result.x += ((rand() % 1024) - 512)/512.0; //   -1.0 .. +1.0
						result.y += ((rand() % 1024) - 512)/512.0; //   -1.0 .. +1.0
						result.z += ((rand() % 1024) - 512)/512.0; //   -1.0 .. +1.0
					}
					
					Vector  nearest_point = p1;
					nearest_point.x += d_forward * f1.x;	nearest_point.y += d_forward * f1.y;	nearest_point.z += d_forward * f1.z;
					// nearest point now holds nearest point on line to center of incident object
					
					Vector outward = nearest_point;
					outward.x -= result.x;	outward.y -= result.y;	outward.z -= result.z;
					outward = unit_vector(&outward);
					// outward holds unit vector through the nearest point on the line from the center of incident object
					
					Vector backward = p1;
					backward.x -= result.x;	backward.y -= result.y;	backward.z -= result.z;
					backward = unit_vector(&backward);
					// backward holds unit vector from center of the incident object to the center of the ship
					
					Vector dd = result;
					dd.x -= p1.x; dd.y -= p1.y; dd.z -= p1.z;
					double current_distance = sqrt (dd.x*dd.x + dd.y*dd.y + dd.z*dd.z);
					
					// sanity check current_distance
					//NSLog(@"Current distance is %.1f CR", current_distance/cr);
					if (current_distance < cr * 1.25)	// 25% safety margin
						current_distance = cr * 1.25;
					if (current_distance > cr * 5.0)	// up to 2 diameters away 
						current_distance = cr * 5.0;
										
					// choose a point that's three parts backward and one part outward
					//
					result.x += 0.25 * (outward.x * current_distance) + 0.75 * (backward.x * current_distance);		// push 'out' by this amount
					result.y += 0.25 * (outward.y * current_distance) + 0.75 * (backward.y * current_distance);
					result.z += 0.25 * (outward.z * current_distance) + 0.75 * (backward.z * current_distance);

					//NSLog(@"Bypassing %@ by going from (%.1f,%.1f,%.1f) to (%.1f,%.1f,%.1f)",e2,p1.x,p1.y,p1.z,result.x,result.y,result.z);
					
				}
			}
		}
	}
	return result;
}

- (NSArray *) getLaserLineEntitiesForEntity:(Entity *) e1 inView:(int) viewdir
{
	NSArray *entlist = [NSArray arrayWithArray:entities];
	NSMutableArray *hitlist = [NSMutableArray arrayWithCapacity:4];
	int i;
	Vector p1 = [e1 getPosition];
	Quaternion q1 = [e1 QRotation];
	if ([e1 isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   //  reverse for player viewpoint

	Vector u1 = vector_up_from_quaternion(q1);
	
	switch (viewdir)
	{
		case VIEW_AFT :
			quaternion_rotate_about_axis(&q1, u1, PI);
			break;
		case VIEW_PORT :
			quaternion_rotate_about_axis(&q1, u1, PI/2.0);
			break;
		case VIEW_STARBOARD :
			quaternion_rotate_about_axis(&q1, u1, -PI/2.0);
			break;
	}
	
	Vector f1 = vector_forward_from_quaternion(q1);
	
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide]))
		{
			Vector p2 = [e2 getPosition];
			p2.x -= p1.x;	p2.y -= p1.y;	p2.z -= p1.z;
			double d_forward = dot_product(p2,f1);
			if (d_forward > 0)
			{
				Vector p0 = [e1 getPosition];
				p0.x += d_forward * f1.x;	p0.y += d_forward * f1.y;	p0.z += d_forward * f1.z;
				// p0 holds nearest point on current course to center of incident object
				Vector epos = [e2 getPosition];
				p0.x -= epos.x;	p0.y -= epos.y;	p0.z -= epos.z;
				// compare with center of incident object
				double  dist2 = p0.x * p0.x + p0.y * p0.y + p0.z * p0.z;
				double cr = [e2 collisionRadius];
				if (dist2 < cr*cr)
					[hitlist addObject:e2];
			}
		}
	}
	return  [hitlist sortedArrayUsingSelector:@selector(compareZeroDistance:)];
}

- (int) getFirstEntityHitByLaserFromEntity:(Entity *) e1
{
	NSArray *entlist = [NSArray arrayWithArray:entities];
	Entity  *hit_entity = nil;
	int		result = NO_TARGET;
	double  nearest = PARTICLE_LASER_LENGTH;
	int i;
	Quaternion q1 = [e1 QRotation];
	if ([e1 isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   //  reverse for player viewpoint
	Vector u1 = vector_up_from_quaternion(q1);
	Vector f1 = vector_forward_from_quaternion(q1);
	Vector r1 = vector_right_from_quaternion(q1);
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide])&&([e2 isKindOfClass:[ShipEntity class]]))
		{
			BoundingBox arbb = [e2 findBoundingBoxRelativeTo:e1 InVectors:r1 :u1 :f1];
			if ((arbb.min_x < 0.0)&&(arbb.max_x > 0.0)&&(arbb.min_y < 0.0)&&(arbb.max_y > 0.0)&&(arbb.min_z > 0.0)&&(arbb.min_z < nearest))
			{
				hit_entity = e2;
				nearest = arbb.min_z;
			}
		}
	}
	if (hit_entity)
		result = [hit_entity universal_id];
	return result;
}

- (int) getFirstEntityHitByLaserFromEntity:(Entity *) e1 inView:(int) viewdir
{
	NSArray *entlist = [NSArray arrayWithArray:entities];
	Entity  *hit_entity = nil;
	int		result = NO_TARGET;
	double  nearest = PARTICLE_LASER_LENGTH;
	int i;
	Quaternion q1 = [e1 QRotation];
	if ([e1 isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   //  reverse for player viewpoint
	Vector u1 = vector_up_from_quaternion(q1);
	switch (viewdir)
	{
		case VIEW_AFT :
			quaternion_rotate_about_axis(&q1, u1, PI);
			break;
		case VIEW_PORT :
			quaternion_rotate_about_axis(&q1, u1, PI/2.0);
			break;
		case VIEW_STARBOARD :
			quaternion_rotate_about_axis(&q1, u1, -PI/2.0);
			break;
	}
	Vector f1 = vector_forward_from_quaternion(q1);
	Vector r1 = vector_right_from_quaternion(q1);
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide])&&([e2 isKindOfClass:[ShipEntity class]]))
		{
			BoundingBox arbb = [e2 findBoundingBoxRelativeTo:e1 InVectors:r1 :u1 :f1];
			if ((arbb.min_x < 0.0)&&(arbb.max_x > 0.0)&&(arbb.min_y < 0.0)&&(arbb.max_y > 0.0)&&(arbb.min_z > 0.0)&&(arbb.min_z < nearest))
			{
				hit_entity = e2;
				nearest = arbb.min_z;
			}
		}
	}
	if (hit_entity)
		result = [hit_entity universal_id];
	return result;
}

- (int) getFirstEntityTargettedFromEntity:(Entity *) e1 inView:(int) viewdir
{
	NSArray *entlist = [NSArray arrayWithArray:entities];
	Entity  *hit_entity = nil;
	int		result = NO_TARGET;
	double  nearest = SCANNER_MAX_RANGE;
	int i;
	Vector p1 = [e1 getPosition];
	Quaternion q1 = [e1 QRotation];
	if ([e1 isKindOfClass:[PlayerEntity class]])
		q1.w = -q1.w;   //  reverse for player viewpoint
	Vector u1 = vector_up_from_quaternion(q1);
	switch (viewdir)
	{
		case VIEW_AFT :
			quaternion_rotate_about_axis(&q1, u1, PI);
			break;
		case VIEW_PORT :
			quaternion_rotate_about_axis(&q1, u1, PI/2.0);
			break;
		case VIEW_STARBOARD :
			quaternion_rotate_about_axis(&q1, u1, -PI/2.0);
			break;
	}
	Vector f1 = vector_forward_from_quaternion(q1);
	Vector r1 = vector_right_from_quaternion(q1);
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide])&&([e2 isKindOfClass:[ShipEntity class]]))
		{
			Vector rp = [e2 getPosition];
			rp.x -= p1.x;	rp.y -= p1.y;	rp.z -= p1.z;
			double df = dot_product(f1,rp);
			if ((df > 0.0)&&(df < nearest))
			{
				double du = dot_product(u1,rp);
				double dr = dot_product(r1,rp);
				double cr = [e2 collisionRadius];
				if (du*du + dr*dr < cr*cr)
				{
					hit_entity = e2;
					nearest = df;
				}
			}
		}
	}
	if (hit_entity)
	{
		result = [hit_entity universal_id];
		//NSLog(@"===> First entity Targetted is %@ %d with collisionRadius %.1f", [(ShipEntity *) hit_entity name], [hit_entity universal_id], [hit_entity collisionRadius]);
	}
	return result;
}

- (NSArray *) getEntitiesWithinRange:(double) range1 ofEntity:(Entity *) e1
{
	NSArray *entlist = [NSArray arrayWithArray:entities];
	NSMutableArray *hitlist = [NSMutableArray arrayWithCapacity:4];
	int i;
	Vector p1 = [e1 getPosition];
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 canCollide]))
		{
			Vector p2 = [e2 getPosition];
			p2.x -= p1.x;	p2.y -= p1.y;	p2.z -= p1.z;
			double cr = range1 + [e2 collisionRadius];
			double d2 = p2.x*p2.x + p2.y*p2.y + p2.z*p2.z - cr*cr;
			if (d2 < 0)
				[hitlist addObject:e2];
		}
	}
	return  [NSArray arrayWithArray:hitlist];
}

- (int) countShipsWithRole:(NSString *) desc inRange:(double) range1 ofEntity:(Entity *)e1
{
	int i, found;
	NSArray *entlist = [NSArray arrayWithArray:entities];
	found = 0;
	Vector p1 = [e1 getPosition];
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if ((e2 != e1)&&([e2 isKindOfClass:[ShipEntity class]])&&([[(ShipEntity *)e2 roles] isEqual:desc]))
		{
			Vector p2 = [e2 getPosition];
			p2.x -= p1.x;	p2.y -= p1.y;	p2.z -= p1.z;
			double cr = range1 + [e2 collisionRadius];
			double d2 = p2.x*p2.x + p2.y*p2.y + p2.z*p2.z - cr*cr;
			if (d2 < 0)
				found++;
		}
	}
	return  found;
}

- (int) countShipsWithRole:(NSString *) desc
{
	int i, found;
	NSArray *entlist = [NSArray arrayWithArray:entities];
	found = 0;
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if (([e2 isKindOfClass:[ShipEntity class]])&&([[(ShipEntity *)e2 roles] isEqual:desc]))
			found++;
	}
	return  found;
}

- (void) sendShipsWithRole:(NSString *) desc messageToAI:(NSString *) ms
{
	int i, found;
	NSArray *entlist = [NSArray arrayWithArray:entities];
	found = 0;
	for (i = 0; i < [entlist count]; i++)
	{
		Entity *e2 = [entlist objectAtIndex:i];
		if (([e2 isKindOfClass:[ShipEntity class]])&&([[(ShipEntity *)e2 roles] isEqual:desc]))
			[[(ShipEntity *)e2 getAI] reactToMessage:ms];
	}
}


- (double) getTime
{
	return universal_time;
}

- (void) findCollisions
{
	Entity *e1,*e2;
	Vector p1, p2;
	double dist, r1, r2, min_dist;
	int i,j;
	//
	// use a non-mutable copy so this can't be changed under us.
	//
	NSArray  *entityList = [[NSArray alloc] initWithArray:entities];	// alloc retains
	//
	int ent_count = [entityList count];
	
	for (i = 0; i < ent_count; i++)
	{
		e1 = (Entity *)[entityList objectAtIndex:i];
		[[e1 collisionArray] removeAllObjects];
	}
	if (ent_count <= 1)
		return;
	for (i = 0; i < ent_count; i++)
	{
		e1 = (Entity *)[entityList objectAtIndex:i];
		if ([e1 canCollide])
		{
			p1 = [e1 getPosition];
			r1 = [e1 collisionRadius];
			for (j = 1; j < ent_count; j++)
			{
				e2 = (Entity *)[entityList objectAtIndex:j];
				if ((j != i)&&([e2 canCollide]))
				{
					p2 = [e2 getPosition];
					r2 = [e2 collisionRadius];
					p2.x -= p1.x;   p2.y -= p1.y;   p2.z -= p1.z;
					dist = p2.x*p2.x + p2.y*p2.y + p2.z*p2.z;
					min_dist = (r1 + r2) * (r1 + r2);
					if (dist < min_dist)
					{
						BOOL	coll1 = [e1 checkCloseCollisionWith:e2];
						BOOL	coll2 = [e2 checkCloseCollisionWith:e1];
						//NSLog(@"Checking close collision between entities [%@:%@]",[e1 getModel],[e2 getModel]);
						//if (coll1)  NSLog(@"%@ ccw %@",[e1 getModel],[e2 getModel]);
						//if (coll2)  NSLog(@"%@ ccw %@",[e2 getModel],[e1 getModel]);
						if ( coll1 && coll2 )
						{
							//NSLog(@"collision!");
							[[e1 collisionArray] addObject:e2];
							[[e2 collisionArray] addObject:e1];
						}
					}
					if (dumpCollisionInfo)
						NSLog(@"Entity %d (%.1f) to entity %d (%.1f)- distance  %.1f (%.1f,%.1f,%.1f)", i, r1, j, r2, sqrt(dist), p2.x, p2.y, p2.z);
				}
			}
		}
	}
	if (dumpCollisionInfo)
		dumpCollisionInfo = NO;
	//
	[entityList release];   // we're done with this now
}

- (void) dumpCollisions
{
	dumpCollisionInfo = YES;
}

- (void) setViewDirection:(int) vd
{
	NSString	*ms = nil;
	
	if ((viewDirection == vd)&&(!displayText))
		return;
	
	switch (vd)
	{
		case VIEW_FORWARD :
			ms = @"Forward View";
			displayText = NO;   // switch off any text displays
			break;
		case VIEW_AFT :
			ms = @"Aft View";
			displayText = NO;   // switch off any text displays
			break;
		case VIEW_PORT :
			ms = @"Port View";
			displayText = NO;   // switch off any text displays
			break;
		case VIEW_STARBOARD :
			ms = @"Starboard View";
			displayText = NO;   // switch off any text displays
			break;
		default :
			break;
	}
	if (viewDirection != vd)
	{
		viewDirection = vd;
		if (ms)
			[self addMessage:ms forCount:3];
	}
	//NSLog(@"Universe viewDir : %d %@",viewDirection,ms);
}

- (int) viewDir
{
	//NSLog(@"Universe viewDir : %d",viewDirection);
	return viewDirection;
}

- (void) displayMessage:(NSString *) text forCount:(int) count
{
	if (![currentMessage isEqual:text])
    {
		if (msSpriteArray[0]) [msSpriteArray[0] release];
		msSpriteArray[0] = [[OpenGLSprite alloc] initWithText:text];	// alloc retains
		
		if (currentMessage)	[currentMessage release];
		currentMessage = [text retain];
    }
	msDuration[0] = (double)count;
	//NSLog(@"-->> Message '%@' %.1f generated.",text,msDuration[0]);
}

- (void) addMessage:(NSString *) text forCount:(int) count
{
	if (![currentMessage isEqual:text])
    {
		// shuffle messages up
		int i;
		if (msSpriteArray[MAX_MESSAGES - 1])	// remove top message if necessary
		{
			[msSpriteArray[MAX_MESSAGES - 1] release];
			msSpriteArray[MAX_MESSAGES - 1] = nil;
		}
		for (i = MAX_MESSAGES - 1 ; i > 0; i--)
		{
			msSpriteArray[i] = msSpriteArray[i - 1];
			msDuration[i] = msDuration[i - 1];
		}
		// insert message at message zero
		msSpriteArray[0] = [[OpenGLSprite alloc] initWithText:text];	// alloc retains
		
		//NSLog(@"-->> Message '%@' %.1f generated.",text,msDuration[0]);
		
		//speech synthesis
		if ([(PlayerEntity *)[self entityZero] speech_on])
		{
			// really ought to collect these together into some sort of translation dictionary really... (TODO)
			NSString *spoken_text = [[text componentsSeparatedByString:@"\n"] componentsJoinedByString:@" "];
			spoken_text = [[spoken_text componentsSeparatedByString:@"\t"] componentsJoinedByString:@" "];
			spoken_text = [[spoken_text componentsSeparatedByString:@" Cr"] componentsJoinedByString:@" credits"];
			spoken_text = [[spoken_text componentsSeparatedByString:@" s."] componentsJoinedByString:@" seconds"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Purse"] componentsJoinedByString:@", Total "];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Mark III"] componentsJoinedByString:@"Mach Three"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Mark II"] componentsJoinedByString:@"Mach Two"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Mark I"] componentsJoinedByString:@"Mach One"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"GalCop"] componentsJoinedByString:@"galactic police"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Witchspace"] componentsJoinedByString:@"Witch+space"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Fer-de-Lance"] componentsJoinedByString:@"fairdehlauns"];
			spoken_text = [[spoken_text componentsSeparatedByString:@"Sidewinder"] componentsJoinedByString:@"Side+winder"];
			if ([speechSynthesizer isSpeaking])
				[speechSynthesizer stopSpeaking];
			[speechSynthesizer startSpeakingString:spoken_text];
		}
		
		if (currentMessage)	[currentMessage release];
		currentMessage = [text retain];
    }
	msDuration[0] = (double)count;
}

- (void) update:(double) delta_t
{
    int i;
	PlayerEntity	*player = (PlayerEntity *)[self entityZero];
	//
	// use a non-mutable copy so this can't be changed under us.
	//
	NSArray  *entityList = [[NSArray alloc] initWithArray:entities];	// alloc retains
    //
	universal_time += delta_t;
	//
	if ((demo_stage)&&([player getStatus] == STATUS_DEMO)&&(universal_time > demo_stage_time)&&([player gui_screen] == GUI_SCREEN_INTRO2))
	{
		if ([entityList count] > 1)
		{
			ShipEntity  *demo_ship = (ShipEntity *)[entityList objectAtIndex:1];
			Vector  vel;
			Quaternion		q2;
			q2.x = 0.0;   q2.y = 0.0;   q2.z = 0.0; q2.w = 1.0;
			quaternion_rotate_about_y(&q2,PI);
			switch (demo_stage)
			{
				case DEMO_FLY_IN :
					vel.x = 0.0;	vel.y = 0.0;	vel.z = 0.0;
					[demo_ship setVelocity:vel];
					demo_stage = DEMO_SHOW_THING;
					demo_stage_time = universal_time + 6.0;
					break;
				case DEMO_SHOW_THING :
					vel.x = 0.0;	vel.y = 0.0;	vel.z = 3.6*[demo_ship collisionRadius]*100.0;
					[demo_ship setVelocity:vel];
					demo_stage = DEMO_FLY_OUT;
					demo_stage_time = universal_time + 1.5;
					break;
				case DEMO_FLY_OUT :
					// change the demo_ship here
					demo_ship_index++;
					demo_ship_index %= [demo_ships count];
					[demo_ship setUpShipFromDictionary:[self getDictionaryForShip:[demo_ships objectAtIndex:demo_ship_index]]];
					[[demo_ship getAI] setStateMachine:@"nullAI.plist"];
					[demo_ship setQRotation:q2];
					[demo_ship setPosition:0.0:0.0:3.6*[demo_ship collisionRadius]*100.0];
					vel.x = 0.0;	vel.y = 0.0;	vel.z = -3.6*[demo_ship collisionRadius]*100.0;
					[demo_ship setVelocity:vel];
					[demo_ship setScanClass: CLASS_NO_DRAW];
					[demo_ship setRoll:PI/5.0];
					[demo_ship setPitch:PI/10.0];
					[gui setText:[demo_ship name] forRow:19 align:GUI_ALIGN_CENTER];
					[gui generateDisplay];
					[self guiUpdated];
					demo_stage = DEMO_FLY_IN;
					demo_stage_time = universal_time + 1.5;
					break;
			}
		}
	}
				
	//
	for (i = 0; i < [entityList count]; i++)
    {
        Entity *thing = [[entityList objectAtIndex:i] retain];
        
		[thing update:delta_t];
		
		if ([thing isKindOfClass:[ShipEntity class]])
		{
			AI* theShipsAI = [(ShipEntity *)thing getAI];
			if ((universal_time > [theShipsAI nextThinkTime])||([theShipsAI nextThinkTime] == 0.0))
			{
				[theShipsAI setNextThinkTime:universal_time + [theShipsAI thinkTimeInterval]];
				[theShipsAI think];
			}
		}
		
		[thing release];
    }
	//
	//
	//
	[self findCollisions];
	//
	//
	//
	int ms = 0;
	while ((ms < MAX_MESSAGES)&&(msDuration[ms] > 0.0))
	{
		msDuration[ms] -= delta_t;
		if (msDuration[ms] <= 0.0)
		{
			msDuration[ms] = 0.0;
			[msSpriteArray[ms] release];
			msSpriteArray[ms] = nil;
			if (ms == 0)
			{
				[currentMessage release];
				currentMessage = nil;
			}
		}
		ms++;
	}
	//
	//
	//
	fpsDisplayDuration -= delta_t;
	if (fpsDisplayDuration < 0.0)
	{
		fpsDisplayDuration = 0.0;
		[fpsSprite setText:[NSString stringWithFormat:@"FPS %d\nObjs %d",(int)(1.0 / delta_t),[entityList count]]];

		fpsDisplayDuration = 1.5;
	}
	//
	// dispose of the non-mutable copy and everything it references neatly
	//
	[entityList release];
}

- (void) setSystemTo:(Random_Seed) s_seed
{
	NSDictionary*   systemData;
	PlayerEntity*   player = (PlayerEntity *)[self entityZero];
	
	galaxy_seed = [player galaxy_seed];
	system_seed = s_seed;
	target_system_seed = s_seed;
	
	systemData =		[[self generateSystemData:target_system_seed] retain];  // retained
	int economy =		[(NSNumber *)[systemData objectForKey:@"economy"] intValue];
	
	[self generateEconomicDataWithEconomy:economy andRandomFactor:[player random_factor]];
	
	[systemData release];   // released
}

- (NSDictionary *) descriptions
{
	return descriptions;
}

- (NSDictionary *) missiontext
{
	return missiontext;
}

- (NSDictionary *) generateSystemData:(Random_Seed) s_seed
{
	NSMutableDictionary* systemdata = [[NSMutableDictionary alloc] initWithCapacity:8];
	
	int government = (s_seed.c / 8) & 7;
	
	int economy = s_seed.b & 7;
	if (government < 2)
		economy = economy | 2;
	
	int techlevel = (economy ^ 7) + (s_seed.d & 3) + (government / 2) + (government & 1);
	
	int population = (techlevel * 4) + government + economy + 1;
	
	int productivity = ((economy ^ 7) + 3) * (government + 4) * population * 8;
	
	int radius = (((s_seed.f & 15) + 11) * 256) + s_seed.d;
	
	[systemdata setObject:[NSNumber numberWithInt:government]		forKey:@"government"];
	[systemdata setObject:[NSNumber numberWithInt:economy]			forKey:@"economy"];
	[systemdata setObject:[NSNumber numberWithInt:techlevel]		forKey:@"techlevel"];
	[systemdata setObject:[NSNumber numberWithInt:population]		forKey:@"population"];
	[systemdata setObject:[NSNumber numberWithInt:productivity]		forKey:@"productivity"];
	[systemdata setObject:[NSNumber numberWithInt:radius]			forKey:@"radius"];
	[systemdata setObject:[self generateSystemName:s_seed]			forKey:@"name"];
	[systemdata setObject:[self generateSystemInhabitants:s_seed]	forKey:@"inhabitants"];
	[systemdata setObject:[self generateSystemDescription:s_seed]	forKey:@"description"];
	
	//NSLog(@"Generated system data is :\n%@",[systemdata description]);
	
	return [NSDictionary dictionaryWithDictionary:[systemdata autorelease]];
}

- (NSString *) generateSystemName:(Random_Seed) s_seed
{
	int i;
		
	NSString*			digrams = @"ABOUSEITILETSTONLONUTHNOALLEXEGEZACEBISOUSESARMAINDIREA?ERATENBERALAVETIEDORQUANTEISRION";
	NSMutableString*	name = [NSMutableString stringWithString:@""];
	int size = 4;
	
	if ((s_seed.a & 0x40) == 0)
		size = 3;
	
	for (i = 0; i < size; i++)
	{
		NSString *c1, *c2;
		int x = s_seed.f & 0x1f;
		if (x != 0)
		{
			x += 12;	x *= 2;
			c1 = [digrams substringWithRange:NSMakeRange(x,1)];
			c2 = [digrams substringWithRange:NSMakeRange(x+1,1)];
			[name appendString:c1];
			if (![c2 isEqual:@"?"])		[name appendString:c2];
		}
		rotate_seed(&s_seed);
	}
	
	return [NSString stringWithString:[name capitalizedString]];
}

- (NSString *) generateSystemInhabitants:(Random_Seed) s_seed
{
	NSMutableString* inhabitants= [NSMutableString stringWithString:@""];

	if (s_seed.e < 127)
		[inhabitants appendString:@"Human Colonial"];
	else
	{
		int inhab = (s_seed.f / 4) & 7;
		if (inhab < 3)
			[inhabitants appendString:(NSString *)[(NSArray *)[(NSArray *)[descriptions objectForKey:@"inhabitants"] objectAtIndex:0] objectAtIndex:inhab]];
		
		inhab = s_seed.f / 32;
		if (inhab < 6)
		{
			[inhabitants appendString:@" "];
			[inhabitants appendString:(NSString *)[(NSArray *)[(NSArray *)[descriptions objectForKey:@"inhabitants"] objectAtIndex:1] objectAtIndex:inhab]];
		}

		inhab = (s_seed.d ^ s_seed.b) & 7;
		if (inhab < 6)
		{
			[inhabitants appendString:@" "];
			[inhabitants appendString:(NSString *)[(NSArray *)[(NSArray *)[descriptions objectForKey:@"inhabitants"] objectAtIndex:2] objectAtIndex:inhab]];
		}

		inhab = (inhab + (s_seed.f & 3)) & 7;
		[inhabitants appendString:@" "];
		[inhabitants appendString:(NSString *)[(NSArray *)[(NSArray *)[descriptions objectForKey:@"inhabitants"] objectAtIndex:3] objectAtIndex:inhab]];
	}
	[inhabitants appendString:@"s"];
	//	
	return [NSString stringWithString:inhabitants];
}


- (Random_Seed) findSystemAtCoords:(NSPoint) coords withGalaxySeed:(Random_Seed) gal_seed
{
	Random_Seed system;
	int distance, dx, dy;
	int i;
    int min_dist = 10000;
	Random_Seed g_seed = gal_seed;

	for (i = 0; i < 256; i++)
	{
		dx = abs(coords.x - g_seed.d);
		dy = abs(coords.y - g_seed.b);

		if (dx > dy)
			distance = (dx + dx + dy) / 2;
		else
			distance = (dx + dy + dy) / 2;

		if (distance < min_dist)
		{
			min_dist = distance;
			system = g_seed;
		}
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
	}

	return system;
}

- (int) findSystemNumberAtCoords:(NSPoint) coords withGalaxySeed:(Random_Seed) gal_seed
{
	int system;
	int distance, dx, dy;
	int i;
    int min_dist = 10000;
	Random_Seed g_seed = gal_seed;

	for (i = 0; i < 256; i++)
	{
		dx = abs(coords.x - g_seed.d);
		dy = abs(coords.y - g_seed.b);

		if (dx > dy)
			distance = (dx + dx + dy) / 2;
		else
			distance = (dx + dy + dy) / 2;

		if (distance < min_dist)
		{
			min_dist = distance;
			system = i;
		}
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
	}
	return system;
}

- (NSPoint) findSystemCoordinatesWithPrefix:(NSString *) p_fix withGalaxySeed:(Random_Seed) gal_seed
{
	NSPoint system_coords = NSMakePoint(0.0,0.0);
	int i;
	Random_Seed g_seed = gal_seed;
	for (i = 0; i < 256; i++)
	{
		if ([[[self generateSystemName:g_seed] lowercaseString] hasPrefix:p_fix])
		{
			system_coords.x = g_seed.d;
			system_coords.y = g_seed.b;
		}
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
		rotate_seed(&g_seed);
	}
	return system_coords;
}

- (NSArray *) equipmentdata
{
	return equipmentdata;
}

- (NSDictionary *) commoditylists
{
	return commoditylists;
}

- (NSArray *) commoditydata
{
	return commoditydata;
}

- (BOOL) generateEconomicDataWithEconomy:(int) economy andRandomFactor:(int) random_factor
{
	StationEntity *some_station = [self station];
	//NSLog(@">>>>> generateEconomicDataWithEconomy:andRandomFactor for System");
	NSString *station_roles = [some_station roles];
	if (![commoditylists objectForKey:station_roles])
		station_roles = @"default";

	NSArray *newcommoditydata = [[self commodityDataForEconomy:economy andStation:some_station andRandomFactor:random_factor] retain];
	[commoditydata release];
	commoditydata = newcommoditydata;
	return YES;
}

- (NSArray *) commodityDataForEconomy:(int) economy andStation:(StationEntity *)some_station andRandomFactor:(int) random_factor
{
	NSString *station_roles = [some_station roles];
	
	//NSLog(@"///// station roles detected as '%@'", station_roles);
	
	if (![commoditylists objectForKey:station_roles])
	{
		//NSLog(@"///// using default economy");
		station_roles = @"default";
	}
	else
	{
		//NSLog(@"///// found a special economy");
	}
		
	NSMutableArray *ourEconomy = [NSMutableArray arrayWithArray:(NSArray *)[commoditylists objectForKey:station_roles]];
	int i;
	
	for (i = 0; i < [ourEconomy count]; i++)
	{
		NSMutableArray *commodityInfo = [[NSMutableArray arrayWithArray:[ourEconomy objectAtIndex:i]] retain];  // retain
		
		int base_price =			[(NSNumber *)[commodityInfo objectAtIndex:MARKET_BASE_PRICE] intValue];
		int eco_adjust_price =		[(NSNumber *)[commodityInfo objectAtIndex:MARKET_ECO_ADJUST_PRICE] intValue];
		int eco_adjust_quantity =	[(NSNumber *)[commodityInfo objectAtIndex:MARKET_ECO_ADJUST_QUANTITY] intValue];
		int base_quantity =			[(NSNumber *)[commodityInfo objectAtIndex:MARKET_BASE_QUANTITY] intValue];
		int mask_price =			[(NSNumber *)[commodityInfo objectAtIndex:MARKET_MASK_PRICE] intValue];
		int mask_quantity =			[(NSNumber *)[commodityInfo objectAtIndex:MARKET_MASK_QUANTITY] intValue];
		
		int price =		(base_price + (random_factor & mask_price) + (economy * eco_adjust_price)) & 255;
		int quantity =  (base_quantity  + (random_factor & mask_quantity) - (economy * eco_adjust_quantity)) & 255;
		
		if (quantity > 127) quantity = 0;
		quantity &= 63;
		
		/*-- ***** special pleading removed after adding separate price and quantity masks *****
		if ([(NSString *)[commodityInfo objectAtIndex:MARKET_NAME]  hasPrefix:@"Alien"])
			quantity = 0;		// No 'Alien' things for sale!
		--*/
		
		[commodityInfo replaceObjectAtIndex:MARKET_PRICE withObject:[NSNumber numberWithInt:price * 4]];
		[commodityInfo replaceObjectAtIndex:MARKET_QUANTITY withObject:[NSNumber numberWithInt:quantity]];
		
		[ourEconomy replaceObjectAtIndex:i withObject:[NSArray arrayWithArray:commodityInfo]];
		[commodityInfo release];	// release, done
	}
	
	return [NSArray arrayWithArray:ourEconomy];
}

- (NSString *) generateSystemDescription:(Random_Seed) s_seed
{
/*--
	if (cmdr.mission == 1)
	{
		mission_text = mission_planet_desc (planet);
		if (mission_text != NULL)
			return mission_text;
	}
--*/
	seed_for_planet_description (s_seed);

	return [self expandDescription:@"[14] is [22]." forSystem:s_seed];
}

- (NSString *) expandDescription:(NSString *) desc forSystem:(Random_Seed)s_seed;
{
	NSMutableString*	partial = [NSMutableString stringWithString:desc];
	
	while ([partial rangeOfString:@"["].location != NSNotFound)
	{
		NSString	*part, *before, *after;
		int			sub, rnd, opt;
		int			p1 = [partial rangeOfString:@"["].location;
		int			p2 = [partial rangeOfString:@"]"].location + 1;
		
		before = [partial substringWithRange:NSMakeRange(0,p1)];
		after = [partial substringWithRange:NSMakeRange(p2,[partial length] - p2)];
		
		sub = [[partial substringWithRange:NSMakeRange(p1 + 1 , p2 - p1 - 2)] intValue];
		
		//NSLog(@"Expanding:\t%@",partial);
		
		rnd = gen_rnd_number();
		opt = 0;
		if (rnd >= 0x33) opt++;
		if (rnd >= 0x66) opt++;
		if (rnd >= 0x99) opt++;
		if (rnd >= 0xCC) opt++;
		
		part = (NSString *)[(NSArray *)[(NSArray *)[descriptions objectForKey:@"system_description"] objectAtIndex:sub] objectAtIndex:opt];
		
		partial = [NSMutableString stringWithFormat:@"%@%@%@",before,part,after];
	}
		
	[partial	replaceOccurrencesOfString:@"%H"
				withString:[self generateSystemName:s_seed]
				options:NSLiteralSearch range:NSMakeRange(0, [partial length])];
	
	[partial	replaceOccurrencesOfString:@"%I"
				withString:[NSString stringWithFormat:@"%@ian",[self generateSystemName:s_seed]]
				options:NSLiteralSearch range:NSMakeRange(0, [partial length])];
				
	[partial	replaceOccurrencesOfString:@"%R"
				withString:[self getRandomDigrams]
				options:NSLiteralSearch range:NSMakeRange(0, [partial length])];

	return [NSString stringWithString:partial]; 
}

- (NSString *) getRandomDigrams
{
	int i;
	int len = gen_rnd_number() & 3;	
	NSString*			digrams = @"ABOUSEITILETSTONLONUTHNOALLEXEGEZACEBISOUSESARMAINDIREA?ERATENBERALAVETIEDORQUANTEISRION";
	NSMutableString*	name = [NSMutableString stringWithString:@""];
	for (i = 0; i <=len; i++)
	{
		int x =  gen_rnd_number() & 0x3e;
		[name appendString:[digrams substringWithRange:NSMakeRange(x,2)]];
	}
	return [NSString stringWithString:[name capitalizedString]]; 
}

- (Vector) getWitchspaceExitPosition
{
	Vector result;
	seed_for_planet_description(system_seed);

	// new system is hyper-centric : witchspace exit point is origin
	result.x = 0.0;
	result.y = 0.0;
	result.z = 0.0;
	//
	result.x += SCANNER_MAX_RANGE*(gen_rnd_number()/256.0 - 0.5);   // offset by a set amount, up to 12.8 km
	result.y += SCANNER_MAX_RANGE*(gen_rnd_number()/256.0 - 0.5);
	result.z += SCANNER_MAX_RANGE*(gen_rnd_number()/256.0 - 0.5);
	//
	return result;
}

- (Quaternion) getWitchspaceExitRotation
{
	// this should be fairly close to {0,0,0,1}
	Quaternion q_result;
	seed_for_planet_description(system_seed);
	//
	q_result.x = (gen_rnd_number() - 128)/1024.0;
	q_result.y = (gen_rnd_number() - 128)/1024.0;
	q_result.z = (gen_rnd_number() - 128)/1024.0;
	q_result.w = 1.0;
	quaternion_normalise(&q_result);
	//
	return q_result;
}

///////////////////////////////////////

- (GuiDisplayGen *) gui
{
	return gui;
}

- (void) guiUpdated
{
	[gui updateGui];
}

- (void) setDisplayText:(BOOL) value
{
	displayText = value;
}

- (BOOL) displayText
{
	return displayText;
}

- (void) setDisplayFPS:(BOOL) value
{
	displayFPS = value;
}

- (BOOL) displayFPS
{
	return displayFPS;
}


@end
